#!/usr/bin/python

import re
import optparse
import urllib2
import os
import errno
import hashlib
import tarfile
import struct
import subprocess
import shutil
import multiprocessing
import glob

conf = None

def log(s):
    print(s)

def log_v(s):
    print(s)

def log_error(s):
    print(s)

def mkdir_p(p):
    try:
        os.makedirs(p)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(p):
            pass
        else:
            raise
    
class build_failed(Exception):
    def __init__(self, m):
        self.w = m

    def what(self):
        return self.w

class conf_t:
    values = {}

    def __init__(self):
        self.defaults()

        args = self.parse_cmd()

        if args.store:
            self.values["store"] = args.store

        if args.arch:
            self.values["arch"] = self.check_arch(args.arch)

        if args.variant:
            self.values["variant"] = self.check_variant(args.variant)

        if args.prefix:
            self.values["prefix"] = args.prefix

        if args.skip_download:
            self.values["skip_download"] = True

        if args.force_download:
            self.values["force_download"] = True

        if args.skip_extract:
            self.values["skip_extract"] = True

        if args.clean_extract:
            self.values["clean_extract"] = True

        if args.skip_build:
            self.values["skip_build"] = True

        if args.jobs:
            self.values["jobs"] = args.jobs

        if args.skip_wxwidgets:
            self.values["skip_wxwidgets"] = args.skip_wxwidgets

        if args.skip_python:
            self.values["skip_python"] = args.skip_python

        if args.skip_boost:
            self.values["skip_boost"] = args.skip_boost

        if args.skip_libtorrent:
            self.values["skip_libtorrent"] = args.skip_libtorrent

        if args.skip_libxml:
            self.values["skip_libxml"] = args.skip_libxml

        if args.skip_curl:
            self.values["skip_curl"] = args.skip_curl

        if args.skip_iconv:
            self.values["skip_iconv"] = args.skip_iconv

        if args.skip_phoenix:
            self.values["skip_phoenix"] = args.skip_phoenix
    
        p = os.path.join(
            self.values["prefix"],
            self.arch(),
            self.variant())

        p = os.path.abspath(p)

        if p == "" or p == "/":
            raise build_failed("bad prefix directory")

        self.values["prefix"] = p

        p = os.path.abspath(self.values["store"])
        if p == "" or p == "/":
            raise build_failed("bad store directory")

        self.values["store"] = p

        mkdir_p(self.store())
        mkdir_p(self.prefix())


    def store(self):
        return self.values["store"]

    def arch(self):
        return self.values["arch"]

    def variant(self):
        return self.values["variant"]

    def prefix(self):
        return self.values["prefix"]

    def skip_download(self):
        return self.values["skip_download"]

    def force_download(self):
        return self.values["force_download"]

    def skip_extract(self):
        return self.values["skip_extract"]

    def clean_extract(self):
        return self.values["clean_extract"]

    def skip_build(self):
        return self.values["skip_build"]

    def jobs(self):
        return self.values["jobs"]

    def skip_wxwidgets(self):
        return self.values["skip_wxwidgets"]

    def skip_python(self):
        return self.values["skip_python"]

    def skip_boost(self):
        return self.values["skip_boost"]

    def skip_libtorrent(self):
        return self.values["skip_libtorrent"]

    def skip_libxml(self):
        return self.values["skip_libxml"]

    def skip_curl(self):
        return self.values["skip_curl"]

    def skip_iconv(self):
        return self.values["skip_iconv"]

    def skip_phoenix(self):
        return self.values["skip_phoenix"]


    def x64(self):
        return (self.arch() == "x64")

    def x86(self):
        return (self.arch() == "x86")

    def debug(self):
        return (self.variant() == "debug")

    def release(self):
        return (self.variant() == "release")


    def defaults(self):
        self.values["store"] = "./deps-store"
        self.values["arch"] = self.get_arch()
        self.values["variant"] = "debug"
        self.values["prefix"] = "./deps"
        self.values["skip_download"] = False
        self.values["force_download"] = False
        self.values["skip_extract"] = False
        self.values["clean_extract"] = False
        self.values["skip_build"] = False
        self.values["jobs"] = multiprocessing.cpu_count()

        self.values["skip_wxwidgets"] = False
        self.values["skip_python"] = False
        self.values["skip_boost"] = False
        self.values["skip_libtorrent"] = False
        self.values["skip_libxml"] = False
        self.values["skip_curl"] = False
        self.values["skip_iconv"] = False
        self.values["skip_phoenix"] = False

    def check_arch(self, a):
        a = a.lower()

        if a == "x86" or a == "x64":
            return a

        return self.get_arch()

    def check_variant(self, v):
        v = v.lower()

        if v == "debug" or v == "release":
            return v

        return "debug"

    def get_arch(self):
        s = 8 * struct.calcsize("P")

        if s == 32:
            return "x86"
        else:
            return "x64"

    def parse_cmd(self):
        parser = optparse.OptionParser()

        parser.add_option(
            "--store", help=
            "directory where the dependencies are put")

        parser.add_option(
            "--arch", help=
            "architecture, either x64 or x86")

        parser.add_option(
            "--variant", help=
            "either debug or release")

        parser.add_option(
            "--prefix", help=
            "directory where the build files are put")

        parser.add_option(
            "--skip-download", action="store_true", help=
            "assumes the archives are present in --store and will " +
            "attempt to extract them if needed")

        parser.add_option(
            "--force-download", action="store_true", help=
            "redownloads the archive even if it is found")

        parser.add_option(
            "--skip-extract", action="store_true", help=
            "assumes the archives have already been downloaded and " +
            "extracted")

        parser.add_option(
            "--clean-extract", action="store_true", help=
            "deletes the extracted directory (if any) and extracts the " +
            "archive again")

        parser.add_option(
            "--skip-build", action="store_true", help=
            "does not build the libraries")

        parser.add_option(
            "--force-configure", action="store_true", help=
            "always call ./configure (if needed), even if Makefile is " +
            "present")

        parser.add_option(
            "--jobs", "-j", type="int", help=
            "overrides the -j used for make")
        

        parser.add_option(
            "--skip-wxwidgets", action="store_true", help=
            "skips wxwidgets")

        parser.add_option(
            "--skip-python", action="store_true", help=
            "skips python")

        parser.add_option(
            "--skip-boost", action="store_true", help=
            "skips boost")

        parser.add_option(
            "--skip-libtorrent", action="store_true", help=
            "skips libtorrent")

        parser.add_option(
            "--skip-libxml", action="store_true", help=
            "skips libxml")

        parser.add_option(
            "--skip-curl", action="store_true", help=
            "skips curl")

        parser.add_option(
            "--skip-iconv", action="store_true", help=
            "skips iconv")

        parser.add_option(
            "--skip-phoenix", action="store_true", help=
            "skips phoenix")

        (opts, args) = parser.parse_args()
        return opts


class library:
    def version(self):
        raise NotImplementedError()

    def version_underscore(self):
        return self.version().replace(".", "_")

    def name(self):
        raise NotImplementedError()

    def url(self):
        raise NotImplementedError()

    def pre_compile(self):
        pass

    def compile(self):
        raise NotImplementedError()

    def install(self):
        raise NotImplementedError()

    def post_compile(self):
        pass

    def build(self):
        self.get_source()

        if not conf.skip_build():
            old_cwd = os.getcwd()
            os.chdir(self.source_directory())

            self.pre_compile()
            self.compile()
            self.install()
            self.post_compile()

            os.chdir(old_cwd)

    def get_source(self):
        if not conf.skip_download():
            self.download()

        if not conf.skip_extract():
            self.decompress()

    def download(self):
        url = self.url()
        target = self.downloaded_file()
       
        if not conf.force_download():
            if os.path.isfile(target):
                if self.check_sha1(target):
                    log(target + " already exists")
                    return
                
                log(target + " already exists, but sha1 is bad")

        log("downloading " + url)

        if os.path.isfile(target):
            os.remove(target)

        self.download_impl(url, target)
        if not self.check_sha1(target):
            log_error("download failed, sha1 doesn't match")

    def decompress(self):
        tar_file = self.downloaded_file()
        target = self.source_directory()
        
        if not conf.clean_extract():
            log("checking for " + target)
            if os.path.isdir(target):
                log("directory " + target + " already exists")
                return

        log("untaring " + tar_file)

        if os.path.isdir(target):
            shutil.rmtree(target)

        t = tarfile.open(tar_file)
        t.extractall(conf.store())
        t.close()

    def downloaded_file(self):
        url = self.url()
        filename = url.split('/')[-1]

        p = filename.find("?")
        if p >= 0:
            filename = filename[0:p]
        
        return os.path.join(conf.store(), filename)

    def source_directory(self):
        f = self.downloaded_file()
        filename = f.split('/')[-1]

        tar_file = os.path.join(conf.store(), filename)
        name, ext = os.path.splitext(filename)

        if ext == ".bz2" or ext == ".gz":
            name, ext = os.path.splitext(name)

        return os.path.join(conf.store(), name)

    def download_impl(self, url, target):
        r = urllib2.urlopen(url)
        f = open(target, "wb")

        while True:
            buffer = r.read(8192)
            if not buffer:
                break

            f.write(buffer)

        f.close()

    def check_sha1(self, target):
        f = open(target, "rb")
        h = hashlib.sha1()

        while True:
            buffer = f.read(65536)
            if not buffer:
                break

            h.update(buffer)
    
        d = h.hexdigest()
        log_v("sha1 is " + d + ", expecting " + self.sha1())

        if d != self.sha1():
            return False
        return True

    def do_configure(self, flags):
        c = self.join_flags(flags)
        
        r = self.run_cmd("./configure " + c)
        if r != 0:
            raise build_failed("configure failed")

    def do_make(self, flags=[]):
        c = self.join_flags(flags)
        c = "-j" + str(conf.jobs()) + " " + c

        r = self.run_cmd("make " + c)
        if r != 0:
            raise build_failed("make failed")

    def do_b2(self, flags=[]):
        c = self.join_flags(flags)
        c = "-j" + str(conf.jobs()) + " " + c

        b = boost()
        b2 = os.path.join(
            b.source_directory(),
            "b2")

        r = self.run_cmd(b2 + " " + c + " install")
        if r != 0:
            raise build_failed("b2 failed")

    def do_make_install(self, flags=[]):
        c = self.join_flags(flags)

        r = self.run_cmd("make " + c + " install")
        if r != 0:
            raise build_failed("make install failed")

    def join_flags(self, d):
        flags = []

        if "common" in d:
            flags.extend(d["common"])

        if conf.debug():
            if "debug" in d:
                flags.extend(d["debug"])
        elif conf.release():
            if "release" in d:
                flags.extend(d["release"])

        if conf.x64():
            if "x64" in d:
                flags.extend(d["x64"])
        elif conf.x86():
            if "x86" in d:
                flags.extend(d["x86"])

        return " ".join(flags)

    def run_cmd(self, c):
        log_v(c)
        return subprocess.call(c, shell=True)


class wxwidgets(library):
    def version(self):
        return "3.0.0"

    def short_version(self):
        return "3.0"

    def name(self):
        return "wxWidgets"

    def url(self):
        url = "http://downloads.sourceforge.net/project/wxwindows/"
        url += self.version() + "/"
        url += "wxWidgets-" + self.version() + ".tar.bz2"

        return url

    def sha1(self):
        return "756a9c54d1f411e262f03bacb78ccef085a9880a"
    
    def compile(self):
        self.do_configure(self.configure_flags())
        self.do_make()

    def install(self):
        self.do_make_install()

        s = os.path.join(
            self.source_directory(),
            "include",
            "wx",
            "preferences.h")

        d = os.path.join(
            conf.prefix(),
            "include",
            "wx-" + self.short_version(),
            "wx")

        shutil.copy(s, d)
    
    def configure_flags(self):
        return {
            "common": [
                "--prefix=" + conf.prefix(),
                "--with-opengl",
                "--enable-monolithic",
                "--enable-prefseditor"],
            "debug": [
                "--enable-debug",
                "--enable-debug_info",
                "--enable-debug_gdb"],
            
            "release": [
                "--disable-debug_flag"]
        }

class python(library):
    def version(self):
        return "3.3.3"

    def short_version(self):
        return "3.3"

    def name(self):
        return "python"

    def executable_name(self):
        return "python" + self.short_version()

    def packages_path(self):
        p = os.path.join(
            conf.prefix(),
            "lib",
            "python" + self.short_version(), "site-packages")

        return p

    def executable_path(self):
        return os.path.join(
            conf.prefix(),
            "bin",
            self.executable_name())

    def include_path(self):
        return os.path.join(
            conf.prefix(),
            "include",
            "python" + self.short_version() + "m")

    def library_path(self):
        return os.path.join(
            conf.prefix(),
            "lib")

    def url(self):
        url = "http://www.python.org/ftp/python/"
        url += self.version() + "/"
        url += "Python-" + self.version() + ".tar.bz2"

        return url

    def sha1(self):
        return "6ff7d34427cbf7bf875e6a456850231e488118ca"

    def compile(self):
        self.do_configure(self.configure_flags())
        self.do_make()
        self.do_make_install()

    def install(self):
        pass

    def configure_flags(self):
        return {
            "common": [
                "--prefix=" + conf.prefix(),
                "CFLAGS=-fPIC",
                "CPPFLAGS=-fPIC"]
        }

class boost(library):
    def version(self):
        return "1.55.0"

    def name(self):
        return "boost"

    def url(self):
	url = "https://downloads.sourceforge.net/project/boost/boost/"
        url += self.version() + "/"
        url += "boost_" + self.version_underscore() + ".tar.bz2"
        url += "?r=&ts=1356557543&use_mirror=hivelocity"

        return url

    def sha1(self):
        return "cef9a0cc7084b1d639e06cd3bc34e4251524c840"

    def set_b2_env(self):
        bp = os.path.join(
            self.source_directory(),
            "tools",
            "build",
            "v2")

        r = self.source_directory()

        os.environ["BOOST_BUILD_PATH"] = bp
        os.environ["BOOST_ROOT"] = r

    def compile(self):
        if not os.path.isfile("b2"):
            r = self.run_cmd("./bootstrap.sh")
            if r != 0:
                raise build_failed("bootstrap failed")

        self.fix_project_config()
        self.do_b2(self.b2_flags())

    def install(self):
        pass

    def fix_project_config(self):
        c = "project-config.jam"
        tc = c + ".tmp"

        shutil.copyfile(c, c + ".orig")

        s = "s/using python/# using python/"
        red = "> project-config.jam.old" 
        sed = "sed '" + s + "' " + c + " > " + tc
         
        r = self.run_cmd(sed)
        if r != 0:
            raise build_failed("sed failed")

        os.remove(c)
        os.rename(tc, c)

        p = python()
        s = "using python : "
        s += p.short_version() + " : "
        s += p.executable_path() + " : "
        s += p.include_path() + " : "
        s += p.library_path() + " ;"

        f = open(c, "a")
        f.write("\n" + s)
        f.close()

    def b2_flags(self):
        return {
            "common": [
                "link=static",
                "runtime-link=shared",
                "threading=multi",
                "cxxflags=-fPIC",
                "--prefix=" + conf.prefix(),
                "--with-chrono",
                "--with-date_time",
                "--with-exception",
                "--with-filesystem",
                "--with-program_options",
                "--with-regex",
                "--with-signals",
                "--with-system",
                "--with-thread",
                "--with-test",
                "--with-python"],
            "debug": [
                "variant=debug",
                "debug-symbols=on"],
            "release": [
                "variant=release",
                "debug-symbols=off"],
            "x64": [
                "address-model=64"],
            "x86": [
                "address-model=32"]
        }

class libtorrent(library):
    def version(self):
        return "0.16.12"

    def name(self):
        return "libtorrent"

    def url(self):
	url = "https://libtorrent.googlecode.com/files/"
        url += "libtorrent-rasterbar-" + self.version() + ".tar.gz"
        
        return url

    def sha1(self):
        return "cd326fd2609c3d00c25dabdd441a0704efbf0207"

    def compile(self):
        b = boost()
        b.set_b2_env()
        
        self.do_b2(self.b2_flags())

    def install(self):
        pass

    def b2_flags(self):
        return {
            "common": [
                "--prefix=" + conf.prefix(),
                "boost=source",
                "boost-link=static",
                "link=static",
                "runtime-link=shared",
                "cxxflags=-fPIC",
                "logging=none",
                "asserts=off",
                "geoip=off",
                "encryption=openssl",
                "invariant-checks=on",
                "resolve-countries=off",
                "deprecated-functions=off",
                "full-stats=off",
                "-q"],
            "debug": [
                "debug-symbols=on",
                "variant=debug"],
            "release": [
                "variant=release",
                "debug-symbols=off"],
            "x86": [
                "address-model=32"],
            "x64": [
                "address-model=64"]
        }

class libxml(library):
    def version(self):
        return "2.9.1"

    def name(self):
        return "libxml"

    def url(self):
	url = "ftp://xmlsoft.org/libxml2/"
        url += "libxml2-" + self.version() + ".tar.gz"

        return url

    def sha1(self):
        return "eb3e2146c6d68aea5c2a4422ed76fe196f933c21"

    def compile(self):
        self.do_configure(self.configure_flags())
        self.do_make()

    def install(self):
       self.do_make_install() 

    def configure_flags(self):
        return {
            "common": [
                "--prefix=" + conf.prefix(),
                "--without-python",
                "CPPFLAGS=-fPIC",
                "CFLAGS=-fPIC"]
        }
        

class curl(library):
    def version(self):
        return "7.30.0"

    def name(self):
        return "curl"

    def url(self):
	url = "http://curl.haxx.se/download/"
        url += "curl-" + self.version() + ".tar.bz2"

        return url

    def sha1(self):
        return "23fdc215558023b943cea9dfab04b86020037b0d"

    def compile(self):
        self.do_configure(self.configure_flags())
        self.do_make()

    def install(self):
        self.do_make_install()

    def configure_flags(self):
        return {
            "common": [
                "--prefix=" + conf.prefix(),
                "--disable-shared",
                "--disable-ldap",
                "--without-libssh2",
                "--without-librtmp"],
            "debug": [
                "--enable-debug",
                "--enable-curldebug"],
            "release": [
                "--enable-optimize"]
        }


class iconv(library):
    def version(self):
        return "1.14"

    def name(self):
        return "libiconv"

    def url(self):
	url = "http://ftp.gnu.org/pub/gnu/libiconv/"
        url += "libiconv-" + self.version() + ".tar.gz"

        return url

    def sha1(self):
        return "be7d67e50d72ff067b2c0291311bc283add36965"

    def compile(self):
        self.do_configure(self.configure_flags())
        self.do_make()

    def install(self):
        self.do_make_install()
        
        p = os.path.join(conf.prefix(), "lib", "libiconv.so*")
        for f in glob.glob(p):
            os.remove(f)

    def configure_flags(self):
        return {
            "common": [
                "--prefix=" + conf.prefix(),
                "--enable-static",
                "--enable-shared=no",
                "CPPFLAGS=-fPIC",
                "CFLAGS=-fPIC"]
        }

class phoenix(library):
    def version(self):
        return "0"

    def name(self):
        return "phoenix"

    def url(self):
        return ""

    def get_source(self):
        name = "phoenix"
        target = os.path.join(conf.store(), name)

        if not os.path.isdir(target):
            s = "https://svn.wxwidgets.org/svn/wx/wxPython/Phoenix/trunk"
        
            r = self.run_cmd("svn co " + s + " " + target)
            if r != 0:
                raise build_failed("svn checkout failed")

        self.fix_source()

    def source_directory(self):
        return os.path.join(conf.store(), "phoenix")

    def compile(self):
        self.do_build(self.build_args("dox"))
        self.do_build(self.build_args("etg sip"))
        self.fix_sip_source()
        self.do_build(self.build_args("build_py"))

    def install(self):
        p = python()

        source = os.path.join(self.source_directory(), "wx")
        target = os.path.join(p.packages_path(), "wx")

        try:
            shutil.rmtree(target)
        except OSError as e:
            pass

        shutil.copytree(source, target)
        
        incs_source = os.path.join(
            self.source_directory(),
            "src",
            "*.h")

        incs_target = os.path.join(
            conf.prefix(),
            "include",
            "phoenix")

        mkdir_p(incs_target)
        for f in glob.glob(incs_source):
            shutil.copy(f, incs_target)

        # this fails, not sure why
        #self.do_build(self.build_args("install_py"))

    def do_build(self, args):
        r = self.run_cmd("python3 build.py " + args)
        if r != 0:
            raise build_failed("build.py failed")


    def build_args(self, targets):
        w = wxwidgets()
        path = os.path.join(conf.prefix(), "bin") 
        os.environ["PATH"] = path + os.pathsep + os.environ["PATH"]
        os.environ["WXWIN"] = w.source_directory()

        p = python()
        flags = self.join_flags(self.compile_flags())
 
        args = p.short_version() + " " + targets + " " + flags

        return args

    def compile_flags(self):
        return {
            "common": [
                "--build_dir=" + os.path.join(conf.prefix(), "bin"),
                "--destdir=" + conf.prefix(),
                "--jobs=" + str(conf.jobs())]
        }

    def fix_source(self):
        self.replace_in_file(
            os.path.join("sphinxtools", "utilities.py"),
            r'^(\s*)modules\.sort\(\)',
            r'\1modules = sorted(modules)',
            '#')

        self.replace_in_file(
            os.path.join("etgtools", "sphinx_generator.py"),
            r'(?<!string_base = )basestring',
            'string_base',
            '#')

        self.replace_in_file(
            "wscript",
            r'^(\s*)(etg = loadETG\(\'etg/_html2.py\'\)$)',
            r'\1"""\2',
            '#')

        self.replace_in_file(
            "wscript",
            r'^(\s*)(makeExtCopyRule\(bld, \'_html2\'\))$',
            r'\1\2"""',
            '#')

    def fix_sip_source(self):
        self.replace_in_file(
            os.path.join("sip", "cpp", "sip_corewxPyApp.cpp"),
            r'(\s*)#include <wx/gtk/private/win_gtk.h>',
            '',
            '//')

        self.replace_in_file(
            os.path.join("sip", "cpp", "sip_corewxWindow.cpp"),
            r'(\s*)#include <wx/gtk/private/win_gtk.h>',
            '',
            '//')

    def replace_in_file(self, file, pattern, subst, comment_char):
        root = self.source_directory()

        in_file = os.path.join(root, file)
        out_file = os.path.join(root, file + ".baretorrent")

        inf = open(in_file, "r")
        outf = open(out_file, "w+")

        e = re.compile(pattern)

        log("fixing source in " + file)
        log("looking for " + pattern)

        comment = comment_char + "baretorrent "

        found = False
        for line in inf:
            if not line.startswith(comment):
                s = e.sub(subst, line)
            else:
                s = line

            if s == line:
                outf.write(line)
            else:
                outf.write(comment + line)
                outf.write(s)
                found = True

        if found:
            os.remove(in_file)
            os.rename(out_file, in_file)
        else:
            log("not found")
            os.remove(out_file)
        
        inf.close()
        outf.close()



def start():
    global conf
    conf = conf_t()

    log(
        "building dependencies for " +
        conf.variant() + " " + conf.arch())

    libs = []

    if not conf.skip_wxwidgets(): libs.append(wxwidgets())
    if not conf.skip_python(): libs.append(python())
    if not conf.skip_boost(): libs.append(boost())
    if not conf.skip_libtorrent(): libs.append(libtorrent())
    if not conf.skip_libxml(): libs.append(libxml())
    if not conf.skip_curl(): libs.append(curl())
    if not conf.skip_iconv(): libs.append(iconv())
    if not conf.skip_phoenix(): libs.append(phoenix())

    print("libraries:")
    for lb in libs:
        print("  " + lb.name() + " " + lb.version())

    try:
        for lb in libs:
            lb.build()
    except build_failed as e:
        log_error("build failed: " + e.what())

start()

