#!/usr/bin/python

import sys
import re
import os
import shutil
import glob
import subprocess

class config_t:
	def __init__(self):
		if sys.platform.startswith('win'):
			self.sevenz = "C:\\Program Files\\7-Zip\\7z.exe"
			self.src = "..\\src"
			self.gen_bin = "..\\bin"
			self.version_file = self.src + "\\version.h"
			self.msvc = "C:\\Program Files (x86)\\Microsoft Visual Studio 11.0\\VC"
			self.plats = ["windows", "xp"]
		else:
			self.sevenz = "7z"
			self.src = "../src"
			self.gen_bin = "../bin"
			self.version_file = self.src + "/version.h"
			self.plats = ["ubuntu", "debian"]

		self.archs = ["x86", "x64"]
		self.verbose = False

config = config_t()


def usage():
	print("usage: python release.py all")
	print("       python release.py PLAT ARCH")
	print("")
	print("options")
	print("PLAT    one of " + ", ".join(config.plats))
	print("ARCH    one of " + ", ".join(config.archs))

def get_version():
	print("checking " + config.version_file)
	
	major = re.compile("BARETORRENT_VERSION_MAJOR ([0-9]+)")
	minor = re.compile("BARETORRENT_VERSION_MINOR ([0-9]+)")
	maintenance = re.compile("BARETORRENT_VERSION_MAINTENANCE ([0-9]+)")
	
	v = [-1, -1, 0]
	
	for line in open(config.version_file):
		m = major.search(line)
		if m != None:
			v[0] = m.group(1)
		else:
			m = minor.search(line)
			if m != None:
				v[1] = m.group(1)
			else:
				m = maintenance.search(line)
				if m != None:
					v[2] = m.group(1)

	if -1 in v:
		print("bad version file")
		return None
	
	print("version is " + ".".join(v))
	return v

def make_config(plat, arch):
	if plat == "windows":
		return "v110"
	elif plat == "xp":
		return "v110_xp"
	elif plat == "debian":
		return ""
	elif plat == "ubuntu":
		return ""
	
def make_bin(plat, arch):
	if sys.platform.startswith("win"):
		return "..\\bin." + arch + "." + make_config(plat, arch)
	elif sys.platform.startswith("linux"):
		return "../bin." + arch

def copy_dir(src, dest):
	shutil.copytree(src, os.path.join(dest, os.path.basename(src)))

def copy_file(src, dest):
	shutil.copy(src, os.path.join(dest, os.path.basename(src)))

def copy_files(src, pat, dest):
	files = glob.iglob(os.path.join(src, pat))
	for f in files:
		if os.path.isfile(f):
			shutil.copy(f, dest)

def make_installer(v, plat, arch):
	defs = {
		"ARCH": arch,
		"PLATFORM": make_config(plat, arch),
		"PNAME": plat,
		"VERSION_STRING": ".".join(v),
		"VERSION_MAJOR": v[0],
		"VERSION_MINOR": v[1],
		"VERSION_RES": ".".join(v) + ".0" }
	
	args = ["baretorrent.nsi"]

	if config.verbose:
		args.insert(0, "/V4")
	else:
		args.insert(0, "/V2")
	
	for n,v in defs.iteritems():
		args.insert(0, "/D" + n + "=" + v)

	args.insert(0, "makensis")
	
	print("making installer")
	
	fnull = None
	if not config.verbose:
		fnull = open(os.devnull, "w")

	if subprocess.call(args, stdout=fnull) != 0:
		return False
	
	return True

def do_7z(src_dir, target_file):
	args = [config.sevenz, "a", target_file, src_dir]

	fnull = None
	if not config.verbose:
		fnull = open(os.devnull, "w")
	
	if subprocess.call(args, stdout=fnull) != 0:
		return False
	
	return True

def make_win_archive(v, plat, arch, bin, target):
	redist = os.path.join(config.msvc, "redist", arch)
	msvcp = os.path.join(redist, "Microsoft.VC110.CRT\\msvcp110.dll")
	msvcr = os.path.join(redist, "Microsoft.VC110.CRT\\msvcr110.dll")
	
	copy_file(os.path.join(bin, "baretorrent.exe"), target)
	copy_file(os.path.join(bin, "baretorrent.dll"), target)
	copy_file(os.path.join(bin, "boost_python-vc110-mt-1_52.dll"), target)
	copy_file(msvcp, target)
	copy_file(msvcr, target)
	copy_file(os.path.join(bin, "python33.dll"), target)
	copy_file(os.path.join(bin, "wxmsw294u_vc_custom.dll"), target)
	copy_files(bin, "(.*?)(?<!_d)\\.pyd$", target)

def make_linux_archive(v, plat, arch, bin, target):
	copy_file(os.path.join(bin, "baretorrent"), target)
	copy_files(bin, "*.so", target)
	copy_file(os.path.join(bin, "libwx_gtk2u-2.9.so.4"), target)
	copy_dir(os.path.join(config.src, "res"), target)

def make_archive(v, plat, arch):
	target = "baretorrent-" + ".".join(v) + "-" + plat + "-" + arch
	archive = target + ".7z"
	bin = make_bin(plat, arch)

	print("")
	print("doing " + plat + " " + arch)
	print("copying to " + target)
	
	if os.path.exists(target):
		shutil.rmtree(target)
	
	os.mkdir(target)
	
	if sys.platform.startswith("win"):
		make_win_archive(v, plat, arch, bin, target)
	elif sys.platform.startswith("linux"):
		make_linux_archive(v, plat, arch, bin, target)
	
	copy_dir(os.path.join(config.gen_bin, "locales"), target)
	copy_file(os.path.join("..", "CHANGELOG"), target)
	copy_file(os.path.join("..", "COPYING"), target)
	copy_file(os.path.join(bin, "python33.zip"), target)

	print("archiving to " + archive)
	
	if not do_7z(target, archive):
		return False
	
	shutil.rmtree(target)
	
	return True

def do_one(v, plat, arch):
	if not make_archive(v, plat, arch):
		return False
	
	if sys.platform.startswith("win"):
		if not make_installer(v, plat, arch):
			return False
	
	return True

def do_all(v):
	for p in config.plats:
		for a in config.archs:
			if not do_one(v, p, a):
				return False
	
	return True
	

if __name__ == "__main__":
	v = get_version()
	if v == None:
		quit()
	
	if len(sys.argv) == 2:
		if sys.argv[1] == "all":
			do_all(v)
		else:
			usage()
	elif len(sys.argv) == 3:
		p = sys.argv[1]
		a = sys.argv[2]
		
		if p not in config.plats or a not in config.archs:
			usage()
		else:
			do_one(v, p, a)
	else:
		usage()

