import threading
import time
import inspect
import copy
import locale
import types

import wx
import wx.xrc
import baretorrent as bt
import libtorrent

from stats import graph

g_panel = None

def ui_file():
	return "ui.xrc"

def panel_name():
	return "panel"
	
def format_int(n):
	return locale.format("%d", n, True)
	
def format_float(n):
	return locale.format("%.1f", n, True)

class section:
	status_fields = []
	cache_fields = []
	
	last_status = None
	last_cache = None
	lock = threading.Lock()
	
	def status(self):
		return self.last_status
	
	def cache(self):
		return self.last_cache
	
	def set_status_fields(self, parent, v):
		self.set_fields_impl(parent, v, self.status_fields)
	
	def set_cache_fields(self, parent, v):
		self.set_fields_impl(parent, v, self.cache_fields)
	
	def set_fields_impl(self, parent, v, list):
		for i in v:
			n = i[0]
			w = wx.xrc.XRCCTRL(parent, i[0])
			a = i[1]
			s = i[2]
			
			if w is None:
				bt.log_error("widget " + n + " not found")
			else:
				list.append({
					"name": n, "value": None, "attr": a,
					"widget": w, "stringer": s})
		
	def update_data(self, ss, c):
		self.last_status = ss
		self.last_cache = c
		
		self.update_status(ss)
		self.update_cache(c)
		
		self.update_data_impl(ss, c)
	
	def update_data_impl(self, ss, c):
		pass
	
	def update_status(self, ss):
		fs = None
		with self.lock:
			fs = copy.copy(self.status_fields)
		
		self.process_fields(fs, ss)
	
		with self.lock:
			self.status_fields = fs

	def update_cache(self, c):
		fs = None
		with self.lock:
			fs = copy.copy(self.cache_fields)
		
		self.process_fields(fs, c)
	
		with self.lock:
			self.cache_fields = fs
			
	def process_fields(self, fs, data):
		for f in fs:
			if f["attr"] is None:
				continue
			
			attrs = f["attr"].split(".")
			
			o = data
			for a in attrs:
				if not hasattr(o, a):
					bt.log_error(f["attr"] + " not found")
					o = None
					break
				else:
					o = getattr(o, a)
			
			if o is not None:
				f["value"] = o
		
	def update_widgets(self):
		sfs = None
		cfs = None
		
		with self.lock:
			sfs = copy.copy(self.status_fields)
			cfs = copy.copy(self.cache_fields)
		
		for f in sfs:
			s = self.make_string(f)
			if f["widget"].GetLabel() != s:
				f["widget"].SetLabel(s)
		
		for f in cfs:
			s = self.make_string(f)
			if f["widget"].GetLabel() != s:
				f["widget"].SetLabel(s)

		self.update_widgets_impl()
		
	def update_widgets_impl(self):
		pass

	def make_string(self, f):
		bad = "???"
		
		if f["attr"] is None:
			if self.last_status is None or self.last_cache is None:
				return bad
				
			return f["stringer"](self.last_status, self.last_cache)
		
		if f["value"] is None:
			return bad
		
		if f["stringer"] is None:
			return format_int(f["value"])
		
		return f["stringer"](f["value"])

class transfers(section):
	graph = None
	
	def __init__(self, parent):
		gw = wx.xrc.XRCCTRL(parent, "graph")
		if not gw:
			bt.log_error("graph was not found")
		else:
			self.graph = graph.graph_meh(gw)
		
		self.set_status_fields(parent, [
			["transfers_ul_rate", "upload_rate", bt.make_loc_rate],
			["transfers_ul_pl_rate", "payload_upload_rate", bt.make_loc_rate],
			["transfers_ul", "total_upload", bt.make_loc_size],
			["transfers_ul_pl", "total_payload_upload", bt.make_loc_size],
			["transfers_dl_rate", "download_rate", bt.make_loc_rate],
			["transfers_dl_pl_rate", "payload_download_rate", bt.make_loc_rate],
			["transfers_dl", "total_download", bt.make_loc_size],
			["transfers_dl_pl", "total_payload_download", bt.make_loc_size],
			["transfers_redundant", "total_redundant_bytes", bt.make_loc_size],
			["transfers_failed", "total_failed_bytes", bt.make_loc_size]])
	
	def update_data_impl(self, ss, c):
		if not self.graph:
			return

	def update_widgets_impl(self):
		if not self.graph:
			return
		
	

class ip_overhead(section):
	def __init__(self, parent):
		self.set_status_fields(parent, [
			["ip_overhead_ul_rate", "ip_overhead_upload_rate", bt.make_loc_rate],
			["ip_overhead_dl_rate", "ip_overhead_download_rate", bt.make_loc_rate],
			["ip_overhead_total_ul", "total_ip_overhead_upload", bt.make_loc_size],
			["ip_overhead_total_dl", "total_ip_overhead_download", bt.make_loc_size]])
			
			
class dht(section):
	active_requests = None
	active_requests_label = None
	routing_table = None
	routing_table_label = None
	
	def __init__(self, parent):
		self.set_status_fields(parent, [
			["dht_ul_rate", "dht_upload_rate", bt.make_loc_rate],
			["dht_dl_rate", "dht_download_rate", bt.make_loc_rate],
			["dht_total_ul", "total_dht_upload", bt.make_loc_size],
			["dht_total_dl", "total_dht_download", bt.make_loc_size],
			["dht_nodes", None, self.make_nodes],
			["dht_torrents", "dht_torrents", None],
			["dht_global_nodes", "dht_global_nodes", None],
			["dht_total_alloc", "dht_total_allocations", None]])
	
		self.set_active_requests(parent)
		self.set_routing_table(parent)
	
	def set_active_requests(self, parent):
		list = wx.xrc.XRCCTRL(parent, "dht_active_requests")
		label = wx.xrc.XRCCTRL(parent, "dht_active_requests_label")
		
		self.active_requests = list
		self.active_requests_label = label
		
		list.AppendColumn("Type")
		list.AppendColumn("OR")
		list.AppendColumn("Timeouts")
		list.AppendColumn("Responses")
		list.AppendColumn("BF")
		list.AppendColumn("Nodes left")
		list.AppendColumn("Last sent")
		list.AppendColumn("First TO")
		
		list.SetMinSize(wx.Size(10, 100))
		
	def set_routing_table(self, parent):
		list = wx.xrc.XRCCTRL(parent, "dht_routing_table")
		label = wx.xrc.XRCCTRL(parent, "dht_routing_table_label")
		
		self.routing_table = list
		self.routing_table_label = label
		
		list.AppendColumn("Nodes")
		list.AppendColumn("Replacements")
		list.AppendColumn("Last active")

		list.SetMinSize(wx.Size(10, 100))

	def make_nodes(self, ss, c):
		return format_int(ss.dht_nodes) + " (" + format_int(ss.dht_node_cache) + ")"
	
	def update_widgets_impl(self):	
		if self.status() is not None:
			self.update_dht_active_requests(self.status())
			self.update_dht_routing_table(self.status())
	
	def update_dht_active_requests(self, ss):
		lb = "Active requests (" + format_int(len(ss.active_requests)) + ")"
		if (self.active_requests_label.GetLabel() != lb):
			self.active_requests_label.SetLabel(lb)
		
		list = self.active_requests
		
		list.DeleteAllItems()
		
		i = 0
		for r in ss.active_requests:
			list.InsertItem(i, r.type)
			list.SetItem(i, 1, format_int(r.outstanding_requests))
			list.SetItem(i, 2, format_int(r.timeouts))
			list.SetItem(i, 3, format_int(r.responses))
			list.SetItem(i, 4, format_int(r.branch_factor))
			list.SetItem(i, 5, format_int(r.nodes_left))
			list.SetItem(i, 6, format_int(r.last_sent))
			list.SetItem(i, 7, format_int(r.first_timeout))
			++i
	
	def update_dht_routing_table(self, ss):
		lb = "Routing table (" + format_int(len(ss.dht_routing_table)) + ")"
		if (self.routing_table_label.GetLabel() != lb):
			self.routing_table_label.SetLabel(lb)
		
		list = self.routing_table
		
		list.DeleteAllItems()
		
		i = 0
		for b in ss.dht_routing_table:
			list.InsertItem(i, format_int(b.num_nodes))
			list.SetItem(i, 1, format_int(b.num_replacements))
			list.SetItem(i, 2, format_int(b.last_active))
			++i
	
class trackers(section):
	def __init__(self, parent):
		self.set_status_fields(parent, [
			["trackers_ul_rate", "tracker_upload_rate", bt.make_loc_rate],
			["trackers_dl_rate", "tracker_download_rate", bt.make_loc_rate],
			["trackers_total_ul", "total_tracker_upload", bt.make_loc_size],
			["trackers_total_dl", "total_tracker_download", bt.make_loc_size]])
	
class utp(section):
	def __init__(self, parent):
		self.set_status_fields(parent, [
			["utp_idle_connected", None, self.make_idle_connected],
			["utp_syn_sent", None, self.make_syn_sent],
			["utp_fin_sent", None, self.make_fin_sent],
			["utp_close_wait", None, self.make_close_wait]])
	
	def make_idle_connected(self, ss, c):
		return \
			format_int(ss.utp_stats["num_connected"]) + " (" + \
			format_int(ss.utp_stats["num_idle"]) + ")"
	
	def make_syn_sent(self, ss, c):
		return format_int(ss.utp_stats["num_syn_sent"])
	
	def make_fin_sent(self, ss, c):
		return format_int(ss.utp_stats["num_fin_sent"])
	
	def make_close_wait(self, ss, c):
		return format_int(ss.utp_stats["num_close_wait"])
	
class peers(section):
	def __init__(self, parent):
		self.set_status_fields(parent, [
			["peers_total_unchoked", None, self.make_total_unchoked],
			["peers_upload_slots", "allowed_upload_slots", None],
			["peers_ul_queue", "up_bandwidth_queue", None],
			["peers_dl_queue", "down_bandwidth_queue", None],
			
			["peers_opt_unchoke", "optimistic_unchoke_counter", bt.make_loc_duration_abbr],
			["peers_unchoke", "unchoke_counter", bt.make_loc_duration_abbr],
			["peers_ul_queue_size", "up_bandwidth_bytes_queue", bt.make_loc_size],
			["peers_dl_queue_size", "down_bandwidth_bytes_queue", bt.make_loc_size],
			
			["peers_disk_write_queue", "disk_write_queue", None],
			["peers_disk_read_queue", "disk_read_queue", None]])
	
	def make_total_unchoked(self, ss, c):
		return format_int(ss.num_peers) + " (" + format_int(ss.num_unchoked) + ")"

class cache(section):
	def __init__(self, parent):
		self.set_cache_fields(parent, [
			["cache_write_hit", None, self.make_write_hit],
			["cache_blocks_written", "blocks_written", None],
			["cache_writes", "writes", None],
			
			["cache_read_hit", "blocks_read_hit", None],
			["cache_blocks_read", "blocks_read", None],
			["cache_reads", "reads", None],
			
			["cache_size", "cache_size", self.make_blocks],
			["cache_size_read", "read_cache_size", self.make_blocks],
			["cache_size_write", None, self.make_size_write],
			["cache_used_buffers", "total_used_buffers", self.make_blocks],
			
			["cache_queue_time", "average_queue_time", self.make_us],
			["cache_read_time", "average_read_time", self.make_us],
			["cache_write_time", "average_write_time", self.make_us],
			["cache_hash_time", "average_hash_time", self.make_us],
			["cache_job_time", "average_job_time", self.make_us],
			["cache_sort_time", "average_sort_time", self.make_us],
			["cache_job_queue", "job_queue_length", None]])
	
	def make_write_hit(self, ss, c):
		f = 0.0
		
		if c.blocks_written > 0:
			f = float(c.blocks_written - c.writes) / c.blocks_written
		
		return format_float(round(f * 100, 1)) + "%"

	def make_blocks(self, n):
		return format_int(n) + " bl."
		
	def make_size_write(self, ss, c):
		return self.make_blocks(c.cache_size - c.read_cache_size)
	
	def make_us(self, n):
		return format_float(float(n) / 1000) + " ms"
		
class stats_panel(wx.Panel):
	thread = None
	res = None
	panel = None
	timer = None
	must_stop = False
	lock = threading.Lock()
	sections = []
	
	
	def __init__(self, parent):
		global g_panel
		
		wx.Panel.__init__(self, parent)
		g_panel = self
		
		bt.log("creating stats panel")
		
		if not self.load_res():
			return
		
		self.make_sections()
		
		bt.log("starting timer")
		self.timer = wx.Timer(self)
		self.Bind(wx.EVT_TIMER, self.on_timer, self.timer)
		self.timer.Start(1000)
		
		bt.log("starting thread")
		self.thread = threading.Thread(target=self.thread_fun)
		self.thread.start()

	def stop(self):
		bt.log("stopping")
		self.must_stop = True
		
		if self.timer:
			bt.log("stopping timer")
			self.timer.Stop()
		else:
			bt.log("timer is not running")
	
	def join(self):
		if self.thread:
			self.thread.join()
	
	def tab_name(self):
		return "Statistics"
	
	def thread_fun(self):
		try:
			bt.log("starting thread")
			
			session = bt.manager().session()
			
			while not self.must_stop:
				ss = session.status()				
				c = session.get_cache_status()
				
				self.update_data(ss, c)
				
				time.sleep(1)
		except:
			bt.log_error("exception in update thread:", sys.exc_info()[0])
	
	def on_timer(self, e):
		try:
			self.update_widgets()
		
		except:
			bt.log_error("exception in timer:", sys.exc_info()[0])
		
	def load_res(self):
		bt.log("loading " + ui_file())
		no_log = wx.LogNull()
		self.res = wx.xrc.XmlResource()
		
		if not self.res.LoadFile(ui_file()):
			bt.log_error("can't load xrc file")
			return False
		
		bt.log("loading panel " + panel_name())
		self.panel = self.res.LoadPanel(self, panel_name())
		
		if not self.panel:
			bt.log_error("can't find panel " + panel_name())
			return False
				
		del no_log
		
		scroll = wx.xrc.XRCCTRL(self.panel, "scroll")
		if not scroll:
			bt.log("can't find scroll")
			return False
		
		self.SetSizer(wx.BoxSizer(wx.VERTICAL))
		self.GetSizer().Add(self.panel, 1, wx.EXPAND)
			
		return True
	
	def make_sections(self):
		self.sections.append(transfers(self.panel))
		self.sections.append(peers(self.panel))
		self.sections.append(ip_overhead(self.panel))
		self.sections.append(dht(self.panel))
		self.sections.append(trackers(self.panel))
		self.sections.append(utp(self.panel))
		self.sections.append(cache(self.panel))
	
	def update_data(self, ss, c):
		with self.lock:
			for s in self.sections:
				s.update_data(ss, c)
	
	def update_widgets(self):
		with self.lock:
			for s in self.sections:
				s.update_widgets()
		

def init():
	try:
		locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
	except locale.Error:
		locale.setlocale(locale.LC_ALL, "")
	
	bt.add_panel("stats_panel")

def cleanup():
	if g_panel:
		g_panel.stop()
		g_panel.join()
	
	bt.remove_panel("stats_panel")
