<?php
if (!defined("bt_include_is_in_script"))
{
	header("Location: ../index.html");
	exit;
}

$g_page_start_time = gettimeofday(true);

require_once(__DIR__ . "/libs/SqlFormatter.php");

if (!defined("NO_DATABASE"))
	require_once(__DIR__ . "/config.php");

require_once(__DIR__ . "/forum/utility.php");
	
date_default_timezone_set("UTC");

function format_sql($q)
{
	return "<blockquote>" .
		SqlFormatter::format($q) .
		"</blockquote>";
}

$g_debug_queries = null;//array();

function debug_add_query($q)
{
	global $g_debug_queries;
	if ($g_debug_queries === null)
		return;
	
	array_push($g_debug_queries,
		array("q" => $q, "bt" => get_backtrace()));
}

function debug_dump_queries()
{
	global $g_debug_queries;
	if ($g_debug_queries === null)
		return;
	
	echo
		"<div style=\"white-space: pre-wrap; font-family: monospace;\">" .
		"\n\n" . count($g_debug_queries) . " queries\n\n";
	
	foreach ($g_debug_queries as $q)
		echo format_sql($q["q"]) . "\n" . $q["bt"] . "\n\n";
	
	echo "</div>";
}


class db_exception extends \Exception
{
	protected $message = "";
	protected $query = "";
	
	public function __construct($message, $q)
	{
		$this->message = $message;
		$this->query = $q;
	}
	
	public function html()
	{
		$s = $this->message;
		
		if ($this->query != null)
			$s .= format_sql($this->query);
		
		return $s;
	}
}

function handle_db_error($message, $q = null)
{
	throw new db_exception($message, $q);
}

class pdo_statement_wrapper extends \PDOStatement
{
	private $pdo = null;
	private $used_impl = false;
	private $next_bind = 1;
	private $debug_bind = array();
	
	protected function __construct($pdo)
	{
		$this->pdo = $pdo;
	}
	
	public function rowCount()
	{
		die(
			"do not use rowCount() as it may not be implemented
			on all dbms");
	}

	public function reset()
	{
		$this->closeCursor();
		$this->next_bind = 1;
		$this->debug_bind = array();
	}
	
	public function bind($value)
	{
		if ($this->used_impl)
			bad_mix();
		
		if (parent::bindValue($this->next_bind, $value) === FALSE)
			$this->check_error($this->queryString);
	
		array_push($this->debug_bind, $value);
		++$this->next_bind;
		
		return TRUE;
	}
	
	public function bindParam(
		$p, &$v, $t = \PDO::PARAM_STR, $len = 0, $d = "")
	{
		if ($this->next_bind > 1)
			bad_mix();
			
		$r = parent::bindParam($p, $v, $t, $len, $d);

		array_push($this->debug_bind, $v);
		$this->used_impl = true;
		
		return $r;
	}
	
	public function bindValue($p, $v, $t = \PDO::PARAM_STR)
	{
		if ($this->next_bind > 1)
			bad_mix();
			
		$r = parent::bindValue($p, $v, $t);

		array_push($this->debug_bind, $v);
		$this->used_impl = true;
		
		return $r;
	}
	
	public function fetch($s = null, $c = \PDO::FETCH_ORI_NEXT, $o = 0)
	{
		if ($s == null)
			$s = \PDO::FETCH_ASSOC;
		
		return parent::fetch($s, $c, $o);
	}
		
	public function fetchAll($s = null, $a = null, $ctor = null)
	{
		if ($s == null)
			$s = \PDO::FETCH_ASSOC;
		
		// todo: I'm getting a
		//   SQLSTATE[HY000]: General error:
		//      Extraneous additional parameters 
		// when I pass $a and $ctor
		
		return parent::fetchAll($s);
	}

	public function execute($params = null)
	{
		if (parent::execute($params) === FALSE)
			$this->check_error($this->queryString);
		return TRUE;
	}
	
	private function bad_mix()
	{
		die(
			"bind() and pdo's native features (bindParam() and
			bindValue()) cannot be mixed on the same statement");
	}
	
	public function check_error($q = null)
	{
		// todo: message encoding

		$i = $this->errorInfo();
			
		$s = "$i[2] (SQLSTATE:$i[0] code:$i[1])";
		handle_db_error($s, $q);
	}
	
	public function dump()
	{
		$q = $this->queryString;
		
		$qb = $q;
		foreach ($this->debug_bind as $v)
		{
			$i = strpos($qb, "?");
			if ($i === FALSE)
				break;
				
			$qb = substr($qb, 0, $i) . "'$v'" . substr($qb, $i + 1);
		}
		
		echo "
			<div class=\"bt_db-message\">original query:" .
			format_sql($this->queryString) .
			"query with params (not necessarily how the server sees it):" .
			format_sql($qb) .
			"params:<ul style=\"padding: 10px;\">";

		foreach ($this->debug_bind as $v)
			echo "<li style=\"list-style: disc inside;\">$v";
			
		echo "</ul></div>";
	}
}

abstract class database
{
	private $pdo = null;
	private $last_result = null;
	
	private $dsn = null;
	private $username = null;
	private $password = null;
	private $message_encoding = null;
	
	function __construct($dsn, $username, $password, $message_encoding)
	{
		$this->dsn = $dsn;
		$this->username = $username;
		$this->password = $password;
		$this->message_encoding = $message_encoding;
		
		$this->connect();
	}
	
	static public function test(
		$dsn, $username, $password, $message_encoding)
	{
		try
		{
			$pdo = new PDO($dsn, $username, $password);
			return "";
		}
		catch(PDOException $e)
		{
			return iconv($message_encoding, "utf-8", $e->getMessage());
		}
	}
	
	private function connect()
	{
		if ($this->pdo != null)
			return;

		try
		{
			$this->pdo = new \PDO(
				$this->dsn, $this->username, $this->password);
				
			$this->pdo->setAttribute(
				\PDO::ATTR_STATEMENT_CLASS,
				array("pdo_statement_wrapper", array($this)));
				
			$this->on_connected();
		}
		catch(\PDOException $e)
		{
			handle_db_error(iconv(
				$this->message_encoding, "utf-8", $e->getMessage()),
				null, $this);
		}
	}
		
	function execute($q)
	{
		debug_add_query($q);
		
		$this->connect();
		$this->reset_result();
	
		$r = $this->pdo->exec($q);
		$this->check_error($q);

		return $r;
	}
		
	public function query($q)
	{
		debug_add_query($q);
		
		$this->connect();
		$this->reset_result();

		$this->last_result = $this->pdo->query($q);
		$this->check_error($q);

		return $this->last_result;
	}
	
	public function query_value($q, $name)
	{
		$res = $this->query($q);
		$row = $res->fetch();
		
		if ($row === FALSE)
		{
			handle_db_error(
				"query_value: nothing to fetch", $q, $this);
		}
			
		if (!isset($row[$name]))
		{
			handle_db_error(
				"query_value: $name column not found", null, $this);
		}
	
		return $row[$name];
	}
	
	public function last_insert_id()
	{
		$this->connect();
		return $this->pdo->lastInsertId();
	}
	
	public function prepare($q)
	{
		debug_add_query($q);
		
		$this->connect();
		$this->reset_result();

		$s = $this->pdo->prepare($q);
		$this->check_error($q);

		return $s;
	}
	
	public function begin()
	{
		$this->connect();
		$this->pdo->beginTransaction();
	}
	
	public function commit()
	{
		$this->connect();
		$this->pdo->commit();
	}
	
	public function rollback()
	{
		$this->connect();
		$this->pdo->rollback();
	}
	
	public function timestamp($dt)
	{
		if ($dt === null)
			return 0;
		
		return strtotime($dt);
	}
	
	public function date_time($ts)
	{
		return strftime("%Y-%m-%d %H:%M:%S", $ts);
	}
	
	public function now()
	{
		return gmdate("Y-m-d H:i:s");
	}
	
	public function never()
	{
		return "0000-00-00 00:00:00";
	}
	
	public function dump()
	{
		$s = "dsn: $this->dsn\nlogin: $this->username";
		if ($this->password === "")
			$s .= " (no password)";
		else
			$s .= ":$this->password";
		
		return $s;
	}
	
	private function check_error($q = null)
	{
		if ($this->pdo->errorCode() != 0)
		{
			$i = $this->pdo->errorInfo();
			
			$s = "$i[2] (SQLSTATE:$i[0] code:$i[1])";
			$s = iconv($this->message_encoding, "utf-8", $s);
			
			handle_db_error($s, $q, $this);
		}
	}
	
	private function reset_result()
	{
		if ($this->last_result != null)
		{
			$this->last_result->closeCursor();
			$this->last_result = null;
		}
	}
}

class mysql_database extends database
{
	function __construct($host, $db, $username, $password)
	{
		$dsn = "mysql:host=$host;dbname=$db";
		parent::__construct($dsn, $username, $password, "utf-8");
	}
	
	static public function test($host, $name, $username, $password)
	{
		$dsn = "mysql:host=$host;dbname=$name";
		return database::test($dsn, $username, $password, "utf-8");
	}
	
	public function make_date($ts)
	{
		return date("Y-m-d H:i:s", $ts);
	}
	
	protected function on_connected()
	{
		$this->execute("set names 'utf8'");
	}
}


$g_database = null;

function test_connection($host, $name, $username, $password)
{
	return mysql_database::test($host, $name, $username, $password);
}

function db()
{
	global $g_database;
	
	if ($g_database == null)
	{
		$g_database = new mysql_database(
			DATABASE_HOST, DATABASE_NAME,
			DATABASE_USERNAME, DATABASE_PASSWORD);
	}
	
	return $g_database;
}

?>
