<?php
require_once(__DIR__ . "/db.php");
require_once(__DIR__ . "/libs/Markdown.php");
require_once(__DIR__ . "/queries.inc.php");

use \Michelf\Markdown;

class bad_guid extends \Exception {}

define("ext_permission_none" , 0);
define("ext_permission_edit", 1);

class extension_already_exists extends \Exception
{
	private $e = null;
	
	public function __construct($e)
	{
		$this->e = $e;
	}
	
	public function extension()
	{
		return $this->e;
	}
}

class guid_already_exists extends \Exception
{
	private $e = null;
	
	public function __construct($e)
	{
		$this->e = $e;
	}
	
	public function extension()
	{
		return $this->e;
	}
}

function make_ext_link($id, $page = "", $rev = "")
{
	$s = "ext.html?id=$id";
	
	if ($page != "")
		$s .= "&page=$page";
	
	if ($rev != "")
		$s .= "&rev=$rev";
	
	return $s;
}

function make_history_link($id, $page, $rev = "")
{
	$s = "history.html?id=$id&page=$page";
	
	if ($rev != "")
		$s .= "&rev=$rev";
	
	return $s;
}

function make_diff_link($id, $page, $rev)
{
	return "history.html?id=$id&page=$page&diff=$rev";
}

function make_undo_link($id, $page, $rev)
{
	return "history.html?id=$id&page=$page&undo=$rev";
}

class version
{
	private $major = null;
	private $minor = null;
	private $maintenance = null;
	
	public static function parse($s)
	{
		$v = new version;
		
		$vs = explode(".", $s);
		
		if (count($vs) >= 1)
			$v->major = (int)$vs[0];
		if (count($vs) >= 2)
			$v->minor = (int)$vs[1];
		if (count($vs) >= 3)
			$v->maintenance = (int)$vs[2];
		
		return $v;
	}
	
	public function as_string()
	{
		if ($this->major == null)
			return "1.0";
			
		$s = "" . $this->major;
		
		if ($this->minor != null)
		{
			$s .= ".$this->minor";
		
			if ($this->maintenance != null)
				$s .= ".$this->maintenance";
		}
		else
		{
			$s .= ".0";
		}
		
		return $s;
	}
}

class extension_version
{
	private $id = 0;
	private $extension_id = 0;
	private $version = null;
	private $requires = "";
	private $release_date = 0;

	public static function fields(
		$t = "extension_versions", $p = "ext_ver_")
	{
		return
			"$t.id {$p}id, " .
			"$t.extension {$p}extension, " .
			"$t.version {$p}version, " .
			"$t.file {$p}file, " .
			"$t.requires {$p}requires, " .
			"$t.release_date {$p}release_date ";
	}
	
	public static function from_db($r, $p = "ext_ver_")
	{
		$e = new extension_version;
		
		$e->id = (int)$r["{$p}id"];
		$e->extension_id = (int)$r["{$p}extension"];
		$e->version = version::parse($r["{$p}version"]);
		$e->requires = $r["{$p}requires"];
		$e->release_date = db()->timestamp($r["{$p}release_date"]);
		
		return $e;
	}
	
	public function as_string()
	{
		return $this->version->as_string();
	}
	
	public function release_date()
	{
		return $this->release_date;
	}
}

class extension
{
	private $id = 0;
	private $guid = "";
	private $name = "";
	private $author = "";
	private $description = "";
	private $email = "";
	private $website = "";
	private $icon = null;
	private $version = null;
	
	public static function fields($t = "extensions", $p = "ext_")
	{
		return
			"$t.id {$p}id, " .
			"$t.guid {$p}guid, " .
			"$t.name {$p}name, " .
			"$t.author {$p}author, " .
			"$t.description {$p}description, " .
			"$t.email {$p}email, " .
			"$t.website {$p}website ";
	}
	
	public static function from_db($r, $p = "ext_")
	{
		$e = new extension;
		
		$e->id = (int)$r["{$p}id"];
		$e->guid = $r["{$p}guid"];
		$e->name = $r["{$p}name"];
		$e->author = $r["{$p}author"];
		$e->description = $r["{$p}description"];
		$e->email = $r["{$p}email"];
		$e->website = $r["{$p}website"];
		$e->version = extension_version::from_db($r);
		
		return $e;
	}
	
	
	public function id()
	{
		return $this->id;
	}
	
	public function guid()
	{
		return $this->guid;
	}
	
	public function name()
	{
		return $this->name;
	}
	
	public function description()
	{
		return $this->description;
	}
	
	public function icon()
	{
		if ($this->icon === null)
		{
			$ps = query::extension_medium_icon($this->id);
			
			$row = $ps->fetch();
			if ($row !== null)
				$this->icon = $row["medium_icon"];
		}
		
		return $this->icon;
	}
	
	public function has_icon()
	{
		return query::extension_has_medium_icon($this->id);
	}
	
	public function version()
	{
		return $this->version;
	}
	
	public function wiki_page()
	{
		return $this->name . "/";
	}
}

function get_extensions()
{
	$ps = query::get_extensions();
	
	$es = array();
	foreach ($ps as $row)
		array_push($es, extension::from_db($row));
	
	return $es;
}

function get_extension($id)
{
	$ps = query::get_extension($id);
	
	$row = $ps->fetch();
	if ($row === FALSE)
		return null;
	
	return extension::from_db($row);
}

function find_extension_by_guid($guid)
{
	$ps = query::find_extension_by_guid($guid);

	$row = $ps->fetch();
	if ($row === FALSE)
		return null;
	
	return extension::from_db($row);
}

function find_extension_by_name($name)
{
	$ps = query::find_extension_by_name($name);

	$row = $ps->fetch();
	if ($row === FALSE)
		return null;
	
	return extension::from_db($row);
}

function find_extension_version($id, $v)
{
	$ps = query::find_extension_version($id, $v);
	
	$row = $ps->fetch();
	if ($row === FALSE)
		return null;
	
	return "";
}

function delete_extension($id)
{
	query::delete_extension($id);
}

function parse_guid($s)
{
	$m = null;
	if (preg_match('/^\s*guid\s*=\s*(.*)$/m', $s, $m) != 1)
		throw bad_guid();
	
	return $m[1];
}

function add_extension($name, $filename)
{
	$btx = get_7z_file($filename, "btx.txt");
	$guid = parse_guid($btx);
	
	$e = find_extension_by_guid($guid);
	if ($e)
		throw new guid_already_exists($e);
	
	$e = find_extension_by_name($name);
	if ($e)
		throw new extension_already_exists($e);
	
	$small_icon = get_optional_7z_file($filename, "icon_16.png");
	$medium_icon = get_optional_7z_file($filename, "icon_32.png");
	$large_icon = get_optional_7z_file($filename, "icon.png");
	
	$id = query::add_extension($guid, $name);
	
	query::add_extension_version(
		$id, 1, file_get_contents($filename), "");
	
	query::add_extension_icons(
		$id, $small_icon, $medium_icon, $large_icon);
	
	return $id;
}

function extension_permissions($user_id, $ext_id)
{
	$ps = query::extension_permissions($user_id, $ext_id);
	
	$row = $ps->fetch();
	if ($row === FALSE)
		return ext_permission_none;
	
	return (int)$row["permissions"];
}

function output_ext($e)
{
	$dl = "https://launchpad.net/baretorrent/trunk/httpd-0.1/+download/httpd-0.1.btx";
	$m = "ext.html?id=" . $e->id();

	$rd = make_simple_date($e->version()->release_date());
	if ($rd != "")
	{
		$rd = sprintf(loc("ext_release"), $rd);
		$rd = "<span class=\"ext-release-date\">$rd</span>";
	}
	
	echo "
		<a href=\"$m\" class=\"ext-entry\">
		<div class=\"\">
		<img
			src=\"api.php?action=icon&size=medium&id=" . $e->id() . "\"
			alt=\"icon of " . $e->name() . "\"
			class=\"ext-icon\" />
				<span class=\"ext-name\">" . $e->name() . "</span>
				<span class=\"ext-version\">" . $e->version()->as_string() . "</span>
				$rd
			</div>
			<div class=\"ext-description\">" . $e->description() . "</div>
		</a>
		";
}

?>