// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "app.h"
#include "main_window.h"
#include "torrent.h"
#include "manager.h"
#include "locale.h"
#include "utility.h"

namespace baretorrent
{

// time between libtorrent ticks
const boost::posix_time::milliseconds app_timer_delay(300);

// time between saving options
const boost::posix_time::milliseconds options_timer_delay(2000);

// time before checking for updates on first run
const boost::posix_time::minutes first_run_update_delay(5);


app::app() :
	window_(nullptr), update_delay_(0), update_stop_(false),
	update_disabled_(false), quitting_(false), manager_(nullptr)
{
}

app::~app()
{
	if (update_thread_.joinable())
	{
		update_stop_ = true;
		update_thread_.interrupt();
		update_thread_.join();
	}
}

void app::set(baretorrent::manager& m)
{
	bt_assert(manager_ == nullptr);
	manager_ = &m;
}

void app::add_torrents(const std::vector<utf8_string>& v)
{
	if (window_)
		window_->add_torrents(v);
	else
		new_ = v;
}

main_window& app::window()
{
	bt_assert(window_);
	return *window_;
}

void app::create_window()
{
	// the window is created on the heap here and put in an auto_ptr
	// which is released at the end because wxWidgets deletes it

	std::auto_ptr<main_window> w(new main_window(*this));
	window_ = w.get();

	// tells wxWidgets that this is the main window
	SetTopWindow(window_);

	w.release();

	window_->add_torrents(new_);
	new_.clear();

	manager_->start();
}

void app::show_window()
{
	window_->show_window();
}

bool app::OnInit()
{
	bt_assert(manager_);

	try
	{
		// makes sure the wxWidgets dll matches the settings used when
		// compiling baretorrent
		CheckBuildOptions(WX_BUILD_OPTIONS_SIGNATURE, "baretorrent");

		// all images from bittorrent are pngs, but some platforms
		// might need more (like tiff on apple for the system bitmaps)
		wxImage::AddHandler(new wxPNGHandler);
		platform::add_image_handlers();
	}
	catch(std::exception& e)
	{
		// this probably won't give a localized message to the user,
		// but it's better than nothing

		const utf8_string t(loc("dialog.error_title"));
		const utf8_string m((locf("dialog.error") % e.what()).str());

		log_error() << e.what() << "\n";
		wxMessageBox(to_wxstring(m).c_str(), to_wxstring(t).c_str());
	}
	
	// starting timers
	timer_.start(boost::bind(
		&app::on_timer, this), app_timer_delay);

	options_timer_.start(boost::bind(
		&app::on_options_timer, this), options_timer_delay);

	return true;
}

void app::check_associations()
{
	platform::assoc a = platform::associated();

	if (!a.okay)
		fire(missing_association, a.name);
}

void app::reload_settings()
{
	if (opt<bool>("bt.system.firewall_exception"))
		platform::add_firewall_exception();
	
	if (opt<bool>("bt.update") && !update_disabled_)
	{
		const int ud = opt<int>("bt.update_delay");

		if (update_thread_.joinable() && update_delay_ != ud)
		{
			// update delay changed and it was enabled before; kill the
			// thread and restart it
			log_v()
				<< "stopping update thread because delay changed from "
				<< update_delay_ << " to " << ud << "\n";

			update_stop_ = true;
			update_thread_.interrupt();
			update_thread_.join();
		}

		if (!update_thread_.joinable())
		{
			update_delay_ = ud;
			update_stop_ = false;
			update_thread_ = boost::thread(
				boost::bind(&app::check_update_thread, this));
		}
	}
	else if (!update_disabled_)
	{
		if (update_thread_.joinable())
		{
			log_v()
				<< "stopping update thread because updates were "
				<< "disabled\n";

			update_stop_ = true;
			update_thread_.interrupt();
		}
	}

	if (opt<bool>("bt.assoc"))
		check_associations();
}

void app::disable_updates()
{
	update_disabled_ = true;
	update_stop_ = true;
	update_thread_.interrupt();
}

void app::check_update_thread()
{
	bool f = true;

	try
	{
		if (opt<bool>("bt.first_run"))
		{
			log()
				<< "first run: waiting " << first_run_update_delay
				<< " before checking for updates\n";

			boost::this_thread::sleep(first_run_update_delay);
		}

		for (;;)
		{
			if (f && opt<bool>("bt.update_on_startup"))
			{
				if (check_update())
					return;
			}

			boost::this_thread::sleep(boost::posix_time::seconds(
				opt<int>("bt.update_delay")));

			if (check_update())
				return;

			f = false;
		}
	}
	catch(boost::thread_interrupted&)
	{
	}
}

bool app::update_cb(double)
{
	return !update_stop_;
}

bool app::check_update()
{
	log() << "checking for update\n";

	fire(checking_update);

	const utf8_string url = opt("bt.update_url");
	const byte_array ba = get_remote_file(
		url, boost::bind(&app::update_cb, this, _1));

	if (ba.empty())
	{
		log_error() << "checking for update failed\n";

		fire(update_done, false);
		return false;
	}

	std::string v(ba.data(), ba.size());
	boost::trim(v);

	const baretorrent_version nv = baretorrent_version::parse(v);

	if (!new_version(nv))
	{
		log()
			<< "version '" << nv << "' is the same as or older than '"
			<< version().string << "'\n";

		fire(update_done, true);
		return false;
	}

	log()
		<< "new version: '" << nv << "' is newer than "
		<< "'" << version().string << "'\n";

	update_available(nv);
	fire(update_done, true);

	return true;
}

baretorrent::manager* app::manager_ptr()
{
	return manager_;
}

manager& app::manager()
{
	bt_assert(manager_);
	return *manager_;
}

void app::quit()
{
	if (quitting_)
		return;

	log() << "exiting...\n";

	opt("bt.first_run", false);
	save_options();

	if (update_thread_.joinable())
	{
		update_stop_ = true;
		update_thread_.interrupt();
	}

	// this is picked up in the timer which will wait until the
	// manager is done
	manager_->quit();
	quitting_ = true;
}

bool app::quitting() const
{
	return quitting_;
}

bool app::on_timer()
{
	if (quitting_)
	{
		if (!manager_ || manager_->done())
		{
			// if the manager is done, close the window, which was
			// only hidden when closed
			if (window_)
				window_->Close(true);
			
			ExitMainLoop();
			return false;
		}
	}

	for (;;)
	{
		// tick() will return false when there's no more alerts to
		// process from libtorrent
		if (!manager_ || !manager_->tick())
			break;
	}

	return true;
}

bool app::on_options_timer()
{
	if (quitting_)
		return false;

	save_options();
	return true;
}

#if BARETORRENT_APPLE
void app::MacOpenFiles(const wxArrayString& files)
{
	if (!manager_)
	{
		log_error() << "mac: received files before manager is up\n";
		return;
	}

	log() << "mac: received " << files.size() << " files; adding\n";

	foreach (const wxString& s, files)
		manager_->add_torrent(from_wxstring(s));
}

void app::MacOpenURL(const wxString& url)
{
	if (!manager_)
	{
		log_error() << "mac: received url before manager is up\n";
		return;
	}

	const utf8_string u = from_wxstring(url);
	log() << "mac: received url '" << u << "'\n";

	manager_->add_torrent(u);
}
#endif


app& get_app()
{
	return *dynamic_cast<app*>(wxTheApp);
}

void quit()
{
	app* a = dynamic_cast<app*>(wxTheApp);
	if (a)
		a->quit();
}

}	// namespace baretorrent
