// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#define BARETORRENT_SKIP_PYTHON
#include "pch.h"

#if BARETORRENT_APPLE

#include "platform.h"
#include "utility.h"
#include "widgets.h"

// gcc on apple defines a macro "check" that fails with boost
#define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0

#import <AppKit/NSImage.h>
#import <Foundation/NSData.h>
#import <Foundation/Foundation.h>
#import <Foundation/NSPathUtilities.h>
#include <ApplicationServices/ApplicationServices.h>
#include <mach-o/dyld.h>
#include <iconv.h>

namespace baretorrent
{
namespace platform
{

// sort indicator stuff comes from
// https://groups.google.com/forum/?fromgroups=
// 				#!topic/wx-users/Upz7xB-8lsU

static const char* sort_asc_xpm[] = {
	"16 16 2 1",
	". c #ACA899",
	"  c #FF0000",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"        .       ",
	"       ...      ",
	"      .....     ",
	"     .......    ",
	"    .........   ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                "};

static const char* sort_desc_xpm[] = {
	"16 16 2 1",
	". c #ACA899",
	"  c #FF0000",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"    .........   ",
	"     .......    ",
	"      .....     ",
	"       ...      ",
	"        .       ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                "};

typedef std::map<images::image, utf8_string> image_map;


platform::platform()
{
	// noop
}

platform::~platform()
{
}

utf8_string system_message(OSStatus s)
{
	NSError* r = [NSError errorWithDomain:NSOSStatusErrorDomain
		code:s userInfo:nil];

	NSString* d = [r localizedDescription];

	return utf8_string([d UTF8String]);
}

long toolbar_style()
{
	return wxBORDER_NONE | wxTB_FLAT | wxTB_TEXT;
}

void init_list_control(wxListCtrl* w)
{
	std::auto_ptr<wxImageList> il(new wxImageList);
	il->Add(wxBitmap(sort_asc_xpm), wxColor(255, 0, 0));
	il->Add(wxBitmap(sort_desc_xpm), wxColor(255, 0, 0));

	w->SetImageList(il.release(), wxIMAGE_LIST_SMALL);
}

void destroy_list_control(wxListCtrl* w)
{
	delete w->GetImageList(wxIMAGE_LIST_SMALL);
}

void set_list_control_style(wxListCtrl*)
{
	// noop
}

void set_tree_control_style(wxTreeCtrl*)
{
	// noop
}

void set_list_sort_column(wxListCtrl* w, std::size_t c, int d)
{
	wxListItem item;
	item.SetMask(wxLIST_MASK_IMAGE);

	for (int i=0; i<w->GetColumnCount(); ++i)
	{
		if (i == static_cast<int>(c))
		{
			if (d > 0)
				item.SetImage(0);
			else
				item.SetImage(1);
		}
		else
		{
			item.SetImage(-1);
		}

		w->SetColumn(i, item);
	}
}

void open_directory(const utf8_path& p)
{
	const utf8_string s(
		"open \"" + p.as_utf8_string().std_string() + "\"");
	system(s.c_str());
}

void open_file(const utf8_path& p)
{
	const utf8_string s(
		"open \"" + p.as_utf8_string().std_string() + "\"");
	system(s.c_str());
}

void add_to_path(const utf8_path& p)
{
	// todo
}

// this is called by download_directory() when the known download
// directory isn't found; it will look for non-localized, hardcoded
// directory names (such as "Downloads") before bailing and returning
// root (which is probably the home directory)
//
// this is somewhat crap, but it might be useful
//
utf8_path try_download_dirs(const utf8_path& root)
{
	const utf8_string downloads("Downloads");

	if ((root / downloads).exists())
	{
		log()
			<< "there's a " << downloads << " directory in "
			<< root << ", taking it\n";

		return root / downloads;
	}
	
	return root;
}

utf8_path home_directory()
{
	// from what I can understand, this cannot fail (and if it does,
	// the user has other problems than an unhappy baretorrent)
	NSString* s = NSHomeDirectory();
	return utf8_path(utf8_string([s UTF8String]));
}

utf8_path download_directory()
{
	// again, from what I understand, this shouldn't fail (although
	// the directory might not exist, but this is taken care of
	// elsewhere)
	NSArray* paths = NSSearchPathForDirectoriesInDomains(
		NSDownloadsDirectory, NSUserDomainMask, true);
	
	if ([paths count] > 0)
	{
		NSString* s = [paths objectAtIndex:0];
		return utf8_path(utf8_string([s UTF8String]));
	}

	// if it does fail, this will most likely use the home directory

	log_error()
		<< "search for download directory path failed, "
		<< "trying home\n";

	return try_download_dirs(home_directory());
}

utf8_path apple_application_support()
{
	NSArray* paths = NSSearchPathForDirectoriesInDomains(
		NSApplicationSupportDirectory, NSUserDomainMask, true);

	if ([paths count] > 0)
	{
		NSString* s = [paths objectAtIndex:0];
		return utf8_path(utf8_string([s UTF8String]));
	}

	return utf8_path();
}

utf8_path conf_directory()
{
	// this shouldn't fail, but it will use the home directory if it
	// does

	utf8_path p = apple_application_support();
	if (!p.empty())
		return p;

	log_error()
		<< "search for application support directory failed, "
		<< "using home\n";

	return home_directory();
}

utf8_string conf_directory_name()
{
	// if the application support directory is found, the directory
	// is just created in it; if it doesn't, it will be hidden in the
	// home directory

	if (!apple_application_support().empty())
		return utf8_string("baretorrent");
	else
		return utf8_string(".baretorrent");
}

utf8_path executable_directory()
{
	char dummy = 0;
	uint32_t size = 0;

	_NSGetExecutablePath(&dummy, &size);
	
	boost::shared_array<char> path(new char[size + 1]);
	std::memset(path.get(), 0, size + 1);

	if (_NSGetExecutablePath(path.get(), &size) == 0)
		return utf8_path(utf8_string(path.get())).parent();
	
	log_error()
		<< "could not get executable directory, returning pwd\n";
		
	return current_path();
}

bool this_url(CFURLRef url)
{
	NSURL* this_url = [[NSBundle mainBundle] bundleURL];
	NSURL* app_url = (NSURL*)url;

	NSString* app_ss = [app_url absoluteString];
	const utf8_string app_s([app_ss UTF8String]);

	if (![app_url isEqual:this_url])
	{
		NSString* this_ss = [this_url absoluteString];
		const utf8_string this_s([this_ss UTF8String]);

		log()
			<< "mime: got '" << app_s << "' "
			<< "instead of '" << this_s << "'\n";

		return false;
	}

	return true;
}

bool check_scheme(const utf8_string& scheme)
{
	log() << "scheme: checking " << scheme << "\n";

	utf8_string url = scheme + utf8_string(":");

	NSString* url_s =
		[[NSString alloc] initWithUTF8String:url.c_str()];

	CFURLRef url_ref = (CFURLRef)[NSURL URLWithString:url_s];

	CFURLRef app;
	OSStatus s = LSGetApplicationForURL(
		url_ref, kLSRolesAll, NULL, &app);

	if (s != 0)
	{
		if (s != kLSApplicationNotFoundErr)
		{
			log_error()
				<< "scheme: error while looking for application, "
				<< system_message(s) << "\n";
		}
		else
		{
			log_error()
				<< "scheme: no application for " << scheme << "\n";
		}

		return false;
	}

	if (!this_url(app))
		return false;

	log() << "scheme: " << scheme << " okay\n";

	return true;
}

void set_scheme(const utf8_string& scheme)
{
	log() << "scheme: setting handler for " << scheme << "\n";

	NSString* scheme_s =
		[[NSString alloc] initWithUTF8String:scheme.c_str()];

	NSString* bid = [[NSBundle mainBundle] bundleIdentifier];

	OSStatus s = LSSetDefaultHandlerForURLScheme(
		(CFStringRef)scheme_s, (CFStringRef)bid);

	if (s != 0)
	{
		log_error()
			<< "scheme: can't set default handler for " << scheme << ", "
			<< system_message(s) << "\n";

		return;
	}

	log() << "scheme: " << scheme << " set\n";
}

bool check_mime(const utf8_string& mime)
{
	log() << "mime: checking " << mime << "\n";

	NSString* mime_s =
		[[NSString alloc] initWithUTF8String:mime.c_str()];

	CFURLRef url;
	OSStatus s = LSCopyApplicationForMIMEType(
		(CFStringRef)mime_s, kLSRolesAll, &url);

	if (s != 0)
	{
		if (s != kLSApplicationNotFoundErr)
		{
			log_error()
				<< "mime: error while looking for application, "
				<< system_message(s) << "\n";
		}
		else
		{
			log_error() << "mime: no application for " << mime << "\n";
		}

		return false;
	}

	if (!this_url(url))
		return false;
		
	log() << "mime: " << mime << " okay\n";
	
	return true;
}

void set_mime(const utf8_string& mime)
{
	log() << "mime: setting handler for " << mime << "\n";

	NSString* mime_s =
		[[NSString alloc] initWithUTF8String:mime.c_str()];

	NSString* bid = [[NSBundle mainBundle] bundleIdentifier];

	CFStringRef uti = UTTypeCreatePreferredIdentifierForTag(
		kUTTagClassMIMEType, (CFStringRef)mime_s, NULL);

	if (!uti)
	{
		log_error() << "mime: can't create uti\n";
		return;
	}

	OSStatus r = LSSetDefaultRoleHandlerForContentType(
		uti, kLSRolesAll, (CFStringRef)bid);

	if (r != 0)
	{
		log_error()
			<< "mime: can't set default handler, "
			<< system_message(r) << "\n";
		return;
	}

	log() << "mime: " << mime << " set\n";
}

assoc associated()
{
	if (!check_mime(utf8_string("application/x-bittorrent")))
		return assoc(false);

	if (!check_scheme(utf8_string("magnet")))
		return assoc(false);

	return assoc(true);
}

void associate()
{
	log() << "assoc: associating\n";

	set_mime(utf8_string("application/x-bittorrent"));
	set_scheme(utf8_string("magnet"));

	log() << "assoc: rechecking\n";

	if (!associated().okay)
		log_error() << "assoc: still not associated\n";
	else
		log() << "assoc: all okay\n";
}


void add_image_handlers()
{
	// osx uses tiffs for system bitmaps
	wxImage::AddHandler(new wxTIFFHandler);
}

image_map get_image_map()
{
	image_map m;

#define a(e, s) m.insert(std::make_pair(images::e, utf8_string(s)))

	// all images are 24x24
	a(add,				"add_24");
	a(remove,			"remove_24");
	a(resume,			"resume_24");
	a(pause,				"pause_24");
	a(up,					"up_24");
	a(down,				"down_24");
	a(top,				"top_24");
	a(bottom,			"bottom_24");
	a(active,			"active_24");
	a(upload,			"upload_24");
	a(download,			"download_24");
	a(inactive,			"inactive_24");
	a(all,				"all_24");
	a(creeper,			"creeper");
	a(icon_16,			"icon_16");
	a(icon_24,			"icon_24");
	a(icon_32,			"icon_32");
	a(icon_48,			"icon_48");
	a(icon_64,			"icon_64");
	a(icon_128,			"icon_128");

	a(close_search,	"close_search");

#undef a

	return m;
}

wxBitmap make_image_unsafe(images::image i)
{
	const image_map m = get_image_map();
	
	image_map::const_iterator itor = m.find(i);
	bt_assert(itor != m.end());
	if (itor == m.end())
		return make_bad_image();

	const utf8_string& name(itor->second);

	log_redirector lg;

	// this will load the png from the bundle
	wxBitmap b(to_wxstring(name), wxBITMAP_TYPE_PNG_RESOURCE);

	if (!b.IsOk())
	{
		log_error()
			<< "couldn't load bitmap '" << name << "', "
			<< lg.what() << "\n";
	}

	return b;
}

wxBitmap make_image(images::image i)
{
	wxBitmap b = make_image_unsafe(i);
	if (!b.IsOk())
		return make_bad_image();
	
	return b;
}

wxIcon make_icon(images::image i)
{
	// no difference between an icon and a bitmap
	wxIcon icon;
	icon.CopyFromBitmap(make_image(i));

	return icon;
}

utf8_string make_time(const utf8_string& format, std::time_t s)
{
	const std::size_t buffer_size = 500;
	char buffer[buffer_size + 1] = {0};

	std::tm t = *std::localtime(&s);

	const std::size_t n = std::strftime(
		buffer, buffer_size, format.c_str(), &t);

	if (n == 0)
	{
		log_error() << "strftime failed\n";
		return utf8_string();
	}

	return utf8_string(buffer);
}

wxBitmap system_bitmap(const utf8_string& n)
{
	// todo; this is for debug, it always returns the quick look
	// image
	NSImage* i = [NSImage imageNamed:NSImageNameQuickLookTemplate];
	NSData* data = [i TIFFRepresentation];

	wxMemoryInputStream stream([data bytes], [data length]);
	wxImage wxi(stream, wxBITMAP_TYPE_TIF);

	return wxBitmap(wxi);
}

void add_firewall_exception()
{
	// noop
}

// from http://stackoverflow.com/a/6105338
//
OSStatus send_apple_event(AEEventID id)
{
	const ProcessSerialNumber psn = {0, kSystemProcess};
	AppleEvent reply = {typeNull, NULL};
	AppleEvent event = {typeNull, NULL};

	AEAddressDesc desc;
	OSStatus s = AECreateDesc(typeProcessSerialNumber, &psn, sizeof(psn), &desc);

	if (s != noErr)
		return s;

	s = AECreateAppleEvent(
			kCoreEventClass, id, &desc, kAutoGenerateReturnID,
			kAnyTransactionID, &event);

	AEDisposeDesc(&desc);

	if (s != noErr)
		return s;

	s = AESendMessage(&event, &reply, kAENormalPriority, kAEDefaultTimeout);

	AEDisposeDesc(&event);

	if (s != noErr)
		return s;

	AEDisposeDesc(&reply);

	return s;
}

void power(power_actions p)
{
	OSStatus s = noErr;

	switch (p)
	{
		case shutdown_power:
		{
			s = send_apple_event(kAEShutDown);
			break;
		}

		case suspend_power:
		{
			s = send_apple_event(kAESleep);
			break;
		}

		case hibernate_power:	// fall-through
		case no_power_action:
		default:
		{
			break;
		}
	}
}

bool has_power(power_actions p)
{
	switch (p)
	{
		case shutdown_power:
			return true;

		case suspend_power:
			return true;

		case hibernate_power:
			return false;

		case no_power_action:	// fall-through
		default:
			return false;
	}
}

bool has_power_management()
{
	return true;
}

void detach_console()
{
	// noop
}

void create_console()
{
	// noop
}

void log_to_console(
	std::ostream* out, log_t::levels, const utf8_string& s)
{
	if (out)
		*out << s.std_string();
}

utf8_string name()
{
	utf8_string s = shell_output(utf8_string(
		"system_profiler -detailLevel -2 | "
		"grep \"System Version\" | "
		"cut -d : -f 2"));

	if (s.empty())
		s = utf8_string("some mac");

	return s;
}

}	// namespace platform
}	// namespace baretorrent

#endif	// BARETORRENT_APPLE

