// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "archiver.h"
#include "utility.h"
#include "platform.h"

#include "7z/Types.h"
#include "7z/7zAlloc.h"
#include "7z/7zFile.h"
#include "7z/7zCrc.h"
#include "7z/7z.h"

// this is adapted from C/Util/7z/7zMain.c
//
// I wanted to clean it up a bit, but it's not the best api I've seen;
// at least now it's exception safe

namespace baretorrent
{
namespace archiver
{

struct file_info : boost::noncopyable
{
	ISzAlloc alloc;
	CSzArEx db;
	int i;
	const CSzFileItem* f;
	UInt32 block_index;
	Byte* out_buffer;
	size_t out_buffer_size;
	CLookToRead look;
	CFileInStream archive;

	file_info()
	{
		std::memset(this, 0, sizeof(file_info));
		alloc.Alloc = SzAlloc;
		alloc.Free = SzFree;
	}

	~file_info()
	{
		SzArEx_Free(&db, &alloc);
		IAlloc_Free(&alloc, out_buffer);
		File_Close(&archive.file);
	}
};

typedef boost::shared_ptr<file_info> file_info_ptr;

utf8_string lzma_error(SRes r)
{
	utf8_string m;

	switch(r)
	{
		case SZ_ERROR_DATA:
			m = utf8_string("SZ_ERROR_DATA");
			break;

		case SZ_ERROR_MEM:
			m = utf8_string("SZ_ERROR_MEM");
			break;

		case SZ_ERROR_CRC:
			m = utf8_string("SZ_ERROR_CRC");
			break;

		case SZ_ERROR_UNSUPPORTED:
			m = utf8_string("SZ_ERROR_UNSUPPORTED");
			break;

		case SZ_ERROR_PARAM:
			m = utf8_string("SZ_ERROR_PARAM");
			break;

		case SZ_ERROR_INPUT_EOF:
			m = utf8_string("SZ_ERROR_INPUT_EOF");
			break;

		case SZ_ERROR_OUTPUT_EOF:
			m = utf8_string("SZ_ERROR_OUTPUT_EOF");
			break;

		case SZ_ERROR_READ:
			m = utf8_string("SZ_ERROR_READ");
			break;

		case SZ_ERROR_WRITE:
			m = utf8_string("SZ_ERROR_WRITE");
			break;

		case SZ_ERROR_PROGRESS:
			m = utf8_string("SZ_ERROR_PROGRESS");
			break;

		case SZ_ERROR_FAIL:
			m = utf8_string("SZ_ERROR_FAIL");
			break;

		case SZ_ERROR_THREAD:
			m = utf8_string("SZ_ERROR_THREAD");
			break;

		case SZ_ERROR_ARCHIVE:
			m = utf8_string("SZ_ERROR_ARCHIVE");
			break;

		case SZ_ERROR_NO_ARCHIVE:
			m = utf8_string("SZ_ERROR_NO_ARCHIVE");
			break;

		default:
			m = utf8_string("unknown error ") + as_utf8_string(r);
			break;
	}

	return utf8_string("7z says: ") + m;
}

void extract_file(file_info* fi, const utf8_path& root)
{
	size_t len = SzArEx_GetFileNameUtf16(&fi->db, fi->i, NULL);
	boost::shared_array<UInt16> temp(new UInt16[len + 1]);
	std::memset(temp.get(), 0, (len + 1) * sizeof(UInt16));

	SzArEx_GetFileNameUtf16(&fi->db, fi->i, temp.get());

	size_t offset = 0;
	size_t outSizeProcessed = 0;

	if (!fi->f->IsDir)
	{
		SRes res = SzArEx_Extract(
			&fi->db, &fi->look.s, fi->i,
			&fi->block_index, &fi->out_buffer, &fi->out_buffer_size,
			&offset, &outSizeProcessed,
			&fi->alloc, &fi->alloc);

		if (res != SZ_OK)
		{
			log_error()
				<< "could not extract data, "
				<< lzma_error(res) << "\n";

			throw exception();
		}
	}


	UInt16 *name = temp.get();
	const UInt16 *destPath = (const UInt16 *)name;

	for (size_t j = 0; name[j] != 0; j++)
	{
		if (name[j] == '/')
		{
			name[j] = 0;

			const byte_array ba = convert(
				name, len * sizeof(UInt16),
				utf16_encoding, utf8_encoding);

			if (ba.empty())
			{
				log_error()
					<< "archiver: failed to convert filename from "
					<< "utf16 to utf8\n";

				throw exception();
			}

			const utf8_path p(utf8_string(ba.data()));

			log_v() << "archiver: creating " << root / p << "\n";
			create_directories(root / p);

			name[j] = CHAR_PATH_SEPARATOR;
		}
	}

	if (fi->f->IsDir)
	{
		const byte_array ba = convert(
			destPath, len * sizeof(UInt16),
			utf16_encoding, utf8_encoding);

		if (ba.empty())
		{
			log_error()
				<< "archiver: failed to convert directory name from "
				<< "utf16 to utf8\n";
			
			throw exception();
		}

		const utf8_path p(utf8_string(ba.data()));

		log_v() << "archiver: creating " << root / p << "\n";
		create_directories(root / p);
		return;
	}

	CSzFile outFile;
	
	const byte_array ba = convert(
		destPath, len * sizeof(UInt16),
		utf16_encoding, utf8_encoding);

	if (ba.empty())
	{
		log_error()
			<< "archiver: failed to convert destination path from "
			<< "utf16 to utf8\n";

		throw exception();
	}

	const utf8_path p = root / utf8_string(ba.data());

	log_v()
		<< "archiver: extracting " << p << " (" << outSizeProcessed
		<< " bytes)\n";

#if BARETORRENT_WINDOWS
	WRes wres = OutFile_OpenW(&outFile, p.native().c_str());
#else
	WRes wres = OutFile_Open(&outFile, p.native().c_str());
#endif

	if (wres != 0)
	{
		log_error()
			<< "could not open output file " << p << ", "
			<< platform::system_message(wres) << "\n";

		throw exception();
	}

	size_t processedSize = outSizeProcessed;
	wres = File_Write(
		&outFile, fi->out_buffer + offset, &processedSize);
	
	if (wres != 0 || processedSize != outSizeProcessed)
	{
		log_error()
			<< "could not write to output file " << p << ", "
			<< platform::system_message(wres) << "\n";

		throw exception();
	}

	wres = File_Close(&outFile);

	if (wres != 0)
	{
		log_error()
			<< "could not close output file, "
			<< platform::system_message(wres) << "\n";

		throw exception();
	}
}

file_info_ptr initialize_archive(const utf8_path& what)
{
	file_info_ptr fi(new file_info);

#if BARETORRENT_WINDOWS
	WRes wres = InFile_OpenW(&fi->archive.file, what.native().c_str());
#else
	WRes wres = InFile_Open(&fi->archive.file, what.native().c_str());
#endif

	if (wres != 0)
	{
		log_error()
			<< "couldn't open archive " << what << ", "
			<< platform::system_message(wres) << "\n";

		throw exception();
	}

	FileInStream_CreateVTable(&fi->archive);

	LookToRead_CreateVTable(&fi->look, False);
	fi->look.realStream = &fi->archive.s;
	LookToRead_Init(&fi->look);

	CrcGenerateTable();

	SzArEx_Init(&fi->db);
	
	SRes res = SzArEx_Open(
		&fi->db, &fi->look.s, &fi->alloc, &fi->alloc);
	
	if (res != SZ_OK)
	{
		log_error()
			<< "couldn't create stream, " << lzma_error(res) << "\n";

		throw exception();
	}

	fi->block_index = 0xFFFFFFFF; 
	fi->out_buffer_size = 0;

	return fi;
}

void extract(const utf8_path& what, const utf8_path& where)
{
	log_v() << "extracting " << what << " into " << where << "\n";
	
	file_info_ptr fi = initialize_archive(what);

	for (UInt32 i = 0; i < fi->db.db.NumFiles; i++)
	{
		fi->i = i;
		fi->f = fi->db.db.Files + i;

		extract_file(fi.get(), where);
	}
}

}	// namespace archiver
}	// namespace baretorrent
