// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "connections_panel.h"
#include "locale.h"
#include "torrent.h"
#include "dialogs.h"

namespace baretorrent
{

enum ids
{
	// trackers list
	trackers_id = wxID_HIGHEST + 100,

	// peers list
	peers_id,

	// add tracker menu item
	add_tracker_id,

	// remove tracker menu item
	remove_tracker_id
};

// time between the retrieving the connection information from
// libtorrent in the thread
const boost::posix_time::milliseconds update_thread_delay(1000);

// update timer length
const boost::posix_time::milliseconds update_timer_delay(2000);


// returns -1, 1 or 0
//
int pred(int a, int b)
{
	if (a < b) return -1;
	if (a > b) return 1;
	return 0;
}

// lexical sort on the tracker urls
//
int url_sorter(bt::announce_entry* left, bt::announce_entry* right)
{
	return natural_compare(
		utf8_string(left->url), utf8_string(right->url));
}

// lexical sort on the tracker messages
//
int message_sorter(bt::announce_entry* left, bt::announce_entry* right)
{
	return natural_compare(
		utf8_string(left->message), utf8_string(right->message));
}

// lexical sort on the peer ip addresses
//
int ip_sorter(bt::peer_info* left, bt::peer_info* right)
{
	return natural_compare(
		utf8_string(left->ip.address().to_string()),
		utf8_string(right->ip.address().to_string()));
}

// sort on the download rate
//
int dl_rate_sorter(bt::peer_info* left, bt::peer_info* right)
{
	return pred(left->payload_down_speed, right->payload_down_speed);
}

// sort on the upload rate
int up_rate_sorter(bt::peer_info* left, bt::peer_info* right)
{
	return pred(left->payload_up_speed, right->payload_up_speed);
}


// popup menu on the tracker list
//
class trackers_popup : public wxMenu
{
public:
	trackers_popup()
	{
		Append(add_tracker_id, loc_wx("connections.trackers.add"));
		Append(remove_tracker_id, loc_wx("connections.trackers.remove"));
	}
};


connections_panel::connections_panel(wxWindow* parent) :
	wxPanel(parent),
	trackers_list_(nullptr), trackers_caption_(nullptr),
	peers_list_(nullptr), peers_caption_(nullptr),
	trackers_popup_(nullptr), initial_(false), showing_initial_(false)
{
	wxBoxSizer* sizer = new wxBoxSizer(wxVERTICAL);

	// splitter
	splitter* split =  new splitter(this, "bt.ui.connections_splitter");
	sizer->Add(split, 1, wxEXPAND);

	// lists
	split->SplitHorizontally(
		create_trackers(split),
		create_peers(split));

	SetSizer(sizer);

	// update timer
	timer_.start(boost::bind(
		&connections_panel::on_timer, this), update_timer_delay);

	// update thread
	thread_ = boost::thread(
		boost::bind(&connections_panel::update_thread, this));
}

connections_panel::~connections_panel()
{
	timer_.Stop();
	thread_.interrupt();
	thread_.join();

	// deletes the announce_entry and peer_info pointers that were
	// added to the list
	trackers_list_->delete_data();
	peers_list_->delete_data();
}

wxWindow* connections_panel::create_trackers(wxWindow* parent)
{
	wxPanel* p = new wxPanel(parent);
	trackers_list_ = new custom_list<bt::announce_entry>(
		p, trackers_id, "bt.ui.trackers_list");

	trackers_list_->add_column(
		loc("torrent.trackers.url"), &url_sorter);
	trackers_list_->add_column(
		loc("torrent.trackers.message"), &message_sorter);

	wxFlexGridSizer* trackers_sizer = new wxFlexGridSizer(1);

	trackers_caption_ = new wxStaticText(
		p, -1, to_wxstring(locf("connections.trackers") % 0));
	trackers_sizer->Add(trackers_caption_);
	trackers_sizer->Add(trackers_list_, 1, wxEXPAND);

	p->SetSizer(trackers_sizer);

	trackers_sizer->AddGrowableRow(1);
	trackers_sizer->AddGrowableCol(0);

	trackers_popup_ = new trackers_popup;
	trackers_list_->context(trackers_popup_);

	// no sort, keep the original order by default

	return p;
}

wxWindow* connections_panel::create_peers(wxWindow* parent)
{
	wxPanel* p = new wxPanel(parent);
	peers_list_ = new custom_list<bt::peer_info>(
		p, peers_id, "bt.ui.peers_list");

	peers_list_->add_column(loc("torrent.peers.ip"), &ip_sorter);

	peers_list_->add_column(
		loc("torrent.download_rate.caption"), &dl_rate_sorter);

	peers_list_->add_column(
		loc("torrent.upload_rate.caption"), &up_rate_sorter);

	// sort by ip
	peers_list_->sort(0, false);

	wxFlexGridSizer* peers_sizer = new wxFlexGridSizer(1);

	peers_caption_ = new wxStaticText(
		p, -1, to_wxstring(locf("connections.peers") % 0));
	peers_sizer->Add(peers_caption_);
	peers_sizer->Add(peers_list_, 1, wxEXPAND);

	p->SetSizer(peers_sizer);
	peers_sizer->AddGrowableRow(1);
	peers_sizer->AddGrowableCol(0);

	return p;
}

void connections_panel::update_thread()
{
	try
	{
		for (;;)
		{
			torrents ts;

			{
				// makes a copy of the selected torrent so the progress
				// and priorities can be updated without blocking the
				// ui
				boost::recursive_mutex::scoped_lock
					lock(current_mutex_);

				ts = current_;
			}
			
			// only update the the info if there's a single selection
			// with metadata
			if (ts.size() == 1)
			{
				// trackers
				const std::vector<bt::announce_entry> trackers =
					ts[0]->trackers();

				// check if the thread has been interrupted
				boost::this_thread::interruption_point();

				std::vector<bt::peer_info> peers = ts[0]->peers();
				
				std::vector<bt::peer_info>::iterator itor=peers.begin();
				while (itor != peers.end())
				{
					if (itor->payload_down_speed > 0 ||
						itor->payload_up_speed > 0)
					{
						// good
						++itor;
					}
					else
					{
						itor = peers.erase(itor);
					}
				}

				// the two calls above may be slow, but the mutex is
				// only locked when the info has been retrieved
				boost::mutex::scoped_lock lock(info_mutex_);

				// there will be used in the update timer to update
				// the tree
				trackers_ = trackers;
				peers_ = peers;
				initial_ = false;
			}

			boost::this_thread::sleep(update_thread_delay);
		}
	}
	catch(boost::thread_interrupted&)
	{
	}
}

bool connections_panel::on_timer()
{
	update_impl(current_);
	return true;
}

void connections_panel::clear()
{
	const wxString tc = to_wxstring(locf("connections.trackers") % 0);

	if (trackers_caption_->GetLabel() != tc)
		trackers_caption_->SetLabel(tc);

	
	const wxString cc = to_wxstring(locf("connections.peers") % 0);

	if (peers_caption_->GetLabel() != cc)
		peers_caption_->SetLabel(cc);

	trackers_list_->clear();
	peers_list_->clear();

	// current_ is also used in the update thread
	boost::recursive_mutex::scoped_lock lock(current_mutex_);
	current_.clear();
}

void connections_panel::update(const torrents& ts)
{
	update_impl(ts);

	// remembering the selection so the lists can be updated in the
	// timer
	current_ = ts;
}

void connections_panel::update_impl(const torrents& ts)
{
	// this will block the update thread, but it shouldn't take long
	boost::recursive_mutex::scoped_lock lock(current_mutex_);

	// freezes the list so they don't flicker (too much)
	freezer trackers_freezer(trackers_list_);
	freezer peers_freezer(peers_list_);

	if (ts.empty())
	{
		// no selection, clear
		if (ts != current_)
			clear();
	}
	else if (ts.size() == 1)
	{
		if (current_.size() != ts.size() ||
			ts[0] != current_[0] ||
			ts[0]->status() == torrent::removing_status)
		{
			// different selection, clear it immediately since it
			// might take a while before the thread updates the
			// values
			boost::mutex::scoped_lock lock(info_mutex_);

			trackers_.clear();
			peers_.clear();
			initial_ = true;

			initial_update();
		}
		else if (!initial_)
		{
			// a single torrent is selected
			single_torrent();
		}
	}
	else if (ts.size() > 1)
	{
		multiple_selection();
	}
}

void connections_panel::initial_update()
{
	clear();

	// adding a dummy "updating" item to both lists

	trackers_list_->append(utf8_string("updating"));
	peers_list_->append(utf8_string("updating"));

	showing_initial_ = true;
}

void connections_panel::single_torrent()
{
	update_trackers();
	update_peers();

	if (!initial_)
		showing_initial_ = false;
}

void connections_panel::update_tracker(
	const bt::announce_entry& e, int i)
{
	bt::announce_entry* f = trackers_list_->data(i);

	*f = e;
	trackers_list_->col(i, 0, utf8_string(e.url));

	// message
	utf8_string s;

	if (e.last_error)
		s = loc_error(e.last_error);
	else if (e.verified)
		s = loc("torrent.trackers.ok");
	else if (e.updating)
		s = loc("torrent.trackers.updating");
	// too much detail
	//else if (e.start_sent)
	//	s = loc_wx(utf8_string("torrent.trackers.start_sent"));
	//else if (e.complete_sent)
	//	s = loc_wx(utf8_string("torrent.trackers.complete_sent"));
	else if (!e.message.empty())
		s = utf8_string(e.message);

	trackers_list_->col(i, 0, utf8_string(e.url));
	trackers_list_->col(i, 1, s);
}

bool same_tracker(
	const bt::announce_entry& left, const bt::announce_entry& right)
{
	if (left.url == right.url && left.tier == right.tier &&
		left.source == right.source)
	{
		return true;
	}

	return false;
}

void connections_panel::update_trackers()
{
	std::vector<bt::announce_entry> trackers;

	{
		boost::mutex::scoped_lock lock(info_mutex_);
		trackers = trackers_;
	}
	
	const std::vector<bt::announce_entry> otrackers = trackers;

	const wxString s = to_wxstring(
		locf("connections.trackers") % trackers.size());

	if (trackers_caption_->GetLabel() != s)
		trackers_caption_->SetLabel(s);

	if (showing_initial_)
	{
		// there's currently an "updating" label in the list
		trackers_list_->clear();
	}

	if (trackers.size() > 1000)
		trackers.erase(trackers.begin() + 1000, trackers.end());
	
	// for each item in the list; note that j is incremented inside the
	// loop because items may be deleted while iterating
	for (int j=0; j<trackers_list_->GetItemCount(); )
	{
		bool found = false;
		bt::announce_entry* f = trackers_list_->data(j);
		
		// for each tracker
		for (std::size_t i=0; i<trackers.size(); ++i)
		{
			const bt::announce_entry& e = trackers[i];

			if (same_tracker(e, *f))
			{
				update_tracker(e, j);
				found = true;

				// this tracker was found in the list; those that are
				// not erased will be added below
				trackers.erase(trackers.begin() + i);
				break;
			}
		}

		if (!found)
		{
			// this tracker disappeared
			trackers_list_->DeleteItem(j);
		}
		else
		{
			// this tracker is still there, move on
			++j;
		}
	}

	// trackers that were in the list were removed from the vector;
	// those that are left must be added
			
	for (std::size_t i=0; i<trackers.size(); ++i)
	{
		const bt::announce_entry& e = trackers[i];

		const int j = trackers_list_->append();
		trackers_list_->data(j, new bt::announce_entry(e));
		update_tracker(e, j);
	}

	// trackers are not sorted by default (in which case this is a
	// noop), but the user may have clicked a column
	trackers_list_->sort();
}

void connections_panel::update_peer(const bt::peer_info& p, int i)
{
	bt::peer_info* q = peers_list_->data(i);

	*q = p;

	peers_list_->col(i, 0, utf8_string(p.ip.address().to_string()));
	peers_list_->col(i, 1, make_loc_rate(p.payload_down_speed));
	peers_list_->col(i, 2, make_loc_rate(p.payload_up_speed));
}

bool same_peer(const bt::peer_info& left, const bt::peer_info& right)
{
	if (left.source == right.source && left.ip == right.ip)
		return true;

	return false;
}

void connections_panel::update_peers()
{
	std::vector<bt::peer_info> peers;

	{
		boost::mutex::scoped_lock lock(info_mutex_);
		peers = peers_;
	}
	
	const std::vector<bt::peer_info> opeers = peers;

	const wxString s = to_wxstring(
		locf("connections.peers") % peers.size());

	if (peers_caption_->GetLabel() != s)
		peers_caption_->SetLabel(s);

	if (showing_initial_)
	{
		// there's currently an "updating" label in the list
		peers_list_->clear();
	}

	// for each item in the list; note that j is incremented inside the
	// loop because items may be deleted while iterating
	for (int j=0; j<peers_list_->GetItemCount(); )
	{
		bool found = false;
		bt::peer_info* q = peers_list_->data(j);

		// for each peer
		for (std::size_t i=0; i<peers.size(); ++i)
		{
			const bt::peer_info& p = peers[i];

			if (same_peer(p, *q))
			{
				update_peer(p, j);
				found = true;

				// this peer was found in the list; those that are
				// not erased will be added below
				peers.erase(peers.begin() + i);
				break;
			}
		}

		if (!found)
		{
			// this peer disappeared
			peers_list_->DeleteItem(j);
		}
		else
		{
			// this peer is still there, move on
			++j;
		}
	}

	// peers that were in the list were removed from the vector;
	// those that are left must be added
	
	for (std::size_t i=0; i<peers.size(); ++i)
	{
		const bt::peer_info& p = peers[i];

		const int j = peers_list_->append();
		peers_list_->data(j, new bt::peer_info(p));
		update_peer(p, j);
	}

	const std::size_t peer_count = opeers.size();
	const int item_count = peers_list_->GetItemCount();

	bt_assert(static_cast<int>(peer_count) == item_count);
	
	peers_list_->sort();
}

void connections_panel::multiple_selection()
{
	trackers_list_->clear();
	peers_list_->clear();

	const wxString tc = to_wxstring(locf("connections.trackers") % 0);

	if (trackers_caption_->GetLabel() != tc)
		trackers_caption_->SetLabel(tc);
		
	const wxString cc = to_wxstring(locf("connections.peers") % 0);

	if (peers_caption_->GetLabel() != cc)
		peers_caption_->SetLabel(cc);

	// adding a dummy "multiple selection" item to both lists

	trackers_list_->append(loc("multiple_selection"));
	peers_list_->append(loc("multiple_selection"));
}

void connections_panel::on_tracker_popup_update(wxUpdateUIEvent& e)
{
	if (e.GetId() == add_tracker_id)
		e.Enable(current_.size() == 1);
	if (e.GetId() == remove_tracker_id)
		e.Enable(trackers_list_->has_selection() && !initial_);
}

void connections_panel::on_add_tracker(wxCommandEvent&)
{
	torrent_ptr t;

	{
		// current_ is also used in the update thread; this only
		// locks it so the first torrent in the list (which should
		// be the only one) is retrieved
		boost::recursive_mutex::scoped_lock lock(current_mutex_);

		// todo (add always enabled, no?)
		bt_assert(current_.size() == 1);
		if (current_.size() != 1)
			return;

		t = current_[0];
	}

	// getting top-level window
	wxWindow* tl = this;
	while (tl->GetParent())
		tl = tl->GetParent();

	add_trackers_dialog* d = new add_trackers_dialog(tl);
	if (d->ShowModal() != wxID_OK)
		return;

	std::vector<utf8_string> v = d->trackers();

	// this will be picked up in the next update
	foreach(const utf8_string& s, v)
		t->add_tracker(s);
}

void connections_panel::on_remove_tracker(wxCommandEvent&)
{
	// todo: multiple selection?
	bt_assert(current_.size() == 1);
	if (current_.size() != 1)
		return;

	typedef std::vector<bt::announce_entry> ts;

	// the announce_entry in the list is a copy on the heap so the
	// url and tier need to be compared to find the right one

	ts v = current_[0]->trackers();
	std::vector<const bt::announce_entry*> s =
		trackers_list_->selection();

	// for each tracker in the torrent
	for (ts::iterator itor=v.begin(); itor!=v.end(); )
	{
		bool ok = false;

		// for each tracker in the list
		foreach(const bt::announce_entry* e, s)
		{
			if (itor->url == e->url && itor->tier == e->tier)
			{
				// same one, remove it from the vector, which is used
				// directly below
				itor = v.erase(itor);

				ok = true;
				break;
			}
		}

		// if the tracker as found, the iterator now points to the next
		// element
		if (!ok)
			++itor;
	}

	// setting the new tracker list
	current_[0]->trackers(v);
}

BEGIN_EVENT_TABLE(connections_panel, wxPanel)
	EVT_UPDATE_UI(add_tracker_id, connections_panel::on_tracker_popup_update)
	EVT_UPDATE_UI(remove_tracker_id, connections_panel::on_tracker_popup_update)
	EVT_MENU(add_tracker_id, connections_panel::on_add_tracker)
	EVT_MENU(remove_tracker_id, connections_panel::on_remove_tracker)
END_EVENT_TABLE()

}	// namespace baretorrent
