// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "custom_list.h"

namespace baretorrent
{

custom_list_base::custom_list_base(
	wxWindow* parent, wxWindowID id,
	const std::string& option_name, long style) :
		wxListCtrl(
			parent, id, wxDefaultPosition, wxDefaultSize,
			wxLC_REPORT | style),
		option_(option_name),
		context_(nullptr),
		il_(nullptr)
{
	std::memset(image_indices_, 0, sizeof(image_indices_));

	platform::init_list_control(this);
	platform::set_list_control_style(this);

	// for sorting
	Connect(
		wxEVT_COMMAND_LIST_COL_CLICK,
		wxListEventHandler(custom_list_base::on_column_clicked));

	// for the context menu
	Connect(
		wxEVT_CONTEXT_MENU,
		wxContextMenuEventHandler(custom_list_base::on_context_menu));

	// to save the new column widths
	Connect(
		wxEVT_COMMAND_LIST_COL_END_DRAG,
		wxListEventHandler(custom_list_base::on_column_resized));

	// check
	Connect(
		wxEVT_LEFT_DOWN,
		wxMouseEventHandler(custom_list_base::on_left_down));

	// selection
	Connect(
		wxEVT_COMMAND_LIST_ITEM_SELECTED,
		wxListEventHandler(custom_list_base::on_selected));

	// selection
	Connect(
		wxEVT_COMMAND_LIST_ITEM_DESELECTED,
		wxListEventHandler(custom_list_base::on_deselected));

	// activated
	Connect(
		wxEVT_COMMAND_LIST_ITEM_ACTIVATED,
		wxListEventHandler(custom_list_base::on_activated));

	// select all
	Connect(
		wxID_SELECTALL, wxEVT_COMMAND_MENU_SELECTED,
		wxMenuEventHandler(custom_list_base::on_select_all));

	acc_.push_back(wxAcceleratorEntry(
		wxACCEL_CTRL, 'A', wxID_SELECTALL));

	wxAcceleratorTable at(static_cast<int>(acc_.size()), &acc_[0]);
	SetAcceleratorTable(at);

	// this remembers the initial width of every column and the
	// initial sort from the options so that they can be restored
	// while adding the columns
	if (!option_.empty())
	{
		owidths_ = explode<double>(
			opt(option_ + ".widths"),
			utf8_string(" "));
			
		const std::vector<int> s = explode<int>(
			opt(option_ + ".sort"), utf8_string(" "));
			
		if (s.size() >= 1)
			osort_.col = s[0];
		if (s.size() >= 2)
			osort_.dir = s[1];
	}

	visible_.start(
		this, boost::bind(&custom_list_base::on_visible, this));
}

custom_list_base::~custom_list_base()
{
	delete il_;
	platform::destroy_list_control(this);
	delete context_;
}

int custom_list_base::append()
{
	return append(utf8_string(""));
}

int custom_list_base::append(const utf8_string& c)
{
	InsertItem(GetItemCount(), to_wxstring(c));
	return GetItemCount() - 1;
}

void custom_list_base::add_accelerator(const wxAcceleratorEntry& e)
{
	acc_.push_back(e);

	wxAcceleratorTable at(static_cast<int>(acc_.size()), &acc_[0]);
	SetAcceleratorTable(at);
}

void custom_list_base::use_checkboxes()
{
	set_images();
}

void custom_list_base::context(wxMenu* m)
{
	context_ = m;
}

bool custom_list_base::checked(long i)
{
	wxListItem item;
	item.SetId(i);
	item.SetMask(wxLIST_MASK_IMAGE);

	if (GetItem(item))
		return (item.GetImage() == image_indices_[checked_image]);

	return false;
}

void custom_list_base::checked(long i, bool b)
{
	image_indices ii = (b ? checked_image : unchecked_image);
	SetItemImage(i, image_indices_[ii]);
}

wxImageList* custom_list_base::get_image_list()
{
#if BARETORRENT_WINDOWS
	const int iw = ::GetSystemMetrics(SM_CXMENUCHECK);
	const int ih = ::GetSystemMetrics(SM_CXMENUCHECK);
#else
	const int iw = 16;
	const int ih = 16;
#endif

	bt_assert(iw > 0 && ih > 0);

	wxImageList* il = GetImageList(wxIMAGE_LIST_SMALL);
	if (!il)
	{
		il_ = new wxImageList;
		il_->Create(iw, ih);
		SetImageList(il_, wxIMAGE_LIST_SMALL);

		il = il_;
	}	

	return il;
}

void custom_list_base::set_image(int i, const wxBitmap& b)
{
	SetItemImage(i, get_image_list()->Add(b));
}

void custom_list_base::set_images()
{
	wxImageList* il = get_image_list();

	int iw=0, ih=0;
	il->GetSize(0, iw, ih);

	wxBitmap unchecked_bmp(iw, ih);
	wxBitmap checked_bmp(iw, ih);
	wxBitmap unchecked_disabled_bmp(iw, ih);
	wxBitmap checked_disabled_bmp(iw, ih);
 
	// Bitmaps must not be selected by a DC for addition to the
	// image list but I don't see a way of diselecting them in
	// wxMemoryDC so let's just use a code block to end the scope
	{
		wxMemoryDC renderer_dc;
 
		// Unchecked
		renderer_dc.SelectObject(unchecked_bmp);
		renderer_dc.SetBackground(
			*wxTheBrushList->FindOrCreateBrush(
				GetBackgroundColour(), wxSOLID));
		renderer_dc.Clear();
		wxRendererNative::Get().DrawCheckBox(
			this, renderer_dc, wxRect(0, 0, iw, ih), 0);
 
		// Checked
		renderer_dc.SelectObject(checked_bmp);
		renderer_dc.SetBackground(
			*wxTheBrushList->FindOrCreateBrush(
				GetBackgroundColour(), wxSOLID));
		renderer_dc.Clear();
		wxRendererNative::Get().DrawCheckBox(
			this, renderer_dc, wxRect(0, 0, iw, ih),
			wxCONTROL_CHECKED);
 
		// Unchecked and Disabled
		renderer_dc.SelectObject(unchecked_disabled_bmp);
		renderer_dc.SetBackground(
			*wxTheBrushList->FindOrCreateBrush(
				GetBackgroundColour(), wxSOLID));
		renderer_dc.Clear();
		wxRendererNative::Get().DrawCheckBox(
			this, renderer_dc, wxRect(0, 0, iw, ih),
			wxCONTROL_DISABLED);
 
		// Checked and Disabled
		renderer_dc.SelectObject(checked_disabled_bmp);
		renderer_dc.SetBackground(
			*wxTheBrushList->FindOrCreateBrush(
				GetBackgroundColour(), wxSOLID));
		renderer_dc.Clear();
		wxRendererNative::Get().DrawCheckBox(
			this, renderer_dc, wxRect(0, 0, iw, ih),
			wxCONTROL_CHECKED | wxCONTROL_DISABLED);
	}
 
	image_indices_[unchecked_image] = il->Add(unchecked_bmp);
	image_indices_[checked_image] = il->Add(checked_bmp);
	image_indices_[unchecked_disabled_image] =
	   	il->Add(unchecked_disabled_bmp);
	image_indices_[checked_disabled_image] =
		il->Add(checked_disabled_bmp);
}

void custom_list_base::row_color(int i, const wxColour& c)
{
	SetItemTextColour(i, c);
}

void custom_list_base::row_font(int i, const wxFont& f)
{
	SetItemFont(i, f);
}

void custom_list_base::on_visible()
{
	log_vv()
		<< "custom list is visible, size is "
		<< GetSize().x << " x " << GetSize().y << "\n";

	set_widths();
	save_widths();
}

void custom_list_base::on_column_clicked(wxListEvent& e)
{
	if (e.GetColumn() == -1)
		return;

	const std::size_t c = e.GetColumn();
		
	if (c == sort_.col)
	{
		// the user clicked twice on the same column, which reverses
		// the sort direction
		sort_.dir = -sort_.dir;
	}
	else
	{
		// the user clicked on this column for the first time; the
		// direction is ascending by default
		sort_.dir = 1;
	}

	sort_.col = c;
	save_sort();
	sort();
}

void custom_list_base::on_column_resized(wxListEvent&)
{
	save_widths();
}

void custom_list_base::on_context_menu(wxContextMenuEvent& e)
{
	if (!context_)
		return;

	//int x=0, y=0;
	//wxGetMousePosition(&x, &y);

	wxPoint p = ScreenToClient(e.GetPosition());
	PopupMenu(context_, p);
}

void custom_list_base::on_left_down(wxMouseEvent& e)
{
	if (e.LeftDown())
	{
		int f;
		long i = HitTest(e.GetPosition(), f);

		if (i >= 0 && (f & wxLIST_HITTEST_ONITEMICON))
		{
			const bool b = !checked(i);
			checked(i, b);
			on_checked(i, b);
		}
		else
		{
			e.Skip();
		}
	}
	else
	{
		e.Skip();
	}
}

void custom_list_base::on_selected(wxListEvent&)
{
	// noop
}

void custom_list_base::on_deselected(wxListEvent&)
{
	// noop
}

void custom_list_base::on_activated(wxListEvent&)
{
	// noop
}

void custom_list_base::sort(std::size_t column, bool asc)
{
	sort_.col = column;

	if (asc)
		sort_.dir = 1;
	else
		sort_.dir = -1;

	save_sort();
	sort();
}

void custom_list_base::set_widths()
{
	for (int i=0; i<GetColumnCount(); ++i)
	{
		if (i < static_cast<int>(owidths_.size()))
		{
			int w = 0;
			
			if (owidths_[i] > 0.0 && owidths_[i] < 1.0)
				w = static_cast<int>(owidths_[i] * GetSize().x);
			else
				w = static_cast<int>(owidths_[i]);
				
			// this uses the original widths that were loaded from
			// the options when the list was created
			SetColumnWidth(i, w);
			log_v()
				<< "setting column width "
				<< i << " to " << w << "\n";
		}
		else
		{
			// if the number of widths from the options is too
			// small, it might be because there's no options yet
			// (and this is an empty string) or there's a problem
			// somewhere; just save the widths of the columns so far
			// and bail out
			break;
		}
	}
}

void custom_list_base::save_widths() const
{
	// no option for the widths was specified
	if (option_.empty())
		return;

	// whitespace-separated string of ints
	std::ostringstream oss;
	for (int i=0; i<GetColumnCount(); ++i)
		oss << GetColumnWidth(i) << " ";

	opt(option_ + ".widths", utf8_string(oss.str()));
}

void custom_list_base::save_sort()
{
	if (option_.empty())
		return;

	std::ostringstream oss;
	oss << sort_.col << " " << sort_.dir;

	opt(option_ + ".sort", utf8_string(oss.str()));
}

void custom_list_base::on_select_all(wxMenuEvent&)
{
	long f = GetWindowStyle();

	if ((f & wxLC_SINGLE_SEL) != 0)
		return;

	freezer fz(this);
	
	const int c = GetItemCount();
	for (int i=0; i<c; ++i)
		SetItemState(i, wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);
}

utf8_string custom_list_base::col(int r, int c) const
{
	return utf8_string(GetItemText(r, c).ToUTF8().data());
}

void custom_list_base::col(int r, int c, const utf8_string& s)
{
	const wxString wxs = to_wxstring(s);
	if (wxs != GetItemText(r, c))
		SetItem(r, c, wxs);
}

int custom_list_base::single_selection_index() const
{
	return GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
}

void custom_list_base::unselect(int i)
{
	SetItemState(i, 0, wxLIST_STATE_SELECTED);
}

void custom_list_base::select(int i)
{
	SetItemState(i, wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);
}

void custom_list_base::ensure_visible(int i)
{
	EnsureVisible(i);
}

bool custom_list_base::has_selection() const
{
	return (single_selection_index() != -1);
}

}	// namespace baretorrent
