// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_CUSTOM_LIST_INCLUDED
#define BARETORRENT_CUSTOM_LIST_INCLUDED

#include "path.h"
#include "widgets.h"
#include "platform.h"

namespace baretorrent
{

class custom_list_base : public wxListCtrl
{
public:
	typedef sig::signal<void (int, bool)> check_handler;
	check_handler on_checked;

	~custom_list_base();

	// sets the context menu; may be null
	//
	void context(wxMenu* m);

	bool checked(long i);
	void checked(long i, bool b);

	// sorts the list on the given column
	//
	void sort(std::size_t column, bool asc);
	virtual void sort() = 0;

	void use_checkboxes();
	void headers(bool b);

	void add_accelerator(const wxAcceleratorEntry& e);

	void row_color(int i, const wxColour& c);
	void row_font(int i, const wxFont& f);

	void set_image(int i, const wxBitmap& b);

	int append();
	int append(const utf8_string& c);

	utf8_string col(int r, int c) const;
	void col(int r, int c, const utf8_string& s);

	int single_selection_index() const;
	void unselect(int i);

	void select(int i);
	void ensure_visible(int i);
	bool has_selection() const;

protected:
	struct sort_info
	{
		std::size_t col;
		int dir;

		sort_info()
			: col(std::numeric_limits<std::size_t>::max()), dir(1)
		{
		}
	};

	// indices in the image list
	enum image_indices
	{
		unchecked_image = 0,
		checked_image,
		unchecked_disabled_image,
		checked_disabled_image
	};
	

	// the name of the option containing the widths
	std::string option_;

	// the list if widths read from the options
	std::vector<double> owidths_;

	// the sort information read from the options
	sort_info osort_;

	// the context menu, may be null
	wxMenu* context_;

	// this is needed to set the column widths since they may depend
	// on the width of the control
	visible_timer visible_;
	
	// current sort
	sort_info sort_;

	wxImageList* il_;
	int image_indices_[4];
	std::vector<wxAcceleratorEntry> acc_;


	custom_list_base(
		wxWindow* parent, wxWindowID id=-1,
		const std::string& option_name = "",
		long style=0);

	void set_images();
	wxImageList* get_image_list();

	// fired when the list first becomes visible; sets the column
	// widths
	//
	void on_visible();

	// fired when the user clicked a column; sorts the list using this
	// column's sorter
	//
	void on_column_clicked(wxListEvent& e);

	// fired when the user resizes a column
	//
	void on_column_resized(wxListEvent&);

	// fired when the user releases the right mouse button over the
	// list; pops up the context menu, if any
	//
	void on_context_menu(wxContextMenuEvent& e);

	void on_left_down(wxMouseEvent& e);

	// sets the initial column widths based on the options
	//
	void set_widths();

	// constructs a string with the width of all columns and saves it
	// in the options
	//
	void save_widths() const;

	// constructs a string with the sort column and direction and
	// saves it in the options
	//
	void save_sort();

	// accelerator
	//
	void on_select_all(wxMenuEvent&);

	virtual void on_selected(wxListEvent& e);
	virtual void on_deselected(wxListEvent& e);
	virtual void on_activated(wxListEvent& e);
};

// this is a wrapper around a list control, it is templated on the type
// of user data set in each item, handles sorting by clicking on
// columns, saves and loads the width/sort of columns and can popup a
// context menu
//
// the prefix of the option name that has the widths and sort is given
// in the constructor
//
template <class T>
class custom_list : public custom_list_base
{
public:
	typedef sig::signal<void (T*)> handler;

	handler selected;
	handler deselected;
	handler activated;


	// a sorter takes two pointers to the user data and returns -1, 0
	// or 1
	typedef boost::function<int (T*, T*)> sorter;

	// this is a dummy sorter which shouldn't be used on lists where
	// the columns are visible and clickable (unless you really want to
	// sort items by addresses)
	//
	static int default_sort(T* a, T* b)
	{
		if (a < b)
			return -1;
		else if (a > b)
			return 1;
		return 0;
	}


	// widths_option is the name of the option that contains the list
	// of column widths
	//
	custom_list(
		wxWindow* parent, wxWindowID id=-1,
		const std::string& option_name="",
		long style=0) :
			custom_list_base(parent, id, option_name, style)
	{
	}

	// deletes all the items from the list and their associated data
	//
	void clear()
	{
		delete_data();
		DeleteAllItems();
	}

	// deletes the user data from every item
	//
	void delete_data()
	{
		for (int i=0; i<GetItemCount(); ++i)
			delete reinterpret_cast<T*>(data(i));
	}

	using custom_list_base::sort;

	// sorts the list on the last column that was given to the other
	// sort()
	//
	virtual void sort()
	{
		// this happens when no column was selected yet
		if (sort_.col == std::numeric_limits<std::size_t>::max())
			return;

		// remembering selection and focus because the generic
		// implementation resets it
		std::vector<const T*> s = selection();
		const T* f = focused();

		// this will call sort_trampoline() which is charge of calling
		// the sorter that was supplied when the column was added
		SortItems(
			&custom_list::sort_trampoline,
			reinterpret_cast<wxIntPtr>(this));

		// some platforms need to set the arrows on the column headers
		// manually
		platform::set_list_sort_column(this, sort_.col, sort_.dir);

		// setting back the selection and focus
		select(s);
		focus(f);
	}

	// appends a new column with the given localized caption and sorter
	//
	void add_column(const utf8_string& caption, sorter s=default_sort)
	{
		const int i = GetColumnCount();

		AppendColumn(to_wxstring(caption));

		// the right sorter will be used when sort() is called
		sorters_.push_back(s);

		if (i == static_cast<int>(osort_.col))
			sort(osort_.col, (osort_.dir > 0));
	}

	// returns the item data for the given index
	//
	T* data(int i)
	{
		return reinterpret_cast<T*>(GetItemData(i));
	}

	// returns the item data for the given index
	//
	const T* data(int i) const
	{
		return reinterpret_cast<const T*>(GetItemData(i));
	}

	// overwrites the item data for the given index; the previous data
	// is not deleted
	//
	void data(int i, T* t)
	{
		SetItemPtrData(i, reinterpret_cast<wxUIntPtr>(t));
	}

	int find(T* t)
	{
		const int n = GetItemCount();
		for (int i=0; i<n; ++i)
		{
			if (data(i) == t)
				return i;
		}

		return -1;
	}

	// returns a list of all the items
	//
	std::vector<const T*> items() const
	{
		return find_items(wxLIST_STATE_DONTCARE);
	}
	
	// returns the list of selected items
	//
	std::vector<const T*> selection() const
	{
		return find_items(wxLIST_STATE_SELECTED);
	}

	// returns the focused item, may be null
	//
	const T* focused() const
	{
		std::vector<const T*> v = find_items(wxLIST_STATE_FOCUSED);
		if (!v.empty())
			return v[0];

		return nullptr;
	}

	// selects all the given items
	//
	void select(const std::vector<const T*>& v)
	{
		set_items(v, wxLIST_STATE_SELECTED);
	}

	using custom_list_base::select;
	
	// focuses the given item
	//
	void focus(const T* i)
	{
		std::vector<const T*> v;
		v.push_back(i);

		set_items(v, wxLIST_STATE_FOCUSED, false);
	}

	// returns the first selected item
	//
	const T* single_selection() const
	{
		const int i = single_selection_index();
		if (i != -1)
			return data(i);

		return nullptr;
	}

	int selection_size() const
	{
		return GetSelectedItemCount();
	}

	virtual void on_selected(wxListEvent& e)
	{
		const int i = e.GetIndex();
		if (i >= 0 && i < GetItemCount())
			fire(selected, data(i));
	}

	virtual void on_deselected(wxListEvent& e)
	{
		const int i = e.GetIndex();
		if (i >= 0 && i < GetItemCount())
			fire(deselected, data(i));
	}

	virtual void on_activated(wxListEvent& e)
	{
		const int i = e.GetIndex();
		if (i >= 0 && i < GetItemCount())
			fire(activated, data(i));
	}

private:
	// one sorter per column
	std::vector<sorter> sorters_;


	// called repeatedly by wxWidgets with the user data from two
	// elements and a pointer that was given when SortItems() was
	// called, which is 'this' (because this is a static function)
	//
	static int wxCALLBACK sort_trampoline(
		wxIntPtr a, wxIntPtr b, wxIntPtr data)
	{
		custom_list* list = reinterpret_cast<custom_list*>(data);

		T* left = reinterpret_cast<T*>(a);
		T* right = reinterpret_cast<T*>(b);

		bt_assert(left);
		bt_assert(right);

		// result of the comparison
		const int c = list->sorters_[list->sort_.col](left, right);

		// inverting the result if the direction is descending
		return list->sort_.dir * c;
	}
	// returns a list of items having the given state
	//
	std::vector<const T*> find_items(int state) const
	{
		std::vector<const T*> v;

		int i = -1;
		for (;;)
		{
			i = GetNextItem(i, wxLIST_NEXT_ALL, state);
			if (i == -1)
				break;

			v.push_back(
				reinterpret_cast<const T*>(GetItemData(i)));
		}

		return v;
	}

	// sets the state for all the given items
	//
	void set_items(
		const std::vector<const T*>& v, int state,
		bool reset=true)
	{
		const int c = GetItemCount();
		for (int i=0; i<c; ++i)
		{
			const T* a = reinterpret_cast<const T*>(GetItemData(i));

			bool found = false;
			for (std::size_t j=0; j<v.size(); ++j)
			{
				const T* b = v[static_cast<std::size_t>(j)];

				if (a == b)
				{
					SetItemState(i, state, state);
					found = true;
					break;
				}
			}

			if (!found && reset)
				SetItemState(i, 0, state);
		}
	}
};

}	// namespace baretorrent

#endif	// BARETORRENT_CUSTOM_LIST_INCLUDED
