// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BAREOTRRENT_DATABASE_DB_DEFS_INCLUDED
#define BAREOTRRENT_DATABASE_DB_DEFS_INCLUDED

#ifndef BARETORRENT_DATABASE_IMPLEMENTATION_INCLUDED
	// this is an internal header file and shouldn't be included
	// directly
	#error include the dbms-header directly, such as "db/sqlite.h"
#endif

// these are the inline definitions of generic.h; most of them are
// just forwarding to the dbms-specific implementation

namespace baretorrent
{
namespace db
{

resultset::resultset()
{
}

resultset::resultset(details::resultset_impl_ptr impl)
	: impl_(impl)
{
}

template <class T>
inline T resultset::get(const std::string& name) const
{
	bt_assert(!empty());
	return impl_->get<T>(impl_->column_index(name));
}

template <class T>
inline T resultset::get(int column) const
{
	bt_assert(!empty());
	return impl_->get<T>(column);
}

inline utf8_string resultset::get_string(const std::string& name) const
{
	bt_assert(!empty());
	return impl_->get_string(impl_->column_index(name));
}

inline utf8_string resultset::get_string(int column) const
{
	bt_assert(!empty());
	return impl_->get_string(column);
}

std::vector<std::string> resultset::names() const
{
	bt_assert(!empty());
	return impl_->names();
}

int resultset::column_index(const std::string& name) const
{
	bt_assert(!empty());
	return impl_->column_index(name);
}

bool resultset::done() const
{
	return empty();
}

void resultset::next()
{
	bt_assert(!empty());
	impl_->next();
}

bool resultset::empty() const
{
	if (impl_.get())
		return impl_->empty();
	return true;
}

resultset::iterator resultset::begin()
{
	return iterator(this);
}

resultset::iterator resultset::end()
{
	return iterator();
}

void resultset::close()
{
	impl_.reset();
}


prepared_statement::prepared_statement(
	details::connection_ptr& con, const utf8_string& sql)
{
	impl_ = details::prepared_statement::create(con, sql);
}

template <class T>
prepared_statement& prepared_statement::bind(const T& t)
{
	impl_->bind(t);
	return *this;
}

prepared_statement& prepared_statement::bind(
	const char* cs, std::size_t len)
{
	impl_->bind(cs, len);
	return *this;
}

prepared_statement& prepared_statement::bind(const utf8_string& s)
{
	impl_->bind(s.c_str(), s.size());
	return *this;
}

resultset prepared_statement::execute()
{
	if (impl_->execute())
		return resultset(details::resultset::create(impl_));
	return resultset();
}

void prepared_statement::reset()
{
	impl_->reset();
}


transaction::transaction(details::connection_ptr con, bool start)
	: tn_(new details::transaction(con))
{
	if (start)
		tn_->begin();
}

void transaction::begin()
{
	tn_->begin();
}

void transaction::commit()
{
	tn_->commit();
}

void transaction::rollback()
{
	tn_->rollback();
}


database::database()
{
}

database::database(details::connection_ptr c)
	: con_(c)
{
}

void database::disconnect()
{
	con_->disconnect();
}

prepared_statement database::prepare(const utf8_string& sql)
{
	return prepared_statement(con_, sql);
}

baretorrent::db::transaction database::transaction(bool start)
{
	return baretorrent::db::transaction(con_, start);
}

void database::execute(const utf8_string& q)
{
	details::execute(*con_, q);
}

resultset database::query(const utf8_string& q)
{
	prepared_statement stmt(con_, q);
	return stmt.execute();
}

primary_key database::last_insert_id()
{
	return details::last_insert_id(con_);
}

utf8_string database::now()
{
	return query(utf8_string("select ") + details::now_string())
		.get<utf8_string>(0);
}

traits::handle database::handle()
{
	return con_->handle();
}


resultset_iterator::resultset_iterator()
	: rs_(0)
{
}

resultset_iterator::resultset_iterator(resultset* rs)
	: rs_(rs)
{
	if (rs_->done())
		rs_ = 0;
}

resultset* resultset_iterator::operator->() const
{
	bt_assert(rs_);
	return rs_;
}

resultset& resultset_iterator::operator*() const
{
	bt_assert(rs_);
	return *rs_;
}

resultset_iterator& resultset_iterator::operator++()
{
	bt_assert(rs_);
	rs_->next();

	if (rs_->done())
		rs_ = 0;

	return *this;
}

inline baretorrent::db::primary_key bad_primary_key()
{
	return details::bad_primary_key();
}

inline utf8_string now_string()
{
	return details::now_string();
}

inline utf8_string timestamp(boost::posix_time::ptime p)
{
	return details::timestamp(p);
}

inline void execute(details::connection& con, const utf8_string& q)
{
	return details::execute(con, q);
}

inline baretorrent::db::resultset query(
	details::connection_ptr con, const utf8_string& sql)
{
	baretorrent::db::prepared_statement s(con, sql);
	return s.execute();
}

inline void initialize()
{
	details::initialize();
}

inline void terminate()
{
	details::terminate();
}


inline bool operator==(
	const resultset_iterator& left, const resultset_iterator& right)
{
	return left.rs_ == right.rs_;
}

inline bool operator!=(
	const resultset_iterator& left, const resultset_iterator& right)
{
	return !(left.rs_ == right.rs_);
}

}   // namespace db
}   // namespace baretorrent

#endif	// BAREOTRRENT_DATABASE_DB_DEFS_INCLUDED
