// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_DATABASE_DB
#define BARETORRENT_DATABASE_DB

#ifndef BARETORRENT_DATABASE_IMPLEMENTATION_INCLUDED
	// this is an internal header file and shouldn't be included
	// directly
	#error include the dbms-header directly, such as "db/sqlite.h"
#endif

#include "../utility.h"
#include "../log.h"

// this is the generic implementation of database features; it
// encapsulates dbms-specific calls
//
// everything is inline because multiple configurations would be needed
// to conditonally compile a specific dbms, resulting in various
// different library files; by putting everything in headers, as long
// as unused dbms's are not included, no conditional compilation is
// needed
//
// the definition of most generic functions is in defs.h (some are here)
// and the dbms-specific stuff is in the dbms headers, such as
// "db/sqlite.h"

// there are five major classes in there: prepared_statement, resultset,
// row, transaction and database
//
//
// prepared_statement
// this is an sql statement with placeholders (question marks) that can
// be filled in at runtime; this is the safest way to construct an
// sql statement since it handles conversions and escapes strings
// correctly
//
// resultset
// this is returned by queries, such as when executing a prepared
// statement, and can be iterated through
//
// row
// this is used when iterating a resultset and has operator[] and
// various get() functions; note that a resultset is also a row, which
// is useful for queries that return a single row
//
// transaction
// a simple class that starts a transaction in the constructor and
// rollbacks in the destructor; calling commit() on it will commit the
// transaction and inhibit the rollback
// 
// database
// represents a database connection created by calling a dbms-specific
// connect() function (see db::mysql::connect() for example) and is
// used to create prepared statements, transactions and queries; note
// that db::initialize() (which is not thread-safe) must be called
// once before any connection is attempted
//
//
// db::initialize() must be called once when starting: it
// initializes the dbms but is not thread safe; db::terminate()
// must be called after all connections are closed
//
// there are also convenient classes for creating where conditons and
// various free functions
//
// prepared statements, transactions and simple queries must be created
// from a database object because they are bound to a specific
// connection
//
// all operations may throw an db::exception and what() will
// return a dbms-specific, non-localized message
//
// everything that is internal is declared in the details namespace
//
// internally, a 'row' is just a typedef for a resultset which handles
// walking through the rows

namespace baretorrent
{
namespace db
{
namespace details
{

// a connection is a dbms-specific class that handles connections
// to dbms's; this shouldn't be used directly as its purpose is to
// experience fear, fear in the face of certain death, to accept that
// fear and maintain control of oneself and one's crew, and also to
// initialize a 'database' object
class connection;
typedef boost::shared_ptr<connection> connection_ptr;

class resultset;
typedef boost::shared_ptr<details::resultset> resultset_impl_ptr;

class prepared_statement;
typedef boost::shared_ptr<details::prepared_statement>
	prepared_statement_impl_ptr;

class transaction;
typedef boost::shared_ptr<transaction> transaction_ptr;

}	// namespace details


// type of a primary key
typedef traits::primary_key primary_key;

// value for a bad primary key; this is usually 0, but may be
// different on some dbms's
//
inline baretorrent::db::primary_key bad_primary_key();

class resultset;

// this is for iterating over a resultset; remember that 'rows' are
// really just a typedef for a resultset
//
class resultset_iterator :
	public std::iterator<std::forward_iterator_tag, resultset>
{
public:
	// an invalid iterator, equivalent to end() in a container; can be
	// used to detect when an iterator is done
	//
	inline resultset_iterator();

	// an iterator to the current row in the resultset
	//
	inline resultset_iterator(resultset* rs);

	// returns the resultset
	//
	inline resultset* operator->() const;

	// returns the resultset
	//
	inline resultset& operator*() const;

	// goes to the next row; invalid if '*this == resultset_iterator()'
	//
	inline resultset_iterator& operator++();

	friend bool operator==(
		const resultset_iterator& left,
		const resultset_iterator& right);
	friend bool operator!=(
		const resultset_iterator& left,
		const resultset_iterator& right);

private:
	resultset* rs_;
};


// has the results of a query and can be iterated to get the values
// of all the rows; note that it is not possible to get the number of
// rows directly
//
// using indices for columns is more efficient than names since names
// need to be mapped to indices before passing it through the dbms;
// for large queries, using names() or column_index() to get the
// indices before a loop so they can be passed to operator[] or get()
// inside the loop will be faster
//
class resultset
{
	friend class prepared_statement;

public:
	typedef resultset_iterator iterator;

	// empty resultset
	//
	inline resultset();

	// returns the value for the given column; make sure T is the
	// right type or this will fail/crash
	//
	template <class T>
	inline T get(const std::string& name) const;

	// returns the value for the given column; make sure T is the
	// right type or this will fail/crash
	//
	template <class T> inline T get(int column) const;

	// returns the value for the given column as a string
	//
	inline utf8_string get_string(const std::string& name) const;

	// returns the value for the given column index as a string
	//
	inline utf8_string get_string(int column) const;

	// returns get_string(i)
	//
	utf8_string operator[](int i)
	{
		return get_string(i);
	}

	// returns get_string(n)
	//
	utf8_string operator[](const std::string& n)
	{
		return get_string(n);
	}

	// returns the column names for this query
	//
	inline std::vector<std::string> names() const;

	// returns the index of the given column for this query
	//
	inline int column_index(const std::string& name) const;

	// whether there are any results left; this is the same as empty(),
	// but empty() sounds better for testing whether the query returned
	// anything
	//
	inline bool done() const;

	// moves to the next row; invalid if done() is true
	//
	inline void next();
	
	// whether there are any results in this query; this is the same
	// as done(), but done() sounds better for testing if there's any
	// results left in the query while iterating it
	//
	inline bool empty() const;

	// iterator to the first row
	//
	inline iterator begin();

	// iterator past the last row
	//
	inline iterator end();

	// closes the resultset immediately; this is not normally needed
	// since the destructor handles it
	//
	inline void close();

private:
	// dbms-specific resultset
	details::resultset_impl_ptr impl_;


	// a resultset for the given query
	//
	inline resultset(details::resultset_impl_ptr impl);

	// returns the value as a string for the given column
	//
	inline utf8_string get_impl(int column) const;
};

// a row is just a fancy name for a resultset because
//		foreach (db::row r, rs)
// makes more sense than
//		foreach (db::resultset r, rs)
//
typedef resultset row;


// a prepared statement
//
// always call reset() before binding new values, even though some
// dbms's do not require it (in which case reset() is a no-op)
//
// bind() must be called as many times as there are placeholders in
// the query and in the same order; there is no way of calling bind
// in a different order than the placeholders
//
// calling bind() with anything else than a string will use
// operator<< to convert the value to a string; wide character strings
// must be converted to utf8 first
//
class prepared_statement
{
	friend class database;

	friend baretorrent::db::resultset query(
		details::connection_ptr con, const utf8_string& sql);

public:
	// converts the given value to a string using operator<< and
	// binds it
	//
	template <class T>
	inline prepared_statement& bind(const T& t);

	// binds the given string
	//
	inline prepared_statement& bind(const char* cs, std::size_t len);

	// binds the given string
	inline prepared_statement& bind(const utf8_string& s);

	// executes this prepared statement and returns a resultset for it
	//
	inline resultset execute();

	// resets the prepared statement so it can be filled and executed
	// again; this might be a no-op on certain dbms's but should
	// always be called just in case
	//
	inline void reset();

private:
	// dbms-specific prepared statement
	details::prepared_statement_impl_ptr impl_;
	
	// creates a statement for the given sql string
	//
	inline prepared_statement(
		details::connection_ptr& con, const utf8_string& sql);
};

// begins a transaction in the constructor and rolls it back in the
// destructor; calling commit() before destroying the transaction will
// commit the changes and inhibit the rollback
//
// transaction objects cannot be reused
//
class transaction
{
	friend class database;

public:
	// begins the transaction; this is invalid if the transaction has
	// already been started
	//
	inline void begin();

	// commits the transaction; this is invalid if the transaction has
	// already been commited or hasn't been started yet
	//
	inline void commit();

	// rolls back the transaction; this is invalid if the transaction
	// has already been commited or hasn't been started yet
	inline void rollback();

private:
	details::transaction_ptr tn_;

	// begins the transaction on the given connection if 'start' is true
	//
	inline transaction(details::connection_ptr con, bool start);
};

// a connection to a dbms; the connection will be lazily made when first
// needed
//
// creating a connection is dbms-specific and should be made using
// a connect() function in the dbms-specific namespace (such as
// baretorrent::db::mysql::connect()) which returns a database object
//
// note that baretorrent::db::initialize() must be called once before
// any call to a dbms-specific connect()
//
class database
{
public:
	// a database with no connection; any operation on it will fail
	//
	inline database();

	// a database with the given connection
	//
	inline database(details::connection_ptr c);
	
	// manually disconnects from the dbms; the database will
	// automatically reconnect later if needed
	//
	inline void disconnect();


	// creates a prepared statement with the given sql string
	//
	inline prepared_statement prepare(const utf8_string& sql);
	
	// creates a transaction; if 'start' is true, the transaction is
	// started immediately
	//
	inline baretorrent::db::transaction transaction(bool start=true);


	// executes the given statement and ignores any resultsets; this
	// might be more efficient than query() on certain dbms's if the
	// resultset is not needed
	//
	inline void execute(const utf8_string& q);

	// executes the given statement and returns a resultset; execute()
	// might be more efficient on certain dbms's if the resultset is
	// not needed
	//
	inline resultset query(const utf8_string& q);

	// returns the id of the last insert operation; this is guaranteed
	// to return bad_primary_key() in case no row has been inserted
	// successfully on this connection
	//
	inline primary_key last_insert_id();

	// returns a dbms-specific timestamp as a string which the dbms
	// could then accept in a subsequent query (such as an update); this
	// is basically doing a select with what now_string() returns
	//
	inline utf8_string now();

	// returns the dbms-specific handle of the connection
	//
	inline traits::handle handle();

private:
	// dbms-specific connection
	details::connection_ptr con_;
};

// iterators are the same if they're pointing to the same resultset;
// when a resultset is empty or when an iterator is incremented past
// the last row, the resultset is reset to null
//
inline bool operator==(
	const resultset_iterator& left, const resultset_iterator& right);

inline bool operator!=(
	const resultset_iterator& left, const resultset_iterator& right);


class where_condition;
typedef boost::shared_ptr<where_condition> where_condition_ptr;
typedef std::vector<where_condition_ptr> where_conditions;

class compound_where;
typedef boost::shared_ptr<compound_where> compound_where_ptr;


// used to build complex where conditions
//
// this is the base class: 'condition' is an sql string that contains
// a single placeholder and 'value' is a value that will be bound to a
// prepared statement
//
// calling sql() will return an sql string suitable for a "where"
// condition (but without the "where")
//
// multiple conditions may be joined by using the compound_condition
// class below
//
//	void f(db::database& db)
//	{
//		db::where_condition_ptr wc =
//			db::where_condition::create("item_id=?", 42);
//
//		db::prepared_statement ps = db.prepare(
//			"select * from items where " + wc->sql());
//		wc->bind(ps);
//
//		db::resultset rs = ps.execute();
//		// ...
//	}
//
class where_condition
{
public:
	// constructs a condition where "condition" contains a placeholder
	// (such as "item_id=?") and "value" is a string that will be
	// bound when bind() is called
	//
	static where_condition_ptr create(
		utf8_string& condition, const utf8_string& value)
	{
		return where_condition_ptr(
			new where_condition(condition, value));
	}

	// constructs a condition where "condition" contains a placeholder
	// (such as "item_id=?") and "value" is converted to a string
	// using operator<< and will be bound when bind() is called
	//
	template <class T>
	static where_condition_ptr create(
		const utf8_string& condition, T value)
	{
		return where_condition_ptr(new where_condition(
			condition, baretorrent::as_utf8_string(value)));
	}

	// nothing
	//
	virtual ~where_condition()
	{
		// noop
	}

	// sql string for this condition (without 'where' or anything)
	//
	virtual utf8_string sql() const
	{
		return condition_;
	}

	// binds the value in the given prepared statement
	//
	virtual void bind(baretorrent::db::prepared_statement& ps) const
	{
		if (!empty())
			ps.bind(value_);
	}

	// whether this condition has anything in it
	//
	virtual bool empty() const
	{
		return (condition_.empty() || value_.empty());
	}

protected:
	// empty condition; this is used by compound_where since it
	// doesn't have a condition itself
	//
	where_condition()
		: condition_(""), value_("")
	{
	}

	// a where condition
	//
	where_condition(
		const utf8_string& condition, const utf8_string& value)
			: condition_(condition), value_(value)
	{
	}

private:
	// the condition (value=?)
	utf8_string condition_;
	
	// the value to bind
	utf8_string value_;
};

// a where condition that contains multiple conditions separated by the
// given separator (such as "and" or "or")
//
//	void f(db::database& db)
//	{
//		db::where_condition_ptr wc1 =
//			db::where_condition::create("item_id=?", 42);
//		where_condition_ptr wc2 =
//			db::where_condition::create("type=?", "3");
//		db::where_condition_ptr wc3 =
//			db::where_condition::create("result=?", 2);
//
//		db:compound_where_ptr inner =
//			db::compound_where::create("and");
//		
//		inner->add(wc1);
//		inner->add(wc2);
//
//		db::compound_where_ptr outer =
//			db::compound_where::create("or");
//
//		outer->add(inner);
//		outer->add(wc3);
//
//		db::prepared_statement ps(
//			"select * from items where " + outer->sql());
//
//		outer->bind(ps);
//
//		// the statement will look like:
//		//
//		// select * from items
//		// where ((item_id=42 and type=3) or (result=2))
//
//		db::resultset rs = ps.execute();
//		// ...
//	}
//
class compound_where : public where_condition
{
public:
	// a compound with the given separator
	//
	static compound_where_ptr create(const utf8_string& sep)
	{
		return compound_where_ptr(new compound_where(sep));
	}

	// implodes all the child conditions with the separator, surrounded
	// by parenthesis, or an empty string if it has no children
	//
	virtual utf8_string sql() const
	{
		if (conditions_.empty())
			return utf8_string("");

		// getting the sql string for all the child conditions
		std::vector<utf8_string> parts;
		foreach(where_condition_ptr p, conditions_)
		{
			if (!p->empty())
				parts.push_back(p->sql());
		}

		if (parts.empty())
			return utf8_string("");


		utf8_string sep = utf8_string(" ") + sep_ + utf8_string(" ");
		std::ostringstream oss;
		oss << "(" << implode(parts, sep).std_string() << ")";

		return utf8_string(oss.str());
	}

	// asks all the child conditions to bind their value; a compound
	// condition doesn't have conditions per se
	//
	virtual void bind(baretorrent::db::prepared_statement& ps) const
	{
		foreach(where_condition_ptr c, conditions_)
			c->bind(ps);
	}

	// adds the given condition to this compound
	//
	void add(where_condition_ptr c)
	{
		conditions_.push_back(c);
	}

	// adds the given condition to this compound; t is converted to a
	// string with operator<<
	//
	template <class T>
	void add(const utf8_string& sql, T t)
	{
		add(where_condition_ptr(new where_condition(sql, t)));
	}

	// whether this compound has any child conditions
	//
	virtual bool empty() const
	{
		return conditions_.empty();
	}

protected:
	// a compound with the given separator
	//
	compound_where(const utf8_string& sep)
		: sep_(sep)
	{
	}

private:
	// the separator
	utf8_string sep_;

	// child conditions
	where_conditions conditions_;
};


// one-time initialization of the dbms; this call is not thread-safe and
// will throw on errors
//
void initialize();

// one-time cleanup of the dbms; this call is not thread-safe and is
// nothrow
void terminate();


// returns a dbms-specific sql string that can be used in a select
// statement to get the current timestamp
//
inline utf8_string now_string();

// returns a dbms-specific timestamp string for the given posix
// time which the dbms could then accept in a subsequent query (such
// as an update)
//
inline utf8_string timestamp(boost::posix_time::ptime p);

// returns the property name for identity (auto increment) columns;
// for example, mysql is "auto_increment" and sql server is "identity"
//
inline utf8_string identity_property();

// returns whether this dbms supports unsigned types
//
inline bool supports_unsigned();


// executes the given query on the database connection and returns a
// resultset for it
//
inline baretorrent::db::resultset query(
	details::connection_ptr con, const utf8_string& sql);

// executes the given query on the database connection and returns a
// resultset for it; this may be more efficient than query() on
// certain dbms's for queries that do not return resultsets
//
void execute(details::connection& con, const utf8_string& q);

}	// namespace db
}	// namespace baretorrent

// these are for BOOST_FOREACH
//
namespace boost
{
	template <>
	struct range_mutable_iterator<baretorrent::db::resultset>
	{
		typedef baretorrent::db::resultset_iterator type;
	};

	template <>
	struct range_const_iterator<baretorrent::db::resultset>
	{
		typedef const baretorrent::db::resultset_iterator type;
	};
}

#endif	// BARETORRENT_DATABASE_DB
