// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_DB_SQLITE_INCLUDED
#define BARETORRENT_DB_SQLITE_INCLUDED

#include "../path.h"

namespace baretorrent
{
namespace db
{

struct traits
{
	// type of a database handle for sqlite
	typedef sqlite3* handle;

	// type of the primary key used in the database
	typedef uint64_t primary_key;
};

}	// namespace db
}	// namespace baretorrent

#define BARETORRENT_DATABASE_IMPLEMENTATION_INCLUDED
#include "generic.h"

namespace baretorrent
{
namespace db
{
namespace details
{

// throws a db::exception with make_error(s)
//
inline void throw_exception(
	details::connection* con, const utf8_string& s);

class connection : boost::noncopyable
{
public:
	connection(const utf8_path& p, bool readonly, bool create) :
		handle_(nullptr), path_(p),
		readonly_(readonly), create_(create)
	{
	}

	~connection()
	{
		try
		{
			disconnect();
		}
		catch(...)
		{
		}
	}

	void disconnect()
	{
		try
		{
			if (is_connected())
			{
				log_v() << "db: analyzing\n";
				execute(*this, utf8_string("analyze"));

				log_v() << "db: vacuuming\n";
				execute(*this, utf8_string("vacuum"));
			}

			if (handle_)
			{
				if (sqlite3_close(handle_) != SQLITE_OK)
				{
					throw_exception(this, utf8_string(
						"couldn't close sqlite handle"));
				}
			}
		}
		catch(db::exception& e)
		{
			log_error()
				<< "sqlite: error while disconnecting: "
				<< e.what() << "\n";
		}
	}

	bool is_connected() const
	{
		return (handle_ != nullptr);
	}

	sqlite3* handle()
	{
		if (!handle_)
			connect();

		return handle_;
	}

	void reconnect()
	{
		disconnect();
		connect();
	}

	void connect()
	{
		log_v() << "sqlite: connecting\n";

		if (!path_.exists() && !create_)
		{
			throw baretorrent::db::exception(
				utf8_string("database file '") + path_.as_utf8_string() +
				utf8_string("' doesn't exist"));
		}

		int flags = 0;
	
		if (create_)
			flags |= SQLITE_OPEN_CREATE;

		if (readonly_)
			flags |= SQLITE_OPEN_READONLY;
		else
			flags |= SQLITE_OPEN_READWRITE;

		int r = sqlite3_open_v2(
			path_.as_utf8_string().c_str(), &handle_, flags, nullptr);

		if (r != SQLITE_OK)
		{
			utf8_string m =
				utf8_string("could not open db '") +
				path_.as_utf8_string() + utf8_string("'");

			details::throw_exception(this, m);
		}
	}

	const utf8_path& path() const
	{
		return path_;
	}

	bool readonly() const
	{
		return readonly_;
	}

	bool create() const
	{
		return create_;
	}

private:
	sqlite3* handle_;
	utf8_path path_;
	bool readonly_;
	bool create_;
};

class resultset;

class prepared_statement : boost::noncopyable
{
	friend class resultset;

public:
	static prepared_statement_impl_ptr create(
		details::connection_ptr con, const utf8_string& sql)
	{
		return prepared_statement_impl_ptr(
			new prepared_statement(con, sql));
	}

	~prepared_statement()
	{
		if (impl_)
			sqlite3_finalize(impl_);
	}

	void bind(const char* cs, std::size_t len)
	{
		int r = sqlite3_bind_text(
			impl_, next_,
			cs, static_cast<int>(len),
			SQLITE_TRANSIENT);

		if (r != SQLITE_OK)
			fail(utf8_string("could not bind string parameter"));

		++next_;
	}

	void bind(const char* cs)
	{
		bind(cs, strlen(cs));
	}

	void bind(const utf8_string& s)
	{
		bind(s.c_str(), s.length());
	}

	void bind(uint64_t i)
	{
		if (sqlite3_bind_int64(
			impl_, next_, static_cast<int64_t>(i)) != SQLITE_OK)
		{
			fail(utf8_string("could not bind size_t parameter"));
		}

		++next_;
	}

	void bind(int64_t i)
	{
		if (sqlite3_bind_int64(impl_, next_, i) != SQLITE_OK)
			fail(utf8_string("could not bind size_t parameter"));

		++next_;
	}

	// std::size_t is a uint64_t on x64 platforms
#if !BARETORRENT_X64
	void bind(std::size_t i)
	{
		bind(static_cast<uint64_t>(i));
	}
#endif

	void bind(bool b)
	{
		bind(static_cast<uint64_t>(b ? 1 : 0));
	}

	void bind(int i)
	{
		bind(static_cast<int64_t>(i));
	}

	void bind(const byte_array& a)
	{
		if (a.empty())
			bind("", 0);
		else
			bind(a.data(), a.size());
	}

	// undefined on purpose; other types are not supported
	template <class T>
	void bind(T);

	bool execute()
	{
		int r = sqlite3_step(impl_);

		if (r != SQLITE_DONE && r != SQLITE_ROW)
		{
			fail(
				utf8_string("could not execute prepared statement '") +
				sql_ + utf8_string("'"));
		}

		if (r != SQLITE_ROW)
			return false;

		return true;
	}

	void reset()
	{
		next_ = 1;

		if (impl_)
		{
			if (sqlite3_clear_bindings(impl_) != SQLITE_OK)
				fail(utf8_string("could not clear bindings"));

			if (sqlite3_reset(impl_) != SQLITE_OK)
				fail(utf8_string("could not reset prepared statement"));
		}
	}

	sqlite3_stmt* handle()
	{
		return impl_;
	}

private:
	// database connection
	details::connection_ptr con_;

	// sql string
	utf8_string sql_;

	// sqlite statement handle
	sqlite3_stmt* impl_;

	// next bind index
	int next_;


	prepared_statement(
		details::connection_ptr con, const utf8_string& sql) :
			con_(con),
			sql_(sql),
			impl_(nullptr),
			next_(1)
	{
		int r = sqlite3_prepare_v2(
			con->handle(),
			sql.c_str(), static_cast<int>(sql.length() + 1),
			&impl_, 0);

		if (r != SQLITE_OK)
		{
			fail(
				utf8_string("could not prepare statement '") +
				sql + utf8_string("'"));
		}
	}


	// called from the resultset
	//
	bool next()
	{
		bt_assert(impl_);

		int r = sqlite3_step(impl_);

		if (r != SQLITE_DONE && r != SQLITE_ROW)
			fail(utf8_string("could not get next row"));

		if (r == SQLITE_DONE)
		{
			reset();
			return false;
		}

		return true;
	}

	void fail(const utf8_string& s)
	{
		throw_exception(con_.get(), s);
	}
};

class resultset
{
public:
	static resultset_impl_ptr create(prepared_statement_impl_ptr s)
	{
		return resultset_impl_ptr(new resultset(s));
	}

	utf8_string get_string(int i)
	{
		bt_assert(impl_);

		const std::size_t size = static_cast<std::size_t>(
			sqlite3_column_bytes(impl_->handle(), i));

		const char* r = reinterpret_cast<const char*>(
			sqlite3_column_text(impl_->handle(), i));

		if (!r)
			fail(utf8_string("could not get column text"));

		return utf8_string(r, size);
	}

	// undefined on purpose; other types are not supported
	template <class T>
	T get(int i);

	void next()
	{
		bt_assert(impl_.get());
		bt_assert(impl_->handle());

		int r = sqlite3_step(impl_->handle());

		if (r != SQLITE_DONE && r != SQLITE_ROW)
			fail(utf8_string("could not get next row"));

		if (r == SQLITE_DONE)
			impl_.reset();
	}

	bool empty() const
	{
		if (impl_.get() && impl_->handle())
			return false;
		return true;
	}

	int column_index(const std::string& name)
	{
		int n = -1;
		for (std::size_t i=0; i<names_.size(); ++i)
		{
			if (names_[i] == name)
			{
				n = static_cast<int>(i);
				break;
			}
		}

		bt_assert(n != -1);
		return n;
	}

	std::vector<std::string> names() const
	{
		return names_;
	}

private:
	prepared_statement_impl_ptr impl_;
	std::vector<std::string> names_;


	resultset(prepared_statement_impl_ptr s)
		: impl_(s)
	{
		if (!empty())
		{
			const int cols = sqlite3_column_count(impl_->handle());
			for (int i=0; i<cols; ++i)
			{
				const char* s = sqlite3_column_name(impl_->handle(), i);
				if (!s)
					fail(utf8_string("could not get column name"));

				names_.push_back(s);
			}
		}
	}

	void fail(const utf8_string& s)
	{
		throw_exception(impl_->con_.get(), s);
	}
};

template <>
inline uint64_t resultset::get<uint64_t>(int i)
{
	bt_assert(impl_.get());
	bt_assert(impl_->handle());

	return sqlite3_column_int64(impl_->handle(), i);
}

template <>
inline int64_t resultset::get<int64_t>(int i)
{
	bt_assert(impl_.get());
	bt_assert(impl_->handle());

	return sqlite3_column_int64(impl_->handle(), i);
}

template <>
inline bool resultset::get<bool>(int i)
{
	bt_assert(impl_.get());
	bt_assert(impl_->handle());

	return (sqlite3_column_int(impl_->handle(), i) == 0 ? false : true);
	
}

template <>
inline byte_array resultset::get<byte_array>(int i)
{
	utf8_string s = get_string(i);
	return byte_array(
		reinterpret_cast<const unsigned char*>(s.c_str()),
		s.size_in_bytes());
}


class transaction
{
public:
	transaction(connection_ptr c)
		: con_(c), done_(false), started_(false)
	{
	}

	~transaction()
	{
		try
		{
			if (!done_ && started_)
				rollback();
		}
		catch(db::exception& e)
		{
			log_error()
				<< "couldn't rollback transaction: "
				<< e.what() << "\n";
		}
		catch(...)
		{
		}
	}

	void begin()
	{
		bt_assert(!started_);
		execute(*con_, utf8_string("begin transaction"));
		started_ = true;
	}

	void rollback()
	{
		bt_assert(!done_);
		bt_assert(started_);

		execute(*con_, utf8_string("rollback transaction"));
		done_ = true;
	}

	void commit()
	{
		bt_assert(!done_);
		bt_assert(started_);

		execute(*con_, utf8_string("commit transaction"));
		done_ = true;
	}

private:
	// connection to which this transaction belongs
	connection_ptr con_;

	// whether the transaction has been committed or rolled back
	bool done_;

	// whether the transaction has been started
	bool started_;
};

// initializes the sqlite library
//
inline void initialize()
{
	if (sqlite3_initialize() != SQLITE_OK)
	{
		throw_exception(nullptr, utf8_string(
			"could not initialize sqlite"));
	}
}

// cleans up sqlite
//
inline void terminate()
{
	sqlite3_shutdown();
}

// returns an sqlite error for the given connection (if any) with the
// prefix
//
inline utf8_string make_error(
	details::connection* c, const utf8_string& prefix)
{
	// no connection, just return the prefix
	if (!c || !c->is_connected())
		return prefix;

	std::ostringstream oss;

	oss
		<< prefix << ": " << utf8_string(sqlite3_errmsg(c->handle()))
		<< " (" << sqlite3_errcode(c->handle()) <<  ")";

	return utf8_string(oss.str());
}

// throws a baretorrent::db::exception with make_error(s)
//
inline void throw_exception(
	details::connection* con, const utf8_string& s)
{
	const utf8_string& e = make_error(con, s);
	throw db::exception(e);
}

// executes the query on the given connection; sqlite ignores resultsets
// when no callback is given to sqlite3_exec
//
inline void execute(details::connection& c, const utf8_string& q)
{
	log_vv() << "executing query '" << q << "'\n";

	char* e = nullptr;
	int r = sqlite3_exec(c.handle(), q.c_str(), nullptr, nullptr,  &e);

	if (r != SQLITE_OK)
	{
		std::ostringstream oss;
		oss << "could not execute statement '" << q << "'";
		
		if (e)
		{
			oss << "(" << utf8_string(e) << ")";
			sqlite3_free(e);
		}

		throw_exception(&c, utf8_string(oss.str()));
	}
}

inline primary_key last_insert_id(details::connection_ptr c)
{
	return static_cast<primary_key>(
		sqlite3_last_insert_rowid(c->handle()));
}

// this in a select statement will return a timestamp for the
// current time
//
inline utf8_string now_string()
{
	return utf8_string("datetime('now', 'localtime')");
}

// converts the given time to a a timestamp string that sqlite
// understands
//
inline utf8_string timestamp(boost::posix_time::ptime p)
{
	using namespace boost::posix_time;

	time_facet* facet = new time_facet("%Y-%m-%d %H:%M:%S");

	std::ostringstream oss;
	oss.imbue(std::locale(oss.getloc(), facet));
  
	oss << p;

	return utf8_string(oss.str());
}

// 0 is a bad primary key on sqlite
//
inline db::primary_key bad_primary_key()
{
	return 0;
}

}   // namespace details

namespace sqlite
{

inline database connect(const utf8_path& p, bool readonly, bool create)
{
	return database(details::connection_ptr(new details::connection(
		p, readonly, create)));
}


}   // namespace sqlite

}   // namespace db
}   // namespace baretorrent

#define BARETORRENT_DB_IMPL_INCLUDED
#include "defs.h"

#endif    // BARETORRENT_DB_SQLITE_INCLUDED
