// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "../pch.h"
#include "ext.h"
#include "converters.h"
#include "exporters.h"
#include "../log.h"
#include "../locale.h"
#include "../utility.h"
#include "../options.h"
#include "../widgets.h"
#include "../archiver.h"
#include "../platform.h"

namespace baretorrent
{
namespace ext
{

static instance_list g_instances;
static std::set<utf8_string> g_valid_events;

// name of the main file in the extensions directory
const utf8_string main_file("main.py");

// name of the index file in the extensions directory
const utf8_string index_file("btx.txt");

// name of the file that contains a list of directories when
// developing extensions
const utf8_string dev_file("dev.txt");


extension_handler g_extension_loaded;

extension_handler& extension_loaded()
{
	return g_extension_loaded;
}


instance::ext_pwd::ext_pwd(const utf8_path& p)
{
	try
	{
		old_ = current_path();

		log_vv() << "setting current directory to " << p << "\n";

		boost::system::error_code ec;
		current_path(p, ec);

		bpl::object os = bpl::import("os");
		bpl::object wp(bpl::handle<>(bpl::borrowed(
			PyUnicode_FromString(p.as_utf8_string().c_str()))));

		os.attr("chdir")(wp);
	}
	catch(bpl::error_already_set&)
	{
		log_error()
			<< "can't change the extension's pwd, "
			<< python_error() << "\n";

		PyErr_Clear();
	}
}

instance::ext_pwd::~ext_pwd()
{
	// because anything below can clear or set the exception, this
	// will reset the original because it might have been the source
	// of the error (if any)
	error_restorer er;

	try
	{
		log_vv()
			<< "resetting current directory back to "
			<< old_ << "\n";

		boost::system::error_code ec;
		current_path(old_, ec);

		bpl::object os = bpl::import("os");
		bpl::object wp(bpl::handle<>(bpl::borrowed(
			PyUnicode_FromString(old_.as_utf8_string().c_str()))));

		os.attr("chdir")(wp);
	}
	catch(bpl::error_already_set&)
	{
		log_error()
			<< "can't change the extension's pwd back to old, "
			<< python_error() << "\n";

		PyErr_Clear();
	}
}


// this will raise in exception in the currently executing script so
// it can be forcefully stopped it it takes too long
//
int stop_script(void*)
{
	PyErr_SetInterrupt();
	return -1;
}

instance::script_checker::script_checker()
	: ok_(false), killed_(false)
{
}

void instance::script_checker::operator()()
{
	/*boost::this_thread::sleep(boost::posix_time::milliseconds(
		opt<int>("bt.ui.extension_time")));

	if (!ok_)
	{
		Py_AddPendingCall(&stop_script, nullptr);
		killed_ = true;
	}
	*/
	ok_ = true;
	delete this;
}

bool instance::script_checker::killed()
{
	return killed_;
}

void instance::script_checker::ok()
{
	ok_ = true;
}


void add_event(const utf8_string& n)
{
	g_valid_events.insert(n);
}

bool valid_event(const utf8_string& n)
{
	return (g_valid_events.find(n) != g_valid_events.end());
}

// converts the given guid to a directory name (strips the hyphens)
//
utf8_string guid_to_dir(const std::string& guid)
{
	return utf8_string(boost::replace_all_copy(guid, "-", ""));
}

// removes the given extension from the list
//
void remove_from_list(instance* i)
{
	bt_assert(i);

	for (
		instance_list::iterator itor=g_instances.begin();
		itor!=g_instances.end();
		++itor)
	{
		if (itor->get() == i)
		{
			g_instances.erase(itor);
			return;
		}
	}

	log_error()
		<< "removing extension '" << i->guid()
		<< "': instance not found in the list\n";
}


instance::instance(const utf8_path& p, bool dev) :
	path_(p), enabled_(false), dev_(dev), bad_(false)
{
	log() << p.stem() << ": loading\n";
	
	read_index();
}

instance::instance(const extension_info& ei) :
	enabled_(false), dev_(false), bad_(false)
{
	info_ = ei;
}

instance::~instance()
{
}

instance_ptr instance::create(const utf8_path& p, bool dev)
{
	instance_ptr i(new instance(p, dev));

	if (i->can_enable())
		i->enabled_ = ext_enabled(i->guid());
	
	return i;
}

bool instance::do_extract(instance_ptr& p, const utf8_path& btx)
{
	utf8_path tmp;

	try
	{
		tmp = temp_directory_path() / unique_path();

		log() << "extracting extension " << btx << " in " << tmp << "\n";

		create_directories(tmp);
		archiver::extract(btx, tmp);

		if (!p)
			p.reset(new instance(tmp, false));
		else
			p->path_ = tmp;

		p->enabled_ = true;

		return true;
	}
	catch(archiver::exception&)
	{
		if (tmp.exists())
			tmp.remove_all();
	
		log_error() << "extension extraction failed\n";
		return false;
	}
	catch(path_exception& e)
	{
		log_error() << "bad extracted extension, " << e.what() << "\n";
		return false;
	}
	catch(ext::exception& e)
	{
		log_error() << "bad extracted extension, " << e.what() << "\n";
		return false;
	}
}

instance_ptr instance::extract(const utf8_path& btx)
{
	instance_ptr p;
	
	if (!do_extract(p, btx))
		return instance_ptr();
	else
		return p;
}

instance_ptr instance::temp(const extension_info& ei)
{
	instance_ptr i(new instance(ei));
	i->enabled_ = true;

	return i;
}

bool instance::download()
{
	boost::format f(opt("bt.ext.fetch_url").std_string());
	const utf8_string url(str(f % info_.guid));

	byte_array file = get_remote_file(
		url, boost::bind(&instance::download_cb, this, _1));

	const utf8_path p = temp_directory_path() / unique_path();

	{
		std::ofstream out(p.as_utf8_string().c_str(), std::ios::binary);
		out.write(file.data(), file.size());	
	}

	instance_ptr me(shared_from_this());
	const bool b = do_extract(me, p);

	boost::system::error_code ec;
	p.remove(ec);

	if (ec)
	{
		log_error()
			<< "couldn't remove downloded file '" << p << "', "
			<< ec.message() << "; ignoring\n";
	}

	if (!b)
		return false;

	{
		boost::mutex::scoped_lock lock(dl_.mutex);
		dl_.done = true;
		dl_.progress = 1.0f;
	}

	return true;
}

bool instance::download_cb(double p)
{
	bool s = false;

	{
		boost::mutex::scoped_lock lock(dl_.mutex);
		dl_.progress = p;
		s = !dl_.stop;
	}

	return s;
}

double instance::download_progress() const
{
	return dl_.progress;
}

void instance::read_index()
{
	const utf8_path index = path_ / index_file;

	if (!index.exists())
	{
		utf8_string s =
			utf8_string("missing ") + index_file + 
			utf8_string(" in extensions directory");

		log_error() << s << "\n";
		throw ext::exception(s);
	}

	log_v() << "reading extension index file " << index << "\n";

	std::ifstream in(index.native().c_str());

	if (!in)
	{
		const utf8_string s =
			utf8_string("could not open index file ") +
			index.as_utf8_string();

		log_error() << s << "\n";
		throw ext::exception(s);
	}
	
	utf8_string icon_filename, small_icon_filename;

	int n = 0;
	for (;;)
	{
		++n;

		std::string line_temp;
		std::getline(in, line_temp);
		if (!in)
			break;

		utf8_string line(line_temp);

		boost::trim(line);
		if (line.empty() || line[0] == '#')
			continue;

		std::size_t p = line.find(utf8_string("="));
		if (p == utf8_string::npos)
		{
			log_error() << "invalid line " << n << "\n";
			continue;
		}

		utf8_string name(line.substr(0, p));
		boost::trim(name);

		utf8_string value(line.substr(p + 1));
		boost::trim(value);

		if (name.empty() || value.empty())
		{
			log_error() << "invalid line " << n << "\n";
			continue;
		}

		if (name == utf8_string("guid"))
			info_.guid = value.std_string();
		else if (name == utf8_string("name"))
			info_.name = value;
		else if (name == utf8_string("author"))
			info_.author = value;
		else if (name == utf8_string("short_description"))
			info_.short_description = value;
		else if (name == utf8_string("long_description"))
			info_.long_description = value;
		else if (name == utf8_string("icon"))
			icon_filename = value;
		else if (name == utf8_string("small_icon"))
			small_icon_filename = value;
		else if (name == utf8_string("email"))
			info_.email = value;
		else if (name == utf8_string("website"))
			info_.website = value;
		else if (name == utf8_string("version"))
			info_.version = extension_version::parse(value);
		else if (name == utf8_string("requires"))
			info_.requires = baretorrent_version::parse(value.std_string());
	}

	if (icon_filename.empty())
	{
		// try icon.png in the extension directory
		const utf8_path s(utf8_string("icon.png"));
		if ((path_ / s).exists())
			icon_filename = s.as_utf8_string();
	}
	
	if (small_icon_filename.empty())
	{
		// try icon.png in the extension directory
		const utf8_path s(utf8_string("icon_16.png"));
		if ((path_ / s).exists())
			small_icon_filename = s.as_utf8_string();
	}

	// attempting to create a bitmap icon from the specified file
	if (!icon_filename.empty())
	{
		info_.icon_data = read_file(path_ / icon_filename);
		info_.icon = create_bitmap(info_.icon_data);

		if (!info_.icon.IsOk())
			error(utf8_string("bad icon"));
	}

	if (!small_icon_filename.empty())
	{
		info_.small_icon_data = read_file(path_ / small_icon_filename);
		info_.small_icon = create_bitmap(info_.small_icon_data);

		if (!info_.small_icon.IsOk())
			error(utf8_string("bad small icon"));
	}

	try
	{
		if (info_.guid.empty())
		{
			const utf8_string s("guid wasn't specified");
			log_error() << s << "\n";

			throw ext::exception(s);
		}

		if (info_.name.empty())
		{
			const utf8_string s("name wasn't specified");
			log_error() << s << "\n";

			throw ext::exception(s);
		}

		log_vv() << path_.stem() << ": checking guid\n";
		boost::uuids::string_generator gen;
		boost::uuids::uuid u = gen(guid());
		u = u;

		log() << dump();
	}
	catch(std::runtime_error&)
	{
		std::ostringstream oss;
		oss  << "invalid guid '" << guid() << "'";

		log_error() << oss.str() << "\n";

		unload();
		throw ext::exception(utf8_string(oss.str()));
	}
}

void instance::import(bool rd)
{
	const utf8_path main = path_ / main_file;
	if (!main.exists())
	{
		utf8_string s = 
			utf8_string("missing ") + main_file + 
			utf8_string(" in extensions directory");

		log_error() << s << "\n";
		throw ext::exception(s);
	}

	python_lock lock;

	try
	{
		// changes the current directory to the extension's directory so
		// it can find files without specifying paths
		ext_pwd pwd(path_);

		const utf8_path main = path_ / main_file;

		log_vv()
			<< path_.stem() << ": importing module "
			<< main.stem() << "\n";

		const utf8_string ep = path_.parent().as_utf8_string();
		
		// dev extensions use the leaf of the path as the module name
		// because it could be anything; regular extensions use the
		// guid as the module name

		utf8_string mod_name;
		if (dev_)
			mod_name = path_.leaf();
		else
			mod_name = guid_to_dir(guid());

		const utf8_string mod_file =
			mod_name + utf8_string(".") + main.stem();

		// adds the extension's directory to the path so the 'main'
		// module can be found
		scoped_ext_path sp(ep);

		if (rd && module_)
		{
			module_ = bpl::object(bpl::handle<>(
				PyImport_ReloadModule(module_.ptr())));
		}
		else
		{
			module_ = bpl::import(mod_file.c_str());
		}

		reload_already_imported(mod_name);

		baretorrent::fire(loaded);

		log_vv(shared_from_this()) << "loaded successfully\n";
	}
	catch(db::exception& e)
	{
		utf8_string s =
			utf8_string("couldn't import extension, ") + e.what();

		log_error() << s << "\n";

		unload();
		throw ext::exception(s);
	}
}

void instance::reload_already_imported(const utf8_string& mod_name)
{
	try
	{
		bpl::object dict = module_.attr("__dict__");

		bpl::exec(
			"import sys\n"
			"import imp\n"
			"import traceback",
			dict, dict);

		bpl::object sys = dict["sys"];
		if (!sys)
		{
			log_error(shared_from_this())
				<< "can't reload other modules, dict has no sys\n";
			return;
		}

		bpl::object modules_o = sys.attr("modules");
		if (!modules_o)
		{
			log_error(shared_from_this())
				<< "can't reload other modules, sys has no modules\n";
			return;
		}

		// sys.modules to dict
		bpl::dict modules = bpl::extract<bpl::dict>(modules_o);

		bpl::list keys = modules.keys();
		if (!keys)
		{
			log_error(shared_from_this())
				<< "can't reload other modules, sys.module.keys() "
				<< "failed\n";
			return;
		}

		const bpl::ssize_t count = bpl::len(keys);
		const utf8_string prefix = mod_name + utf8_string(".");
		const utf8_string main_name = prefix + utf8_string("main");

		for (bpl::ssize_t i=0; i<count; ++i)
		{
			utf8_string k;

			try
			{
				bpl::object o = keys[i];
				k = extract_string(o);

				// main is already reloaded
				if (k == main_name)
					continue;

				// not a module from the extension
				if (!boost::starts_with(k, mod_name))
					continue;

				log(shared_from_this())
					<< "reloading " << k << "\n";

				bpl::object m = modules[o];
				if (!m)
				{
					log_error(shared_from_this())
						<< "can't reload module " << k << ", not "
						<< "found in sys.modules (weird)\n";
					continue;
				}

				PyImport_ReloadModule(m.ptr());
			}
			catch(bpl::error_already_set&)
			{
				log_error(shared_from_this())
					<< "can't reload module " << k << ", "
					<< python_error() << "\n";

				PyErr_Clear();
			}
		}
	}
	catch(bpl::error_already_set&)
	{
		log_error(shared_from_this())
			<< "failed to reload modules, "
			<< python_error() << "\n";

		PyErr_Clear();
	}
}

void instance::load_from(const utf8_path& p)
{
	path_ = p;
	load(false);
}

void instance::load(bool rd)
{
	try
	{
		bt_assert(rd || !module_);

		if (!enabled_)
		{
			log(shared_from_this()) << "disabled\n";
			return;
		}

		bad_ = false;
		import(rd);
		execute(utf8_string("init"));

		baretorrent::fire(changed);
	}
	catch(bpl::error_already_set&)
	{
		const utf8_string s = python_error();

		python_lock lock;
		PyErr_Clear();

		log_error(shared_from_this())
			<< "couldn't load, " << s << "\n";

		throw ext::exception(s);
	}
	catch(path_exception& e)
	{
		const utf8_string s = e.what();
		
		log_error(shared_from_this())
			<< "couldn't load, " << s << "\n";

		throw ext::exception(s);
	}
	catch(db::exception& e)
	{
		const utf8_string s = e.what();

		log_error(shared_from_this())
			<< "couldn't load, " << s << "\n";

		throw ext::exception(s);
	}
}

void instance::cleanup()
{
	execute(utf8_string("cleanup"));
}

void instance::reload()
{
	try
	{
		cleanup();
		read_index();
		load(true);
	}
	catch(bpl::error_already_set&)
	{
		log_error()
			<< "could not reload module, " << python_error() << "\n";

		PyErr_Clear();
	}
}

void instance::unload()
{
	if (module_)
		cleanup();

	module_ = bpl::object();
	events_.clear();
}

utf8_string instance::prefix() const
{
	return utf8_string("[extension ") + name() + utf8_string("]: ");
}

bool instance::misbehaving() const
{
	if (bad_ || !enabled_)
		return false;

	return false;
}

void instance::accept_bad_behaviour()
{
	bad_ = true;
}

std::string instance::guid() const
{
	return info_.guid;
}

utf8_path instance::path() const
{
	return path_;
}

void instance::uninstall()
{
	log()
		<< "uninstalling extension " << info_.name
		<< " (" << info_.guid << ")\n";

	try
	{
		// start by disabling it in case there are problems below
		enabled(false);
	}
	catch(ext::exception& e)
	{
		log_error()
			<< "error while trying to disable before uninstalling, "
			<< e.what() << "; ignoring\n";
	}

	try
	{
		remove_extension(guid());
	}
	catch(db::exception& e)
	{
		log_error()
			<< "database error while uninstalling: "
			<< e.what() << "; will still try removing the files\n";
	}

	log() << "removing directory " << path_ << "\n";
	
	boost::system::error_code ec;
	path_.remove_all(ec);

	if (ec)
	{
		log_error()
			<< "could not delete files, " << ec.message() << "\n";
	}

	remove_from_list(this);

	log() << "extension " <<  info_.name << " uninstalled\n";
	baretorrent::fire(uninstalled);
}

bool instance::installed() const
{
	return ext_installed(guid());
}

bool instance::can_enable() const
{
	if (info_.requires > baretorrent::version())
		return false;

	return true;
}

bool instance::can_update() const
{
	if (!installed())
		return false;

	instance_ptr other = find(guid());

	// this should have been picked up by installed()
	bt_assert(other.get());
	if (!other)
		return false;

	const extension_version self = info_.version;
	const extension_version old = other->info_.version;

	if (self.major > old.major)
		return true;

	if (self.major == old.major)
	{
		if (self.minor > old.minor)
			return true;

		if (self.minor == old.minor)
		{
			if (self.maintenance > old.maintenance)
				return true;
		}
	}

	return false;
}

bool instance::can_uninstall() const
{
	if (!installed() || dev_)
		return false;

	return true;
}

void instance::enabled(bool b)
{
	if (enabled_ == b)
		return;

	try
	{
		if (b)
		{
			log(shared_from_this()) << "enabling\n";

			enabled_ = true;
			load(false);
		}
		else
		{
			log(shared_from_this()) << "disabling\n";
			
			unload();
			enabled_ = false;
		}

		ext_enabled(guid(), b);
		baretorrent::fire(changed);
	}
	catch(...)
	{
		enabled_ = false;
		ext_enabled(guid(), false);
		throw;
	}
}

bool instance::enabled() const
{
	return enabled_;
}

bool instance::dev() const
{
	return dev_;
}

utf8_string instance::name() const
{
	if (info_.name.empty())
		return utf8_string(guid());

	return info_.name;
}

utf8_string instance::author() const
{
	return info_.author;
}

utf8_string instance::version() const
{
	std::ostringstream oss;

	oss << info_.version.major << "." << info_.version.minor;

	if (info_.version.maintenance != 0)
		oss << "." << info_.version.maintenance;

	return utf8_string(oss.str());
}

wxBitmap instance::icon() const
{
	return info_.icon;
}

byte_array instance::icon_data() const
{
	return info_.icon_data;
}

wxBitmap instance::small_icon() const
{
	return info_.small_icon;
}

byte_array instance::small_icon_data() const
{
	return info_.small_icon_data;
}

utf8_string instance::short_description() const
{
	return info_.short_description;
}

utf8_string instance::long_description() const
{
	return info_.long_description;
}

utf8_string instance::email() const
{
	utf8_string s = info_.email;
	if (!boost::starts_with(s, "mailto:"))
		s = utf8_string("mailto:") + s;

	return s;
}

utf8_string instance::website() const
{
	return info_.website;
}

baretorrent_version instance::requires() const
{
	return info_.requires;
}

utf8_string instance::dump() const
{
	std::ostringstream oss;

	oss
		<< prefix() << "guid:    " << info_.guid << "\n"
		<< prefix() << "requires:" << info_.requires << "\n"
		<< prefix() << "name:    " << info_.name << "\n"
		<< prefix() << "author:  " << info_.author << "\n"
		<< prefix() << "version: "
			<< info_.version.major << "." << info_.version.minor << "."
			<< info_.version.maintenance << "\n"
		<< prefix() << "desc:    " << info_.short_description << "\n"
		<< prefix() << "ldesc:   " << info_.long_description << "\n"
		<< prefix() << "email:   " << info_.email << "\n"
		<< prefix() << "website: " << info_.website << "\n";

	return utf8_string(oss.str());
}

void instance::dump_options() const
{
	log()
		<< "\ndumping extension '" << name()
		<< " (" << guid() << " configuration:\n";

	const option_list m = get_ext_options(guid());
	std::pair<std::string, option> o;

	foreach (o, m)
	{
		log()
			<< "  " << o.first << " = " << o.second.value()
			<< " (" << o.second.default_value() << ")\n";
	}
	
	log() << "\n";
}

bool instance::has_attr(const utf8_string& name) const
{
	bt_assert(enabled_);

	python_lock lock;
	return ext::has_attr(module_, name);
}

bool instance::unregister_event(
	const utf8_string& n, const utf8_string& f)
{
	bt_assert(enabled_);

	if (!valid_event(n))
	{
		// that event has never been added by anything
		error(utf8_string("invalid event ") + n);
		return false;
	}

	if (n.empty())
	{
		error(utf8_string(
			"unregister_event(): event name cannot be empty"));
		return false;
	}

	if (f.empty())
	{
		error(utf8_string(
			"unregister_event(): function name cannot be empty"));
		return false;
	}

	if (!has_attr(f))
	{
		error(
			utf8_string("unregister_event(): function ") + f +
			utf8_string(" doesn't exist in the module"));
		return false;
	}

	event e(n, f);
	events::iterator itor =
		std::find(events_.begin(), events_.end(), e);

	if (itor == events_.end())
	{
		log(shared_from_this())
			<< "function '" << f << "' for event '" << n << "' "
			<< "was never registered, ignoring\n";

		return true;
	}
	
	log(shared_from_this())
		<< "unregistering function '" << f << "' for event '"
		<< n << "'\n";

	events_.erase(itor);
	return true;
}

bool instance::register_event(
	const utf8_string& n, const utf8_string& f)
{
	bt_assert(enabled_);

	if (!valid_event(n))
	{
		// that event has never been added by anything
		error(utf8_string("invalid event ") + n);
		return false;
	}

	if (n.empty())
	{
		error(utf8_string(
			"register_event(): event name cannot be empty"));
		return false;
	}

	if (f.empty())
	{
		error(utf8_string(
			"register_event(): function name cannot be empty"));
		return false;
	}

	if (!has_attr(f))
	{
		error(
			utf8_string("register_event(): function ") + f +
			utf8_string(" doesn't exist in the module"));
		return false;
	}

	event e(n, f);

	if (std::find(events_.begin(), events_.end(), e) != events_.end())
	{
		log(shared_from_this())
			<< "function '" << f << "' for event '" << n << "' "
			<< "already registered, ignoring\n";

		return true;
	}
	
	log(shared_from_this())
		<< "registering function '" << f << "' for event '"
		<< n << "'\n";

	events_.push_back(e);

	return true;
}

void instance::error(utf8_string w) const
{
	python_lock lock;

	if (w.empty())
		w = python_error();

	PyErr_Clear();
	log_error(shared_from_this()) << w << "\n";
}

bpl::object get_class_object(
	bpl::object module, const utf8_string& name)
{
	if (!has_attr(module, utf8_string("__dict__")))
	{
		log_error()
			<< "instantiating '" << name << "': "
			<< "module has no __dict__\n";

		return bpl::object();
	}

	bpl::object ns = module.attr("__dict__");
	if (!ns)
	{
		log_error()
			<< "instantiating '" << name << "': "
			<< "couldn't get __dict__ object\n";
		
		return bpl::object();
	}

	if (!has_key(ns, name))
	{
		log_error()
			<< "instantiating '" << name << "': "
			<< "__dict__ doesn't have that key\n";

		return bpl::object();
	}

	bpl::object o = ns[name];
	if (!o)
	{
		log_error()
			<< "instantiating '" << name << "': "
			<< "getting class object failed\n";

		return bpl::object();
	}

	return o;
}

bpl::object check_instantiate(bpl::object o, const utf8_string& name)
{
	if (!o)
	{
		log_error()
			<< "instantiating '" << name << "': "
			<< "returned an empty object\n";
	}

	return o;
}

bpl::object instance::create_object(const utf8_string& name)
{
	bpl::object c = get_class_object(module_, name);
	if (!c)
		return c;

	ext_pwd pwd(path_);
	return check_instantiate(c(), name);
}

bpl::object instance::create_object(
	const utf8_string& name, bpl::object a1)
{
	bpl::object c = get_class_object(module_, name);
	if (!c)
		return c;

	ext_pwd pwd(path_);
	return check_instantiate(c(a1), name);
}

bpl::object instance::create_object(
	const utf8_string& name,
	bpl::object a1, bpl::object a2)
{
	bpl::object c = get_class_object(module_, name);
	if (!c)
		return c;

	ext_pwd pwd(path_);
	return check_instantiate(c(a1, a2), name);
}

bpl::object instance::create_object(
	const utf8_string& name,
	bpl::object a1, bpl::object a2, bpl::object a3)
{
	bpl::object c = get_class_object(module_, name);
	if (!c)
		return c;

	ext_pwd pwd(path_);
	return check_instantiate(c(a1, a2, a3), name);
}




instance_list instances()
{
	return g_instances;
}

instance_ptr load_extension(const utf8_path& p, bool dev)
{
	instance_ptr i;

	try
	{
		i = instance::create(p, dev);

		if (find(i->guid()))
		{
			log_error()
				<< "extension " << i->guid()
				<< " already exists, skipping\n";

			return instance_ptr();
		}

		if (std::find(g_instances.begin(), g_instances.end(), i)
				== g_instances.end())
		{
			g_instances.push_back(i);
		}

		i->load();
		baretorrent::fire(g_extension_loaded, i);
	}
	catch(ext::exception&)
	{
		log_error() << "extension exception; skipping\n";

		if (i)
		{
			// when loading the extension fails, it is removed from the
			// list and forgotten, unless this is a dev extension

			if (!dev)
			{
				remove_from_list(i.get());
				i.reset();
			}
		}

		throw;
	}

	return i;
}

void load_dev()
{
	log_v() << "loading dev extensions\n";

	const utf8_path list = paths::extensions() / dev_file;
	if (!list.exists())
	{
		log_v() << "dev list file " << list << " not found\n";
		return;
	}

	std::ifstream in(list.native().c_str());
	if (!in)
	{
		log_error() << "could not open " << list << "\n";
		return;
	}

	for (;;)
	{
		std::string line_temp;
		std::getline(in, line_temp);
		if (!in)
			break;

		utf8_string line(line_temp);

		boost::trim(line);

		if (line.empty() || line[0] == '#')
			continue;

		const utf8_path p(line);

		if (!p.exists())
		{
			log_error()
				<< "directory " << p
				<< " from dev file doesn't exist\n";

			continue;
		}

		log() << "loading dev extension " << p << "\n";

		try
		{
			instance_ptr i = load_extension(p, true);
			add_extension(i.get());
		}
		catch(ext::exception&)
		{
			log_error() << "could not load extension\n";
			continue;
		}
	}
}

// goes through the database and removes extensions that do no exist
// anymore
//
// this is mainly for dev extensions that are added and removed by
// hand from a text file
//
void cleanup()
{
	std::vector<std::string> guids;

	{
		db::resultset rs = get_database()
			.prepare(utf8_string("select guid from extensions"))
			.execute();

		foreach (db::row r, rs)
			guids.push_back(r["guid"].std_string());
	}

	std::vector<std::string> bad_guids;

	foreach (const std::string& guid, guids)
	{
		bool found = false;

		foreach (instance_ptr i, g_instances)
		{
			if (i->guid() == guid)
			{
				found = true;
				break;
			}
		}

		if (!found)
			bad_guids.push_back(guid);
	}

	if (bad_guids.empty())
		return;

	log()
		<< "the following extensions will be removed from the "
		<< "database because they don't exist anymore:\n";

	foreach (const std::string& guid, bad_guids)
	{
		try
		{
			log() << "  . " << guid << "\n";
			remove_extension(guid);
		}
		catch(db::exception& e)
		{
			log_error()
				<< "could not remove extension, "
				<< e.what() << "\n";
		}
	}
}

void load()
{
	log() << "loading extensions from the database\n";
	std::vector<std::string> guids;

	{
		db::resultset rs = get_database()
			.prepare(utf8_string("select guid from extensions"))
			.execute();

		foreach (db::row r, rs)
			guids.push_back(r["guid"].std_string());
	}

	foreach (const std::string& guid, guids)
	{
		const utf8_path p = paths::extensions() / guid_to_dir(guid);

		log() << "loading " << guid << " from " << p << "\n";

		// those might be dev extensions and will be cleaned up
		// afterwards if they're really errors anyways
		if (!p.exists() || !p.is_directory())
			continue;

		try
		{
			load_extension(p, false);
		}
		catch(ext::exception&)
		{
			log() << "uninstalling extension because of errors\n";
			
			remove_extension(guid);
			
			boost::system::error_code ec;
			p.remove_all(ec);
		}
	}

	load_dev();
	cleanup();
}

void unload()
{
	foreach (instance_ptr i, g_instances)
		i->unload();

	// without this, g_instances is cleared outside of main() and will
	// crash since python is dead at that point
	g_instances.clear();
}

void install(instance_ptr e)
{
	const utf8_path target =
		paths::extensions() / guid_to_dir(e->guid());
	
	try
	{
		if (e->can_update())
		{
			log() << "updating extension " << e->guid() << "\n";

			instance_ptr other = find(e->guid());
			if (!other)
			{
				log_error()
					<< "tried to update extension " << e->guid()
					<< " but it's not in the list (weird); aborting\n";

				throw ext::exception();
			}

			remove_from_list(other.get());

			if (!target.exists())
			{
				log_error()
					<< "the extension directory " << target
					<< " does not exist, just making this an install\n";
			}
			else
			{
				target.remove_all();
			}
		}
		else if (!e->installed())
		{
			log() << "installing extension " << e->guid() << "\n";

			if (target.exists())
			{
				log_error()
					<< "the target extension directory " << target
					<< " already exists\n";

				throw ext::exception();
			}
		}
		else
		{
			log_error()
				<< "extension " << e->guid()
				<< " is already installed\n";

				throw ext::exception();
		}

		g_instances.push_back(e);

		log()
			<< "copying extension directory from " << e->path()
			<< " to " << target << "\n";

		//if (!e->download())
		//	throw ext::exception();

		real_copy_directory(e->path(), target);

		const utf8_path init_path = target / utf8_string("__init__.py");

		if (!init_path.exists())
			std::ofstream(init_path.native().c_str());

		e->load_from(target);
		add_extension(e.get());
	}
	catch(...)
	{
		instance_list::iterator itor =
			std::find(g_instances.begin(), g_instances.end(), e);

		if (itor != g_instances.end())
			g_instances.erase(itor);

		log() << "removing directory " << target << "\n";
	
		boost::system::error_code ec;
		target.remove_all(ec);

		if (ec)
		{
			log_error()
				<< "could not delete files, " << ec.message() << "\n";
		}

		throw;
	}
}

instance_ptr find(const std::string& guid)
{
	foreach(instance_ptr i, g_instances)
	{
		if (i->guid() == guid)
			return i;
	}

	return instance_ptr();
}

utf8_path calling_module_path()
{
	try
	{
		bpl::object main = bpl::import("__main__");
		bpl::object global(main.attr("__dict__"));

		bpl::exec("import inspect\n", global, global);

		bpl::object ins = global["inspect"];
		if (!ins)
		{
			log_error()
				<< "calling module: could not get inspect module\n";
			return utf8_path();
		}

		bpl::object st = ins.attr("stack");
		if (!st)
		{
			log_error()
				<< "calling module: could not get stack function\n";
			return utf8_path();
		}

		bpl::object s = st();
		if (!s)
		{
			log_error()
				<< "calling module: inspect.stack() returned null\n";
			return utf8_path();
		}

		bpl::list list(bpl::handle<>(bpl::borrowed(s.ptr())));
		const bpl::ssize_t count = bpl::len(list);

		if (count == 0)
		{
			log_error()
				<< "calling module: inspect.stack() returned "
				<< "an empty list\n";
			return utf8_path();
		}

		bpl::object o = list[0];
		if (!o)
		{
			log_error()
				<< "calling module: inspect.stack() returned "
				<< "a list with null values\n";
			return utf8_path();
		}

		bpl::tuple t(bpl::handle<>(bpl::borrowed(o.ptr())));

		bpl::object po = t[1];
		if (!po)
		{
			log_error()
				<< "calling module: first stack frame has null name\n";
			return utf8_path();
		}

		const utf8_string path = extract_string(po);
		if (path.empty())
		{
			log_error()
				<< "calling module: first stack name is empty\n";
			return utf8_path();
		}

		return utf8_path(path).parent();
	}
	catch(bpl::error_already_set&)
	{
		PyErr_Clear();
		return utf8_path();
	}
}

instance_ptr calling_instance()
{
	const utf8_path p = calling_module_path();
	if (p.empty())
		return instance_ptr();

	instance_list is = instances();

	foreach (instance_ptr i, is)
	{
		if (i->path() == p)
			return i;
	}

	log_error()
		<< "calling instance: no instance with path "
		<< p << "\n";

	return instance_ptr();
}

BARETORRENT_EXPORT bool register_event(
	const utf8_string& name, const utf8_string& function)
{
	instance_ptr i = calling_instance();
	if (!i)
		return false;

	return i->register_event(name, function);
}

BARETORRENT_EXPORT bool unregister_event(
	const utf8_string& name, const utf8_string& function)
{
	instance_ptr i = calling_instance();
	if (!i)
		return false;

	return i->unregister_event(name, function);
}

BARETORRENT_EXPORT void export_python_module()
{
	python_lock lock;

	try
	{
		export_converters();
		export_misc();
		export_log();
		export_torrent();
		export_manager();
		export_config();
		export_utility();
		export_locale();
		export_platform();
		export_ui();

		bpl::def("register", &register_event);
		bpl::def("unregister", &unregister_event);
	}
	catch(bpl::error_already_set&)
	{
		log_error()
			<< "could not inialize baretorrent python module, "
			<< python_error() << "\n";

		throw;
	}
}

}	// namespace ext
}	// namespace baretorrent
