// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_EXT_INCLUDED
#define BARETORRENT_EXT_INCLUDED

#include "python_wrapper.h"
#include "../utility.h"

namespace baretorrent
{

class torrent;

namespace ext
{

class exception
{
public:
	exception(const utf8_string& what = utf8_string())
		: what_(what)
	{
	}

	const utf8_string& what() const
	{
		return what_;
	}

private:
	utf8_string what_;
};

class instance;
typedef boost::shared_ptr<instance> instance_ptr;
typedef std::vector<instance_ptr> instance_list;

typedef sig::signal<void (instance_ptr)> extension_handler;

extension_handler& extension_loaded();


// this represents one extension
//
class instance :
	boost::noncopyable, public boost::enable_shared_from_this<instance>
{
public:
	typedef sig::signal<void ()> handler;

	handler loaded;
	handler changed;
	handler uninstalled;


	// loads the index file from the given path but does not import
	// the module; if 'dev' is true, this comes from the dev file and
	// prevents some operations such as uninstalling
	//
	static instance_ptr create(const utf8_path& p, bool dev);
	
	// extracts the given btx file into a temporary directory and loads
	// the index file from it; the directory will be removed when the
	// instance is destroyed
	//
	// this is used when installing extensions	
	//
	static instance_ptr extract(const utf8_path& btx);

	// sets the fields manually; this is used when searching for
	// extensions
	//
	static instance_ptr temp(const extension_info& ei);


	// deletes the path if temporary
	//
	~instance();
	
	// removes the extension from the database and deletes the files
	//
	void uninstall();


	bool download();
	double download_progress() const;

	// if enabled, this import the extension and calls init() on it
	//
	void load(bool rd=false);

	void load_from(const utf8_path& p);

	// reimports the module
	//
	void reload();

	// calls cleanup() and unloads the extension; load() may be called
	// again afterwards
	//
	void unload();


	// returns whether an extension with this particular guid is
	// already in the database
	//
	bool installed() const;

	// returns whether this extension can be enabled; this returns true
	// even if the extension is currently enabled
	//
	bool can_enable() const;

	// returns whether there already is an extension with this guid
	// installed, but its version is lower
	//
	bool can_update() const;

	// returns whether this extension can be uninstalled; in particular,
	// dev extensions cannot be uninstalled
	//
	bool can_uninstall() const;

	// returns whether the extension is enabled
	//
	bool enabled() const;

	// sets whether the extension is enabled; enabling an extension will
	// import it and call init() on it
	//
	void enabled(bool b);

	// whether this is a dev extension
	//
	bool dev() const;

	// whether this extension has been causing delays
	//
	bool misbehaving() const;

	// whether it's alright for this extension to be misbehaving
	//
	void accept_bad_behaviour();


	// unique id for this instance
	//
	std::string guid() const;

	// returns the extension's name as given by the script
	//
	utf8_string name() const;

	// returns the extension's short description as given by the script
	//
	utf8_string short_description() const;
	
	// returns the extension's long description as given by the script
	//
	utf8_string long_description() const;

	// returns the extension's author as given by the script
	//
	utf8_string author() const;

	// returns the extension's version as given by the script
	//
	utf8_string version() const;

	// returns the extension's icon, may be empty
	//
	wxBitmap icon() const;

	// returns the extension's small icon, may be empty
	//
	wxBitmap small_icon() const;

	// returns the extension icon data so it can be recreated later
	//
	byte_array icon_data() const;

	// returns the small icon data so it can be recreated later
	//
	byte_array small_icon_data() const;

	// returns the extension's email, may be empty
	//
	utf8_string email() const;

	// returns the extension's website, may be empty
	//
	utf8_string website() const;

	// returns the required baretorrent version
	//
	baretorrent_version requires() const;

	// returns the path in which the extension files are
	//
	utf8_path path() const;

	// returns all the infos from the index file
	//
	utf8_string dump() const;

	// logs all the extension's options
	//
	void dump_options() const;


	// registers an event to call the given function when triggered;
	// returns false if the event couldn't be registered for whatever
	// reason
	//
	bool register_event(const utf8_string& e, const utf8_string& f);

	// unregisters an event; returns false if the event couldn't be
	// unregistered for whatever reason
	bool unregister_event(const utf8_string& n, const utf8_string& f);

	// fires the given event and passes it the parameter
	//
	void fire(const utf8_string& n)
	{
		foreach(const event& e, events_)
		{
			if (e.e != n)
				continue;

			if (e.f.empty() || !has_attr(e.f))
				continue;

			execute(e.f);
		}
	}

	// fires the given event and passes it the parameter
	//
	template <class A1>
	void fire(const utf8_string& n, A1 a1)
	{
		foreach(const event& e, events_)
		{
			if (e.e != n)
				continue;

			if (e.f.empty() || !has_attr(e.f))
				continue;

			execute(e.f, a1);
		}
	}
	
	// fires the given event and passes it the parameter
	//
	template <class A1, class A2>
	void fire(const utf8_string& n, A1 a1, A2 a2)
	{
		foreach(const event& e, events_)
		{
			if (e.e != n)
				continue;

			if (e.f.empty() || !has_attr(e.f))
				continue;

			execute(e.f, a1, a2);
		}
	}

	// returns a prefix used when logging
	//
	utf8_string prefix() const;


	// creates an object of the given class
	//
	bpl::object create_object(const utf8_string& name);

	bpl::object create_object(const utf8_string& name,
		bpl::object a1);

	bpl::object create_object(
		const utf8_string& name,
		bpl::object a1, bpl::object a2);

	bpl::object create_object(
		const utf8_string& name,
		bpl::object a1, bpl::object a2, bpl::object a3);

	bpl::object module()
	{
		return module_;
	}

private:
	// an event
	//
	struct event
	{
		// event name
		utf8_string e;

		// function to call in the script
		utf8_string f;


		event(const utf8_string& e, const utf8_string& f)
			: e(e), f(f)
		{
		}

		bool operator==(const event& right)
		{
			if (e == right.e && f == right.f)
				return true;
			return false;
		}

		bool operator!=(const event& right)
		{
			return !(*this == right);
		}
	};

	// this is a thread that is started just before calling into an
	// extension; after sleeping for a while, it will interrupt the call
	// if it didn't complete
	//
	class script_checker : boost::noncopyable
	{
	public:
		script_checker();

		void operator()();
		bool killed();
		void ok();

	private:
		bool ok_;
		bool killed_;
	};

	// sets the current directory to the extensions directory in the
	// constructor and resets it to the old one in the destructor
	//
	class ext_pwd : boost::noncopyable
	{
	public:
		ext_pwd(const utf8_path& p);
		~ext_pwd();

	private:
		utf8_path old_;
	};

	struct download_info
	{
		boost::thread thread;
		boost::mutex mutex;
		double progress;
		bool done;
		bool stop;

		download_info()
			: progress(0.0), done(false), stop(false)
		{
		}
	};

	typedef std::vector<event> events;


	download_info dl_;

	// path of the script; this is restored as the pwd when scripts are
	// called
	utf8_path path_;

	// python module
	bpl::object module_;

	// list of events registered by this extension
	events events_;

	// cached infos
	extension_info info_;

	// whether this extension is enabled
	bool enabled_;

	// whether this is a dev extension
	bool dev_;

	// whether to accept bad behaviour
	bool bad_;


	// reads the index file
	//
	instance(const utf8_path& p, bool dev);

	// from infos
	//
	instance(const extension_info& ei);

	static bool do_extract(instance_ptr& p, const utf8_path& btx);


	void download_fun(const utf8_string& url);
	bool download_cb(double p);

	// parses the btx.txt file
	//
	void read_index();

	// imports the extension python module
	//
	void import(bool rd=false);

	// calls cleanup() on the extension
	//
	void cleanup();

	// returns whether this extension has the given attribute
	//
	bool has_attr(const utf8_string& name) const;

	// logs an error with the given message or python_error() if
	// empty
	//
	void error(utf8_string w=utf8_string("")) const;

	void reload_already_imported(const utf8_string& mod_name);


	// these are defines because they're used in all the execute_thread
	// functions below
#define EXECUTE_HEAD										\
		log_v() << "execute: locking\n";					\
		python_lock lock;									\
															\
		try													\
		{													\
			if (ext::has_attr(module_, name))				\
			{												\
				log_v() << "execute: has mod\n";			\
				ext_pwd pwd(path_);

#define EXECUTE_TAIL												\
				log_v() << "execute: function returned\n";			\
			}														\
			else													\
			{														\
				log_v() << "execute: no mod\n";						\
			}														\
		}															\
		catch(bpl::error_already_set&)								\
		{															\
			error();												\
		}															\
																	\
		log_v() << "execute thread: done\n";						\

	// calls the given function in the extension; swallows exceptions
	// and interrupts the call if it takes too long
	//
	void execute(const utf8_string& name)
	{
		EXECUTE_HEAD
		module_.attr(name.c_str())();
		EXECUTE_TAIL
	}
	
	// calls the given function in the extension; swallows exceptions
	// and interrupts the call if it takes too long
	//
	template <class A1>
	void execute(const utf8_string& name, A1 a1)
	{
		EXECUTE_HEAD
		module_.attr(name.c_str())(a1);
		EXECUTE_TAIL
	}

	// calls the given function in the extension; swallows exceptions
	// and interrupts the call if it takes too long
	//
	template <class A1, class A2>
	void execute(const utf8_string& name, A1 a1, A2 a2)
	{
		EXECUTE_HEAD
		module_.attr(name.c_str())(a1, a2);
		EXECUTE_TAIL
	}

	// calls the given function in the extension; swallows exceptions
	// and interrupts the call if it takes too long
	//
	template <class A1, class A2, class A3>
	void execute(const utf8_string& name, A1 a1, A2 a2, A3 a3)
	{
		EXECUTE_HEAD
		module_.attr(name.c_str())(a1, a2, a3);
		EXECUTE_TAIL
	}

#undef EXECUTE_THREAD_TAIL
#undef EXECUTE_THREAD_HEAD
};


// loads all the extensions from the database
//
void load();

// cleans up the extensions
//
void unload();

// copies it to the extensions directory and loads it
//
void install(instance_ptr i);


// returns the list of extensions
//
instance_list instances();

// returns the extension having the given guid; may be empty if the
// guid doesn't exist
//
instance_ptr find(const std::string& guid);

// checks the stack frame to figure out which module is calling into
// the dll; returns null if the instance is not found
//
instance_ptr calling_instance();

// adds the given event name to the list of possible events; this is
// used for validation when register_event() is called from scripts
//
void add_event(const utf8_string& n);

// returns whether this event has been registered by any part of
// the program
//
bool valid_event(const utf8_string& n);


// fires the given event on all extensions that registered it
//
template <class A1>
void fire(const utf8_string& e, A1 a1)
{
	if (!python_initialized())
		return;

	bt_assert(valid_event(e));
	if (!valid_event(e))
	{
		log_error() << "trying to fire unknown event '" << e << "'\n";
		return;
	}

	foreach (instance_ptr i, instances())
		i->fire(e, a1);
}

// fires the given event on all extensions that registered it
//
template <class A1, class A2>
void fire(const utf8_string& e, A1 a1, A2 a2)
{
	if (!python_initialized())
		return;

	bt_assert(valid_event(e));
	if (!valid_event(e))
	{
		log_error() << "trying to fire unknown event '" << e << "'\n";
		return;
	}

	foreach (instance_ptr i, instances())
		i->fire(e, a1, a2);
}

}	// namespace ext
}	// namespace baretorrent

#endif	// BARETORRENT_EXT_INCLUDED
