// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "../pch.h"
#include "python_wrapper.h"
#include "../utility.h"
#include "../locale.h"
#include "../log.h"
#include "../locale.h"
#include "ext.h"
#include "../platform.h"

void export_libtorrent_module();

namespace baretorrent
{
namespace ext
{

static PyThreadState* g_main_thread_state = 0;
static bool g_has_python = false;

void export_python_module();

error_restorer::error_restorer()
	: type_(nullptr), value_(nullptr), tb_(nullptr)
{
	// remembering current exception
	PyErr_Fetch(&type_, &value_, &tb_);

	Py_IncRef(type_);
	Py_IncRef(value_);
	Py_IncRef(tb_);
}

error_restorer::~error_restorer()
{
	// restoring old exception
	PyErr_Restore(type_, value_, tb_);
}


python_lock::python_lock()
{
	state_ = PyGILState_Ensure();
}

python_lock::~python_lock()
{
	PyGILState_Release(state_);
}

struct output_redirector
{
    PyObject_HEAD
	bpl::object old;
	std::ostringstream out;
	bool error;
	bool capture;
};

PyObject* output_redirector_write(PyObject* selfp, PyObject* args)
{
    output_redirector* self =
		reinterpret_cast<output_redirector*>(selfp);
	bt_assert(self);

    char* data;
    if (!PyArg_ParseTuple(args, "s", &data))
        return 0;

	const utf8_string s(data);

	if (self->capture)
	{
		self->out << s;
	}
	else
	{
		instance_ptr i = calling_instance();

		if (self->error)
			log_error(i) << s;
		else
			log(i) << s;
	}

	return PyLong_FromSize_t(s.size());
}

PyObject* output_redirector_flush(PyObject*, PyObject*)
{
    // no-op
    return Py_BuildValue("");
}

PyMethodDef output_redirector_methods[] =
{
    {"write", output_redirector_write, METH_VARARGS, ""},
    {"flush", output_redirector_flush, METH_VARARGS, ""},
    {0, 0, 0, 0} // sentinel
};

PyTypeObject output_redirector_type =
{
    PyVarObject_HEAD_INIT(0, 0)
    "baretorrent_impl.output_redirector", /* tp_name */
    sizeof(output_redirector), /* tp_basicsize */
    0, /* tp_itemsize */
    0, /* tp_dealloc */
    0, /* tp_print */
    0, /* tp_getattr */
    0, /* tp_setattr */
    0, /* tp_reserved */
    0, /* tp_repr */
    0, /* tp_as_number */
    0, /* tp_as_sequence */
    0, /* tp_as_mapping */
    0, /* tp_hash */
    0, /* tp_call */
    0, /* tp_str */
    0, /* tp_getattro */
    0, /* tp_setattro */
    0, /* tp_as_buffer */
    Py_TPFLAGS_DEFAULT, /* tp_flags */
    "baretorrent_impl.output_redirector objects", /* tp_doc */
    0, /* tp_traverse */
    0, /* tp_clear */
    0, /* tp_richcompare */
    0, /* tp_weaklistoffset */
    0, /* tp_iter */
    0, /* tp_iternext */
    output_redirector_methods, /* tp_methods */
    0, /* tp_members */
    0, /* tp_getset */
    0, /* tp_base */
    0, /* tp_dict */
    0, /* tp_descr_get */
    0, /* tp_descr_set */
    0, /* tp_dictoffset */
    0, /* tp_init */
    0, /* tp_alloc */
    0, /* tp_new */
};

PyModuleDef baretorrent_impl_module =
{
    PyModuleDef_HEAD_INIT,
    "baretorrent_impl", 0, -1, 0,
};

output_redirector* g_stdout = nullptr;
output_redirector* g_stderr = nullptr;

PyMODINIT_FUNC PyInit_baretorrent_impl()
{
    output_redirector_type.tp_new = PyType_GenericNew;
    if (PyType_Ready(&output_redirector_type) < 0)
        return 0;

    PyObject* m = PyModule_Create(&baretorrent_impl_module);
    if (m)
    {
        Py_INCREF(&output_redirector_type);
        PyModule_AddObject(
			m, "output_redirector",
			reinterpret_cast<PyObject*>(&output_redirector_type));
    }
    return m;
}

output_redirector* redirect_output(
	bpl::object mod, const std::string& what)
{
	try
	{
		bpl::object old;
	
		if (mod)
		{
			bpl::object dict = mod.attr("__dict__");
			bpl::exec("import sys", dict, dict);

			bpl::object sys = mod.attr("sys");
			if (!sys)
			{
				log_error()
					<< "could not redirect output, module "
					<< "doesn't have sys\n";

				return nullptr;
			}
			
			old = sys.attr(what.c_str());
		}
		else
		{
			PyObject* o = PySys_GetObject(what.c_str());

			if (o)
				old = bpl::object(boost::python::handle<>(o));
		}

		PyObject* p = output_redirector_type.tp_new(
			&output_redirector_type, 0, 0);

		if (!p)
		{
			log_error()
				<< "could not redirect output, creating an "
				<< "output_redirector failed\n";

			return nullptr;
		}

		PyObject copy = *p;

		output_redirector* r = reinterpret_cast<output_redirector*>(p);
		new(p) output_redirector;

		r->ob_base = copy;
		r->old = old;
		
		if (what == "stderr")
			r->error = true;

		if (mod)
		{
			bpl::object sys = mod.attr("sys");
			if (!sys)
			{
				log_error()
					<< "could not redirect output, module "
					<< "doesn't have sys\n";

				return nullptr;
			}

			sys.attr(what.c_str()) = boost::python::handle<>(p);
		}
		else
		{
			PySys_SetObject(what.c_str(), p);
		}

		return r;
	}
	catch(bpl::error_already_set&)
	{
		log_error()
			<< "could not redirect output, " << python_error() << "\n";
		
		PyErr_Clear();
		return nullptr;
	}
}

void stop_redirect_output(bpl::object mod, const std::string& what)
{
	try
	{
		output_redirector* r = nullptr;

		if (mod)
		{
			bpl::object sys = mod.attr("sys");
			bpl::object o = sys.attr(what.c_str());
			
#if BARETORRENT_VISUALC
#pragma warning(push)
#pragma warning(disable: 4127)
#endif

			Py_XINCREF(o.ptr());

#if BARETORRENT_VISUALC
#pragma warning(pop)
#endif

			PyObject* p = o.ptr();
			r = reinterpret_cast<output_redirector*>(p);

			sys.attr(what.c_str()) = r->old;
		}
		else
		{
			PyObject* p = PySys_GetObject(what.c_str());
			r = reinterpret_cast<output_redirector*>(p);

			PySys_SetObject(what.c_str(), r->old.ptr());
		}

		if (r && r->ob_base.ob_refcnt == 1)
			r->~output_redirector();

#if BARETORRENT_VISUALC
#pragma warning(push)
#pragma warning(disable: 4127)
#endif

		Py_XDECREF(r);

#if BARETORRENT_VISUALC
#pragma warning(pop)
#endif
	}
	catch(bpl::error_already_set&)
	{
		log_error()
			<< "could not redirect output, " << python_error() << "\n";
		
		PyErr_Clear();
	}
}

scoped_output_redirector::scoped_output_redirector()
	: red_(nullptr)
{
	start();
}

scoped_output_redirector::scoped_output_redirector(
	output_redirector* r)
		: red_(nullptr)
{
	start(r);
}

scoped_output_redirector::~scoped_output_redirector()
{
	stop();
}

void scoped_output_redirector::start(output_redirector* r)
{
	if (r)
	{
		red_ = r;
		red_->capture = true;
	}
	else
	{
		g_stdout->capture = true;
		g_stderr->capture = true;
	}
}

void scoped_output_redirector::stop()
{
	if (red_)
	{
		red_->capture = false;

		what_ = utf8_string(red_->out.str());
		red_->out.str("");
	}
	else
	{
		g_stdout->capture = false;
		stdout_ = utf8_string(g_stdout->out.str());
		g_stdout->out.str("");

		g_stderr->capture = false;
		stderr_ = utf8_string(g_stderr->out.str());
		g_stderr->out.str("");
	}
}

utf8_string scoped_output_redirector::what() const
{
	return what_;
}

utf8_string scoped_output_redirector::out() const
{
	return stdout_;
}

utf8_string scoped_output_redirector::err() const
{
	return stderr_;
}


utf8_string extract_string(bpl::object o)
{
	try
	{
		return bpl::extract<utf8_string>(o);
	}
	catch(bpl::error_already_set&)
	{
		try 
		{
			return utf16_to_utf8(bpl::extract<utf16_string>(o));
		}
		catch(bpl::error_already_set&)
		{
			log_error() << "python: can't convert string\n";
			return utf8_string("");
		}
	}
}

bool extract_bytes(bpl::object o, byte_array& out)
{
	if (!PyBytes_Check(o.ptr()))
	{
		log_error()
			<< "extract bytes: got a " << o.ptr()->ob_type->tp_name
			<< " instead of a bytes object\n";

		return false;
	}

	char* s = PyBytes_AsString(o.ptr());
	if (!s)
	{
		log_error() << "extract bytes: got a null string\n";
		return false;
	}

	std::size_t ss = PyBytes_Size(o.ptr());
	if (ss == 0)
		return true;

	out = byte_array(reinterpret_cast<const unsigned char*>(s), ss);
	return true;
}

utf8_string python_error()
{
	python_lock lock;

	if (g_stderr)
	{
		scoped_output_redirector sor(g_stderr);
		PyErr_Print();
		sor.stop();

		return sor.what();
	}

	return utf8_string();
}

bpl::object wx_to_bpl(wxWindow* w)
{
	const wxClassInfo* info = w->GetClassInfo();
	if (!info)
	{
		log_error() << "wx_to_bpl: can't get class info\n";
		return bpl::object();
	}

	const wxString name = info->GetClassName();
	if (name.empty())
	{
		log_error() << "wx_to_bpl: class name is empty\n";
		return bpl::object();
	}
	
	PyObject* o = wxPyConstructObject(w, name, false);
	if (!o)
	{
		log_error() << "wx_to_bpl: construction failed\n";
		return bpl::object();
	}

	return bpl::object((bpl::handle<>(o)));
}

bool has_attr(bpl::object o, const utf8_string& name)
{
	python_lock lock;

	try
	{
		// this will throw if the attribute doesn't exist
		bpl::object a = o.attr(name.c_str());
		if (!a)
			return false;

		return true;
	}
	catch(bpl::error_already_set&)
	{
		PyErr_Clear();
		return false;
	}
}

bool has_key(bpl::object o, const utf8_string& name)
{
	python_lock lock;

	try
	{
		// this will throw if the key doesn't exist
		bpl::object a = o[name.c_str()];
		if (!a)
			return false;

		return true;
	}
	catch(bpl::error_already_set&)
	{
		PyErr_Clear();
		return false;
	}
}

// makes sure that some basic python features are working correctly
//
bool sanity_check()
{
	log_v() << "python sanity checks:\n";

	log_v() << "python: check 1\n";
	scoped_ext_path sp(utf8_string(""));

	bpl::object main = bpl::import("__main__");
	bpl::object global(main.attr("__dict__"));

	bpl::exec(
		"import re\n"
		"import sys\n"
		"import os\n"
		"\n"
		"dummy__ = 1\n"
		"dummy_str__ = \"hello\"",
		global, global);

	log_v() << "python: check 2\n";
	bpl::object sys = main.attr("sys");
	bpl::object o = sys.attr("version_info");

	bpl::tuple v(bpl::handle<>(bpl::borrowed(o.ptr())));

	log_v() << "python: checking version\n";
	const int exec_major = bpl::extract<int>(v[0]);
	const int exec_minor = bpl::extract<int>(v[1]);
	const int exec_micro = bpl::extract<int>(v[2]);

	const int lib_major = PY_MAJOR_VERSION;
	const int lib_minor = PY_MINOR_VERSION;
	const int lib_micro = PY_MICRO_VERSION;

	if (exec_major != lib_major ||
		exec_minor != lib_minor ||
		exec_micro != lib_micro)
	{
		log_error()
			<< "baretorrent was compiled against python "
			<< lib_major << "." << lib_minor << "." << lib_micro
			<< " but the runtime libraries are for "
			<< exec_major << "." << exec_minor << "." << exec_micro
			<< "\n";

		return false;
	}

	log_v() << "python: versions match\n";

	log_v() << "python sanity checks done\n";

	return true;
}

bool init_python(const std::vector<utf8_path> paths)
{
	try
	{
		if (paths::clear_sys_path())
			Py_SetPath(L"");

		platform::add_to_path(platform::executable_directory());

		// setting the paths before initializing python since it also
		// contains the location of the libs; 'true' will set python's
		// path instead of the extensions
		set_ext_path(paths, true);

		log_v() << "python: adding baretorrent impl\n";
		PyImport_AppendInittab("baretorrent_impl", PyInit_baretorrent_impl);

		log_v() << "python: Py_Initialize()\n";
		Py_Initialize();

		PyImport_ImportModule("baretorrent_impl");

		g_stdout = redirect_output(bpl::object(), "stdout");
		g_stderr = redirect_output(bpl::object(), "stderr");

		if (paths::clear_sys_path())
		{
			PySys_SetPath(L"");
			set_ext_path(paths, false);
		}

		log_v() << "python: PyEval_InitThreads\n";

		// initializes threading and acquires the lock
		PyEval_InitThreads();
		
		log_v() << "python: releasing lock\n";

		// this releases the lock because it will be locked/unlocked
		// below in python_lock; it will be restored at the end, in
		// term_python()
		g_main_thread_state = PyGILState_GetThisThreadState();
		PyEval_ReleaseThread(g_main_thread_state);

		log_v() << "python: importing error handler\n";

		python_lock lock;
		PyImport_ImportModule("output_redirector");

		// todo: this needs to be added before the manager starts because
		// events will be fired and an assert will be triggered because
		// those events don't exist yet
		ext::add_event(utf8_string("torrent.added"));
		ext::add_event(utf8_string("torrent.error"));
		ext::add_event(utf8_string("torrent.metadata"));
		ext::add_event(utf8_string("torrent.complete"));
		ext::add_event(utf8_string("torrent.file_completed"));
		ext::add_event(utf8_string("torrent.updated"));
		ext::add_event(utf8_string("torrent.ready"));
		ext::add_event(utf8_string("option_changed"));
		ext::add_event(utf8_string("icon_clicked"));

		if (!sanity_check())
			return false;

		log_v() << "python initialized successfully\n";

		g_has_python = true;
		return true;
	}
	catch(bpl::error_already_set&)
	{
		log_error()
			<< "couldn't initialize python, "
			<< python_error() << "\n";

		python_lock lock;
		PyErr_Clear();

		return false;
	}
}

void term_python()
{
	if (g_main_thread_state)
	{
		// acquires the global lock again since it was manually released
		// in init_python
		PyEval_RestoreThread(g_main_thread_state);

		Py_Finalize();
	}
}

bool python_initialized()
{
	return g_has_python;
}

#if BARETORRENT_WINDOWS
typedef utf16_string python_string_t;
#else
typedef utf32_string python_string_t;
#endif

void set_ext_path(const utf8_string& s, bool module_path)
{
	const python_string_t temp(Py_GetPath());

	const byte_array ba = convert(
		temp.c_str(), temp.size_in_bytes(),
		python_string_t::encoding, utf8_encoding);
	
	utf8_string np;
	if (!ba.empty())
		np = utf8_string(ba.data());

	if (!boost::starts_with(np.std_string(), env_path_separator()))
		np = env_path_separator() + np;

	np = s + np;

	// Py_SetPath() tells python where its modules are, while
	// PySys_SetPath() is used by import

	const byte_array np_temp = convert(
		np.c_str(), np.size_in_bytes(),
		utf8_encoding, python_string_t::encoding);

	python_string_t wnp;
	if (!np_temp.empty())
	{
		wnp = python_string_t(
			reinterpret_cast<const python_string_t::char_t*>(
				np_temp.data()));
	}

	if (module_path)
	{
		Py_SetPath(const_cast<python_string_t::char_t*>(
			wnp.c_str()));
	}
	else
	{
		PySys_SetPath(const_cast<python_string_t::char_t*>(
			wnp.c_str()));
	}

	std::ostringstream lg;
	std::vector<utf8_string> v = explode(np, env_path_separator());

	lg << "python " << (module_path ? "module " : "sys ") << "path is:\n";
	foreach (const utf8_string& s, v)
		lg << " - " << s << "\n";

	log() << lg.str();
}

void set_ext_path(const std::vector<utf8_path>& ep, bool uninitialized)
{
	set_ext_path(implode(ep, env_path_separator()), uninitialized);
}


scoped_ext_path::scoped_ext_path(const std::vector<utf8_path>& ps)
{
	old_ = utf16_to_utf8(utf16_string(Py_GetPath()));
	set_ext_path(ps, false);
}

scoped_ext_path::scoped_ext_path(const utf8_string& p)
{
	old_ = utf16_to_utf8(utf16_string(Py_GetPath()));
	set_ext_path(p, false);
}

scoped_ext_path::~scoped_ext_path()
{
	error_restorer er;

	try
	{
		log_vv() << "reseting old path\n";
		set_ext_path(old_, false);
	}
	catch(...)
	{
		// ignore
	}
}

}	// namespace ext
}	// namespace baretorrent
