// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "ext_ui.h"
#include "locale.h"
#include "json/json.h"
#include "ext/ext.h"

namespace baretorrent
{

// various widget ids
//
enum ids
{
	// extension list
	ext_list_id = wxID_HIGHEST + 500,

	// opens the options dialog for an extension
	ext_options_id,

	// disable/enable extension
	ext_toggle_id,

	// uninstall extension
	ext_uninstall_id,

	// install extension
	ext_install_id,

	// reloads the extension
	ext_reload_id,

	// search
	ext_search_id
};


ext_install_dialog::ext_install_dialog(
	wxWindow* parent, ext::instance_ptr i) :
		wxDialog(
			parent, -1, loc_wx("extension.install.title"),
			wxDefaultPosition, wxDefaultSize,
			wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER)
{
	extension_infos* p = new extension_infos(this);
	p->rebuild(i);

	wxStdDialogButtonSizer* buttons = new wxStdDialogButtonSizer;
	
	wxButton* cancel = new wxButton(
		this, wxID_CANCEL, loc_wx("cancel"));
	
	wxButton* install = new wxButton(
		this, wxID_OK, loc_wx("extension.install"));

	buttons->AddButton(install);
	buttons->AddButton(cancel);
	buttons->Realize();

	wxFlexGridSizer* box = new wxFlexGridSizer(1);
	box->Add(p, 1, wxEXPAND);

	wxStaticText* more = nullptr;

	if (i->can_enable())
	{
		if (i->can_update())
		{
			more = new wxStaticText(
				this, -1, loc_wx("extension.can_update"));
			install->SetLabel(loc_wx("extension.update"));
		}
		else if (i->installed())
		{
			more = new wxStaticText(this, -1,
				loc_wx("extension.already_installed"));
			install->Enable(false);
		}
	}
	else
	{
		install->Enable(false);
	}

	if (more)
	{
		wxFont f = more->GetFont();
		f.SetWeight(wxFONTWEIGHT_BOLD);
		more->SetFont(f);

		wxBoxSizer* already = new wxBoxSizer(wxVERTICAL);
		already->Add(more);

		box->AddSpacer(10);
		box->Add(already);
	}

	box->AddSpacer(5);
	box->Add(CreateSeparatedSizer(buttons), 1, wxEXPAND);

	box->AddGrowableCol(0);
	box->AddGrowableRow(0);

	cancel->SetDefault();
	cancel->SetFocus();
	
	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(box, 1, wxEXPAND | wxALL, 5);

	SetSizer(root);
	SetSize(500, 300);
	CentreOnParent();
}


int ext_sorter_name(ext::instance* left, ext::instance* right)
{
	return natural_compare(left->name(), right->name());
}

int ext_sorter_version(ext::instance* left, ext::instance* right)
{
	return natural_compare(left->version(), right->version());
}

int ext_sorter_author(ext::instance* left, ext::instance* right)
{
	return natural_compare(left->author(), right->author());
}

int ext_sorter_desc(ext::instance* left, ext::instance* right)
{
	return natural_compare(
		left->short_description(), right->short_description());
}


extension_list::extension_list(wxWindow* parent)
	: super(parent, -1, "bt.ui.ext_list", wxLC_SINGLE_SEL)
{
	add_column(loc("options.extensions.name"), &ext_sorter_name);
	add_column(loc("options.extensions.version"), &ext_sorter_version);
	add_column(loc("options.extensions.description"), &ext_sorter_desc);
}

void extension_list::on_selected(wxListEvent& e)
{
	super::on_selected(e);
	fire(extension_selected, selected_extension());
}

void extension_list::on_activated(wxListEvent& e)
{
	super::on_activated(e);
	fire(extension_activated, selected_extension());
}

void extension_list::rebuild(const ext::instance_list& is)
{
	instances_ = is;
	
	DeleteAllItems();
	cons_.clear();

	foreach(const ext::instance_ptr i, is)
	{
		const int item = GetItemCount();
		
		append();
		data(item, i.get());

		update_item(item, i);

		cons_.add(i->changed.connect(boost::bind(
			&extension_list::on_changed, this, i)));

		cons_.add(i->uninstalled.connect(boost::bind(
			&extension_list::on_removed, this, i)));
	}

	if (!is.empty())
		select(0);
	else
		extension_selected(ext::instance_ptr());
}

ext::instance_ptr extension_list::selected_extension()
{
	const ext::instance* i = single_selection();
	if (!i)
		return ext::instance_ptr();

	foreach (ext::instance_ptr p, instances_)
	{
		if (p.get() == i)
			return p;
	}
	
	log_error()
		<< "bad instance in extension dialog, "
		<< "instance count is " << instances_.size() << "\n";

	return ext::instance_ptr();
}

void extension_list::on_changed(ext::instance_ptr e)
{
	const int i = find(e.get());
	if (i >= 0)
		update_item(i, e);
}

void extension_list::on_removed(ext::instance_ptr e)
{
	const int i = find(e.get());
	if (i >= 0)
		remove_item(i);
}

void extension_list::update_item(int item, ext::instance_ptr i)
{
	col(item, 0, i->name());
	col(item, 1, i->version());
	col(item, 2, i->short_description());

	if (i->enabled() && i->can_enable())
		row_color(item, normal_text_color());
	else
		row_color(item, disabled_extension_text_color());

	if (i->small_icon().IsOk())
		set_image(item, i->small_icon());
}

void extension_list::remove_item(int item)
{
	bt_assert(item >= 0 && item < GetItemCount());
	if (item < 0 || item >= GetItemCount())
		return;

	DeleteItem(item);

	const int n = GetItemCount();
	
	if (n > 0)
	{
		if (item < n)
			select(item);
		else
			select(n - 1);
	}
	else
	{
		fire(extension_selected, ext::instance_ptr());
	}
}


extension_infos::extension_infos(wxWindow* parent)
	: wxPanel(parent)
{
	name_ = new wxStaticText(this, -1, wxEmptyString);
	author_ = new wxHyperlinkCtrl(
		this, -1, to_wxstring(utf8_string("dummy")), wxEmptyString);

	sd_ = new wxStaticText(this, -1, wxEmptyString);
	ld_ = new multiline_text(this, -1, wxEmptyString, wxTE_READONLY);

	icon_ = new wxStaticBitmap(this, -1, wxBitmap());
	website_ = new wxHyperlinkCtrl(
		this, -1, to_wxstring(utf8_string("dummy")), wxEmptyString);

	dev_ = new wxStaticText(this, -1, wxEmptyString);
	wxFont f = dev_->GetFont();
	f.SetWeight(wxFONTWEIGHT_BOLD);
	dev_->SetFont(f);

	// bigger name label
	f = name_->GetFont();
	f.SetPointSize(20);
	name_->SetFont(f);
	
	wxBoxSizer* left_box = new wxBoxSizer(wxVERTICAL);
	left_box->Add(icon_);

	wxBoxSizer* right_box = new wxBoxSizer(wxVERTICAL);
	right_box->Add(name_);
	right_box->Add(author_);
	right_box->Add(website_);

	wxBoxSizer* bottom_box = new wxBoxSizer(wxVERTICAL);
	bottom_box->Add(sd_);
	bottom_box->AddSpacer(10);
	bottom_box->Add(ld_, 1, wxEXPAND);
	bottom_box->AddSpacer(5);
	bottom_box->Add(dev_);

	wxBoxSizer* top_box = new wxBoxSizer(wxHORIZONTAL);
	top_box->Add(left_box);
	top_box->AddSpacer(10);
	top_box->Add(right_box, 1, wxEXPAND);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(top_box);
	box->AddSpacer(15);
	box->Add(bottom_box, 1, wxEXPAND);

	rebuild(ext::instance_ptr());

	SetSizer(box);
}

void extension_infos::rebuild(ext::instance_ptr e)
{
	cons_.clear();

	if (e)
	{
		cons_.add(e->changed.connect(boost::bind(
			&extension_infos::on_changed, this, e)));
	}

	on_changed(e);
}

void extension_infos::on_changed(ext::instance_ptr e)
{
	wxString name;
	wxString author_label;
	wxString author_link;
	wxString sd;
	wxString ld;
	wxBitmap icon;
	wxString website_label;
	wxString website_link;
	wxString dev;

	if (e)
	{
		name = to_wxstring(
			e->name() + utf8_string(" ") + e->version());
		author_label = to_wxstring(
			locf("options.extensions.by") % e->author());
		author_link = to_wxstring(e->email());
		sd = to_wxstring(e->short_description());
		ld = to_wxstring(e->long_description());

		if (e->icon().IsOk())
		{
			icon = e->icon();
			if (!e->enabled() || !e->can_enable())
			{
				wxImage i = icon.ConvertToImage();
				i = i.ConvertToGreyscale();
				icon = wxBitmap(i);
			}
		}

		website_label = to_wxstring(e->website());
		website_link = to_wxstring(e->website());
		
		if (e->dev())
			dev = loc_wx("extension.dev");
	}

	wxColour c = wxNullColour;
	
	if (e && (!e->enabled() || !e->can_enable()))
		c = disabled_extension_text_color();

	name_->SetLabel(name);
	name_->SetForegroundColour(c);

	author_->SetLabel(author_label);
	author_->SetURL(author_link);
	author_->SetToolTip(author_link);
	author_->SetForegroundColour(c);
	author_->SetHoverColour(c);
	author_->SetNormalColour(c);
	author_->SetVisitedColour(c);

	sd_->SetLabel(sd);
	sd_->SetForegroundColour(c);

	ld_->ChangeValue(ld);
	ld_->SetForegroundColour(c);

	if (icon.IsOk())
	{
		icon_->SetBitmap(icon);
		icon_->Show();
	}
	else
	{
		icon_->Hide();
	}

	website_->SetLabel(website_label);
	website_->SetURL(website_link);
	website_->SetToolTip(website_link);
	website_->SetForegroundColour(c);
	website_->SetHoverColour(c);
	website_->SetNormalColour(c);
	website_->SetVisitedColour(c);

	if (dev.empty())
	{
		dev_->Hide();
	}
	else
	{
		dev_->Show();
		dev_->SetLabel(dev);
	}

	Layout();
	Refresh();
}


extension_controls::extension_controls(wxWindow* parent)
	: wxPanel(parent), buttons_(nullptr), more_(nullptr)
{
	buttons_ = new wxBoxSizer(wxHORIZONTAL);
	
	more_ = new wxStaticText(this, -1, wxEmptyString);
	wxFont f = more_->GetFont();
	f.SetWeight(wxFONTWEIGHT_BOLD);
	more_->SetFont(f);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);

	box->Add(more_);
	box->AddSpacer(5);
	box->Add(buttons_);

	SetSizer(box);
}

void extension_controls::add(wxWindow* w)
{
	buttons_->Add(w, 0, wxALIGN_CENTER_VERTICAL);
	buttons_->AddSpacer(10);
}

void extension_controls::rebuild(ext::instance_ptr i)
{
	cons_.clear();
	
	if (i)
	{
		cons_.add(i->changed.connect(boost::bind(
			&extension_controls::on_changed, this, i)));
	}

	sel_ = i;
	on_changed(i);
}

ext::instance_ptr extension_controls::selection()
{
	return sel_;
}

utf8_string extension_controls::check_error(ext::instance_ptr e)
{
	if (e->requires() > version())
	{
		return utf8_string(
			str(locf("options.extensions.install.require")
				% e->requires() % version()));
	}

	return utf8_string();
}

void extension_controls::on_changed(ext::instance_ptr e)
{
	const utf8_string err = (e ? check_error(e) : utf8_string());

	if (err.empty())
	{
		more_->Hide();
	}
	else
	{
		more_->Show();
		more_->SetLabel(to_wxstring(err));
	}

	GetParent()->GetSizer()->Layout();
	on_changed_impl(e);
}


install_extension_controls::install_extension_controls(wxWindow* parent) :
	extension_controls(parent),
	install_(nullptr), gauge_(nullptr), what_(nullptr)
{
	install_ = new wxButton(
		this, ext_install_id, loc_wx("extension.install"));

	gauge_ = new wxGauge(this, -1, 100);
	gauge_->Hide();

	what_ = new wxStaticText(this, -1, wxEmptyString);
	what_->Hide();

	add(install_);
	add(gauge_);
	add(what_);
	
	timer_.start(
		boost::bind(&install_extension_controls::on_timer, this),
		boost::posix_time::milliseconds(200));
}

void install_extension_controls::on_install(wxCommandEvent&)
{
	ext::instance_ptr i = selection();
	if (!i)
		return;

	try
	{
		gauge_->SetValue(0);
		what_->SetLabel(wxEmptyString);

		for (install_infos::iterator itor=infos_.begin(); itor!=infos_.end(); )
		{
			const install_info_ptr ii = *itor;

			if (ii->i->guid() == i->guid())
				itor = infos_.erase(itor);
			else
				++itor;
		}

		install_info_ptr ii(new install_info);
		infos_.push_back(ii);

		ii->i = i;
		ii->thread = boost::thread(boost::bind(
			&install_extension_controls::install_fun, this, ii));
	}
	catch(ext::exception& e)
	{
		extension_error_dialog dlg(this, i, e.what());
		dlg.ShowModal();
	}
}

void install_extension_controls::install_fun(install_info_ptr ii)
{
	try
	{
		ext::install(ii->i);

		{
			boost::mutex::scoped_lock lock(ii->mutex);
			ii->done = true;
			ii->okay = true;
		}
	}
	catch(...)
	{
		{
			boost::mutex::scoped_lock lock(ii->mutex);
			ii->done = true;
			ii->okay = false;
		}
	}
}

utf8_string install_extension_controls::check_error(ext::instance_ptr e)
{
	const utf8_string s = extension_controls::check_error(e);
	if (!s.empty())
		return s;

	if (e->installed())
	{
		if (!e->can_update())
			return loc("extension.already_installed");
		else
			return loc("extension.can_update");
	}

	return utf8_string();
}

bool install_extension_controls::on_timer()
{
	ext::instance_ptr s = selection();
	if (!s)
		return true;

	refresh(s);

	return true;
}

void install_extension_controls::on_changed_impl(ext::instance_ptr e)
{
	refresh(e);
}

void install_extension_controls::refresh(ext::instance_ptr e)
{
	wxString caption;

	if (e)
	{
		bool found = false;
		bool installing = false;

		foreach (install_info_ptr ii, infos_)
		{
			boost::mutex::scoped_lock lock(ii->mutex);

			if (e->guid() == ii->i->guid())
			{
				if (ii->done)
				{
					int p = 0;
					wxString w;

					if (ii->okay)
					{
						p = 100;
						w = loc_wx("extension.install_success");
					}
					else
					{
						p = 0;
						w = loc_wx("extension.install_error");
					}

					if (gauge_->GetValue() != p)
						gauge_->SetValue(p);

					if (what_->GetLabel() != w)
					{
						what_->SetLabel(w);
						GetParent()->GetSizer()->Layout();
					}
				}
				else
				{
					installing = true;

					const int p = static_cast<int>(
						ii->i->download_progress() * 100);

					if (gauge_->GetValue() != p)
						gauge_->SetValue(p);
				}

				bool relayout = false;
				if (!gauge_->IsShown() || !what_->IsShown())
					relayout = true;

				gauge_->Show();
				what_->Show();

				if (relayout)
					GetParent()->GetSizer()->Layout();

				found = true;
				break;
			}
		}

		if (!found)
		{
			gauge_->Hide();
			what_->Hide();
		}

		if (installing)
		{
			install_->Enable(false);
		}
		else if (!e->installed())
		{
			install_->Enable(e->can_enable());
		}
		else
		{
			if (e->can_update())
			{
				caption = loc_wx("extension.update");
				install_->Enable(true);
			}
			else
			{
				install_->Enable(false);
			}
		}
	}
	else
	{
		install_->Enable(false);
		gauge_->Hide();
		what_->Hide();
	}

	if (caption.empty())
		caption = loc_wx("extension.install");

	if (caption != install_->GetLabel())
		install_->SetLabel(caption);
}


options_extension_controls::options_extension_controls(
	wxWindow* parent) :
		extension_controls(parent), options_(nullptr), toggle_(nullptr),
		uninstall_(nullptr), reload_(nullptr)
{
	options_ = new wxButton(
		this, ext_options_id, loc_wx("options.extensions.options"));

	toggle_ = new wxButton(
		this, ext_toggle_id, loc_wx("options.extensions.enable"));

	uninstall_ = new wxButton(
		this, ext_uninstall_id, loc_wx("options.extensions.uninstall"));

	reload_ = new wxButton(
		this, ext_reload_id, loc_wx("options.extensions.reload"));

	add(options_);
	add(toggle_);
	add(uninstall_);
	add(reload_);
}

void options_extension_controls::on_changed_impl(ext::instance_ptr e)
{
	if (e)
	{
		options_->Enable(e->enabled());
		uninstall_->Enable(e->can_uninstall());
		reload_->Show(e->dev());
		toggle_->Enable(true);

		std::string ln;

		if (e->enabled())
		{
			ln = "options.extensions.disable";
		}
		else
		{
			ln = "options.extensions.enable";

			if (!e->can_enable())
				toggle_->Enable(false);
			else
				toggle_->Enable(true);
		}

		toggle_->SetLabel(loc_wx(ln));
	}
	else
	{
		options_->Enable(false);
		toggle_->Enable(false);
		uninstall_->Enable(false);

		if (reload_)
		{
			reload_->Enable(false);
			reload_->Show(false);
		}
	}
}

void options_extension_controls::show_options(ext::instance_ptr e)
{
	extension_options_dialog dlg(this, e);
	dlg.ShowModal();
}

void options_extension_controls::on_options(wxCommandEvent&)
{
	ext::instance_ptr i = selection();
	if (i && i->enabled())
		show_options(i);
}

void options_extension_controls::on_toggle(wxCommandEvent&)
{
	ext::instance_ptr i = selection();
	if (!i)
		return;

	try
	{
		i->enabled(!i->enabled());
	}
	catch(ext::exception& e)
	{
		extension_error_dialog dlg(this, i, e.what());
		dlg.ShowModal();
	}
}

void options_extension_controls::on_uninstall(wxCommandEvent&)
{
	try
	{
		ext::instance_ptr i = selection();
		if (!i)
			return;

		wxMessageDialog b(
			this,
			to_wxstring(locf("options.extensions.uninstall.caption")
				% i->name()),
			loc_wx("options.extensions.uninstall.title"),
			wxOK | wxCANCEL);

		if (b.ShowModal() != wxID_OK)
			return;

		i->uninstall();
	}
	catch(ext::exception&)
	{
		// todo, what now?
	}
}

void options_extension_controls::on_reload(wxCommandEvent&)
{
	try
	{
		ext::instance_ptr i = selection();
		if (!i)
			return;

		i->reload();
	}
	catch(ext::exception&)
	{
		wxMessageBox(
			loc_wx("extension.reload_failed"),
			loc_wx("extension.reload_failed.title"),
			wxOK, this);
	}
}


extension_panel::extension_panel(wxWindow* parent, modes m) :
	wxPanel(parent),
	split_(nullptr), list_(nullptr), info_(nullptr), controls_(nullptr),
	bottom_(nullptr)
{
	split_ = new splitter(this, "bt.ui.ext_splitter");

	list_ = new extension_list(split_);
	list_->extension_selected.connect(
		boost::bind(&extension_panel::on_selected, this, _1));
	list_->extension_activated.connect(
		boost::bind(&extension_panel::on_activated, this, _1));

	bottom_ = new wxPanel(split_);
	bottom_->Hide();

	switch (m)
	{
		case options_mode:
			controls_ = new options_extension_controls(bottom_);
			break;

		case install_mode:
			controls_ = new install_extension_controls(bottom_);
			break;
	}
	
	info_ = new extension_infos(bottom_);

	split_->Initialize(list_);
	split_->SetSashGravity(0.0);

	wxBoxSizer* bottom_sizer = new wxBoxSizer(wxVERTICAL);

	bottom_sizer->Add(info_, 1, wxEXPAND);
	bottom_sizer->AddSpacer(5);
	bottom_sizer->Add(controls_);

	bottom_->SetSizer(bottom_sizer);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(split_, 1, wxEXPAND);

	SetSizer(box);
}

void extension_panel::update(const ext::instance_list& is)
{
	list_->rebuild(is);
}

void extension_panel::on_selected(ext::instance_ptr e)
{
	if (!e)
	{
		if (split_->IsSplit())
			split_->Unsplit(bottom_);
	}
	else
	{
		if (!split_->IsSplit())
			split_->SplitHorizontally(list_, bottom_);
	}

	info_->rebuild(e);
	controls_->rebuild(e);
}

void extension_panel::on_activated(ext::instance_ptr e)
{
	if (e && e->enabled())
	{
		extension_options_dialog dlg(this, e);
		dlg.ShowModal();
	}
}


extension_error_dialog::extension_error_dialog(
	wxWindow* parent, ext::instance_ptr i, const utf8_string& e) :
		wxDialog(
			parent, 0, loc_wx("extension.error.title"),
			wxDefaultPosition, wxDefaultSize,
			wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER)
{
	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
		
	wxTextCtrl* t = new multiline_text(
		this, -1, make_text(i, e), wxTE_READONLY);

	wxSizer* buttons = CreateSeparatedButtonSizer(wxOK);

	box->Add(new wxStaticText(
		this, -1, loc_wx("options.extensions.failed")));
	box->AddSpacer(5);
	box->Add(t, 1, wxEXPAND);
	box->AddSpacer(5);
	box->Add(buttons, 0, wxEXPAND);

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(box, 1, wxEXPAND | wxALL, 5);

	SetSizer(root);
	SetSize(500, 330);
	CenterOnParent();
}

wxString extension_error_dialog::make_text(
	ext::instance_ptr i, const utf8_string& e)
{
	std::ostringstream oss;

	oss << e;
	
	if (i)
	{
		if (!e.empty())
			oss << "\n\n";
		
		oss << i->dump();
	}

	if (!oss.str().empty())
		oss << "\n";

	return oss.str();
}


extension_options_dialog::extension_options_dialog(
	wxWindow* parent, ext::instance_ptr i) :
		wxDialog(
			parent, 0, wxEmptyString,
			wxDefaultPosition, wxDefaultSize,
			wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER),
		advanced_(nullptr),
		instance_(i)
{
	SetTitle(to_wxstring(
		locf("options.extensions.options.title") % i->name()));

	wxNotebook* tabs = new wxNotebook(this, -1);
	
	advanced_ = new option_list_widget(tabs, i->guid());
	advanced_->changed.connect(boost::bind(
		&extension_options_dialog::on_changed, this, _1, _2));

	// panels
	tabs->InsertPage(0, advanced_, loc_wx("options.advanced"));

	// ok/cancel
	wxSizer* buttons = CreateSeparatedButtonSizer(wxOK | wxCANCEL);

	wxFlexGridSizer* root = new wxFlexGridSizer(1);
	root->Add(tabs, 1, wxEXPAND);
	root->AddSpacer(5);
	root->Add(buttons, wxSizerFlags().Expand().Right());
	
	root->AddGrowableCol(0);
	root->AddGrowableRow(0);
		
	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(root, 1, wxEXPAND | wxALL, 5);

	SetSizerAndFit(box);

	SetSize(530, 480);
	CentreOnParent();
}

const option_list& extension_options_dialog::options() const
{
	return new_;
}

void extension_options_dialog::on_changed(const std::string&, option o)
{
	new_[o.name()].set(o.value());
}

void extension_options_dialog::on_ok(wxCommandEvent&)
{
	std::pair<std::string, option> o;
	foreach (o, new_)
		ext_opt(instance_->guid(), o.first, o.second.value());

	if (!new_.empty())
		instance_->dump_options();

	EndModal(wxID_OK);
}


ext_browser::ext_browser(wxWindow* parent) :
	custom_dialog(parent, loc("ext_browser.title")),
	panel_(nullptr), search_(nullptr)
{
	wxPanel* search_panel = new wxPanel(this);
	
	search_ = new wxTextCtrl(search_panel, -1, wxEmptyString);
	wxStaticText* search_label = new wxStaticText(
		search_panel, -1, loc_wx("ext_browser.search_label"));
	wxButton* search_button = new wxButton(
		search_panel, ext_search_id, loc_wx("ext_browser.search"));

	wxFlexGridSizer* search_box = new wxFlexGridSizer(1, 0, 0, 5);
	search_box->Add(search_label, 0, wxALIGN_CENTER_VERTICAL);
	search_box->Add(search_, 1, wxEXPAND);
	search_box->Add(search_button);
	
	search_box->AddGrowableCol(1);

	search_panel->SetSizer(search_box);
	

	panel_ = new extension_panel(this, extension_panel::install_mode);


	wxFlexGridSizer* box = new wxFlexGridSizer(1, 5, 0);
	box->Add(search_panel, 1, wxEXPAND);
	box->Add(panel_, 1, wxEXPAND);

	box->AddGrowableCol(0);
	box->AddGrowableRow(1);


	wxBoxSizer* dbox = new wxBoxSizer(wxVERTICAL);
	dbox->Add(box, 1, wxEXPAND);
	dbox->AddSpacer(5);
	dbox->Add(
		CreateSeparatedButtonSizer(wxCANCEL),
		wxSizerFlags().Align(wxBOTTOM).Expand());

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(dbox, 1, wxEXPAND | wxALL, 5);
	
	SetSizer(root);
	SetSize(630, 550);
	CentreOnParent();

	search(utf8_string(""));
}

ext_browser::~ext_browser()
{
	panel_->update(ext::instance_list());
}

void ext_browser::on_search(wxCommandEvent&)
{
	search(from_wxstring(search_->GetValue()));
}

void ext_browser::search(const utf8_string& s)
{
	boost::format f(opt("bt.ext.search_url").std_string());
	const utf8_string url(str(f % s));

	const byte_array r = get_remote_file(url);
	if (r.empty())
		return;

	Json::Value root;
	Json::Reader reader;

	if (!reader.parse(r.data(), r.data() + r.size(), root))
	{
		log_error()
			<< "ext search: reading json failed: "
			<< reader.getFormatedErrorMessages() << "\n";
		return;
	}

	if (!root.isArray())
	{
		log_error()
			<< "ext search: bad json, root isn't an array\n";
		return;
	}

	ext::instance_list is;

	for (Json::Value::ArrayIndex i=0; i<root.size(); ++i)
	{
		const Json::Value v = root[i];

		extension_info ei;

		ei.guid = v["guid"].asString();
		ei.name = utf8_string(v["name"].asString());
		ei.author = utf8_string(v["author"].asString());
		ei.version = extension_version::parse(utf8_string(v["version"].asString()));
		ei.short_description = utf8_string(v["sd"].asString());
		ei.long_description = utf8_string(v["ld"].asString());
		ei.email = utf8_string(v["email"].asString());
		ei.website = utf8_string(v["website"].asString());
		ei.requires = baretorrent_version::parse(v["requires"].asString());

		if (!v["icon"].empty())
		{
			ei.icon_data = base64_decode(v["icon"].asString());
			ei.icon = create_bitmap(ei.icon_data);
		}

		if (!v["small_icon"].empty())
		{
			ei.small_icon_data = base64_decode(v["small_icon"].asString());
			ei.small_icon = create_bitmap(ei.small_icon_data);
		}

		is.push_back(ext::instance::temp(ei));
	}

	panel_->update(is);
	instances_ = is;
}

BEGIN_EVENT_TABLE(options_extension_controls, wxPanel)
	EVT_BUTTON(ext_options_id, options_extension_controls::on_options)
	EVT_BUTTON(ext_toggle_id, options_extension_controls::on_toggle)
	EVT_BUTTON(ext_uninstall_id, options_extension_controls::on_uninstall)
	EVT_BUTTON(ext_reload_id, options_extension_controls::on_reload)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(install_extension_controls, wxPanel)
	EVT_BUTTON(ext_install_id, install_extension_controls::on_install)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(ext_browser, custom_dialog)
	EVT_BUTTON(ext_search_id, ext_browser::on_search)
END_EVENT_TABLE()

}	// namespace baretorrent
