// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_EXT_BROWSER_INCLUDED
#define BARETORRENT_EXT_BROWSER_INCLUDED

#include "dialogs.h"

namespace baretorrent
{

// extension installation dialog
//
class ext_install_dialog : public wxDialog
{
public:
	ext_install_dialog(wxWindow* parent, ext::instance_ptr i);

private:
};


class extension_list : public custom_list<ext::instance>
{
	typedef custom_list<ext::instance> super;

public:
	typedef sig::signal<void (ext::instance_ptr)> ext_handler;

	ext_handler extension_selected;
	ext_handler extension_activated;


	extension_list(wxWindow* parent);

	void rebuild(const ext::instance_list& is);
	ext::instance_ptr selected_extension();

protected:
	virtual void on_selected(wxListEvent& e);
	virtual void on_activated(wxListEvent& e);

private:
	ext::instance_list instances_;
	signal_connections cons_;

	void on_changed(ext::instance_ptr i);
	void on_removed(ext::instance_ptr i);
	
	void update_item(int item, ext::instance_ptr i);
	void remove_item(int item);
};

// panel used in both the options and install dialogs
//
class extension_infos : public wxPanel
{
public:
	extension_infos(wxWindow* parent);

	// updates all the widgets with the given extension's information
	//
	void rebuild(ext::instance_ptr e);

private:
	// extension name
	wxStaticText* name_;

	// author
	wxHyperlinkCtrl* author_;

	// short description
	wxStaticText* sd_;

	// long description
	wxTextCtrl* ld_;
		
	// icon
	wxStaticBitmap* icon_;

	// website
	wxHyperlinkCtrl* website_;

	// whether this is a dev extension
	wxStaticText* dev_;

	signal_connections cons_;


	void on_changed(ext::instance_ptr e);
};

class extension_controls : public wxPanel
{
public:
	extension_controls(wxWindow* parent);

	void rebuild(ext::instance_ptr i);

protected:
	void add(wxWindow* w);
	ext::instance_ptr selection();
	virtual void on_changed_impl(ext::instance_ptr i) = 0;
	virtual utf8_string check_error(ext::instance_ptr e);

private:
	signal_connections cons_;
	wxBoxSizer* buttons_;
	ext::instance_ptr sel_;

	// error
	wxStaticText* more_;

	void on_changed(ext::instance_ptr i);
};

class install_extension_controls : public extension_controls
{
public:
	install_extension_controls(wxWindow* parent);

protected:
	virtual utf8_string check_error(ext::instance_ptr e);

private:
	DECLARE_EVENT_TABLE()

	struct install_info
	{
		ext::instance_ptr i;
		bool okay;
		bool done;

		boost::thread thread;
		boost::mutex mutex;

		install_info()
			: okay(false), done(false)
		{
		}
	};
	
	typedef boost::shared_ptr<install_info> install_info_ptr;
	typedef std::vector<install_info_ptr> install_infos;

	// install
	wxButton* install_;

	// install progress
	wxGauge* gauge_;

	// install message
	wxStaticText* what_;


	timer timer_;
	install_infos infos_;

	bool on_timer();

	// fired when the user clicks on the install button
	//
	void on_install(wxCommandEvent& e);

	void refresh(ext::instance_ptr e);
	virtual void on_changed_impl(ext::instance_ptr i);
	void install_fun(install_info_ptr ii);
};

class options_extension_controls : public extension_controls
{
public:
	options_extension_controls(wxWindow* parent);

private:
	DECLARE_EVENT_TABLE()

	// options
	wxButton* options_;
		
	// toggle
	wxButton* toggle_;

	// uninstall
	wxButton* uninstall_;

	// reload
	wxButton* reload_;


	// changes the label of the toggle button to 'disable' or 'enabled'
	// depending on the current state of the given extension
	//
	void update_toggle(ext::instance_ptr i);

	// pops up the options dialog for the given extension
	//
	void show_options(ext::instance_ptr e);

	// fired when the user clicks on the options button in the
	// extension tab; opens the options dialog
	//
	void on_options(wxCommandEvent&);

	// fired when the user clicks the enable/disable button in the
	// extensions tab
	//
	void on_toggle(wxCommandEvent&);

	// fired when the user clicks on the uninstall button in the
	// extensions tab
	//
	void on_uninstall(wxCommandEvent&);

	// fired when the reload button is clicked in the extensions tab
	//
	void on_reload(wxCommandEvent&);

	virtual void on_changed_impl(ext::instance_ptr i);
};

class extension_panel : public wxPanel
{
public:
	enum modes
	{
		install_mode = 1,
		options_mode
	};

	extension_panel(wxWindow* parent, modes m);

	void update(const ext::instance_list& is);

private:
	// splitter between extension list and bottom part
	splitter* split_;

	// extension list
	extension_list* list_;

	// panel in the bottom part that has the extension information
	// widgets
	extension_infos* info_;

	// buttons
	extension_controls* controls_;

	// bottom part of the extensions
	wxPanel* bottom_;
	

	// fired when the user selects an extension
	//
	void on_selected(ext::instance_ptr e);

	// fired when the user double clicks an extension, opens the
	// options
	//
	void on_activated(ext::instance_ptr e);
};


// used when extensions report errors, displays the error and
// information about the extension
//
class extension_error_dialog : public wxDialog
{
public:
	extension_error_dialog(
		wxWindow* parent, ext::instance_ptr i, const utf8_string& e);

private:
	wxString make_text(ext::instance_ptr i, const utf8_string& e);
};


// dialog that opens when the options button is clicked on an extension
//
class extension_options_dialog : public wxDialog
{
public:
	extension_options_dialog(wxWindow* parent, ext::instance_ptr i);

	// returns the modified options
	//
	const option_list& options() const;

private:
	DECLARE_EVENT_TABLE()

	option_list_widget* advanced_;

	// options that changed
	option_list new_;

	// extension
	ext::instance_ptr instance_;


	// fired when the user changed an option
	//
	void on_changed(const std::string& guid, option o);

	// fired when the user clicks ok, saves and dump the options
	//
	void on_ok(wxCommandEvent&);
};


class ext_browser : public custom_dialog
{
public:
	ext_browser(wxWindow* parent);
	~ext_browser();

private:
	DECLARE_EVENT_TABLE()
	
	ext::instance_list instances_;
	extension_panel* panel_;
	wxTextCtrl* search_;

	void on_search(wxCommandEvent&);
	void search(const utf8_string& s);
};

}

#endif	// BARETORRENT_EXT_BROWSER_INCLUDED
