// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "files_panel.h"
#include "torrent.h"
#include "locale.h"
#include "widgets.h"
#include "platform.h"
#include "main_window.h"

namespace baretorrent
{

enum ids
{
	// file tree
	files_id = wxID_HIGHEST + 200,

	// skip menu item
	priority_skip_id,

	// normal priority menu item
	priority_normal_id,

	// high priority menu item
	priority_high_id,

	// browse save directory
	browse_id
};

// time between the retrieving the file information from libtorrent in
// the thread
const boost::posix_time::milliseconds update_thread_delay(1000);

// update timer length
const boost::posix_time::milliseconds update_timer_delay(2000);

// context menu for the file tree
//
class file_popup : public wxMenu
{
public:
	file_popup(files_panel& fp)
		: panel_(fp), skip_(nullptr), normal_(nullptr), high_(nullptr)
	{
		wxMenu* priorities = new wxMenu;

		skip_ = priorities->Append(
			priority_skip_id, loc_wx("priority.skip"));

		normal_ = priorities->Append(
			priority_normal_id, loc_wx("priority.normal"));

		high_ = priorities->Append(
			priority_high_id, loc_wx("priority.high"));

		AppendSubMenu(priorities, loc_wx("priority"));
	}

private:
	files_panel& panel_;
	wxMenuItem* skip_;
	wxMenuItem* normal_;
	wxMenuItem* high_;
};


files_panel::files_panel(wxWindow* parent) :
	wxPanel(parent),
	tree_(nullptr), path_(nullptr), grid_(nullptr), popup_(nullptr),
	browse_(nullptr), flat_(false), metadata_(false)
{
	// this is used in create_widgets()
	grid_ = new wxFlexGridSizer(1, 3, 3);

	path_ = new wxTextCtrl(
		this, -1, wxEmptyString, wxDefaultPosition, wxDefaultSize,
		wxTE_READONLY);

	path_->SetBackgroundColour(
		path_->GetClassDefaultAttributes().colBg);

	browse_ = new wxButton(
		this, browse_id, loc_wx("browse"),
		wxDefaultPosition, wxDefaultSize, wxBU_EXACTFIT);
	browse_->Enable(false);

	wxBoxSizer* box = new wxBoxSizer(wxHORIZONTAL);

	box->Add(
		new wxStaticText(this, -1, loc_wx("torrent.save_in")),
		0, wxALIGN_CENTER_VERTICAL);
	box->AddSpacer(3);
	box->Add(path_, 1, wxEXPAND);
	box->AddSpacer(3);
	box->Add(browse_);

	grid_->AddSpacer(1);
	grid_->Add(box, wxSizerFlags().Expand());

	create_widgets(true);
	create_menu();

	grid_->AddGrowableCol(0);
	grid_->AddGrowableRow(2);

	SetSizer(grid_);

	// update timer
	timer_.start(boost::bind(
		&files_panel::on_timer, this), update_timer_delay);

	// update thread
	thread_ = boost::thread(
		boost::bind(&files_panel::update_thread, this));
}

files_panel::~files_panel()
{
	timer_.Stop();
	thread_.interrupt();
	thread_.join();

	delete popup_;
}

void files_panel::create_widgets(bool flat)
{
	// this will be called repeatedly after the panel has been created
	// when changing the selection or when metadata is received

	if (tree_)
		tree_->Destroy();

	// flat is true when there's no file hierarchy (all the files are in
	// the root folder), in which case the lines and buttons won't be
	// displayed
	long style = wxTR_HIDE_ROOT | wxTR_MULTIPLE;
	if (!flat)
		style |= wxTR_HAS_BUTTONS | wxTR_LINES_AT_ROOT;

	tree_ = new wxTreeCtrl(
		this, files_id, wxDefaultPosition, wxDefaultSize, style);
	platform::set_tree_control_style(tree_);

	grid_->Add(tree_, wxSizerFlags().Expand());
	grid_->Layout();

	tree_->AddRoot(wxEmptyString);
	items_.clear();

	flat_ = flat;
}

void files_panel::create_menu()
{
	popup_ = new file_popup(*this);
}

bool files_panel::on_timer()
{
	// this will block the update thread, but it shouldn't take long
	boost::recursive_mutex::scoped_lock lock(current_mutex_);
	update(current_);

	return true;
}

void files_panel::update_thread()
{
	try
	{
		for (;;)
		{
			torrents ts;

			{
				// makes a copy of the selected torrent so the progress
				// and priorities can be updated without blocking the
				// ui
				boost::recursive_mutex::scoped_lock
					lock(current_mutex_);

				ts = current_;
			}
			
			// only update the the info if there's a single selection
			// with metadata
			if (ts.size() == 1 && ts[0]->torrent_status().has_metadata)
			{
				// progress
				const std::vector<bt::size_type> progress =
					ts[0]->file_progress();

				// check if the thread has been interrupted
				boost::this_thread::interruption_point();

				// priorities
				const std::vector<int> priorities =
					ts[0]->file_priorities();

				// the two calls above may be slow, but the mutex is
				// only locked when the info has been retrieved
				boost::mutex::scoped_lock lock(info_mutex_);

				// there will be used in the update timer to update
				// the tree
				progress_ = progress;
				priorities_ = priorities;
			}

			boost::this_thread::sleep(update_thread_delay);
		}
	}
	catch(boost::thread_interrupted&)
	{
	}
}

void files_panel::set_selection_priority(int i)
{
	// called when the user selects a priority from the context menu
	
	torrent_ptr t;

	{
		// current_ is also used in the update thread; this only
		// locks it so the first torrent in the list (which should
		// be the only one) is retrieved
		boost::recursive_mutex::scoped_lock lock(current_mutex_);
		
		bt_assert(current_.size() == 1);
		if (current_.size() != 1)
			return;

		t = current_[0];
	}
	
	// getting the current priorities
	std::vector<int> priorities = t->file_priorities();

	wxArrayTreeItemIds ids;
	tree_->GetSelections(ids);

	// for each selected item
	foreach (wxTreeItemId id, ids)
	{
		// the index of the file is needed here since the tree items
		// may have been sorted
		set_priority(i, id, priorities);
	}

	// setting the new priorities
	t->file_priorities(priorities);
}

void files_panel::set_priority(
	int i, wxTreeItemId id, std::vector<int>& priorities)
{
	const file_t* f = dynamic_cast<const file_t*>(
		tree_->GetItemData(id));

	if (f)
	{
		// changing the priority for the selected item
		bt_assert(f->index < priorities.size());
		priorities[f->index] = i;
	}
	else
	{
		// this is probably a directory, recurse
		wxTreeItemIdValue cookie;
		wxTreeItemId cid = tree_->GetFirstChild(id, cookie);

		while (cid.IsOk())
		{
			set_priority(i, cid, priorities);
			cid = tree_->GetNextChild(cid, cookie);
		}
	}
}

void files_panel::skip()
{
	set_selection_priority(0);
}

void files_panel::normal()
{
	set_selection_priority(1);
}

void files_panel::high()
{
	set_selection_priority(2);
}

void files_panel::clear()
{
	tree_->DeleteAllItems();
	path_->SetValue(wxEmptyString);

	browse_->Enable(false);

	// current_ is also used in the update thread
	boost::recursive_mutex::scoped_lock lock(current_mutex_);
	current_.clear();
}

void files_panel::reset_tree()
{
	tree_->DeleteAllItems();
	items_.clear();

	tree_->AddRoot(wxEmptyString);
}

void files_panel::update(const torrents& tps)
{
	// this may block the update thread, but it shouldn't take long
	// anyways
	boost::recursive_mutex::scoped_lock lock(current_mutex_);

	if (tps.empty())
	{
		// no selection; only clear if there was a selection last time
		if (!current_.empty())
			clear();
	}
	else if (tps.size() == 1)
	{
		single_torrent(*tps[0]);
	}
	else
	{
		multiple_selection(tps);
	}

	current_ = tps;
}

void files_panel::single_torrent(const torrent& t)
{
	// whether the selected torrent is not the same as before
	bool different = false;

	if (current_.size() != 1)
	{
		// there was either no or more than one torrent selected before
		different = true;
	}
	else if (current_.size() == 1 && current_[0].get() != &t)
	{
		// there was also a single selection before, but it was not
		// the same torrent
		different = true;
	}

	if (different)
	{
		// progress_ and priorities_ are also used in the update thread
		boost::mutex::scoped_lock lock(info_mutex_);

		// the selection has changed and progress_ and priorities_ will
		// be filled in the update thread later; these two columns will
		// be empty for now
		progress_.clear();
		priorities_.clear();
	}

	if (!t.torrent_status().has_metadata ||
		t.status() == torrent::removing_status)
	{
		// the torrent has no metadata
		no_metadata(t);
		return;
	}

	bool new_metadata = false;
	if (!metadata_)
	{
		// the torrent has metadata, but didn't have any before; force
		// it to be different so it is updated
		new_metadata = true;
		metadata_ = true;
	}

	if (t.status() != torrent::removing_status)
		browse_->Enable(true);

	bool rebuild = false;
	const files_info fi = make_files_info(t);

	if (new_metadata)
	{
		rebuild = true;
	}
	else if (different)
	{
		if (fi.depth == 1 && !flat_)
			rebuild = true;
		else if (fi.depth != 1 && flat_)
			rebuild = true;
	}

	if (rebuild)
	{
		// the tree was for multiple selection or a torrent without
		// metadata; it needs to be rebuilt so the lines and buttons
		// are shown, but only if there are subdirectories
		create_widgets(fi.depth == 1);
	}
	else if (different)
	{
		reset_tree();
	}

	refresh(t, fi);

	// if the tree changed, expand all the items
	if (different)
		tree_->ExpandAll();
}

void files_panel::no_metadata(const torrent& t)
{
	// the tree has to be recreated to get rid of the lines and buttons
	// if there was metadata before
	if (metadata_)
		create_widgets(true);

	if (tree_->GetCount() == 0)
	{
		wxTreeItemId root = tree_->GetRootItem();
		if (!root.IsOk())
		{
			tree_->AddRoot(wxEmptyString);
			root = tree_->GetRootItem();
		}

		// dummy "no metadata" item
		tree_->InsertItem(root, 0, loc_wx("no_metadata"));
	}

	browse_->Enable(true);
	update_path(t);

	metadata_ = false;
}

void files_panel::multiple_selection(const torrents& tps)
{
	if (current_.size() <= 1)
	{
		// this is only called when the last update wasn't a multiple
		// selection so the tree has to be recreated to get rid of the
		// lines and buttons
		create_widgets(true);

		// dummy "multiple selection" item
		tree_->InsertItem(
			tree_->GetRootItem(), 0, loc_wx("multiple_selection"));
	}

	if (same_path(tps))
	{
		const torrent& t = *tps[0];
		update_path(t);
	}
	else
	{
		path_->SetValue(loc_wx("multiple_paths"));
	}

	browse_->Enable(true);

	metadata_ = false;
}

bool files_panel::same_path(const torrents& tps)
{
	utf8_path p;

	foreach (const torrent_ptr& t, tps)
	{
		if (p.empty())
		{
			p = t->save_path();
			continue;
		}

		if (t->save_path() != p)
			return false;
	}

	return true;
}

bool file_t_sort(
	const files_panel::file_t& left,
	const files_panel::file_t& right)
{
	if (natural_compare(left.name, right.name) < 0)
		return true;

	return false;
}

files_panel::files_info files_panel::make_files_info(const torrent& t)
{
	// progress_ and priorities_ are also used in the update thread and
	// are copied here so the thread doesn't block (although this should
	// be quick)
	
	std::vector<bt::size_type> progress;
	std::vector<int> priorities;

	{
		boost::mutex::scoped_lock lock(info_mutex_);
		progress = progress_;
		priorities = priorities_;
	}

	// the following builds a list of file_t's that will be sorted and
	// then passed to set_file() so update the tree


	// libtorrent uses the platform-specific directory separator to
	// build paths (but doesn't export it, unfortunately)
	const utf8_string sep = directory_separator();

	// this uses the original filenames since renaming the incomplete
	// files shouldn't be shown
	std::vector<torrent::file_info> tf = t.files();

	files_info fi;

	// for each file in the torrent
	for (std::size_t i=0; i<tf.size(); ++i)
	{
		const bt::file_entry& f = tf[i].current;

		// skip pad files
		if (f.pad_file)
			continue;

		bt::size_type pg = std::numeric_limits<bt::size_type>::max();
		int prio = -1;

		// this may happen when the update thread hasn't retrieved the
		// progress and priorities yet
		if (static_cast<std::size_t>(i) < progress.size())
			pg = progress[i];

		if (static_cast<std::size_t>(i) < priorities.size())
			prio = priorities[i];

		fi.files.push_back(file_t(
			i, utf8_string(tf[i].original.path), f.size, pg, prio));

		// this is depth of the file in the directory hierarchy
		fi.depth = std::max(
			fi.depth, explode(utf8_string(f.path), sep).size());
	}

	// sorting by filename
	std::sort(fi.files.begin(), fi.files.end(), file_t_sort);

	return fi;
}

void files_panel::refresh(
	const torrent& t, const files_info& fi)
{
	// this is called when there's a single selection that has metadata;
	// rebuild is true if that wasn't the case before so the widget can
	// be recreated with lines and buttons and tree items can be
	// initialized in set_file()

	// set each file in the tree
	for (std::size_t i=0; i<fi.files.size(); ++i)
		set_file(fi.files[i]);

	update_path(t);
}

void files_panel::update_path(const torrent& t)
{
	const wxString v = to_wxstring(t.save_path().as_utf8_string());
	const wxString old = path_->GetValue();
	
	if (old != v)
		path_->SetValue(v);
}

void files_panel::set_file(const file_t& f)
{
	const utf8_string sep = directory_separator();
	
	// splits the directories and filename
	const std::vector<utf8_string> cs = explode(f.name, sep);

	// this shouldn't happen
	bt_assert(!cs.empty());
	if (cs.empty())
		return;

	// the root is created in create_widgets()
	wxTreeItemId parent = tree_->GetRootItem();
	bt_assert(parent);

	// this loops through all the path components and adds tree items
	// for each; each component is added to the items_ map with its
	// tree item id so for a file "/home/user/file" items_ will
	// contains
	//		"/home"=>id of the "home" tree item
	//		"/home/user"=>id of the "user" tree item (under "home")
	//		"/home/user/file"=>id of the "file" tree item (under "user")
	//
	// items_ is remembered so this works whatever the order of the
	// files
	
	// built by adding the current component plus the separator
	utf8_string current;

	// this will have the id of the last item that was added to the
	// tree, that is, the file itself so its user data can be set
	// below
	wxTreeItemId last;

	// for each path component
	for (std::size_t i=0; i<cs.size(); ++i)
	{
		const utf8_string& s = cs[i];

		// appends the separator if there's already at least one
		// component in it
		if (!current.empty())
			current += sep;

		current += s;

		// text on the item
		const utf8_string caption =
			file_caption(s, f, (i < cs.size() - 1));
		
		// looking for this path in the tree item list
		items::const_iterator itor = items_.find(current);

		if (itor == items_.end())
		{
			// adding the item in the tree under the previous item
			last = tree_->AppendItem(parent, to_wxstring(caption));

			// remembering the id for this particular path
			items_[current] = last;
		}
		else
		{
			last = itor->second;

			// setting caption
			const wxString v = to_wxstring(caption);
			if (tree_->GetItemText(last) != v)
				tree_->SetItemText(last, v);
		}
		
		wxColor c;
		
		// skip is gray, normal is windowtext and high is bold

		if (f.priority == static_cast<int>(torrent::skip_priority))
			c = wxSystemSettings::GetColour(wxSYS_COLOUR_GRAYTEXT);
		else
			c = wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOWTEXT);

		bool bold = false;
		if (f.priority > static_cast<int>(torrent::normal_priority))
			bold = true;

		tree_->SetItemBold(last, bold);
		if (tree_->GetItemTextColour(last) != c)
			tree_->SetItemTextColour(last, c);

		parent = last;
	}

	// just making sure something was inserted
	bt_assert(last.IsOk());

	// deleting the old file_t, setting the new
	delete tree_->GetItemData(last);
	tree_->SetItemData(last, new file_t(f));
}

utf8_string files_panel::file_caption(
	const utf8_string& name, const file_t& f, bool is_dir)
{
	std::ostringstream oss;

	oss << name;

	// if this is a directory or the update thread hasn't had a chance
	// to run yet, the progress will not be on the caption
	if (!is_dir &&
		f.progress != std::numeric_limits<bt::size_type>::max())
	{
		const int p = make_percent(
			static_cast<double>(f.progress),
			static_cast<double>(f.size));
		oss
			<< " ("
			<< (locf("torrent.file.progress") % p).str()
			<< ")";
	}

	return utf8_string(oss.str());
}

void files_panel::on_file_popup_update(wxTreeEvent& e)
{
	if (!metadata_)
		return;

	wxTreeItemId id = e.GetItem();
	if (!id.IsOk())
		return;

	tree_->PopupMenu(popup_, e.GetPoint());
}

void files_panel::on_skip(wxCommandEvent&)
{
	skip();
}

void files_panel::on_normal(wxCommandEvent&)
{
	normal();
}

void files_panel::on_high(wxCommandEvent&)
{
	high();
}

void files_panel::on_browse(wxCommandEvent&)
{
	if (current_.size() < 1)
		return;
	
	move_torrent_storage(current_);
}

void files_panel::on_activated(wxTreeEvent& e)
{
	e.Skip();

	const file_t* f = dynamic_cast<const file_t*>(
		tree_->GetItemData(e.GetItem()));

	// not a file
	if (!f)
		return;

	if (current_.size() != 1)
		return;
	
	const std::vector<torrent::file_info> v = current_[0]->files();
	bt_assert(f->index < v.size());

	const utf8_path root = current_[0]->save_path();

	if (f->index < v.size())
	{
		const torrent::file_info& fi = v[f->index];
		const utf8_path p = root / utf8_path(utf8_string(fi.original.path));
	
		log() << "opening " << p << "\n";
		platform::open_file(p);

		e.Skip(false);
	}
}


BEGIN_EVENT_TABLE(files_panel, wxPanel)
	EVT_TREE_ITEM_RIGHT_CLICK(files_id,
		files_panel::on_file_popup_update)
	EVT_TREE_ITEM_ACTIVATED(files_id, files_panel::on_activated)
	EVT_MENU(priority_skip_id, files_panel::on_skip)
	EVT_MENU(priority_normal_id, files_panel::on_normal)
	EVT_MENU(priority_high_id, files_panel::on_high)
	EVT_BUTTON(browse_id, files_panel::on_browse)
END_EVENT_TABLE()

}	// namespace baretorrent
