// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_FILES_PANEL_INCLUDED
#define BARETORRENT_FILES_PANEL_INCLUDED

#include "utility.h"
#include "widgets.h"

namespace baretorrent
{

class torrent;

// this is the panel on the main window that contains the file list
//
// since fetching the file progress is synchronous and may hang, a
// thread is started that updates an internal list, which is picked up
// in a timer
//
class files_panel : public wxPanel
{
public:
	// creates the panel and its widgets
	//
	files_panel(wxWindow* parent);

	// cleans up the user data in the lists and stops the thread
	//
	~files_panel();
	
	// empties all the widgets
	//
	void clear();

	// updates the widgets for the given torrents
	//
	void update(const torrents& ts);

	// sets the priority of the selected files to skip
	//
	void skip();

	// sets the priority of the selected files to normal
	//
	void normal();

	// sets the priority of the selected files to high
	//
	void high();

private:
	DECLARE_EVENT_TABLE()

	typedef std::map<utf8_string, wxTreeItemId> items;

	// this is set as the user data for each element in the tree and
	// is used when sorting the tree
	//
	struct file_t : public wxTreeItemData
	{
		// index of the file in libtorrent
		std::size_t index;

		// name of the file including the path
		utf8_string name;

		// size of the file in bytes
		bt::size_type size;

		// download progress
		bt::size_type progress;

		// priority from libtorrent (not the priorities enum)
		int priority;

		file_t(
			std::size_t i,  const utf8_string& n, bt::size_type s,
			bt::size_type pr, int p) :
				index(i),
				name(n),
				size(s),
				progress(pr),
				priority(p)
		{
		}
	};

	struct files_info
	{
		std::vector<file_t> files;
		std::size_t depth;

		files_info()
			: depth(0)
		{
		}
	};

	// tree
	wxTreeCtrl* tree_;

	// root save path
	wxTextCtrl* path_;

	// grid in which the file tree is set; needed when rebuilding the
	// tree
	wxFlexGridSizer* grid_;

	// update timer
	timer timer_;

	// file list popup menu
	wxMenu* popup_;

	// browse button
	wxButton* browse_;

	// whether the current tree is flat; this avoids recreating the
	// tree if it's still flat
	bool flat_;

	// protects current_ since it's used in various places, such as
	// timers and the update thread
	boost::recursive_mutex current_mutex_;

	// currently displayed torrents
	torrents current_;

	// used when a single torrent is selected; remembers whether the
	// torrent had metadata on the last update so the whole list can
	// be recreated
	bool metadata_;

	// maps filenames to tree items, used to update specific items
	items items_;


	// runs update_thread()
	boost::thread thread_;

	// protects progress_ and priorities_ since they're used in both
	// the thread and the update timer
	boost::mutex info_mutex_;

	// a copy of the progress list from libtorrent fetched from the
	// thread
	std::vector<bt::size_type> progress_;

	// a copy of the priorities list from libtorrent fetched from the
	// thread
	std::vector<int> priorities_;

	
	// retrieves the progress and priorities from libtorrent and
	// saves them locally so they can be updated from the timer
	void update_thread();

	// creates the various widgets; flat is true when there's no
	// file hierarchy to hide the expand/collapse button and reduce the
	// indentation
	//
	void create_widgets(bool flat);

	// creates the popup menu
	//
	void create_menu();

	// sets up the widgets for torrents that have no metadata yet
	//
	void no_metadata(const torrent& t);

	// sets up the widgets for multiple torrents
	//
	void multiple_selection(const torrents& tps);

	// sets up the widgets for a single torrent; may call no_metadata()
	//
	void single_torrent(const torrent& t);

	// returns whether the given torrents are all downloading in the
	// same path
	//
	bool same_path(const torrents& tps);


	// calls update()
	//
	bool on_timer();

	// fired when the context menu pops up, sets the state of menu items
	//
	void on_file_popup_update(wxTreeEvent& e);

	// called from update(), updates the list with the current progress
	// and priorities
	//
	// if rebuild is true, this will recreate the widget; this will
	// happen if the number of files changes (dubious) or when metadata
	// arrives
	//
	void refresh(const torrent& t, const files_info& fi);

	// sets the priority of all the selected torrents to 'i'
	//
	void set_selection_priority(int i);

	void set_priority(
		int i, wxTreeItemId id, std::vector<int>& priorities);

	// called from refresh(), sets a particular item in the tree based
	// on the given info; when rebuild is true, the file name and user
	// data is also set (which isn't needed once the list is built)
	//
	void set_file(const file_t& f);

	void reset_tree();
	void update_path(const torrent& t);
	files_info make_files_info(const torrent& t);

	// returns the text that will be displayed on the tree item for
	// this file
	utf8_string file_caption(
		const utf8_string& name, const file_t& f, bool is_dir);

	void on_skip(wxCommandEvent&);
	void on_normal(wxCommandEvent&);
	void on_high(wxCommandEvent&);
	void on_browse(wxCommandEvent&);
	void on_activated(wxTreeEvent& e);

	// lexical sort on the file names
	//
	friend bool file_t_sort(const file_t& left, const file_t& right);
};

}	// namespace baretorrent

#endif	// BARETORRENT_FILES_PANEL_INCLUDED
