// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "info_panel.h"
#include "torrent.h"
#include "locale.h"
#include "widgets.h"

namespace baretorrent
{

// the goal of all this crap is to reduce the amount of duplicated code
// and allow for easily adding or removing fields
//
// setters take a torrent_ptr, a torrent_status and a wxWindow (the
// label) and returns a string that will be set on that label; only
// setter_error() needs the label because it has to change its color
//
// the macro below creates the function signature by prepending
// "setter_" in front of the given name, so the "hash" setter below
// would be preprocessed as
//
//		utf8_string setter_hash(
//			const torrent_ptr BARETORRENT_UNUSED(t),
//			const bt::torrent_status& BARETORRENT_UNUSED(ts),
//			wxWindow* BARETORRENT_UNUSED(w))
//		{
//			return to_string(ts.info_hash);
//		}
//
// the BARETORRENT_UNUSED() macro marks all the arguments as possibly
// unused since most of them are unused; however, this doesn't seem to
// work on visual c++ 11, so there's a pragma warning around the whole
// thing
//
// these setters are used below in info_panel::create_widgets()

#define s(name) \
	utf8_string setter_##name( \
		const torrent_ptr BARETORRENT_UNUSED(t), \
		const bt::torrent_status& BARETORRENT_UNUSED(ts), \
		wxWindow* BARETORRENT_UNUSED(w)) \

#if BARETORRENT_VISUALC
#pragma warning(push)
#pragma warning(disable: 4100)
#endif

s(name) { return t->name(); }
s(status) { return status_to_loc(t->status()); }
s(total_size) { return make_loc_size(t->total_size()); }
s(wanted_size) { return make_loc_size(ts.total_wanted); }
s(last_seen) { return make_loc_time(ts.last_seen_complete); }
s(num_peers) { return as_utf8_string(ts.num_peers); }
s(connection_count) { return as_utf8_string(ts.num_connections); }
s(up_queue) { return as_utf8_string(ts.up_bandwidth_queue); }
s(down_queue) { return as_utf8_string(ts.down_bandwidth_queue); }
s(dl_payload_rate) { return make_loc_rate(ts.download_payload_rate); }
s(dl_rate) { return make_loc_rate(ts.download_rate); }
s(dl_payload) { return make_loc_size(ts.total_payload_download); }
s(dl_total) { return make_loc_size(ts.total_download); }
s(dl_all_time) { return make_loc_size(ts.all_time_download); }
s(dl_time_since) { return make_loc_duration(ts.time_since_download); }
s(up_payload_rate) { return make_loc_rate(ts.upload_payload_rate); }
s(up_rate) { return make_loc_rate(ts.upload_rate); }
s(up_payload) { return make_loc_size(ts.total_payload_upload); }
s(up_total) { return make_loc_size(ts.total_upload); }
s(up_all_time) { return make_loc_size(ts.all_time_upload); }
s(up_time_since) { return make_loc_duration(ts.time_since_upload); }
s(active_time) { return make_loc_duration(ts.active_time); }
s(seeding_time) { return make_loc_duration(ts.seeding_time); }
s(added_time) { return make_loc_time(ts.added_time); }
s(completed_time) { return make_loc_time(ts.completed_time); }
s(finished_time) { return make_loc_duration(ts.finished); }
s(ratio) { return round(t->ratio(), 2); }

// these are here because they don't fit on one line

s(hash)
{
	return to_utf8_string(ts.info_hash);
}

s(error)
{
	if (t->error() == 0)
	{
		w->SetForegroundColour(normal_text_color());
		return loc("error.none");
	}
	else
	{
		w->SetForegroundColour(errored_text_color());
		return loc_error(t->error());
	}
}

s(full_copies)
{
	if (ts.distributed_copies == -1)
		return utf8_string("--");
	else
		return as_utf8_string(ts.distributed_copies);
}

s(seeders)
{
	if (ts.num_complete == -1)
		return utf8_string("--");
	else
		return as_utf8_string(ts.num_complete);
}

s(leechers)
{
	if (ts.num_incomplete == -1)
		return utf8_string("--");
	else
		return as_utf8_string(ts.num_incomplete);
}

s(upload_limit)
{
	const int i = ts.uploads_limit;
	if (i < 0)
		return loc("unlimited");
	else
		return as_utf8_string(i);
}

s(connections_limit)
{
	const int i = ts.connections_limit;
	if (i < 0)
		return loc("unlimited");
	else
		return as_utf8_string(i);
}

s(dl_wasted)
{
	return make_loc_size(
		ts.total_failed_bytes + ts.total_redundant_bytes);
}

#if BARETORRENT_VISUALC
#pragma warning(pop)
#endif

#undef s


info_panel::info_panel(wxWindow* parent)
	: wxPanel(parent), progress_(nullptr), name_(nullptr)
{
	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);

	root->Add(create_widgets(), wxSizerFlags().Expand());

	SetSizer(root);
}

wxSizer* info_panel::create_widgets()
{
	wxGridBagSizer* grid = new wxGridBagSizer(0, 10);

	grid->SetEmptyCellSize(wxSize(5, 5));

	int row = 0;
	wxStaticText* wg = nullptr;

	progress_ = new wxGauge(
		this, -1, 100, wxDefaultPosition, wxDefaultSize,
		wxGA_HORIZONTAL | wxGA_SMOOTH);
	
	// the first macro 'a' ("add") creates a new label widget, adds it
	// to the grid at the given coordinates and adds the setter to the
	// list
	//
	// the second macro 'aw' ("add with widget") is meant for widgets
	// that don't have a setter (such as captions) and just adds the
	// given widget to the grid
	//
	// 'row' is incremented between rows so that adding or removing a
	// row doesn't change everything after it

#define a(f, x, y, w, h)											\
		wg = new wxStaticText(										\
			this, -1, loc_wx("not_available"));						\
		grid->Add(wg, wxGBPosition(x, y), wxGBSpan(w, h), wxEXPAND);\
		setters_.push_back(setter(wg, f));
	
#define aw(wg, x, y, w, h)	\
		grid->Add(wg, wxGBPosition(x, y), wxGBSpan(w, h), wxEXPAND);

	// general
	aw(progress_, row, 0, 1, 5);
	row += 2;

	aw(make_text("torrent.name.caption"), row, 0, 1, 1);
	++row;
	name_ = a(setter_name, row, 0, 1, 3);
	row += 2;
	aw(make_text("torrent.hash.caption"), row, 0, 1, 1);
	++row;
	a(setter_hash, row, 0, 1, 3);
	
	row += 2;

	aw(make_text("torrent.status.caption"), row, 0, 1, 1);
	aw(make_text("torrent.total_size.caption"), row, 1, 1, 1);
	aw(make_text("torrent.wanted_size.caption"), row, 2, 1, 1);
	++row;
	a(setter_status, row, 0, 1, 1);
	a(setter_total_size, row, 1, 1, 1);
	a(setter_wanted_size, row, 2, 1, 1);

	row += 2;

	aw(make_text("torrent.error.caption"), row, 0, 1, 3);
	++row;
	a(setter_error, row, 0, 1, 3);

	row += 3;


	// swarm
	aw(make_text("swarm", true), row, 0, 1, 5);
	row += 2;

	aw(make_text("torrent.full_copies.caption"), row, 0, 1, 1);
	aw(make_text("torrent.time.last_seen"), row, 1, 1, 2);
	++row;
	a(setter_full_copies, row, 0, 1, 1);
	a(setter_last_seen, row, 1, 1, 2);

	row += 2;

	aw(make_text("torrent.peers.caption"), row, 0, 1, 1);
	aw(make_text("torrent.seeders.caption"), row, 1, 1, 1);
	aw(make_text("torrent.leechers.caption"), row, 2, 1, 1);
	++row;
	a(setter_num_peers, row, 0, 1, 1);
	a(setter_seeders, row, 1, 1, 1);
	a(setter_leechers, row, 2, 1, 1);

	row += 3;


	// network
	aw(make_text("network", true), row, 0, 1, 5);

	row += 2;

	aw(make_text("torrent.connections.caption"), row, 0, 1, 1);
	aw(make_text("torrent.upload_slots.caption"), row, 1, 1, 1);
	aw(make_text("torrent.max_connections.caption"), row, 2, 1, 1);
	++row;
	a(setter_connection_count, row, 0, 1, 1);
	a(setter_upload_limit, row, 1, 1, 1);
	a(setter_connections_limit, row, 2, 1, 1);

	row += 3;


	// download
	aw(make_text("download", true), row, 0, 1, 5);

	row += 2;

	aw(make_text("torrent.download.rate.caption"), row, 0, 1, 1);
	aw(make_text("torrent.download.session.caption"), row, 1, 1, 1);
	aw(make_text("torrent.download.all_time.caption"), row, 2, 1, 1);
	++row;
	a(setter_dl_payload_rate, row, 0, 1, 1);
	a(setter_dl_payload, row, 1, 1, 1);
	a(setter_dl_all_time, row, 2, 1, 1);
	
	row += 3;


	// upload
	aw(make_text("upload", true), row, 0, 1, 5);

	row += 2;

	aw(make_text("torrent.upload.rate.caption"), row, 0, 1, 1);
	aw(make_text("torrent.upload.session.caption"), row, 1, 1, 1);
	aw(make_text("torrent.upload.all_time.caption"), row, 2, 1, 1);
	++row;
	a(setter_up_payload_rate, row, 0, 1, 1);
	a(setter_up_payload, row, 1, 1, 1);
	a(setter_up_all_time, row, 2, 1, 1);

	row += 2;

	aw(make_text("torrent.upload.time_since.caption"), row, 0, 1, 1);
	aw(make_text("torrent.upload.ratio.caption"), row, 1, 1, 1);
	++row;
	a(setter_up_time_since, row, 0, 1, 1);
	a(setter_ratio, row, 1, 1, 1);

	row += 3;


	// times
	aw(make_text("times", true), row, 0, 1, 5);

	row += 2;

	aw(make_text("torrent.time.active.caption"), row, 0, 1, 1);
	aw(make_text("torrent.time.seeding.caption"), row, 1, 1, 1);
	aw(make_text("torrent.time.finished.caption"), row, 2, 1, 1);
	++row;
	a(setter_active_time, row, 0, 1, 1);
	a(setter_seeding_time, row, 1, 1, 1);
	a(setter_finished_time, row, 2, 1, 1);
	
	row += 2;

	aw(make_text("torrent.time.added.caption"), row, 0, 1, 2);
	aw(make_text("torrent.time.completed.caption"), row, 2, 1, 2);
	++row;
	a(setter_added_time, row, 0, 1, 2);
	a(setter_completed_time, row, 2, 1, 2);

#undef aw
#undef a

	grid->AddGrowableCol(0);
	grid->AddGrowableCol(1);
	grid->AddGrowableCol(2);
	grid->AddGrowableCol(3);
	grid->AddGrowableCol(4);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(grid, 0, wxEXPAND);

	return box;
}

wxStaticText* info_panel::make_text(const std::string& name, bool bold)
{
	wxStaticText* t = new wxStaticText(this, -1, loc_wx(name));
	
	if (bold)
	{
		wxFont f = t->GetFont();
		f.SetWeight(wxFONTWEIGHT_BOLD);
		t->SetFont(f);
	}

	return t;
}

void info_panel::update(const torrents& tps)
{
	if (tps.empty())
		clear();
	else if (tps.size() == 1)
		single_torrent(tps[0]);
	else
		multiple_selection();
}

void info_panel::single_torrent(const torrent_ptr t)
{
	if (t->status() == torrent::removing_status)
	{
		clear();
		return;
	}

	const bt::torrent_status& ts = t->torrent_status();
	
	// goes through each setter and sets its widget
	foreach (setter& s, setters_)
	{
		// asks the setter for a new value
		const wxString n = to_wxstring(s.f(t, ts, s.widget));

		// only set the widget if the value is different, prevents
		// flickering
		if (s.widget->GetLabel() != n)
			s.widget->SetLabel(n);
	}

	progress_->SetValue(make_percent(ts.progress));
}

void info_panel::multiple_selection()
{
	clear();
	name_->SetLabel(loc_wx("multiple_selection"));
}

void info_panel::clear()
{
	progress_->SetValue(0);
	foreach (setter& s, setters_)
	{
		s.widget->SetLabel(to_wxstring(utf8_string("--")));

		// some setters (like error) may change the text color; make
		// sure it's always the normal color
		s.widget->SetForegroundColour(normal_text_color());
	}
}

}	// namespace baretorrent
