// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"

#if BARETORRENT_LINUX

#include <pwd.h>
#include <unistd.h>
#include <gio/gio.h>
#include <gio/gdesktopappinfo.h>
#include <gconf/gconf.h>
#include <gconf/gconf-client.h>
#include "platform.h"
#include "locale.h"
#include "utility.h"
#include "widgets.h"
#include "path.h"

namespace baretorrent
{

namespace platform
{

// todo: icons will (hopefully) be installed in /usr/share somewhere
// eventually
const utf8_path icon_path(utf8_string("./res"));

// sort indicator stuff comes from
// https://groups.google.com/forum/?fromgroups=
// 				#!topic/wx-users/Upz7xB-8lsU

static const char* sort_asc_xpm[] = {
	"16 16 2 1",
	". c #ACA899",
	"  c #FF0000",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"        .       ",
	"       ...      ",
	"      .....     ",
	"     .......    ",
	"    .........   ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                "};

static const char* sort_desc_xpm[] = {
	"16 16 2 1",
	". c #ACA899",
	"  c #FF0000",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"    .........   ",
	"     .......    ",
	"      .....     ",
	"       ...      ",
	"        .       ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                "};

typedef std::map<images::image, utf8_string> image_map;

utf8_path make_image_path(images::image i);


platform::platform()
{
	// noop
}

platform::~platform()
{
	// noop
}

long toolbar_style()
{
	// todo: this looks fine on gtk, but I haven't tested other wms
	// yet
	return wxBORDER_NONE | wxTB_FLAT;
}

void init_list_control(wxListCtrl* w)
{
	std::auto_ptr<wxImageList> il(new wxImageList);
	il->Add(wxBitmap(sort_asc_xpm), wxColor(255, 0, 0));
	il->Add(wxBitmap(sort_desc_xpm), wxColor(255, 0, 0));

	w->SetImageList(il.release(), wxIMAGE_LIST_SMALL);
}

void destroy_list_control(wxListCtrl* w)
{
	delete w->GetImageList(wxIMAGE_LIST_SMALL);
}

void set_list_control_style(wxListCtrl*)
{
	// noop
}

void set_tree_control_style(wxTreeCtrl*)
{
	// noop
}

void set_list_sort_column(wxListCtrl* w, std::size_t c, int d)
{
	wxListItem item;
	item.SetMask(wxLIST_MASK_IMAGE);

	for (int i=0; i<w->GetColumnCount(); ++i)
	{
		if (i == static_cast<int>(c))
		{
			if (d > 0)
				item.SetImage(0);
			else
				item.SetImage(1);
		}
		else
		{
			item.SetImage(-1);
		}

		w->SetColumn(i, item);
	}
}

void open_directory(const utf8_path& p)
{
	// todo: find other ways if xdg isn't available
	const std::string s =
		"xdg-open \"" + p.as_utf8_string().std_string() + "\"";
	system(s.c_str());
}

void open_file(const utf8_path& p)
{
	const std::string s =
		"xdg-open \"" + p.as_utf8_string().std_string() + "\"";
	system(s.c_str());
}

void add_to_path(const utf8_path& p)
{
	// todo
}

// this is called by download_directory() when the known download
// directory isn't found; it will look for non-localized, hardcoded
// directory names (such as "Downloads") before bailing and returning
// root (which is probably the home directory)
//
// this is somewhat crap, but it might be useful
//
utf8_path try_download_dirs(const utf8_path& root)
{
	const utf8_string downloads("Downloads");

	if ((root / downloads).exists())
	{
		log()
			<< "there's a " << downloads << " directory in "
			<< root << ", taking it\n";

		return root / downloads;
	}

	return root;
}

// this tries to find the home directory in case xdg fails
//
utf8_path posix_home_directory()
{
	// HOME environment variable
	const char* e = getenv("HOME");
	if (e)
		return utf8_path(utf8_string(e));
	
	log_error() << "$HOME doesn't exist, trying getpwuid\n";

	// getpwuid; if this fails, the user has other problems than a
	// failing baretorrent
	long buffer_size = sysconf(_SC_GETPW_R_SIZE_MAX);
	if (buffer_size == -1)
		buffer_size = 1024 * 16;

	boost::shared_array<char> buffer(new char[buffer_size]);
	
	passwd pwd;
	passwd* pwd_r = 0;

	int r = getpwuid_r(getuid(), &pwd, buffer.get(), buffer_size, &pwd_r);
	if (!r)
		return utf8_path(utf8_string(pwd.pw_dir));

	log_error()
	   	<< "getpwuid_r() failed, " << std::strerror(r)
	   	<< " (" << r << ")\n";

	return utf8_path();
}

// uses xdg-user-dir to find the home directory
//
utf8_path xdg_home_directory()
{
	utf8_string s = shell_output(utf8_string("xdg-user-dir"));
	if (!s.empty())
		return utf8_path(utf8_string(s));

	log_error() << "xdg-user-dir failed\n";
	return utf8_path();
}

utf8_path home_directory()
{
	utf8_path p = xdg_home_directory();
	if (!p.empty())
		return p;

	log_error() << "xdg failed, trying posix\n";
	p = posix_home_directory();
	if (!p.empty())
		return p;

	log_error() << "posix failed, falling back on /tmp\n";

	return utf8_path(utf8_string("/tmp"));
}

utf8_path download_directory()
{
	utf8_string s =
		shell_output(utf8_string("xdg-user-dir DOWNLOAD"));
	if (!s.empty())
		return utf8_path(utf8_string(s));

	log_error()
		<< "xdg-user-dir for DOWNLOAD failed, "
		<< "looking for home directory\n";

	// this will try stuff like Downloads in the home or fallback on
	// the home
	return try_download_dirs(home_directory());
}

utf8_path conf_directory()
{
	// configuration will be stored in $HOME/.baretorrent
	return home_directory();
}

utf8_string conf_directory_name()
{
	return utf8_string(".baretorrent");
}

utf8_path executable_path()
{
	std::size_t size = 1000;
	boost::shared_array<char> path;

	for (int i=0; i<3; ++i)
	{
		path.reset(new char[size + 1]);
		std::memset(path.get(), 0, size + 1);
		
		ssize_t out_size =
	   		readlink("/proc/self/exe", path.get(), size);

		if (out_size == -1)
		{
			log_error()
			   << "readlink failed, "
			   << std::strerror(errno) << "\n";

			return utf8_path();
		}

		if (static_cast<std::size_t>(out_size) < size)
			break;

		size *= 2;
	}

	return utf8_path(utf8_string(path.get()));
}

utf8_path executable_directory()
{
	return executable_path().parent();
}

utf8_string slurp(const utf8_path& p)
{
	std::ifstream in(p.as_utf8_string().c_str());
	if (!in)
		return utf8_string();

	std::stringstream ss;
	ss << in.rdbuf();

	return utf8_string(ss.str());
}

struct desktop_file
{
	utf8_string content;
	utf8_string name;
	utf8_string exec;
};

utf8_string desktop_file_exec()
{
	return executable_path().as_utf8_string() + utf8_string(" %U");
}

utf8_string desktop_file_name()
{
	return utf8_string("baretorrent.desktop");
}

utf8_path xdg_apps_directory()
{
	return utf8_path(
		home_directory() / utf8_string(".local/share/applications"));
}

utf8_path xdg_global_apps_directory()
{
	return utf8_path(utf8_string("/usr/share/applications"));
}

utf8_string make_desktop_file()
{
	std::ostringstream oss;

	const utf8_string name("baretorrent");
	const utf8_string exec = desktop_file_exec();
	const utf8_string path = executable_directory().as_utf8_string();
	const utf8_string icon =
		make_image_path(images::icon_128).as_utf8_string();

	oss
		<< "[Desktop Entry]\n"
		<< "Type=Application\n"
		<< "Name=baretorrent\n"
		<< "GenericName=BitTorrent Client\n"
		<< "Comment=Share files using BitTorrent\n"
		<< "Icon=" << icon << "\n"
		<< "Exec=" << exec << "\n"
		<< "Path=" << path << "\n"
		<< "Terminal=false\n"
		<< "MimeType=application/x-bittorrent;x-scheme-handler/magnet\n"
		<< "Categories=Network;P2P";

	return utf8_string(oss.str());
}

desktop_file get_desktop_file(
	const utf8_path& root, const utf8_string& file)
{
	log() << "mime: looking for " << file << " in " << root << "\n";

	utf8_path p = root / file;
	if (!p.exists())
	{
		boost::filesystem::recursive_directory_iterator itor(
			root.internal_fs_path());
		boost::filesystem::recursive_directory_iterator end;

		for (; itor != end; ++itor)
		{
			const boost::filesystem::path& pp = *itor;
			if (!is_directory(pp))
				continue;

			p = utf8_path::internal_create(pp) / file;
			if (p.exists())
				break;
		}
	}

	if (!p.exists())
	{
		log() << "assoc: could not find " << file << "\n";
		return desktop_file();
	}

	const utf8_string ds = slurp(p);
	if (ds.empty())
	{
		log() << "assoc: can't read desktop file\n";
		return desktop_file();
	}

	log() << "assoc: desktop file found at '" << p << "'\n";

	desktop_file df;
	df.content = ds;

	boost::regex e("^Name=(.*)$");
	boost::smatch m;
	const boost::match_flag_type flag =
		boost::regex_constants::match_not_dot_newline;

	if (!regex_search(ds.std_string(), m, e, flag))
	{
		log() << "assoc: Name line not found\n";
		return desktop_file();
	}

	df.name = utf8_string(m[1]);

	e = boost::regex("^Exec=(.*)$");
	if (!regex_search(ds.std_string(), m, e, flag))
	{
		log() << "assoc: Exec line not found in desktop file\n";
		return desktop_file();
	}

	df.exec = utf8_string(m[1]);

	return df;
}

assoc check_mime(const utf8_string& mime)
{
	const utf8_string cmd =
		utf8_string("xdg-mime query default ") + mime;
	const utf8_string file = shell_output(cmd);

	log() << "mime: checking " << mime << " with `" << cmd << "`\n";

	if (file.empty())
	{
		log() << "mime: no default for '" << mime << "'\n";
		return assoc(false);
	}

	log() << "assoc: default for '" << mime << "' is '" << file << "'\n";

	desktop_file df = get_desktop_file(xdg_apps_directory(), file);

	if (df.content.empty())
	{
		df = get_desktop_file(xdg_global_apps_directory(), file);

		if (df.content.empty())
		{
			log() << "mime: failed, bailing out\n";
			return assoc(false);
		}
	}

	const utf8_string this_name("baretorrent");
	if (df.name != this_name)
	{
		log()
			<< "mime: Name line has '" << df.name << "' instead of "
			<< "'" << this_name << "'\n";
		return assoc(false, df.name);
	}

	log() << "mime: Name line matches '" << this_name << "'\n";

	const utf8_string this_exe = desktop_file_exec();

	if (this_exe != df.exec)
	{
		log()
			<< "mime: Exec line has '" << df.exec << "' while this exe "
			<< "is '" << this_exe << "'\n";
		return assoc(false);
	}

	log() << "mime: Exec line matches '" << this_exe << "'\n";

	return assoc(true);
}

assoc check_gnome_mime(const utf8_string& mime)
{
	log() << "gnome: checking mime '" << mime << "'\n";

	GAppInfo* app =
		g_app_info_get_default_for_type(mime.c_str(), false);

	if (!app)
	{
		log() << "gnome: nothing associated with " << mime << "\n";
		return assoc(false);
	}

	const utf8_string this_name("baretorrent");
	const utf8_string this_exec(executable_path().as_utf8_string());

	const utf8_string app_name(g_app_info_get_name(app));
	const utf8_string app_exec(g_app_info_get_executable(app));

	if (app_name != this_name)
	{
		log()
			<< "gnome: " << mime << " associated with name "
			<< "'" << app_name << "' instead of '" << this_name << "'\n";
		return assoc(false, app_name);
	}

	if (app_exec != this_exec)
	{
		log()
			<< "gnome: " << mime << " associated with exec "
			<< "'" << app_exec << "' instead of '" << this_exec << "'\n";
		return assoc(false);
	}

	log() << "gnome: " << mime << " associated correctly\n";

	return assoc(true);
}

template <class T>
class g_object_holder : boost::noncopyable
{
public:
	g_object_holder(T* t)
		: t_(t)
	{
	}

	~g_object_holder()
	{
		if (t_)
			g_object_unref(t_);
	}

	operator T*()
	{
		return t_;
	}

private:
	T* t_;
};

class gconf_value_holder : public boost::noncopyable
{
public:
	gconf_value_holder(GConfValue* v)
		: v_(v)
	{
	}

	~gconf_value_holder()
	{
		if (v_)
			gconf_value_free(v_);
	}

	operator GConfValue*()
	{
		return v_;
	}

private:
	GConfValue* v_;
};

utf8_string error_to_string(GError* e)
{
	if (!e)
		return utf8_string();

	utf8_string s =
		utf8_string(" (") + utf8_string(e->message) + utf8_string(")");

	g_error_free(e);

	return s;
}

template <class T>
bool check_gconf_value(
	GConfClient* c, const utf8_string& name, const T& expected);

template <>
bool check_gconf_value<utf8_string>(
	GConfClient* c, const utf8_string& name, const utf8_string& expected)
{
	GError* e = nullptr;
	gconf_value_holder v(gconf_client_get(c, name.c_str(), &e));

	if (!v || e)
	{ 
		log()
			<< "gconf: can't get '" << name << "'"
			<< error_to_string(e) << "\n";
		return false;
	}

	const utf8_string value(gconf_value_get_string(v));

	if (value != expected)
	{
		log()
			<< "gconf: " << name << " is '" << value << "' "
			<< "instead of '" << expected << "'"
			<< error_to_string(e) << "\n";
		return false;
	}

	return true;
}

template <>
bool check_gconf_value<bool>(
	GConfClient* c, const utf8_string& name, const bool& expected)
{
	GError* e = nullptr;
	gconf_value_holder v(gconf_client_get(c, name.c_str(), &e));

	if (!v || e)
	{ 
		log()
			<< "gconf: can't get '" << name << "'"
			<< error_to_string(e) << "\n";
		return false;
	}

	const bool value = (gconf_value_get_bool(v) == 0 ? false : true);

	if (value != expected)
	{
		log()
			<< "gconf: " << name << " is '" << value << "' "
			<< "instead of '" << expected << "'"
			<< error_to_string(e) << "\n";
		return false;
	}

	return true;
}

template <class T>
void set_gconf_value(
	GConfClient* c, const utf8_string& name, const T& v);

template <>
void set_gconf_value<utf8_string>(
	GConfClient* c, const utf8_string& name, const utf8_string& value)
{
	gconf_value_holder v(gconf_value_new(GCONF_VALUE_STRING));
	gconf_value_set_string(v, value.c_str());
	
	GError* e = nullptr;
	gconf_client_set(c, name.c_str(), v, &e);

	if (e)
	{
		log_error()
			<< "gconf: can't set value '" << name << "' to "
			<< "'" << value << "'" << error_to_string(e) << "\n";
		return;
	}

	log() << "gconf: '" << name << "' set to '" << value << "'\n";
}

template <>
void set_gconf_value<bool>(
	GConfClient* c, const utf8_string& name, const bool& value)
{
	gconf_value_holder v(gconf_value_new(GCONF_VALUE_BOOL));
	gconf_value_set_bool(v, value);
	
	GError* e = nullptr;
	gconf_client_set(c, name.c_str(), v, &e);

	if (e)
	{
		log_error()
			<< "gconf: can't set value '" << name << "' to "
			<< "'" << value << "'" << error_to_string(e) << "\n";
		return;
	}

	log() << "gconf: '" << name << "' set to '" << value << "'\n";
}

assoc check_gconf_scheme(const utf8_string& scheme, bool create)
{
	g_object_holder<GConfClient> c(gconf_client_get_default());
	if (!c)
	{
		log_error() << "gconf: can't get default client\n";
		return assoc(false);
	}

	const utf8_string root(
		utf8_string("/desktop/gnome/url-handlers/") + scheme);
	const utf8_string cmd(root + utf8_string("/command"));
	const utf8_string enabled(root + utf8_string("/enabled"));
	const utf8_string term(root + utf8_string("/needs_terminal"));
	
	if (!check_gconf_value(c, cmd, desktop_file_exec()))
	{
		if (create)
			set_gconf_value(c, cmd, desktop_file_exec());
		else
			return assoc(false);
	}

	if (!check_gconf_value(c, enabled, true))
	{
		if (create)
			set_gconf_value(c, enabled, true);
		else
			return assoc(false);
	}

	if (!check_gconf_value(c, term, false))
	{
		if (create)
			set_gconf_value(c, term, false);
		else
			return assoc(false);
	}

	return assoc(true);
}

assoc associated()
{
	log() << "assoc: checking\n";

	if (!has_xdg())
	{
		log_error()
			<< "xdg not available, can't check for associations\n";
		return assoc(true);
	}

	assoc a;

	a = check_mime(utf8_string("application/x-bittorrent"));
	if (!a.okay)
		return a;

	a = check_mime(utf8_string("x-scheme-handler/magnet"));
	if (!a.okay)
		return a;

	a = check_gnome_mime(utf8_string("application/x-bittorrent"));
	if (!a.okay)
		return a;

	a = check_gnome_mime(utf8_string("x-scheme-handler/magnet"));
	if (!a.okay)
		return a;

	a = check_gconf_scheme(utf8_string("magnet"), false);
	if (!a.okay)
		return a;

	log() << "assoc: all okay\n";

	return assoc(true);
}

void write_desktop_file()
{
	const utf8_path root = xdg_apps_directory();

	log() << "assoc: creating " << root << "\n";
	create_directories(root);

	const utf8_path p(root / desktop_file_name());
	log() << "assoc: writing desktop file to " << p << "\n";
	std::ofstream out(p.as_utf8_string().c_str());

	if (!out)
	{
		log_error()
			<< "assoc: couldn't open " << p << " for writing\n";
		return;
	}

	out << make_desktop_file();
	out.close();
}

void xdg_mime_default(const utf8_string& mime)
{
	const utf8_string cmd(
		utf8_string("xdg-mime default ") +
		utf8_string("\"") + desktop_file_name() + utf8_string("\" ") +
		mime);

	log() << "assoc: executing `" << cmd << "`\n";

	const int sr = system(cmd.c_str());
	const int r = WEXITSTATUS(sr);

	if (r != 0)
		log_error() << "assoc: failed\n";
}

void update_desktop_database()
{
	const utf8_string gnome(
		utf8_string("update-desktop-database ") +
		xdg_apps_directory().as_utf8_string());

	log() << "assoc: updating with `" << gnome << "`\n";
	int sr = system(gnome.c_str());
	int r = WEXITSTATUS(sr);

	if (r != 0)
		log_error() << "assoc: update failed (" << r << "), ignoring\n";

	const utf8_string kde(utf8_string("kbuildsycoca4"));
	
	log() << "assoc: updating with `" << kde << "`\n";

	if (r != 0)
		log_error() << "assoc: update failed (" << r << "), ignoring\n";
}

void gnome_mime_default(const utf8_string& mime)
{
	log() << "gnome: setting default for '" << mime << "'\n";

	const utf8_path df = xdg_apps_directory() / desktop_file_name();

	GError* e = nullptr;

	/*GAppInfo* app = g_app_info_create_from_commandline(
			executable_path().as_utf8_string().c_str(), "baretorrent",
			G_APP_INFO_CREATE_SUPPORTS_URIS, &e);*/

	log() << "gnome: getting " << df << "\n";
	GAppInfo* app = (GAppInfo*)g_desktop_app_info_new(
		desktop_file_name().c_str());

	if (!app || e)
	{
		log_error()
			<< "gnome: can't create app info"
			<< error_to_string(e) << "\n";
		return;
	}

	g_app_info_set_as_default_for_type(app, mime.c_str(), &e);

	if (e)
	{
		log_error()
			<< "gnome: can't set as default, "
			<< error_to_string(e) << "\n";
		return;
	}

	log() << "gnome: default set\n";
}

void associate()
{
	try
	{
		write_desktop_file();
		xdg_mime_default(utf8_string("application/x-bittorrent"));
		xdg_mime_default(utf8_string("x-scheme-handler/magnet"));
		gnome_mime_default(utf8_string("application/x-bittorrent"));
		gnome_mime_default(utf8_string("x-scheme-handler/magnet"));
		check_gconf_scheme(utf8_string("magnet"), true);

		update_desktop_database();

		log() << "assoc: rechecking\n";

		if (!associated().okay)
		{
			log_error() << "assoc: failed, still not associated\n";
			return;
		}

		log() << "assoc: done\n";
	}
	catch(path_exception& e)
	{
		log_error() << "assoc: " << e.what() << "\n";
	}
}

void power(power_actions p)
{
	switch (p)
	{
		case shutdown_power:
		{
			const utf8_string c(
				"dbus-send --system --dest=org.freedesktop.ConsoleKit "
				"--type=method_call --print-reply --reply-timeout=2000 "
				"/org/freedesktop/ConsoleKit/Manager "
				"org.freedesktop.ConsoleKit.Manager.Stop");
			
			const int r = system(c.c_str());
			if (!WIFEXITED(r))
			{
				log_error()
					<< "shutdown failed with " << r << ", "
					"errno is " << std::strerror(errno) << "\n";
			}

			break;
		}

		case suspend_power:
		{
			const utf8_string upower(
				"dbus-send --system --dest=org.freedesktop.UPower "
				"--print-reply /org/freedesktop/UPower "
				"org.freedesktop.UPower.Suspend");

			log() << "suspend: " << upower << "\n";

			const int r = system(upower.c_str());
			if (!WIFEXITED(r) || WEXITSTATUS(r))
				log_error() << "suspend failed\n";
			
			break;
		}
	
		case hibernate_power:
		{
			const utf8_string upower(
				"dbus-send --system --dest=org.freedesktop.UPower "
				"--print-reply /org/freedesktop/UPower "
				"org.freedesktop.UPower.Hibernate");

			log() << "hibernate: " << upower << "\n";

			const int r = system(upower.c_str());
			if (!WIFEXITED(r) || WEXITSTATUS(r))
				log_error() << "hibernate failed\n";
			
			break;
		}

		case no_power_action:	// fall-through
		default:
		{
			break;
		}
	}
}

bool has_power(power_actions p)
{
	switch (p)
	{
		case shutdown_power:
		{
			return true;
		}

		case suspend_power:
		{
			const utf8_string c("pm-is-supported --suspend");
			return executable_exists(c);
		}

		case hibernate_power:
		{
			const utf8_string c("pm-is-supported --hibernate");
			return executable_exists(c);
		}

		case no_power_action:	// fall-through
		default:
		{
			return false;
		}
	}
}

bool has_power_management()
{
	const utf8_string upower("which upower");
	return executable_exists(upower);
}

void add_image_handlers()
{
	// noop
}

image_map get_image_map()
{
	image_map m;

#define a(e, s) m.insert(std::make_pair(images::e, utf8_string(s)))

	// todo: find a way of getting the native icon size, this is
	// hardcoded to 16x16, which is fine for gnome
	a(add, "add_16");
	a(remove, "remove_16");
	a(resume, "resume_16");
	a(pause, "pause_16");
	a(up, "up_16");
	a(down, "down_16");
	a(top, "top_16");
	a(bottom, "bottom_16");
	a(active, "active_16");
	a(upload, "upload_16");
	a(download, "download_16");
	a(inactive, "inactive_16");
	a(all, "all_16");
	a(creeper, "creeper");
	a(icon_16, "icon_16");
	a(icon_24, "icon_24");
	a(icon_32, "icon_32");
	a(icon_48, "icon_48");
	a(icon_64, "icon_64");
	a(icon_128, "icon_128");
	a(close_search, "close_search");
	a(pcon_16, "pcon_16");

#undef a

	return m;
}

utf8_path make_image_path(images::image i)
{
	const image_map m = get_image_map();
	image_map::const_iterator itor = m.find(i);
	bt_assert(itor != m.end());

	if (itor == m.end())
		return utf8_path();

	const utf8_string& name = itor->second;

	return
		executable_directory() /
		icon_path / (name + utf8_string(".png"));
}


wxBitmap make_image_unsafe(images::image i)
{
	log_redirector lg;

	const utf8_path p = make_image_path(i);
	if (p.empty())
		return make_bad_image();

	// this gets the png from the icon path
	wxBitmap b(to_wxstring(p.as_utf8_string()), wxBITMAP_TYPE_PNG);

	if (!b.IsOk())
	{
		log_error()
			<< "couldn't load bitmap '" << name << "', "
			<< lg.what() << "\n";
	}
	
	return b;
}

wxBitmap make_image(images::image i)
{
	wxBitmap b = make_image_unsafe(i);
	if (!b.IsOk())
		return make_bad_image();

	return b;
}

wxIcon make_icon(images::image i)
{
	// no difference between images and icons
	wxIcon icon;
	icon.CopyFromBitmap(make_image(i));

	return icon;
}

utf8_string make_time(const utf8_string& format, std::time_t s)
{
	const std::size_t buffer_size = 500;
	char buffer[buffer_size + 1] = {0};

	std::tm t = *std::localtime(&s);

	const std::size_t n = std::strftime(
		buffer, buffer_size, format.c_str(), &t);

	if (n == 0)
	{
		log_error() << "strftime failed\n";
		return utf8_string();
	}

	return utf8_string(buffer);
}

void add_firewall_exception()
{
	// noop
}

void detach_console()
{
	// noop
}

void create_console()
{
	// noop
}

bool colored_terminal()
{
	static bool checked = false;
	static bool result = false;

	if (!checked)
	{
		const utf8_string o =
			shell_output(utf8_string("tput colors"));

		if (o.empty())
		{
			result = false;
		}
		else
		{
			int i = as<int>(o);
			if (i > 0)
				result = true;
		}

		checked = true;
	}

	return result;
}

void log_to_console(
	std::ostream* out, log_t::levels lv, const utf8_string& s)
{
	if (!out)
		return;

	const bool c = colored_terminal();

	if (c && lv == log_t::error_level)
		std::clog << "\33[0;31m";

	*out << s.std_string();

	if (c && lv == log_t::error_level)
		std::clog << "\33[0m";
}

utf8_string name()
{
	// trying lsb
	utf8_string s = shell_output(utf8_string("lsb_release -ds"));
	if (s.empty())
	{
		s = shell_output(utf8_string("uname -sr"));
		if (s.empty())
			s = utf8_string("some linux");
	}

	// if lsb fails, just use uname which doesn't give as much
	// information

	utf8_string a = shell_output(utf8_string("uname -m"));
	if (!a.empty())
		s += utf8_string(" ") + a;

	return utf8_string(s);
}

}	// namespace platform
}	// namespace baretorrent

#endif	// BARETORRENT_LINUX

