// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "log.h"
#include "platform.h"
#include "ext/ext.h"

namespace baretorrent
{

//! default maximum file size is 5MiB
const std::size_t default_max_file_size = 5 * 1024 * 1024;

//! suffix added to the log file's filename for backup
const utf8_string backup_suffix(".0");

// the main log
log_t g_std_log(&std::clog);


int log_t::max(10000);

static boost::mutex g_make_time_mutex;

std::ostream& operator<<(std::ostream& out, const log_t::entry& e)
{
	out << "[" << log_t::make_time(e.time) << "] " << e.s;
	return out;
}

log_t::log_t(
	std::ostream* out, const utf8_path& f,
	const std::string& prefix, bool enabled) :
		enabled_(enabled),
		capture_(false),
		timestamp_(true),
		prefix_(prefix),
		out_(out),
		file_(f),
		newline_(true),
		max_(default_max_file_size),
		level_(normal_level)
{
}

std::string log_t::make_time(boost::posix_time::ptime now)
{
	const boost::posix_time::time_facet* f =
		new boost::posix_time::time_facet("%m/%d %H:%M:%S");

	std::ostringstream oss;
	
	{
		boost::mutex::scoped_lock(g_make_time_mutex);
		oss.imbue(std::locale(oss.getloc(), f));
	}

	oss << now;

	return oss.str();
}

std::string log_t::make_prefix(
	levels lv, ext::const_instance_ptr i) const
{
	// makes
	//		"[timestamp prefix *error*] "
	// (note the whitespace at the end), where each component may be
	// left out

	std::string s, t;

	if (timestamp_)
		t = log_t::make_time(time_);

	if (!t.empty() || !prefix_.empty())
	{
		s += "[" + t;

		if (!t.empty() && !prefix_.empty())
			s += " ";

		s += prefix_;
	
		// this is hardcoded
		if (lv == error_level)
			s += " *error*";
	
		s += "] ";
	}

	if (i)
		s += i->prefix().std_string();

	return s;
}

void log_t::start_capture()
{
	boost::mutex::scoped_lock lock(capture_mutex_);
	capture_ = true;
	entries_.clear();
}
  
log_t::entries log_t::stop_capture()
{
	entries v;

	{
		boost::mutex::scoped_lock lock(capture_mutex_);
		v.swap(entries_);
		capture_ = false;
	}

	return v;
}

log_t::entries log_t::stop_capture_and_connect(
	flushed_handler::slot_function_type s,
	sig::connection* con)
{
	entries v;

	{
		boost::mutex::scoped_lock lock(capture_mutex_);
		boost::mutex::scoped_lock lock2(mutex_);

		v.swap(entries_);
		capture_ = false;
		sig::connection c = flushed.connect(s);

		if (con)
			*con = c;
	}

	return v;
}

void log_t::log(
	levels lv, const std::string& w, ext::const_instance_ptr i)
{
	boost::mutex::scoped_lock lock(mutex_);
	log_impl(lv, w, i);
}

void log_t::log_impl(
	levels lv, const std::string& original_what,
	ext::const_instance_ptr i)
{
	// position of current line
	std::string::size_type start = 0;

	// taking what was output so far and resetting the temporary stream
	std::string what = temp_line_.str() + original_what;
	temp_line_.str("");

	while (start < what.size())
	{
		// finding the next line
		std::string::size_type end = what.find("\n", start);

		// if there's no more \n, the remaining of the string will be
		// used
		if (end == std::string::npos)
			end = what.size();

		std::string original_line = what.substr(start, end - start);
		std::string line = original_line;

		// if the previous output ended with a newline, this output
		// needs to begin with a prefix
		if (newline_)
		{
			// special case: this is a new line, but the current line is
			// empty, which means there probably was a \n\n; the line
			// prefix will not be output
			if (!line.empty())
			{
				time_ = boost::posix_time::second_clock::local_time();
				line = make_prefix(lv, i) + line;
			}

			newline_ = false;
		}

		line_ << original_line;

		if (end != what.size())
		{
			entry e(time_, lv, utf8_string(line_.str()), i);
			
			if (capture_)
			{
				boost::mutex::scoped_lock lock(capture_mutex_);
				if (capture_)
				{
					if (max >= 0)
					{
						for (;;)
						{
							if (entries_.empty())
								break;

							if (static_cast<int>(entries_.size()) < max)
								break;

							entries_.pop_front();
						}
					}

					entries_.push_back(e);
				}
			}

			flushed(e);
	
			line_.str("");

			// the newline was ignored by substr() since 'end' might
			// have been the end of string
			line += "\n";

			// if the end of this line was a newline (instead of the
			// end of the string), a prefix will need to be output the
			// next time
			newline_ = true;
		}

		// the check is done here instead of at the top (which would
		// ignore the whole loop) so that disabling a log will still
		// process newlines. This way, the log will still behave
		// normally wrt prefixes even if it was disabled
		if (enabled_ && lv <= level_)
			flush(lv, line);

		start = end + 1;
	}
}

void log_t::flush(levels lv, const std::string& s)
{
	// outputs to the console
	platform::log_to_console(out_, lv, utf8_string(s));

	// outputs to the file (if any)
	write_to_file(s);
}

std::ostringstream& log_t::internal_output_stream()
{
	return temp_line_;
}

boost::mutex& log_t::internal_mutex()
{
	return mutex_;
}

void log_t::log(
	log_t::levels lv, 
	const std::string& file, int line, const std::string& fun,
	const std::string& what)
{
	std::ostringstream oss;

	// BARETORRENT_TRACE will only display the line info on newlines
	if (newline_)
	{
		// only displays the filename
		std::string::size_type p = file.find_last_of("/");
		if (p == std::string::npos)
			p = file.size();
		else
			++p;

		oss << fun << "@" << file.substr(p) << ":" << line << ": ";
	}

	oss << what;

	log(lv, oss.str());
}


void log_t::write_to_file(const std::string& line)
{
	if (file_.empty())
		return;

	// opening log file
	std::ofstream out(file_.native().c_str(), std::ios::app);
	if (!out)
		return;

	out << line;
	out.close();

	check_log_file_size();
}

void log_t::check_log_file_size()
{
	// no maximum size
	if (max_ == 0)
		return;

	// getting file size
	boost::uintmax_t s = file_.file_size();
	if (s < max_)
		return;

	const utf8_path backup_f(file_.as_utf8_string() + backup_suffix);

	boost::system::error_code ec;

	// this will fail silently if the backup file doesn't exist
	backup_f.remove(ec);

	// renaming the log file
	file_.rename(backup_f, ec);

	// in case the renaming didn't work, truncate the current file
	std::ofstream(file_.native().c_str(), std::ios::trunc);
}

void log_t::enabled(bool b)
{
	enabled_ = b;
}

bool log_t::enabled() const
{
	return enabled_;
}

void log_t::timestamp(bool b)
{
	timestamp_ = b;
}

bool log_t::timestamp() const
{
	return timestamp_;
}

void log_t::prefix(const std::string& s)
{
	prefix_ = s;
}

const std::string& log_t::prefix() const
{
	return prefix_;
}

void log_t::out(std::ostream* out)
{
	out_ = out;
}

std::ostream* log_t::out() const
{
	return out_;
}

void log_t::max_file_size(std::size_t bytes)
{
	max_ = bytes;
}

std::size_t log_t::max_file_size() const
{
	return max_;
}

void log_t::level(levels lv)
{
	level_ = lv;
}

log_t::levels log_t::level() const
{
	return level_;
}

void log_t::file(const utf8_path& f)
{
	file_ = f;

	write_to_file("\n\n\nthis is a utf8 file\n");
	baretorrent::log()
		<< "log redirected to " << f.as_utf8_string() << "\n";
}

utf8_path log_t::file() const
{
	return file_;
}


log_proxy::log_proxy(
	log_t& lg, log_t::levels lv, ext::const_instance_ptr i)
		: lg_(lg), unlock_(true), lv_(lv), instance_(i)
{
	lg_.internal_mutex().lock();
}

log_proxy::~log_proxy()
{
	if (unlock_)
		lg_.internal_mutex().unlock();
}

log_proxy::log_proxy(const log_proxy& other) :
	lg_(other.lg_), unlock_(true), lv_(other.lv_),
	instance_(other.instance_)
{
	other.unlock_ = false;
}

void log_proxy::log(const std::string& s) const
{
	lg_.log_impl(lv_, s, instance_);
}

log_t& log_proxy::internal_log() const
{
	return lg_;
}

log_t::levels log_proxy::level() const
{
	return lv_;
}

ext::const_instance_ptr log_proxy::instance() const
{
	return instance_;
}


log_proxy log(log_t& where, log_t::levels lv, const std::string& what)
{
	log_proxy p(where, lv);
	p.log(what);

	return p;
}

log_proxy log(log_t::levels lv, const std::string& what)
{
	log_proxy p(g_std_log, lv);
	p.log(what);
  
	return p;
}

log_proxy log(ext::const_instance_ptr i)
{
	return log_proxy(g_std_log, log_t::normal_level, i);
}

log_proxy log_error()
{
	return log(log_t::error_level);
}

log_proxy log_error(ext::const_instance_ptr i)
{
	return log_proxy(g_std_log, log_t::error_level, i);
}

log_proxy log_v()
{
	return log(log_t::verbose_level);
}

log_proxy log_v(ext::const_instance_ptr i)
{
	return log_proxy(g_std_log, log_t::verbose_level, i);
}

log_proxy log_vv()
{
	return log(log_t::very_verbose_level);
}

log_proxy log_vv(ext::const_instance_ptr i)
{
	return log_proxy(g_std_log, log_t::very_verbose_level, i);
}

}   // namespace baretorrent
