// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_LOG_INCLUDED
#define BARETORRENT_LOG_INCLUDED

#include "path.h"

// these are logging facilities used throughout the program
//
// log_t supports writing to any output stream as well as duplicating
// the output to a file with automatic backup when the file gets too
// large; it also adds timestamps and an optional prefix to each line
//
// using operator<<() with a log_t returns a proxy that will lock a
// mutex until destroyed at the end of the statement; this allows for
// logging lines individually in multiple threads
//
// g_std_log is a global instance of log_t and is probably the only
// one needed; it is used as the default log when calling functions
// such as log() and log_error()
//
// each log_t supports multiple levels of logging that can be turned on
// or off, such as quiet and verbose; using log() by itself logs under
// the normal level while log_error() uses the error level

namespace baretorrent
{

namespace ext
{

class instance;
typedef boost::shared_ptr<instance> instance_ptr;
typedef boost::shared_ptr<const instance> const_instance_ptr;
typedef std::vector<instance_ptr> instance_list;

}

class log_proxy;

// a log
//
class BARETORRENT_EXPORT log_t
{
	friend class log_proxy;

public:
	enum levels
	{
		//! this will append a string (such as "*error*") at the end of
		//! the prefix
		error_level = -2,

		// shouldn't be used
		no_level = -1,

		quiet_level = 0,
		normal_level,
		verbose_level,
		very_verbose_level
	};

	//! an entry in the log
	//!
	struct entry
	{
		boost::posix_time::ptime time;
		levels lv;
		utf8_string s;
		ext::const_instance_ptr i;

		entry(
			boost::posix_time::ptime time, levels lv,
			const utf8_string& s,
			const ext::const_instance_ptr i
				= ext::const_instance_ptr()) :
					time(time), lv(lv), s(s), i(i)
		{
		}
	};

	typedef std::deque<entry> entries;

	//! maximum number of entries retained while capturing is enabled;
	//! also used as the number of items in the log dialog list
	static int max;

	//! a callback executed every time a line is flushed
	typedef sig::signal<void (const entry&)> flushed_handler;
	flushed_handler flushed;

	//! returns the current time formatted as "dd/mm/yy hh:mm:ss.mm" or an
	//! empty string if it fails
	//!
	static std::string make_time(boost::posix_time::ptime now);


	//! creates an log writing to the given stream and log file; the log
	//! is enabled and has timestamps. Both @a file and @a out can be
	//! empty, in which case their corresponding stream isn't used.
	//!
	log_t(
		std::ostream* out, const utf8_path& file=utf8_path(),
		const std::string& prefix="", bool enabled=true);


	//! splits @a what on newlines and outputs each line preceded by
	//! timestamp/prefix on both the output stream and the log file (if
	//! opened). noop if the stream is not enabled
	//!
	void log(
		levels lv, const std::string& what,
		ext::const_instance_ptr i = ext::const_instance_ptr());

	//! prepends "function() @ file:line: " to the given string and
	//! calls log_t::log(const std::string&)
	//!
	void log(
		levels lv,
		const std::string& file, int line, const std::string& function,
		const std::string& what);

	// internal: doesn't lock the mutex
	//
	void log_impl(
		levels lv, const std::string& what,
		ext::const_instance_ptr i = ext::const_instance_ptr());

	//! sets whether this log is enabled
	//!
	void enabled(bool b);

	//! returns whether this log is enabled
	//!
	bool enabled() const;


	//! remembers all the logs until stop_capture() is called
	//!
	void start_capture();

	//! stops the capture and returns all the entries logged
	//!
	entries stop_capture();

	//! stops the capture, connects the flushed handler and returns all
	//! the entries logged; this makes no entries are lost in between
	//!
	entries stop_capture_and_connect(
		flushed_handler::slot_function_type s,
		sig::connection* con = nullptr);


	//! sets whether timestamps are output at the beginning of each line
	//!
	void timestamp(bool b);

	//! returns whether timestamps are enabled
	//!
	bool timestamp() const;


	//! sets the prefix that is output at the beginning of each line
	//!
	void prefix(const std::string& s);

	//! returns the prefix for this log
	//!
	const std::string& prefix() const;


	//! redirects this log to a different output stream
	//!
	void out(std::ostream* out);

	//! returns this log's output stream
	//!
	std::ostream* out() const;


	//! closes the current log file (if any), opens the given filename
	//! and duplicates all subsequent logs to it; disables file logging
	//! if the filename is empty
	//!
	//! the file will be truncated unless it cannot be opened, in which
	//! case this will silently fail
	//!
	void file(const utf8_path& f);

	//! returns the filename of the log file, if any
	//!
	utf8_path file() const;


	//! sets the maximum size in bytes of the log file; if 0, the log
	//! file is never emptied (scary)
	//!
	void max_file_size(std::size_t bytes);

	//! returns the maximum size in bytes of the log file or 0 if the
	//! log file has no maximum size
	//!
	std::size_t max_file_size() const;


	//! sets the highest level that will be logged
	//!
	void level(levels lv);

	//! returns the highest level that will be logged
	//!
	log_t::levels level() const;
  

	//! returns a reference to the internal stringstream used to
	//! construct the current line before it is flushed; this is used
	//! by the log proxy and is internal
	//!
	//! a stringstream is used internally so that manipulators can be
	//! used
	//!
	std::ostringstream& internal_output_stream();

	//! returns a reference to the mutex used by this log; this is
	//! used by the log proxy to serialize outputs and is internal
	//!
	boost::mutex& internal_mutex();

private:
	//! whether the log is enabled
	bool enabled_;

	//! whether entries are captured
	bool capture_;

	//! captured entries
	entries entries_;

	//! protects entries
	boost::mutex capture_mutex_;

	//! whether the log prints timestamps on new lines
	bool timestamp_;

	//! an arbitrary string printed next to the timestamp
	std::string prefix_;

	//! the output stream
	std::ostream* out_;

	//! log file
	utf8_path file_;

	//! whether the next output needs the line prefix
	bool newline_;

	//! maximum size of the log file; if 0, the file is never replaced
	std::size_t max_;

	//! current line
	std::ostringstream line_;

	//! temporary stream to which values are output until log("") is
	//! called, which retrieves the string and adds it to line_; this
	//! allows for using manipulators
	std::ostringstream temp_line_;

	//! highest level that will be logged
	levels level_;

	//! used both by log_t itself in log() and locked manually by the
	//! log proxy until the end of the statement
	boost::mutex mutex_;

	//! time of the last new line
	//!
	boost::posix_time::ptime time_;


	//! returns the line prefix that contains the timestamp and log
	//! prefix. if both are empty, this returns an empty string. If not,
	//! the resulting string is enclosed in brackets, such as
	//! "[timestamp prefix]".
	//!
	std::string make_prefix(
		levels lv, ext::const_instance_ptr i) const;

	//! writes the given string to the log file (if any)
	//!
	void write_to_file(const std::string& s);

	//! renames the log file if it gets too large
	//!
	void check_log_file_size();

	//! does the actual write; this can be modified to write to
	//! different places
	//!
	void flush(levels lv, const std::string& s);
};

std::ostream& operator<<(std::ostream& out, const log_t::entry& e);


//! returned by operator<<(); locks the log mutex in the constructor,
//! unlocks it in the destructor, unless it was copied to another
//! instance
//!
//! the log_proxy remembers the log level and uses it
//!
class BARETORRENT_EXPORT log_proxy
{
public:
	//! locks lg's mutex
	//!
	log_proxy(
		log_t& lg, log_t::levels lv,
		ext::const_instance_ptr i = ext::const_instance_ptr());

	//! unlocks lg's mutex
	~log_proxy();
	
	//! transfers the lock to this proxy
	//!
	log_proxy(const log_proxy& other);

	//! disabled, doesn't make sense
	//!
	log_proxy& operator=(const log_proxy& other);


	//! logs the given string with the logging level specified
	//!
	void log(const std::string& s) const;

	//! returns the log_t used by this proxy
	//!
	log_t& internal_log() const;

	//! returns the log level
	//!
	log_t::levels level() const;

	//! returns the extension (if any)
	//!
	ext::const_instance_ptr instance() const;

private:
	//! log
	log_t& lg_;

	//! whether the destructor should unlock the mutex; this is false
	//! when this proxy has been copied to another
	mutable bool unlock_;

	//! level of this proxy
	log_t::levels lv_;

	//! extension (if any)
	ext::const_instance_ptr instance_;
};


//! a log that prints to std::clog with timestamps and an empty
//! prefix
//!
extern log_t g_std_log;


//! calls where.log(lv, what) and returns a reference to @a where
//!
BARETORRENT_EXPORT log_proxy log(
	log_t& where, log_t::levels lv, const std::string& what);

//! calls g_std_log.log(lv, what) and returns a reference to g_std_log
//!
BARETORRENT_EXPORT log_proxy log(
	log_t::levels lv=log_t::normal_level,
	const std::string& what="");

BARETORRENT_EXPORT log_proxy log(ext::const_instance_ptr i);

//! returns a proxy that logs with a log_t::error_level
//!
BARETORRENT_EXPORT log_proxy log_error();
BARETORRENT_EXPORT log_proxy log_error(ext::const_instance_ptr i);

//! returns a proxy that logs with a log_t::verbose_level
//!
BARETORRENT_EXPORT log_proxy log_v();
BARETORRENT_EXPORT log_proxy log_v(ext::const_instance_ptr i);

//! returns a proxy that logs with a log_t::very_verbose_level
//!
BARETORRENT_EXPORT log_proxy log_vv();
BARETORRENT_EXPORT log_proxy log_vv(ext::const_instance_ptr i);

// the following are operator<<'s for log_proxy

//! outputs the given string
//!
inline log_proxy operator<<(
	const log_proxy& out, const char* s)
{
	// the mutex has already been locked by log_proxy's constructor
	out.internal_log().internal_output_stream() << s;

	// because this is using the stringstream directly, log("") is
	// called so log_t can check it
	out.internal_log().log_impl(out.level(), "", out.instance());

	return out;
}

//! outputs the given string
//!
inline log_proxy operator<<(
	const log_proxy& out, const std::string& s)
{
	return out << s.c_str();
}

//! outputs the given object; operator<< must be available for T.
//!
template <class T>
BARETORRENT_EXPORT log_proxy operator<<(
	const log_proxy& out, const T& t)
{
	// the mutex has already been locked by log_proxy's constructor
	out.internal_log().internal_output_stream() << t;

	// because this is using the stringstream directly, log_impl("") is
	// called so log_t can check it
	out.internal_log().log_impl(out.level(), "", out.instance());

	return out;
}

//! outputs the given manipulator
//!
inline log_proxy operator<<(
	const log_proxy& out,
	std::ios_base& (*pf)(std::ios_base&))
{
	// the mutex has already been locked by log_proxy's constructor
	out.internal_log().internal_output_stream() << pf;

	// because this is using the stringstream directly, log_impl("") is
	// called so log_t can check it
	out.internal_log().log_impl(out.level(), "", out.instance());

	return out;
}


// the following are operator<<'s for log_t

//! outputs the given string
//!
inline log_proxy operator<<(
	log_t& out, const char* s)
{
	// the mutex will be locked in the constructor
	log_proxy lp(out, log_t::normal_level);

	out.internal_output_stream() << s;

	// because this is using the stringstream directly, log_impl("") is
	// called so log_t can check it
	out.log_impl(log_t::normal_level, "");

	return lp;
}

//! outputs the given string
//!
inline log_proxy operator<<(
	log_t& out, const std::string& s)
{
	return out << s.c_str();
}

//! outputs the given object; operator<< must be available for T
//!
template <class T>
BARETORRENT_EXPORT log_proxy operator<<(log_t& out, const T& t)
{
	// the mutex will be locked in the constructor
	log_proxy lp(out, log_t::normal_level);

	out.internal_output_stream() << t;

	// because this is using the stringstream directly, log_impl("") is
	// called so log_t can check it
	out.log_impl(log_t::normal_level, "");

	return lp;
}

//! outputs the given manipulator
//!
inline log_proxy operator<<(
	log_t& out, std::ios_base& (*pf)(std::ios_base&))
{
	// the mutex will be locked in the constructor
	log_proxy lp(out, log_t::normal_level);

	out.internal_output_stream() << pf;

	// because this is using the stringstream directly, log_impl("") is
	// called so log_t can check it
	out.log_impl(log_t::normal_level, "");

	return lp;
}


// calls g.log() with the current file name, line number and function
// name
#define BARETORRENT_TRACE_LOG(g, what) g.log( \
			BARETORRENT_MACRO_FILE, BARETORRENT_MACRO_LINE, \
			BARETORRENT_MACRO_FUNC, what);

// forwards to BARETORRENT_TRACE_LOG with g_std_log
#define BARETORRENT_TRACE(what) \
			BARETORRENT_TRACE_LOG(g_std_log, what)

}   // namespace baretorrent

#endif   // BARETORRENT_LOG_INCLUDED
