// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "log_dialog.h"
#include "widgets.h"

namespace baretorrent
{

enum ids
{
	browse_id = wxID_HIGHEST + 800,

	// log list
	log_list_id,

	// log level
	log_level_id,

	// clear log
	clear_log_id,

	// save log
	save_log_id,

	// extension list
	log_ext_list_id
};

int log_date_sorter(
	const log_t::entry* left, const log_t::entry* right)
{
	if (left->time < right->time)
		return -1;
	else if (left->time > right->time)
		return 1;
	else
		return 0;
}

int log_message_sorter(
	const log_t::entry* left, const log_t::entry* right)
{
	return natural_compare(left->s, right->s);
}

int log_ext_name_sorter(
		const ext::instance* left, const ext::instance* right)
{
	return natural_compare(left->name(), right->name());
}


log_dialog::log_dialog(wxWindow*) :
	custom_frame("bt.ui.log_window"),
	list_(nullptr), ext_(nullptr), level_(log_t::normal_level)
{
	SetTitle(loc_wx("log.title"));
	
	splitter* split = new splitter(this, "bt.ui.log.split");

	list_ = new custom_list<log_t::entry>(
		split, log_list_id, "bt.ui.log_list");

	split->SetSashGravity(1.0);
	split->SplitVertically(list_, controls(split));

	list_->add_column(loc("log.time"), log_date_sorter);
	list_->add_column(loc("log.text"), log_message_sorter);

	sig::connection con;
	new_entries_ = g_std_log.stop_capture_and_connect(
		boost::bind(&log_dialog::on_flushed, this, _1), &con);

	cons_.add(con);

	timer_.start(
		boost::bind(&log_dialog::on_timer, this), 
		boost::posix_time::milliseconds(500));

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(split, 1, wxEXPAND);

	SetSizer(box);

	creating_ = false;

	list_->add_accelerator(wxGetStockAccelerator(wxID_COPY));

	ext::extension_loaded().connect(boost::bind(
		&log_dialog::on_extension_loaded, this, _1));
}

void log_dialog::on_extension_loaded(ext::instance_ptr e)
{
	for (int i=0; i<ext_->GetItemCount(); ++i)
	{
		if (ext_->data(i) == e.get())
			return;
	}

	add_extension(e);
}

void log_dialog::on_extension_uninstalled(ext::instance_ptr e)
{
	for (int i=0; i<ext_->GetItemCount(); ++i)
	{
		if (ext_->data(i) == e.get())
		{
			ext_->DeleteItem(i);
			return;
		}
	}
}

wxPanel* log_dialog::controls(wxWindow* parent)
{
	wxPanel* panel = new wxPanel(parent);

	wxGridBagSizer* grid = new wxGridBagSizer(5, 5);

	wxArrayString choices;
	choices.Add(loc_wx("log.level.error"));
	choices.Add(loc_wx("log.level.normal"));
	choices.Add(loc_wx("log.level.verbose"));
	choices.Add(loc_wx("log.level.very_verbose"));
	choices.Add(loc_wx("log.level.ext"));

	wxComboBox* lvs = new wxComboBox(
		panel, log_level_id, wxEmptyString,
		wxDefaultPosition, wxDefaultSize, choices,
		wxCB_READONLY);

	lvs->SetSelection(1);

	grid->Add(
		new wxStaticText(panel, -1, loc_wx("log.level")),
		wxGBPosition(0, 0), wxGBSpan(1, 1), wxALIGN_CENTER_VERTICAL);
	
	grid->Add(lvs, wxGBPosition(0, 1), wxGBSpan(1, 1), wxEXPAND);
	
	grid->Add(
		new wxStaticText(panel, -1, loc_wx("log.extensions")),
		wxGBPosition(1, 0), wxGBSpan(1, 2));


	ext_ = new custom_list<ext::instance>(
		panel, log_ext_list_id, "bt.ui.log_ext", 0);

	ext_->use_checkboxes();
	ext_->on_checked.connect(
		boost::bind(&log_dialog::on_ext_checked, this, _1, _2));

	ext_->add_column(loc("log.extensions.name"), log_ext_name_sorter);

	grid->Add(ext_, wxGBPosition(2, 0), wxGBSpan(1, 2), wxEXPAND);

	grid->Add(
		new multiline_text(panel, -1, to_wxstring(diag_info())),
		wxGBPosition(3, 0), wxGBSpan(1, 2), wxEXPAND);

	grid->Add(
		new wxButton(panel, clear_log_id, loc_wx("log.clear")),
		wxGBPosition(4, 0), wxGBSpan(1, 1));

	grid->Add(
		new wxButton(panel, save_log_id, loc_wx("log.save")),
		wxGBPosition(4, 1), wxGBSpan(1, 2));

	grid->AddGrowableCol(1);
	grid->AddGrowableRow(2);

	panel->SetSizer(grid);

	rebuild_ext();

	return panel;
}

void log_dialog::on_flushed(const log_t::entry& e)
{
	boost::mutex::scoped_lock lock(entries_mutex_);
	new_entries_.push_back(e);
}

void log_dialog::on_ext_checked(int, bool)
{
	rebuild(level_);
}

bool log_dialog::on_timer()
{
	log_t::entries v;

	{
		boost::mutex::scoped_lock lock(entries_mutex_);
		v = new_entries_;

		entries_.insert(entries_.end(), v.begin(), v.end());
		new_entries_.clear();

		if (log_t::max >= 0 &&
			static_cast<int>(entries_.size()) > log_t::max)
		{
			log_t::entries nv(
				entries_.begin() + entries_.size() - log_t::max,
				entries_.end());

			entries_.swap(nv);
		}
	}

	insert_items(v);
	return true;
}

void log_dialog::insert_items(const log_t::entries& v)
{
	bool scroll = false;
	int sel = -1;

	if (list_->GetItemCount() == 0)
	{
		scroll = true;
	}
	else if (list_->selection_size() == 1)
	{
		const int i = list_->single_selection_index();
		if (i == (list_->GetItemCount() - 1))
		{
			sel = i;
			scroll = true;
		}
	}

	freezer fz(list_);

	if (log_t::max >= 0)
	{
		int c = list_->GetItemCount();

		for (;;)
		{
			if (c == 0)
				break;
			
			if (static_cast<int>(c + v.size()) > log_t::max)
			{
				delete list_->data(0);
				list_->DeleteItem(0);
			}
			else
			{
				break;
			}

			--c;

			if (scroll && sel >= 0)
				--sel;
		}
	}
	
	std::vector<ext::instance*> si;
	for (int i=0; i<ext_->GetItemCount(); ++i)
	{
		if (ext_->checked(i))
			si.push_back(ext_->data(i));
	}

	foreach (const log_t::entry& e, v)
	{
		if (level_ != log_t::quiet_level && e.lv > level_)
			continue;

		if (e.i)
		{
			if (std::find(si.begin(), si.end(), e.i.get()) == si.end())
				continue;
		}
		else if (level_ == log_t::quiet_level)
		{
			continue;
		}
		
		const int i = list_->GetItemCount();

		list_->append(utf8_string(log_t::make_time(e.time)));
		list_->col(i, 1, e.s);
		list_->data(i, new log_t::entry(e));

		if (e.lv == log_t::error_level)
			list_->row_color(i, errored_text_color());
	}

	if (scroll && list_->GetItemCount() > 0)
	{
		if (sel >= 0)
			list_->unselect(sel);
		
		list_->select(list_->GetItemCount() - 1);
		list_->ensure_visible(list_->GetItemCount() - 1);
	}
}

void log_dialog::rebuild(log_t::levels lv)
{
	{
		boost::mutex::scoped_lock lock(entries_mutex_);

		log_t::entries v = entries_;
		v.insert(v.begin(), new_entries_.begin(), new_entries_.end());

		new_entries_.clear();
	}

	level_ = lv;
	list_->clear();
	insert_items(entries_);
}

void log_dialog::rebuild_ext()
{
	ext_->clear();

	ext::instance_list is = ext::instances();
	foreach (const ext::instance_ptr i, is)
		add_extension(i);
}

void log_dialog::add_extension(ext::instance_ptr i)
{
	i->uninstalled.connect(boost::bind(
		&log_dialog::on_extension_uninstalled, this, i));

	const int item = ext_->append(i->name());
	ext_->data(item, i.get());
	ext_->checked(item, true);
}

void log_dialog::on_clear_log(wxCommandEvent&)
{
	{
		boost::mutex::scoped_lock lock(entries_mutex_);
		entries_.clear();
		new_entries_.clear();
	}

	list_->clear();
}

void log_dialog::on_level(wxCommandEvent& e)
{
	log_level_indices i =
		static_cast<log_level_indices>(e.GetSelection());

	log_t::levels lv;
	switch(i)
	{
		case log_level_error:
			lv = log_t::error_level;
			break;

		case log_level_normal:
			lv = log_t::normal_level;
			break;

		case log_level_verbose:
			lv = log_t::verbose_level;
			break;

		case log_level_very_verbose:
			lv = log_t::very_verbose_level;
			break;

		case log_level_ext:
			lv = log_t::quiet_level;
			break;

		default:
			lv = log_t::normal_level;
			break;
	}

	rebuild(lv);
}

void log_dialog::on_close(wxCloseEvent&)
{
	Hide();
	opt("bt.ui.log", false);
}

void log_dialog::on_copy(wxCommandEvent&)
{
	std::ostringstream oss;

	const std::vector<const log_t::entry*> s = list_->selection();
	foreach (const log_t::entry* e, s)
		oss << *e << "\n";

	clipboard cb;
	cb.set_text(utf8_string(oss.str()));
}

void log_dialog::on_save_log(wxCommandEvent&)
{
	wxFileDialog dlg(
		this, loc_wx("log.save.title"), wxEmptyString, wxEmptyString,
		"*.*", wxFD_SAVE | wxFD_OVERWRITE_PROMPT);

	if (dlg.ShowModal() != wxID_OK)
		return;

	const utf8_path p(utf8_string(dlg.GetPath().ToUTF8().data()));
	if (p.empty())
		return;
	
	std::ofstream out(p.native().c_str(), std::ios::binary);
	if (!out)
		return;

	out
		<< to_simple_string(
			boost::posix_time::second_clock::local_time()) << "\n"
		<< diag_info()
		<< "\n\n"
		<< "extensions: ";

	ext::instance_list is = ext::instances();
	
	if (is.empty())
	{
		out << "none\n";
	}
	else
	{
		out << "\n";

		foreach (const ext::instance_ptr i, is)
			out << i->dump() << "\n\n";
	}

	out << "log:\n";

	const std::vector<const log_t::entry*> s = list_->items();
	foreach (const log_t::entry* e, s)
		out << *e << "\n";
}

BEGIN_EVENT_TABLE(log_dialog, wxFrame)
	EVT_BUTTON(clear_log_id, log_dialog::on_clear_log)
	EVT_BUTTON(save_log_id, log_dialog::on_save_log)
	EVT_COMBOBOX(log_level_id, log_dialog::on_level)
	EVT_CLOSE(log_dialog::on_close)
	EVT_MENU(wxID_COPY, log_dialog::on_copy)
END_EVENT_TABLE()

}	// namespace baretorrent
