// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "utility.h"
#include "log.h"
#include "manager.h"
#include "app.h"
#include "options.h"
#include "ext/ext.h"
#include "db/sqlite.h"
#include "platform.h"

namespace baretorrent
{

// for --noext
static bool g_noext = false;

namespace help
{

// these return the help text for command line options

utf8_string add()
{
	return utf8_string(
		"adds a new torrent, can be specified multiple times; NAME "
		"can be a file, a URL or a magnet link");
}

utf8_string quit()
{
	return utf8_string(
		"with -a, quits baretorrent once the operation is complete");
}

utf8_string resp()
{
	return utf8_string(
		"reads command line options from the given file; @path may "
		"also be used");
}

utf8_string config()
{
	return 
		utf8_string(
			"sets the path of the configuration file, defaults to ") +
		paths::conf().as_utf8_string();
}

utf8_string lib()
{
	std::ostringstream oss;

	oss
		<< "sets the path of the python library files, defaults to "
		<< paths::python_path() << "; may be specified "
		<< "multiple times";

	return utf8_string(oss.str());
}

utf8_string clear_sys_path()
{
	return utf8_string(
		"clears sys.path on startup and use --lib and extensions "
		"path");
}

utf8_string set()
{
	return utf8_string(
		"sets the value of an option, 'arg' must be NAME=VALUE");
}

utf8_string no_console()
{
	return utf8_string(
		"hides the console even if -v is specified; to get anything "
		"logged, --log is needed");
}

utf8_string verbose()
{
	utf8_string s(
		"verbose, can be specified multiple times, defaults to quiet");

#if BARETORRENT_WINDOWS
	s += utf8_string(
		"; specifying at least one -v or --verbose will open a console "
		"window");
#endif

	return s;
}

utf8_string logfile()
{
	return utf8_string(
		"duplicates the logs to the given file up to 5MiB, with up to "
		"one backup");
}

utf8_string maxlog()
{
	return utf8_string(
		"maximum number of log entries retained in the log dialog, "
		"-1 for unlimited");
}

utf8_string noext()
{
	return utf8_string("disables python and extensions");
}

utf8_string version()
{
	return utf8_string("ouputs version information and exits");
}

utf8_string help()
{
	return utf8_string("outputs this message and exits");
}

}	// namespace help

struct ipc_data
{
	static const std::string object_name;
	
	enum
	{
		// 8k ought to be enough for everybody 
		max_size = 1024 * 8
	};

	char s[max_size];
	ipc::interprocess_mutex mutex;
};

const std::string ipc_data::object_name("baretorrent-data");
static ipc::mapped_region g_ipc_region;
static ipc_data* g_ipc_data = nullptr;

bool parse_command_line(
	const std::vector<std::string>& v, po::variables_map& vm,
	bool first);
bool handle_commands(app& m, po::variables_map& vm);


// called from the main instance to retrieve anything set by other
// instances
//
std::vector<std::string> ipc_args()
{
	std::vector<std::string> v;

	if (!g_ipc_data)
		return v;

	std::string s;
	for (std::size_t i=0; i<ipc_data::max_size; ++i)
	{
		if (g_ipc_data->s[i] == 0)
		{
			if (s.empty())
			{
				// last argument is followed by two \0
				break;
			}
			else
			{
				v.push_back(s);
				s = "";
			}
		}
		else
		{
			s.append(1, g_ipc_data->s[i]);
		}
	}

	return v;
}

void kill_ipc()
{
	try
	{
		ipc::shared_memory_object::remove(
			ipc_data::object_name.c_str());
	}
	catch(ipc::interprocess_exception& e)
	{
		std::cerr << "couldn't remove ipc object, " << e.what() << "\n";
	}
}

void create_ipc()
{
	try
	{
		kill_ipc();

		ipc::shared_memory_object o(
			ipc::create_only, ipc_data::object_name.c_str(),
			ipc::read_write);

		o.truncate(sizeof(ipc_data));
		g_ipc_region = ipc::mapped_region(o, ipc::read_write);

		void* p = g_ipc_region.get_address();
		g_ipc_data = new(p) ipc_data;
	}
	catch(ipc::interprocess_exception& e)
	{
		std::cerr << "couldn't create ipc object, " << e.what() << "\n";
	}
}

void set_ipc(int argc, char** argv)
{
	try
	{
		ipc::shared_memory_object o(
			ipc::open_only, ipc_data::object_name.c_str(),
			ipc::read_write);

		ipc::mapped_region region(o, ipc::read_write);
		void* p = region.get_address();
	
		ipc_data* d = reinterpret_cast<ipc_data*>(p);
		if (!d)
		{
			std::cerr << "ipc data is null\n";
			return;
		}

		ipc::scoped_lock<ipc::interprocess_mutex> lock(d->mutex);

		std::memset(d->s, 0, ipc_data::max_size);

		// appending all the arguments, separated by \0
		std::string s;
		for (int i=1; i<argc; ++i)
		{
			s.append(argv[i], argv[i] + std::strlen(argv[i]));
			s.append(1, '\0');
		}

		// last argument has two nulls
		s.append(1, '\0');
		s.copy(d->s, ipc_data::max_size);
	}
	catch(ipc::interprocess_exception& e)
	{
		std::cerr << "couldn't set ipc data, " << e.what() << "\n";
	}
}

std::vector<std::string> parse_ipc()
{
	std::vector<std::string> v;

	if (!g_ipc_data)
		return v;
		
	ipc::scoped_lock<ipc::interprocess_mutex> lock(g_ipc_data->mutex);

	std::string s;

	for (std::size_t i=0; i<ipc_data::max_size; ++i)
	{
		if (g_ipc_data->s[i] == '\0')
		{
			// end of argument
			if (s.empty())
			{
				// last argument has two nulls
				break;
			}

			v.push_back(s);
			s = "";
		}
		else
		{
			s.append(1, g_ipc_data->s[i]);
		}
	}

	std::memset(g_ipc_data->s, 0, ipc_data::max_size);

	return v;
}

void check_ipc()
{
	try
	{
		std::vector<std::string> v = parse_ipc();
		if (v.empty())
			return;

		log()
			<< "received command line from a second instance: "
			<< implode(v, utf8_string(" ")) << "\n";

		po::variables_map vm;
		if (!parse_command_line(v, vm, false))
			return;

		handle_commands(*dynamic_cast<app*>(wxTheApp), vm);
	}
	catch(...)
	{
		log_error() << "unhandled exception while checking ipc\n";
	}
}


// outputs the given description
//
void output_help(const po::options_description& d)
{
	std::cout << d;
}

// outputs the version information to the given stream
//
template <class Out>
void output_version(Out& out)
{
	out << versions();
}

// boost doesn't natively support multiple options with zero tokens,
// like having multiple -v options to increase the verbosity level
//
// this is a custom parser given to parse_command_line() so it
// figure out the -v option
//
// this will replace all "-v" options by a "--verbose 1" and something
// like -vvv by a "--verbose 3"; the value of this --verbose option
// will be picked up in handle_options and added to the current level
//
// therefore, the --verbose option actually takes a string containing
// the value to add to the verbose level and value_with_name::name()
// has a hack to prevent the 'arg' from appearing the usage, since it is
// internal
// 
// apple: this also ignores the -psn option given by osx for the
// process serial number
//
std::pair<std::string, std::string> custom_parser(const std::string& s)
{
	if (boost::starts_with(s, "-v"))
	{
		// supports -vvvvv
		std::size_t count = s.size() - 1;
		return std::make_pair(
			"verbose", boost::lexical_cast<std::string>(count));
	}
	else if (s == "--verbose")
	{
		return std::make_pair("verbose", "1");
	}
	else if (s[0] == '@')
	{
		return std::make_pair("response-file", s.substr(1));
	}
	else if (boost::starts_with(s, "-psn_"))
	{
		return std::make_pair("dummy", "");
	}

	return std::make_pair("", "");
}

// this wraps typed_value so the argument name can be specified instead
// of always being displayed as 'arg'; boost doesn't natively support
// it
//
template <class T>
class value_with_name : public po::typed_value<T>
{
public:
	value_with_name(const std::string& name="")
		: po::typed_value<T>(&value_), name_(name)
	{
	}

	// typed_value::name() will return the string after the option
	// name, including the default value; the first word in that
	// string will be 'arg' if that option takes an argument and it will
	// be replaced by the name given or left as is if the name wasn't
	// specified
	//
	// this also hacks --verbose so the 'arg' won't be displayed at
	// all (see verbosity())
	//
	virtual std::string name() const
	{
		if (name_ == "VERBOSE HACK")
			return "";

		std::string s = po::typed_value<T>::name();
		if (!name_.empty() && boost::starts_with(s, po::arg))
			s = name_ + s.substr(po::arg.size());
		
		return s;
	}

private:
	T value_;
	std::string name_;
};

// creates a value_with_name on the heap; these are deleted by
// ~options_description()
//
template <class T>
value_with_name<T>* value2(const std::string& s = "")
{
	return new value_with_name<T>(s);
}

// parses the response file (if any) and adds it to the options
//
bool parse_response_file(
	const utf8_path& p,
	po::options_description& d, po::variables_map& vm)
{
	std::ifstream ifs(p.native().c_str());
	if (!ifs)
	{
		log_error() << "couldn't open the response file " << p << "\n";
		return false;
	}

	// reading the file
	std::stringstream ss;
	ss << ifs.rdbuf();

	// splitting
	std::string s = ss.str();
	boost::char_separator<char> sep(" \n\r");
	boost::tokenizer<boost::char_separator<char> > tok(s, sep);

	std::vector<std::string> args;
	std::copy(tok.begin(), tok.end(), std::back_inserter(args));
			
	// parsing
	po::store(po::command_line_parser(args)
		.options(d)
		.extra_parser(&custom_parser)
		.run(), vm);

	return true;
}

// adds all the possible options
//
void create_options(
	po::positional_options_description& p,
	po::options_description& to,
	po::options_description& go,
	po::options_description& ho)
{
	// torrent options
	to.add_options()
		("add,a", value2<std::vector<std::string> >("NAME"),
			help::add().c_str());

	// general options
	go.add_options()
		("response-file", value2<std::string>("PATH"),
			help::resp().c_str())
		("config,c", value2<std::string>("PATH"),
			help::config().c_str())
		("lib,l", value2<std::vector<std::string> >("PATH"),
			help::lib().c_str())
		("clear-sys-path", value2<bool>()->zero_tokens(),
			help::clear_sys_path().c_str())
		("set,s", value2<std::string>(),
			help::set().c_str())
		("quit,q", value2<bool>()->zero_tokens(),
			help::quit().c_str())
		("verbose,v", value2<std::vector<std::size_t> >("VERBOSE HACK"),
			help::verbose().c_str())
		("logfile", value2<std::string>("PATH"),
			help::logfile().c_str())
		("maxlog", value2<int>()->default_value(log_t::max),
			help::maxlog().c_str())
		("noext", value2<bool>()->zero_tokens(),
			help::noext().c_str())
		("no-console", value2<bool>()->zero_tokens(),
			help::no_console().c_str())
		("version", value2<bool>()->zero_tokens(),
			help::version().c_str())
		("help", value2<bool>()->zero_tokens(),
			help::help().c_str());

	// hidden options
	ho.add_options()
		("ver", value2<bool>()->zero_tokens(), "")
		("dummy", value2<std::vector<int> >()->zero_tokens(), "");

	p.add("add", -1);
}

// handles the various commands (see handle_options())
//
bool handle_commands(app& a, po::variables_map& vm)
{
	bool op = false;

	if (vm.count("add") > 0)
	{
		std::vector<std::string> v =
			vm["add"].as<std::vector<std::string> >();

		std::vector<utf8_string> vs;
		foreach (const std::string& s, v)
			vs.push_back(utf8_string(s));

		a.add_torrents(vs);
		op = true;
	}

	if (op && vm.count("quit"))
		return false;

	return true;
}
// handles the configuration options (see handle_commands())
//
bool handle_options(
	po::options_description d, po::variables_map& vm, bool first)
{
	if (vm.count("help"))
	{
		output_help(d);
		return false;
	}
	else if (vm.count("version") || vm.count("ver"))
	{
		output_version(std::cout);
		return false;
	}

	if (first)
	{
		if (vm.count("config"))
		{
			paths::set_configuration_path(utf8_path(utf8_string(
				vm["config"].as<std::string>())));
		}

		if (vm.count("lib"))
		{
			std::vector<std::string> temp =
				vm["lib"].as<std::vector<std::string> >();

			std::vector<utf8_path> libs;
			foreach (const std::string& s, temp)
				libs.push_back(utf8_path(utf8_string(s)));

			paths::set_python_path(libs);
		}

		if (vm.count("clear-sys-path"))
			paths::clear_sys_path(true);
	}

	bool force_no_console = false;
	
	if (vm.count("no-console") > 0)
	{
		force_no_console = true;
		if (first)
			platform::detach_console();
	}

	if (first && vm.count("verbose") == 0)
	{
		platform::detach_console();
	}
	else if (vm.count("verbose") > 0)
	{
		const std::vector<std::size_t> vs =
			vm["verbose"].as<std::vector<std::size_t> >();

		std::size_t v = 0;
		foreach (std::size_t s, vs)
			v += s;

		if (v > 0 && !force_no_console)
			platform::create_console();

		if (v == 2)
			g_std_log.level(log_t::verbose_level);
		else if (v >= 3)
			g_std_log.level(log_t::very_verbose_level);
	}

	if (vm.count("logfile"))
	{
		const utf8_path p(utf8_string(vm["logfile"].as<std::string>()));

		boost::system::error_code ec;
		create_directories(p.parent(), ec);

		if (ec)
			std::cerr << "could not create log file '" << p << "'\n";
		else
			g_std_log.file(p);
	}

	g_std_log.max = vm["maxlog"].as<int>();

	if (vm.count("noext") > 0)
		g_noext = true;
	
	return true;
}

// parses the given command line into the variables map (including the
// response file); returns false in case of problems
//
bool parse_command_line(
	const std::vector<std::string>& args, po::variables_map& vm,
	bool first)
{
	po::options_description go("General"), to("Torrents"), ho;
	po::positional_options_description pod;
	create_options(pod, to, go, ho);

	po::options_description all_options;
	all_options.add(to).add(go).add(ho);

	po::options_description visible_options;
	visible_options.add(to).add(go);

	po::store(po::command_line_parser(args)
		.options(all_options)
		.extra_parser(&custom_parser)
		.positional(pod)
		.run(), vm);

	if (vm.count("response-file"))
	{
		if (!parse_response_file(
				utf8_path(utf8_string(
					vm["response-file"].as<std::string>())),
				all_options, vm))
		{
			return false;
		}
	}

	return handle_options(visible_options, vm, first);
}

// created in start(), the destructor cleans up
//
struct initializer : boost::noncopyable
{
	bool has_python;
	bool has_curl;

	initializer(int argc, char** argv)
		: has_python(false), has_curl(false)
	{
		log() << "starting initializer\n";

		log() << "testing libxml\n";
		LIBXML_TEST_VERSION

		log() << "initializing libxml\n";
		xmlInitParser();

		log() << "initializing curl\n";
		CURLcode c = curl_global_init(CURL_GLOBAL_DEFAULT);

		if (c != CURLE_OK)
			log_error() << "curl failed with " << c << "\n";
		else
			has_curl = true;

		if (g_noext)
		{
			log() << "skipping python\n";
		}
		else
		{
			log() << "initializing python\n";

			std::vector<utf8_path> ext_paths;
			ext_paths.push_back(utf8_path(paths::python_path()));
			ext_paths.push_back(paths::extensions());
			ext_paths.push_back(platform::executable_directory());
			ext_paths.push_back(utf8_path(paths::python_zip()));

			has_python = ext::init_python(ext_paths);
		}

		log() << "initializing db\n";
		db::initialize();

		log() << "initializing wxWidgets\n";
		wxEntryStart(argc, argv);
		
		log() << "initializer done\n";
	}

	~initializer()
	{
		try
		{
			wxEntryCleanup();
			db::terminate();
			ext::term_python();
			xmlCleanupParser();

			if (has_curl)
				curl_global_cleanup();
		}
		catch(...)
		{
			std::cerr << "unhandled exception in ~initializer\n";
		}
	}
};

struct program_instance : boost::noncopyable
{
	bool first;
	ipc::file_lock flock;

	program_instance(int argc, char** argv)
		: first(true)
	{
		try
		{
			const utf8_path flock_path =
				paths::conf() / utf8_string("baretorrent-lock");

			if (!flock_path.exists())
				std::ofstream(flock_path.native().c_str());

#if BARETORRENT_WINDOWS
			const utf8_string fs = flock_path.as_utf8_string();

			const byte_array ba = convert(
				fs.c_str(), fs.size_in_bytes(),
				utf8_encoding, current_locale_encoding);

			if (ba.empty())
			{
				throw std::exception(
					"could not conver flock path from utf8 to "
					"current locale encoding");
			}

			const std::string p = ba.data();
#else
			const std::string p = flock_path.as_utf8_string().std_string();
#endif

			// mutex shared between instances; only the first instance
			// can lock it
			flock = ipc::file_lock(p.c_str());

			if (flock.try_lock())
			{
				// lock succeeded, this is the first instance
				create_ipc();
			}
			else
			{
				// locking fails if there's already an instance running;
				// saving the arguments in an ipc structure and exiting
				set_ipc(argc, argv);
				first = false;
			}
		}
		catch(std::exception& e)
		{
			std::cerr
				<< "couldn't not lock ipc mutex: " << e.what() << "\n";
		}

		// checking failed; assume that the user will be careful not to
		// start multiple instances
	}

	~program_instance()
	{
		try
		{
			flock.unlock();
		}
		catch(...)
		{
			// ignore
		}
	}
};

BARETORRENT_EXPORT int start(int argc, char** argv)
{
	try
	{
		platform::set_signals();

		// the application object
		baretorrent::app* a = new baretorrent::app;
		wxApp::SetInstance(a);

		// command line
		po::variables_map vm;
		
		std::vector<std::string> args;
		for (int i=1; i<argc; ++i)
			args.push_back(argv[i]);

		if (!parse_command_line(args, vm, true))
			return EXIT_SUCCESS;

		set_cmd_line(implode(args, utf8_string(" ")));
		std::cout
			<< "command line is '" << cmd_line() << "'\n"
			<< "verbose level is " << g_std_log.level() << "\n";

		g_std_log.start_capture();

		output_version(g_std_log);
		log()
			<< "command line: " << cmd_line() << "\n"
			<< "configuration path: " << paths::conf() << "\n"
			<< "extensions path: " << paths::extensions() << "\n";

		// creates the configuration path if needed
		paths::create_all();

		initializer init(argc, argv);
		
		// init
		load_options();
		update_options(
			baretorrent_version::parse(opt("bt.version").std_string()),
			version());

		load_locale();

		program_instance inst(argc, argv);

		// return if this is not the first instance
		if (!inst.first)
			return EXIT_SUCCESS;

		// CallOnInit(), OnRun() and OnExit() need to be called
		// manually because the application is not created using
		// wxWidget's macro

		{
			manager m;
			a->set(m);
			a->CallOnInit();

			if (handle_commands(*a, vm))
			{
				a->create_window();

				if (init.has_python)
					ext::load();

				a->show_window();
			}
			else
			{
				a->quit();
			}

			a->OnRun();

			ext::unload();
		}

		a->OnExit();

		log() << "baretorrent exited correctly\n";

		return EXIT_SUCCESS;
	}
	catch(db::exception& e)
	{
		std::cerr << "database exception: " << e.what() << "\n";
	}
	catch(std::exception& e)
	{
		std::cerr << e.what() << "\n";
	}
	catch(platform::exception& e)
	{
		std::cerr << "platform exception: " << e.what() << "\n";
	}
	catch(...)
	{
		std::cerr << "unhandled exception\n";
	}

	return EXIT_FAILURE;
}

}	// namespace baretorrent
