// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "main_window.h"
#include "app.h"
#include "torrent.h"
#include "locale.h"
#include "utility.h"
#include "resources.h"
#include "dialogs.h"
#include "files_panel.h"
#include "info_panel.h"
#include "connections_panel.h"
#include "options_dialog.h"
#include "ext_ui.h"
#include "ext/ext.h"
#include "log_dialog.h"
#include "python_console.h"

namespace baretorrent
{

// various widget ids
//
enum ids
{
	// quit menu item
	quit_id = wxID_EXIT,

	// options menu item
	options_id = wxID_PREFERENCES,

	// about menu item
	about_id = wxID_ABOUT,

	// install extension menu item
	install_ext_id = wxID_HIGHEST + 300,

	// python console menu item
	pcon_id,

	// shutdown after downloads
	ad_shutdown_id,
	ad_suspend_id,
	ad_hibernate_id,
	ad_power_cancel_id,

	// taskbar icon
	taskbar_id,

	log_dialog_id,

	// show menu on taskbar icon
	taskbar_show_id,

	// quit menu on taskbar icon
	taskbar_quit_id,

	// categories (for both menus and the toolbar)
	all_torrents_id,
	downloading_id,
	uploading_id,
	active_id,
	inactive_id,

	search_id,

	// bad extension bar
	bad_ext_yes_id,
	bad_ext_no_id,

	// update panel
	update_not_now_id,
	update_link_id,

	// associations panel
	assoc_yes_id,
	assoc_no_id,

	// accelerators
	reload_ext_acc_id,
	remove_torrent_acc_id
};

// time between update requests
const boost::posix_time::milliseconds update_timer_delay(2000);

// time after which updates are reposted
const boost::posix_time::milliseconds update_monitor_timer_delay(5000);

// timer after which the misc timer is fired
const boost::posix_time::milliseconds misc_timer_delay(500);

// timer after which to start checking for incoming connections
const boost::posix_time::minutes incoming_timer_delay(5);

// suffix of extension files
const utf8_string extension_suffix("btx");


// taskbar icon
//
class taskbar_icon : public wxTaskBarIcon
{
public:
	taskbar_icon(main_window* w)
		: window_(w)
	{
		bt_assert(IsOk());
	}

	void update(const wxString& s, wxIconBundle& icons)
	{
		// todo: I've had this return -1 on gtk (iirc)
		int is = wxSystemSettings::GetMetric(wxSYS_SMALLICON_X);
		if (is == -1)
		{
			// whatever
			is = 16;
		}

		SetIcon(icons.GetIcon(wxSize(is, is)), s);
	}
	
private:
	DECLARE_EVENT_TABLE()

	main_window* window_;

	// fired when double clicking the icon; shows (if needed) and
	// activates the main window
	//
	void on_double_click(wxTaskBarIconEvent&)
	{
		show();
	}

	// fired when the user clicks on the icon to open the context menu
	//
	virtual wxMenu* CreatePopupMenu()
	{
		wxMenu* m = new wxMenu;

		wxMenuItem* show = new wxMenuItem(
			m, taskbar_show_id, loc_wx("window.open"));

		// todo: SetFont() is only on windows
#if BARETORRENT_WINDOWS
		wxFont f = window_->GetFont();
		f.SetWeight(wxFONTWEIGHT_BOLD);
		show->SetFont(f);
#endif

		m->Append(show);

		m->AppendSeparator();
		m->Append(taskbar_quit_id, loc_wx("menu.file.quit"));

		return m;
	}

	void on_show(wxCommandEvent&)
	{
		show();
	}

	void on_quit(wxCommandEvent&)
	{
		quit();
	}

	void show()
	{
		window_->Show();
		window_->Iconize(false);
		window_->Raise();
	}
};


main_window::main_window(app& a) :
	custom_frame("bt.ui.main_window"),
	app_(a),
	taskbar_(nullptr),
	sb_(nullptr),
	list_(nullptr),
	info_(nullptr),
	files_(nullptr),
	connections_(nullptr),
	tabs_(nullptr),
	cat_(all_categories),
	tb_(nullptr),
	bad_ext_panel_(nullptr),
	notification_(nullptr),
	update_(nullptr), assoc_(nullptr), power_(nullptr),
	bad_ext_(nullptr),
	log_(nullptr),
	pcon_(nullptr),
	check_incoming_(false)
{
	ext::extension_loaded().connect(
		boost::bind(&main_window::on_extension_loaded, this, _1));

	create_ui();
	load_icons();

	if (opt<bool>("bt.ui.log"))
		open_log();

	SetDropTarget(new dnd(*this));

	list_->category(cat_);
	connect_events();
	update_options();

	update_titlebar(0, 0);

	// future events are from the user
	creating_ = false;

	(new queue_dialog(this, app_.manager()))->Show();
}

main_window::~main_window()
{
	delete taskbar_;
}

void main_window::open_log()
{
	if (!log_)
	{
		log_ = new log_dialog(this);
		log_->SetIcons(icons_);
	}

	log_->Show(true);
}

void main_window::connect_events()
{
	manager& m = app_.manager();

	m.adding.connect(boost::bind(
		&main_window::on_torrent_adding, this, _1));
	
	m.duplicate.connect(boost::bind(
		&main_window::on_torrent_duplicate, this, _1));

	m.added.connect(boost::bind(
		&main_window::on_torrent_added, this, _1));

	m.removed.connect(boost::bind(
		&main_window::on_torrent_removed, this, _1));

	m.add_progress.connect(boost::bind(
		&main_window::on_add_progress, this, _1, _2));

	m.power_action_changed.connect(boost::bind(
		&main_window::on_power_action_changed, this, _1));

	m.power_action.connect(boost::bind(
		&main_window::on_power_action, this, _1));

	app_.checking_update.connect(
		boost::bind(&main_window::on_checking_update, this));

	app_.update_available.connect(
		boost::bind(&main_window::on_update_available, this, _1));

	app_.update_done.connect(
		boost::bind(&main_window::on_update_done, this, _1));

	app_.missing_association.connect(
		boost::bind(&main_window::on_missing_association, this, _1));
}

void main_window::on_extension_loaded(ext::instance_ptr i)
{
	i->changed.connect(boost::bind(
		&main_window::on_extension_changed, this, i));
}

void main_window::on_extension_changed(ext::instance_ptr i)
{
	async_call(boost::bind(
		&main_window::on_extension_changed_impl, this, i));
}

void main_window::on_extension_changed_impl(ext::instance_ptr i)
{
	sb_->remove_icon(i);
}

void main_window::restart_update_timer()
{
	// this will be restarted when the update timer fires
	update_monitor_timer_.Stop();

	update_timer_.start(boost::bind(
		&main_window::on_update_timer, this), update_timer_delay);
}

void main_window::update_titlebar(int dl_rate, int ul_rate)
{
	if (dl_rate < opt<int>("bt.torrent.stalled_rate"))
		dl_rate = 0;

	if (ul_rate < opt<int>("bt.torrent.stalled_rate"))
		ul_rate = 0;

	wxString s = GetTitle();
	wxString ns = to_wxstring(locf("window.title")
		% make_loc_rate(dl_rate)
		% make_loc_rate(ul_rate));

	if (!ext::python_initialized())
		ns += " - " + loc_wx("window.title.no_python");

	if (ns != s)
	{
		SetTitle(ns);

		if (taskbar_)
			taskbar_->update(ns, icons_);
	}
}

void main_window::update_options()
{
	// this is called on startup and when the options dialog is opened

	create_taskbar();
	
	// updating monitor
	utf8_path m(opt("bt.torrent.auto_load"));
	app_.manager().monitor(m);

	// updating manager and app
	app_.manager().reload_settings();
	app_.reload_settings();
}

void main_window::iconized(bool b)
{
	if (b)
	{
		if (opt<bool>("bt.ui.minimize_notification"))
			Hide();
	}
}

void main_window::create_ui()
{
	wxPanel* panel = new wxPanel(this);

	// main splitter between the list and panels
	splitter* split = new splitter(panel, "bt.ui.window_splitter");

	list_ = new torrent_list(split, app_.manager());

	list_->selection_changed.connect(boost::bind(
		&main_window::on_selection_changed, this, _1));

	list_->activated.connect(boost::bind(
		&main_window::on_torrent_activated, this, _1));

	create_panels(split);
	create_menu();
	create_toolbar();
	create_status_bar();
	create_bad_ext(panel);
	create_notifications(panel);

	split->SplitVertically(list_, tabs_);
	split->SetSashGravity(1.0);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(notification_, 0, wxEXPAND);
	box->Add(split, 1, wxEXPAND);
	box->Add(bad_ext_panel_, 0, wxEXPAND);

	panel->SetSizer(box);

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(panel, 1, wxEXPAND);
	
	SetSizer(root);
	root->Layout();

	log_vv() << "main window create ui done\n";

	std::vector<wxAcceleratorEntry> ae;
	
	ae.push_back(wxAcceleratorEntry(
		wxACCEL_NORMAL, WXK_F2, reload_ext_acc_id));

	ae.push_back(wxAcceleratorEntry(
		wxACCEL_NORMAL, WXK_DELETE, remove_torrent_acc_id));

	SetAcceleratorTable(wxAcceleratorTable(
		static_cast<int>(ae.size()), &ae[0]));
}

void main_window::show_window()
{
	if (opt<bool>("bt.ui.minimized"))
	{
		// the option to minimize the window on startup is on

		if (opt<bool>("bt.ui.minimize_notification"))
		{
			// the option to minimize to the notification area is on
			Hide();
		}
		else
		{
			// minimize it in the taskbar
			Iconize();
			Show();
		}
	}
	else
	{
		// the minimize state is not remembered on purpose because
		// starting an application in the taskbar/notification area
		// when the option hasn't been set manually by the user may be
		// confusing ("where's the window?" or more likely "when I
		// click the program on my computer the program doesn't work")

		Show();
	}

	Layout();

	restart_update_timer();

	misc_timer_.start(boost::bind(
		&main_window::on_misc_timer, this), misc_timer_delay);

	incoming_timer_.start(boost::bind(
		&main_window::on_incoming_timer, this), incoming_timer_delay);

	if (opt<bool>("pcon.visible"))
		open_pcon();
}

void main_window::create_taskbar()
{
	const bool n = opt<bool>("bt.ui.minimize_notification");

	if (taskbar_ && n)
	{
		// the window is minimized to the notification area and the
		// icon is already there
		return;
	}

	// resets the notification; if the option is on, it will be created
	// below (in which case the following two lines are a noop because
	// there wasn't any icon before)
	delete taskbar_;
	taskbar_ = nullptr;

	if (n)
	{
		taskbar_ = new taskbar_icon(this);
		taskbar_->update(GetTitle(), icons_);
	}
}

void main_window::load_icons()
{
	// the platform will choose the right icon to display depending on
	// the context (window icon, notification area, alt-tab, etc.)

	icons_.AddIcon(platform::make_icon(images::icon_16));
	icons_.AddIcon(platform::make_icon(images::icon_24));
	icons_.AddIcon(platform::make_icon(images::icon_32));
	icons_.AddIcon(platform::make_icon(images::icon_48));
	icons_.AddIcon(platform::make_icon(images::icon_64));

	SetIcons(icons_);
}

status_bar& main_window::status()
{
	bt_assert(sb_);
	return *sb_;
}

void main_window::create_menu()
{
	wxMenuBar* mb = new wxMenuBar;

	// this assumes that for a menu M, a locale "menu.M" has the menu
	// caption and "menu.M.help" has the help string displayed in the
	// status bar

	// a regular menu item
#define a(m, id, n) m->Append( \
			id, loc_wx("menu." n), loc_wx("menu." n ".help"));

	// a checked menu item
#define ac(m, id, n) m->AppendCheckItem( \
			id, loc_wx("menu." n), loc_wx("menu." n ".help"));

	// a checked menu item with a shortcut
#define ac_s(m, id, n, s) m->AppendCheckItem( \
			id, loc_wx("menu." n) + "\t" + s, loc_wx("menu." n ".help"));

	// a radio menu item
#define ar(m, id, n) m->AppendRadioItem( \
			id, loc_wx("menu." n), loc_wx("menu." n ".help"));

	// file
	wxMenu* file = new wxMenu;

	a(file, add_id, "file.add");
	file->AppendSeparator();
	a(file, quit_id, "file.quit");

	mb->Append(file, loc_wx("menu.file"));
	
	
	// view
	wxMenu* view = new wxMenu;

	ar(view, all_torrents_id, "view.all");
	ar(view, downloading_id, "view.downloading");
	ar(view, uploading_id, "view.uploading");
	ar(view, active_id, "view.active");
	ar(view, inactive_id, "view.inactive");
	view->AppendSeparator();
	ac_s(view, search_id, "view.search", "Ctrl+F");

	mb->Append(view, loc_wx("menu.view"));

	
	// torrent	
	wxMenu* torrent = new wxMenu;

	a(torrent, resume_id, "torrent.resume");
	a(torrent, pause_id, "torrent.pause");
	torrent->AppendSeparator();
	a(torrent, up_id, "torrent.up");
	a(torrent, down_id, "torrent.down");
	a(torrent, top_id, "torrent.top");
	a(torrent, bottom_id, "torrent.bottom");
	torrent->AppendSeparator();
	a(torrent, remove_id, "torrent.remove");
	a(torrent, recheck_id, "torrent.recheck");

	mb->Append(torrent, loc_wx("menu.torrent"));


	// tools
	wxMenu* tools = new wxMenu;

	if (ext::python_initialized())
	{
		a(tools, install_ext_id, "tools.install_extension");
		
		if (opt<bool>("pcon.enabled"))
			ac(tools, pcon_id, "tools.pcon");
	}
	
	tools->AppendSeparator();

	// after downloads submenu
	wxMenu* ad = new wxMenu;

	if (platform::has_power(shutdown_power))
		ac(ad, ad_shutdown_id, "tools.after_downloads.shutdown");

	if (platform::has_power_management())
	{
		if (platform::has_power(suspend_power))
			ac(ad, ad_suspend_id, "tools.after_downloads.suspend");

		if (platform::has_power(hibernate_power))
			ac(ad, ad_hibernate_id, "tools.after_downloads.hibernate");
	}
	else
	{
#if BARETORRENT_LINUX
		ad->AppendSeparator();

		wxMenuItem* i = a(ad, -1, "tools.after_downloads.linux_pm");
		i->Enable(false);
#endif
	}

	tools->AppendSubMenu(ad, loc_wx("menu.tools.after_downloads"));

	a(tools, options_id, "tools.options");

	mb->Append(tools, loc_wx("menu.tools"));


	// help
	wxMenu* help = new wxMenu;

	ac(help, log_dialog_id, "help.log");

#if !BARETORRENT_APPLE
	help->AppendSeparator();
#endif

	a(help, about_id, "help.about");
	mb->Append(help, loc_wx("menu.help"));

#undef ac
#undef a

	SetMenuBar(mb);
}

void main_window::create_panels(wxWindow* parent)
{
	tabs_ = new wxNotebook(parent, -1);
	
	info_ = new info_panel(tabs_);
	files_ = new files_panel(tabs_);
	connections_ = new connections_panel(tabs_);
	
	tabs_->InsertPage(0, info_, loc_wx("general"));
	tabs_->InsertPage(1, files_, loc_wx("files"));
	tabs_->InsertPage(2, connections_, loc_wx("connections"));
}

void main_window::create_toolbar()
{
	tb_ = CreateToolBar(platform::toolbar_style());

	// adds a regular toolbar button
#define t(id, s, i) tb_->AddTool( \
			id, loc_wx(s), platform::make_image(i), loc_wx(s));

	// adds a radio toolbar button
#define rt(id, s, i) tb_->AddTool(					\
			id, loc_wx(s), platform::make_image(i),	\
			loc_wx(s), wxITEM_RADIO);

	// adds a toggle toolbar button
#define ct(id, s, i) tb_->AddTool(					\
			id, loc_wx(s), platform::make_image(i),	\
			loc_wx(s), wxITEM_CHECK);
	
	t(add_id, "torrent.add", images::add);
	t(remove_id, "torrent.remove", images::remove);
	
	tb_->AddSeparator();
	
	t(resume_id, "torrent.resume", images::resume);
	t(pause_id, "torrent.pause", images::pause);
	t(up_id, "torrent.up", images::up);
	t(down_id, "torrent.down", images::down);
	t(top_id, "torrent.top", images::top);
	t(bottom_id, "torrent.bottom", images::bottom);
	
	tb_->AddSeparator();
	
	rt(all_torrents_id, "category.all", images::all);
	rt(downloading_id, "category.downloading", images::download);
	rt(uploading_id, "category.uploading", images::upload);
	rt(active_id, "category.active", images::active);
	rt(inactive_id, "category.inactive", images::inactive);

	if (opt<bool>("pcon.enabled"))
	{
		tb_->AddSeparator();
		ct(pcon_id, "menu.tools.pcon", images::pcon_16);
	}

#undef ct
#undef rt
#undef t

	// this needs to be called after tools are added
	tb_->Realize();
}

void main_window::create_status_bar()
{
	sb_ = new status_bar(this);
	SetStatusBar(sb_);

	if (app_.manager().ready())
		sb_->clear_status();
	else
		sb_->set_status(loc("status.resuming_session"));
}

void main_window::create_bad_ext(wxWindow* parent)
{
	bad_ext_panel_ = new wxPanel(parent);

	wxButton* yes = new wxButton(
		bad_ext_panel_, bad_ext_yes_id, loc_wx("yes"));

	wxButton* no = new wxButton(
		bad_ext_panel_, bad_ext_no_id, loc_wx("no"));

	bad_ext_ = new wxStaticText(bad_ext_panel_, -1, wxEmptyString);

	wxBoxSizer* box = new wxBoxSizer(wxHORIZONTAL);
	box->Add(bad_ext_, 0, wxALIGN_CENTER_VERTICAL);
	box->AddSpacer(10);
	box->Add(yes);
	box->AddSpacer(10);
	box->Add(no);

	wxBoxSizer* root = new wxBoxSizer(wxHORIZONTAL);
	root->Add(box, 0, wxALL, 5);

	bad_ext_panel_->SetSizer(root);
	bad_ext_panel_->Hide();
}

void main_window::create_notifications(wxWindow* parent)
{
	notification_ = new notification_panel(parent);
	update_ = new update_notification(notification_);
	assoc_ = new assoc_notification(notification_);
	power_ = new power_notification(notification_);
}

void main_window::on_power_action_changed(power_actions p)
{
	async_call(boost::bind(
		&main_window::on_power_action_changed_impl, this, p));
}

void main_window::on_power_action_changed_impl(power_actions p)
{
	if (p == no_power_action)
	{
		notification_->close(power_);
	}
	else
	{
		power_->set(p);
		notification_->show(power_);
	}
}

void main_window::on_power_action(power_actions p)
{
	async_call(boost::bind(
		&main_window::on_power_action_impl, this, p));
}

void main_window::on_power_action_impl(power_actions p)
{
	log() << "power: showing dialog\n";

	Show();
	Iconize(false);
	Raise();
	RequestUserAttention(wxUSER_ATTENTION_ERROR);

	notification_->close(power_);

	power_dialog d(this, p);
	if (d.ShowModal() != wxID_OK)
	{
		log() << "power: action cancelled\n";
		return;
	}

	log() << "power: accepted\n";
	platform::power(p);
}

void main_window::quit()
{
	// this is called twice:
	//   1) when the user wants to exit (through the menu item, the
	//      close button, etc.): the application is notified, which
	//      stops the manager and waits for it to finish
	//
	//   2) after the application is done, in which case the window
	//      needs to be destroyed (it was merely hidden before)

	// hides the window immediately even though it might take a while
	// before libtorrent finishes cleaning up
	Hide();
	
	// killing timers
	update_timer_.Stop();
	update_monitor_timer_.Stop();


	if (app_.quitting())
	{
		// quit() has already been called once and the application has
		// now finished cleaning up
		
		log_v() << "destroying window\n";
		Destroy();
	}
	else
	{
		// this is in response to a user action; tells the application
		// to stop the manager, clean up and call Close() on this

		log_v() << "telling app to quit\n";
		app_.quit();
	}
}

void main_window::on_checking_update()
{
	async_call(boost::bind(
		&main_window::on_checking_update_impl, this));
}

void main_window::on_checking_update_impl()
{
	sb_->set_status(loc("status.checking_update"));
}

void main_window::on_update_done(bool b)
{
	async_call(boost::bind(
		&main_window::on_update_done_impl, this, b));
}

void main_window::on_update_done_impl(bool b)
{
	if (!b)
		sb_->set_status(loc("status.checking_update_failed"));
	else
		sb_->clear_status();
}

void main_window::on_update_available(const baretorrent_version& v)
{
	async_call(boost::bind(
		&main_window::on_update_available_impl, this, v));	
}

void main_window::on_update_available_impl(
	const baretorrent_version& v)
{
	update_->set(v);
	notification_->show(update_);
}

void main_window::on_missing_association(const utf8_string& other)
{
	async_call(boost::bind(
		&main_window::on_missing_association_impl, this, other));
}

void main_window::on_missing_association_impl(const utf8_string& other)
{
	assoc_->set(other);
	notification_->show(assoc_);
}

void main_window::on_quit(wxCommandEvent&)
{
	quit();
}

void main_window::on_install_ext(wxCommandEvent&)
{
	wxFileDialog dlg(
		this, loc_wx("browse.open_file"), wxEmptyString, wxEmptyString,
		loc_wx("wildcard.extensions"),
		wxFD_OPEN | wxFD_FILE_MUST_EXIST);

	if (dlg.ShowModal() != wxID_OK)
		return;

	utf8_path p(utf8_string(dlg.GetPath().ToUTF8().data()));
	if (p.empty())
		return;
	
	install_ext(p);

	//ext_browser b(this);
	//b.ShowModal();
}

void main_window::install_ext(const utf8_path& p)
{
	ext::instance_ptr i;

	try
	{
		i = ext::instance::extract(p);
		if (!i)
		{
			wxMessageBox(
				loc_wx("options.extensions.install.bad"),
				loc_wx("options.extensions.install.failed.title"),
				wxOK, this);

			return;
		}

		ext_install_dialog dlg(this, i);
		if (dlg.ShowModal() != wxID_OK)
			return;
		
		ext::install(i);
	}
	catch(ext::exception& e)
	{
		extension_error_dialog dlg(this, i, e.what());
		dlg.ShowModal();
	}
}

void main_window::on_reload_ext(wxCommandEvent&)
{
	ext::instance_list is = ext::instances();
	foreach(ext::instance_ptr i, is)
	{
		try
		{
			if (i->dev())
				i->reload();
		}
		catch(ext::exception&)
		{
			// ignore
		}
	}
}

void main_window::on_options(wxCommandEvent&)
{
	// options_dialog modifies a copy of the current options which is
	// used below to change all the options at once

	options_dialog dlg(this);
	if (dlg.ShowModal() != wxID_OK)
		return;
	
	// sets the new options
	const std::map<std::string, option_list> m = dlg.options();

	// no changes
	if (m.empty())
		return;

	std::pair<std::string, option_list> e;
	foreach (e, m)
	{
		std::pair<std::string, option> p;
		foreach (p, e.second)
		{
			if (e.first.empty())
				opt(p.first, p.second.value());
			else
				ext_opt(e.first, p.first, p.second.value());
		}
	}

	update_options();

	dump_options();
	app_.manager().dump_settings();
}

void main_window::on_ad_power_update(wxUpdateUIEvent& e)
{
	manager& m = app_.manager();

	e.Enable(m.downloading());

	switch (e.GetId())
	{
		case ad_shutdown_id:
		{
			e.Check(m.power() == shutdown_power);
			break;
		}

		case ad_suspend_id:
		{
			e.Check(m.power() == suspend_power);
			break;
		}

		case ad_hibernate_id:
		{
			e.Check(m.power() == hibernate_power);
			break;
		}
	}
}

void main_window::toggle_after_downloads(power_actions p)
{
	manager& m = app_.manager();
	
	if (m.power() == p)
		m.power(no_power_action);
	else
		m.power(p);
}

void main_window::on_ad_shutdown(wxCommandEvent&)
{
	toggle_after_downloads(shutdown_power);
}

void main_window::on_ad_suspend(wxCommandEvent&)
{
	toggle_after_downloads(suspend_power);
}

void main_window::on_ad_hibernate(wxCommandEvent&)
{
	toggle_after_downloads(hibernate_power);
}

void main_window::on_log(wxCommandEvent&)
{
	if (!log_)
	{
		open_log();
		opt("bt.ui.log", true);
	}
	else
	{
		const bool b = !log_->IsShown();
		log_->Show(b);

		if (b)
			log_->Raise();

		opt("bt.ui.log", b);
	}
}

void main_window::on_log_update(wxUpdateUIEvent& e)
{
	e.Check(log_ && log_->IsShown());
}

void main_window::on_about(wxCommandEvent&)
{
	about_dialog dlg(this);
	dlg.ShowModal();
}

void main_window::on_close(wxCloseEvent&)
{
	quit();
}

void main_window::on_torrent_updated(torrent_ptr t)
{
	if (!list_)
		return;

	list_->update(cat_, t);
	
	if (list_->is_selected(t))
	{
		// this torrent is selected so panels and the toolbar need to
		// be updated

		update_panels();
		update_toolbar();
	}

	// asks for updates again; if the manager still has more updates,
	// this will reset the timer
	restart_update_timer();
}

void main_window::on_bad_ext_no(wxCommandEvent&)
{
	ext::instance_list is = ext::instances();

	foreach(ext::instance_ptr i, is)
	{
		if (i->misbehaving())
		{
			i->accept_bad_behaviour();
			break;
		}
	}

	bad_ext_panel_->Hide();
	Layout();
}

void main_window::on_bad_ext_yes(wxCommandEvent&)
{
	ext::instance_list is = ext::instances();

	foreach(ext::instance_ptr i, is)
	{
		if (i->misbehaving())
		{
			try
			{
				i->enabled(false);
			}
			catch(ext::exception&)
			{
				// ignore, just go away
			}

			break;
		}
	}

	bad_ext_panel_->Hide();
	Layout();
}

void main_window::on_selection_changed(const torrents&)
{
	update_panels();
	update_toolbar(); 
}

void main_window::on_torrent_activated(const torrent_ptr t)
{
	open_directory(t);
}

bool main_window::on_update_timer()
{
	manager& m = app_.manager();

	update_titlebar(m.download_rate(), m.upload_rate());

	// this will eventually fire when no updates are received after
	// a while
	update_monitor_timer_.start(
		boost::bind(&main_window::on_update_monitor_timer, this),
		update_monitor_timer_delay);

	// asks libtorrent to post updates
	m.post_update();

	if (check_incoming_)
	{
		// don't display the message when there are no torrents
		if (m.has_active() || m.remote_peer_connected())
			sb_->set_incoming_connection(false);
		else
			sb_->set_incoming_connection(true);
	}

	sb_->set_remaining(m.total_remaining());

	// the timer will be restarted when updates are received or by the
	// update monitor timer if no updates are received for a while
	// (which means no torrents have changed)
	return false;
}

bool main_window::on_update_monitor_timer()
{
	// if this fires, if means libtorrent has not posted updates in the
	// last few seconds, which probably means there's either no torrents
	// or they are all stalled
	//
	// this restarts the update timer to request updates again
	restart_update_timer();

	return false;
}

void main_window::update_panels()
{
	if (!list_)
		return;

	const torrents s = list_->selection();

	// during debugging, the panels might be disabled just to test
	// stuff; this will prevent from crashing
	//
	// normally, those are always set before update_panels() is called

	if (info_)
		info_->update(s);

	if (files_)
		files_->update(s);

	if (connections_)
		connections_->update(s);
}

void main_window::update_toolbar()
{
	if (!tb_ || !list_)
		return;

	const torrents s = list_->selection();

	if (s.empty())
	{
		// no selection, disabling everything that relates to torrents

		tb_->EnableTool(remove_id, false);
		tb_->EnableTool(resume_id, false);
		tb_->EnableTool(pause_id, false);
		tb_->EnableTool(up_id, false);
		tb_->EnableTool(down_id, false);
		tb_->EnableTool(top_id, false);
		tb_->EnableTool(bottom_id, false);

		return;
	}

	tb_->EnableTool(resume_id, list_->selection_can_resume());
	tb_->EnableTool(pause_id, list_->selection_can_pause());

	// at least one torrent is selected, torrent-related buttons need
	// to be enabled
	tb_->EnableTool(remove_id, true);
	tb_->EnableTool(up_id, true);
	tb_->EnableTool(down_id, true);
	tb_->EnableTool(top_id, true);
	tb_->EnableTool(bottom_id, true);
}

void main_window::category(torrent_categories c)
{
	if (cat_ == c)
		return;

	cat_ = c;
	list_->category(c);
}

void main_window::async_call(callback_fun f)
{
	boost::mutex::scoped_lock lock(async_mutex_);
	async_.push_back(f);
}

void main_window::open_directory(const torrent_ptr t)
{
	platform::open_directory(t->root_path());
}

void main_window::on_add(wxCommandEvent&)
{
	add_dialog dlg(this);
	if (dlg.ShowModal() != wxID_OK)
		return;
	
	std::vector<utf8_string> ps = dlg.what();
	if (ps.empty())
		return;

	std::vector<utf8_string> ts;
	foreach (utf8_string s, ps)
	{
		// skip empty lines
		boost::trim(s);
		if (s.empty())
			continue;

		ts.push_back(s);
	}

	add_torrents(ts);
}

void main_window::add_torrents(const std::vector<utf8_string>& ps)
{
	new_.insert(new_.end(), ps.begin(), ps.end());
}

void main_window::on_remove(wxCommandEvent&)
{
	if (!list_)
		return;

	torrents s = list_->selection();
	if (s.empty())
		return;
	
	bool keep_files = opt<bool>("bt.ui.remove_keeps_files");

	if (opt<bool>("bt.ui.remove_confirm"))
	{
		remove_dialog dlg(this);
		if (dlg.ShowModal() != wxID_OK)
			return;

		keep_files = dlg.keep_files();
	}

	foreach (torrent_ptr t, s)
	{
		// the manager will fire on_torrent_removed immediately, which
		// will remove the item from the list
		app_.manager().remove_torrent(t, keep_files);
	}
}

void main_window::on_recheck(wxCommandEvent&)
{
	foreach (torrent_ptr t, list_->selection())
		t->recheck();
}

void main_window::on_move_storage(wxCommandEvent&)
{
	torrents ts = list_->selection();
	if (ts.empty())
		return;

	move_torrent_storage(ts);
}

void main_window::on_resume(wxCommandEvent&)
{
	// forces torrents to start even if they were errored
	foreach (torrent_ptr t, list_->selection())
		t->start(true);
}

void main_window::on_pause(wxCommandEvent&)
{
	foreach (torrent_ptr t, list_->selection())
		t->pause();
}

void main_window::on_up(wxCommandEvent&)
{
	foreach (torrent_ptr t, list_->selection())
		t->up();
}

void main_window::on_down(wxCommandEvent&)
{
	foreach (torrent_ptr t, list_->selection())
		t->down();
}

void main_window::on_top(wxCommandEvent&)
{
	foreach (torrent_ptr t, list_->selection())
		t->top();
}

void main_window::on_bottom(wxCommandEvent&)
{
	foreach (torrent_ptr t, list_->selection())
		t->bottom();
}

void main_window::on_open_directory(wxCommandEvent&)
{
	torrents ts = list_->selection();
	if (ts.empty())
		return;

	open_directory(ts[0]);
}

void main_window::on_category(wxCommandEvent& e)
{
	// switches on the button id to find which category has been
	// clicked and sets it (which will rebuild the list)

	switch (e.GetId())
	{
		case all_torrents_id:
			category(all_categories);
			break;

		case downloading_id:
			category(downloading_category);
			break;

		case uploading_id:
			category(uploading_category);
			break;

		case active_id:
			category(active_category);
			break;

		case inactive_id:
			category(inactive_category);
			break;

		default:
			log_error() << "unknown category " << e.GetId() << "\n";
			category(all_categories);
			break;
	}
}

void main_window::on_torrent_tool_update(wxUpdateUIEvent& e)
{
	switch (e.GetId())
	{
		case resume_id:
			e.Enable(list_->selection_can_resume());
			break;

		case pause_id:
			e.Enable(list_->selection_can_pause());
			break;

		case remove_id:		// fall-through
		case recheck_id:
		case move_storage_id:
		case up_id:
		case down_id:
		case top_id:
		case bottom_id:
		case open_directory_id:
			e.Enable(!list_->selection().empty());
			break;
	}
}

void main_window::on_category_update(wxUpdateUIEvent& e)
{
	// switches on the button id and checks it if it corresponds to the
	// current category

	switch (e.GetId())
	{
		case all_torrents_id:
			e.Check(cat_ == all_categories);
			break;

		case downloading_id:
			e.Check(cat_ == downloading_category);
			break;

		case uploading_id:
			e.Check(cat_ == uploading_category);
			break;

		case active_id:
			e.Check(cat_ == active_category);
			break;

		case inactive_id:
			e.Check(cat_ == inactive_category);
			break;

		// shouldn't happen
		default:
			bt_assert(false);
			e.Check(false);
	}
}

void main_window::on_search(wxCommandEvent&)
{
	list_->search(!list_->searching());
}

void main_window::on_search_update(wxUpdateUIEvent& e)
{
	e.Check(list_->searching());
}

void main_window::on_pcon(wxCommandEvent&)
{
	if (!pcon_)
	{
		open_pcon();
		opt("pcon.visible", true);
	}
	else
	{
		const bool b = !pcon_->IsShown();
		pcon_->Show(b);

		if (b)
			pcon_->Raise();

		opt("pcon.visible", b);
	}
}

void main_window::on_pcon_update(wxUpdateUIEvent& e)
{
	e.Check(pcon_ && pcon_->IsShown());
}

void main_window::open_pcon()
{
	if (!pcon_)
	{
		pcon_ = new python_console(this);
		pcon_->SetIcons(icons_);
	}

	pcon_->Show(true);
}

bool main_window::on_incoming_timer()
{
	log() << "starting to check for incoming connections\n";
	check_incoming_ = true;

	// this is a one-shot
	return false;
}

bool main_window::on_misc_timer()
{
	if (!new_.empty())
	{
		std::vector<utf8_string> v = new_;
		new_.clear();

		handle_new_files(v);
	}
	
	ext::instance_list is = ext::instances();

	foreach(ext::instance_ptr i, is)
	{
		if (i->misbehaving())
		{
			const wxString s =
				to_wxstring(locf("status.bad_ext") % i->name());

			if (bad_ext_->GetLabel() != s)
				bad_ext_->SetLabel(s);

			bad_ext_panel_->Show();
			Layout();

			break;
		}
	}
	
	std::vector<callback_fun> cf;
	
	{
		boost::mutex::scoped_lock lock(async_mutex_);
		cf = async_;
		async_.clear();
	}

	foreach (callback_fun f, cf)
		f();

	return true;
}

void main_window::handle_new_files(const std::vector<utf8_string>& v)
{
	manager& mg = app_.manager();

	if (v.size() == 1)
	{
		try
		{
			utf8_path p(v[0]);
			if (p.extension() == utf8_string(".") + extension_suffix)
			{
				// this is an extension file
				install_ext(p);
				return;
			}
		}
		catch(path_exception&)
		{
			// ignore
		}
	}

	std::vector<utf8_string> bad;
	std::vector<utf8_string> added;
	bool bad_path = false;

	foreach (const utf8_string& s, v)
	{
		try
		{
			mg.add_torrent(s);
			added.push_back(s);
		}
		catch(bt::invalid_torrent_file& e)
		{
			bad.push_back(s);
			log_error()
				<< "invalid torrent file " << s << ", "
				<< e.what() << "\n";
		}
		catch(path_exception& e)
		{
			bad_path = true;
			log_error()
				<< "path error for " << s << ", " << e.what() << "\n";
		}
	}

	if (!bad.empty())
	{
		bad_torrents_dialog dlg(this, bad);
		dlg.ShowModal();
	}
	else if (bad_path)
	{
		wxMessageBox(
			loc_wx("torrent.add.path_error"),
			loc_wx("torrent.add.path_error.title"),
			wxOK, this);
	}
}

void main_window::on_torrent_adding(torrent_ptr t)
{
	async_call(boost::bind(
		&main_window::on_torrent_adding_impl, this, t));
}

void main_window::on_torrent_adding_impl(torrent_ptr t)
{
	sb_->set_status(locf("torrent.adding") % t->name());
}

void main_window::on_torrent_duplicate(torrent_ptr t)
{
	async_call(boost::bind(
		&main_window::on_torrent_duplicate_impl, this, t));
}

void main_window::on_torrent_duplicate_impl(torrent_ptr t)
{
	sb_->set_status(locf("torrent.duplicate") % t->name());
}

void main_window::on_torrent_added(torrent_ptr t)
{
	async_call(boost::bind(
		&main_window::on_torrent_added_impl, this, t));
}

void main_window::on_torrent_added_impl(torrent_ptr t)
{
	// the slots will be disconnected when the torrent_item is deleted,
	// which happens when torrents are removed
	t->updated.connect(boost::bind(
		&main_window::on_torrent_updated,
		boost::ref(*this), t));

	t->moving.connect(boost::bind(
		&main_window::on_torrent_moving,
		boost::ref(*this), t));

	t->move_failed.connect(boost::bind(
		&main_window::on_torrent_move_failed,
		boost::ref(*this), t));

	t->moved.connect(boost::bind(
		&main_window::on_torrent_moved,
		boost::ref(*this), t));


	list_->add(t);

	// ready() will return true when all the torrents have been resumed,
	// in which case "ready" will be displayed on the status bar
	if (app_.manager().ready())
		sb_->clear_status();
}

void main_window::on_torrent_moving(torrent_ptr t)
{
	async_call(boost::bind(
		&main_window::on_torrent_moving_impl, this, t));
}

void main_window::on_torrent_moving_impl(torrent_ptr t)
{
	sb_->set_status(locf("torrent.moving") % t->name() % t->save_path());
}

void main_window::on_torrent_move_failed(torrent_ptr t)
{
	async_call(boost::bind(
		&main_window::on_torrent_move_failed_impl, this, t));
}

void main_window::on_torrent_move_failed_impl(torrent_ptr t)
{
	sb_->set_status(locf("torrent.move_failed")
		% t->name() % loc_error(t->error()));
}

void main_window::on_torrent_moved(torrent_ptr t)
{
	async_call(boost::bind(
		&main_window::on_torrent_moved_impl, this, t));
}

void main_window::on_torrent_moved_impl(torrent_ptr)
{
	sb_->clear_status();
}

void main_window::on_torrent_removed(torrent_ptr t)
{
	async_call(boost::bind(
		&main_window::on_torrent_removed_impl, this, t));
}

void main_window::on_torrent_removed_impl(torrent_ptr t)
{
	list_->remove(t);

	if (app_.manager().ready())
		sb_->set_status(locf("torrent.removed") % t->name());
}

void main_window::on_add_progress(std::size_t c, std::size_t t)
{
	async_call(boost::bind(
		&main_window::on_add_progress_impl, this, c, t));
}

void main_window::on_add_progress_impl(std::size_t c, std::size_t t)
{
	sb_->set_add_progress(c, t);
}


main_window::dnd::dnd(main_window& w)
	: w_(w)
{
}

bool main_window::dnd::OnDropFiles(
	wxCoord, wxCoord, const wxArrayString& fs)
{
	if (fs.empty())
		return true;
	
	std::vector<utf8_string> ps;
	for (std::size_t i=0; i<fs.Count(); ++i)
	{
		const utf8_string s(fs[i].ToUTF8());
		ps.push_back(s);
	}

	w_.add_torrents(ps);
	return true;
}


main_window::update_notification::update_notification(
	notification_panel* p)
		: notification(p), link_(nullptr)
{
	link_ = new wxHyperlinkCtrl(
		this, update_link_id, to_wxstring(website_url()),
		wxEmptyString, wxDefaultPosition, wxDefaultSize);

	wxButton* no = new wxButton(
		this, update_not_now_id, loc_wx("update.no"));

	link_->SetToolTip(link_->GetURL());

	wxFlexGridSizer* box = new wxFlexGridSizer(2, 0, 5);

	box->Add(link_, 0, wxALIGN_CENTER_VERTICAL);
	box->Add(no, 0, wxALIGN_CENTER_VERTICAL);

	box->AddGrowableCol(0);

	wxBoxSizer* root = new wxBoxSizer(wxHORIZONTAL);
	root->Add(box, 1, wxEXPAND | wxALL, 4);

	SetSizer(root);
}

void main_window::update_notification::set(const baretorrent_version& v)
{
	link_->SetLabel(to_wxstring(
		locf("update.available") % v.string));
}

void main_window::update_notification::on_update_link(
	wxHyperlinkEvent& e)
{
	close();
	e.Skip();
}

void main_window::update_notification::on_update_not_now(
	wxCommandEvent&)
{
	close();
	get_app().disable_updates();
}


main_window::assoc_notification::assoc_notification(
	notification_panel* p)
		: notification(p), label_(nullptr)
{
	label_ = new wxStaticText(this, -1, wxEmptyString);
	wxButton* yes = new wxButton(
		this, assoc_yes_id, loc_wx("assoc.yes"));
	wxButton* no = new wxButton(
		this, assoc_no_id, loc_wx("assoc.no"));

	wxFlexGridSizer* box = new wxFlexGridSizer(4, 0, 5);

	box->Add(label_, 0, wxALIGN_CENTER_VERTICAL);
	box->Add(yes, 0, wxALIGN_CENTER_VERTICAL);
	box->Add(no, 0, wxALIGN_CENTER_VERTICAL);

	box->AddGrowableCol(0);

	wxBoxSizer* root = new wxBoxSizer(wxHORIZONTAL);
	root->Add(box, 1, wxEXPAND | wxALL, 4);

	SetSizer(root);
}

void main_window::assoc_notification::set(const utf8_string& other)
{
	if (other.empty())
	{
		label_->SetLabelMarkup(loc_wx("assoc.label"));
	}
	else
	{
		label_->SetLabelMarkup(
			to_wxstring(locf("assoc.label_other") % other));
	}
}

void main_window::assoc_notification::on_yes(wxCommandEvent&)
{
	platform::associate();
	close();
}

void main_window::assoc_notification::on_no(wxCommandEvent&)
{
	opt("bt.assoc", false);
	close();
}


main_window::power_notification::power_notification(
	notification_panel* p)
		: notification(p), label_(nullptr)
{
	label_ = new wxStaticText(this, -1, wxEmptyString);
	wxButton* cancel = new wxButton(
		this, ad_power_cancel_id, loc_wx("power.cancel"));

	wxFlexGridSizer* box = new wxFlexGridSizer(2, 0, 5);

	box->Add(label_, 0, wxALIGN_CENTER_VERTICAL);
	box->Add(cancel, 0, wxALIGN_CENTER_VERTICAL);

	box->AddGrowableCol(0);

	wxBoxSizer* root = new wxBoxSizer(wxHORIZONTAL);
	root->Add(box, 1, wxEXPAND | wxALL, 4);

	SetSizer(root);
}

void main_window::power_notification::set(power_actions p)
{
	switch (p)
	{
		case shutdown_power:
		{
			label_->SetLabel(loc_wx("power.shutdown.label"));
			break;
		}

		case suspend_power:
		{
			label_->SetLabel(loc_wx("power.suspend.label"));
			break;
		}

		case hibernate_power:
		{
			label_->SetLabel(loc_wx("power.hibernate.label"));
			break;
		}

		case no_power_action:
		default:
		{
			break;
		}
	}
}

void main_window::power_notification::on_cancel(wxCommandEvent&)
{
	get_app().manager().power(no_power_action);
	close();
}


void move_torrent_storage(torrents ts)
{
	if (ts.empty())
		return;

	const wxString initial =
		to_wxstring(ts[0]->save_path().as_utf8_string());
	
	wxDirDialog dlg(
		get_app().GetTopWindow(),
		loc_wx("browse.open_directory"), initial);

	if (dlg.ShowModal() != wxID_OK)
		return;

	try
	{
		const utf8_path p(utf8_string(dlg.GetPath().ToUTF8()));
		if (p.empty())
			return;

		foreach (torrent_ptr& t, ts)
			t->move(p);
	}
	catch(path_exception& e)
	{
		log_error()
			<< "couldn't move storage, " << e.what() << "\n";
	}
}

BEGIN_EVENT_TABLE(taskbar_icon, wxTaskBarIcon)
	EVT_TASKBAR_LEFT_DCLICK(taskbar_icon::on_double_click)
	EVT_MENU(taskbar_show_id, taskbar_icon::on_show)
	EVT_MENU(taskbar_quit_id, taskbar_icon::on_quit)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(main_window, wxFrame)
	EVT_MENU(quit_id, main_window::on_quit)
	EVT_MENU(install_ext_id, main_window::on_install_ext)
	EVT_MENU(options_id, main_window::on_options)
	EVT_MENU(ad_shutdown_id, main_window::on_ad_shutdown)
	EVT_MENU(ad_suspend_id, main_window::on_ad_suspend)
	EVT_MENU(ad_hibernate_id, main_window::on_ad_hibernate)
	EVT_MENU(log_dialog_id, main_window::on_log)
	EVT_MENU(about_id, main_window::on_about)
	EVT_MENU(reload_ext_acc_id, main_window::on_reload_ext)
	EVT_MENU(remove_torrent_acc_id, main_window::on_remove)

	EVT_UPDATE_UI(ad_shutdown_id, main_window::on_ad_power_update)
	EVT_UPDATE_UI(ad_suspend_id, main_window::on_ad_power_update)
	EVT_UPDATE_UI(ad_hibernate_id, main_window::on_ad_power_update)

	EVT_CLOSE(main_window::on_close)
	EVT_TOOL(add_id, main_window::on_add)
	EVT_TOOL(remove_id, main_window::on_remove)
	EVT_TOOL(recheck_id, main_window::on_recheck)
	EVT_TOOL(move_storage_id, main_window::on_move_storage)
	
	EVT_TOOL(resume_id, main_window::on_resume)
	EVT_TOOL(pause_id, main_window::on_pause)
	EVT_TOOL(up_id, main_window::on_up)
	EVT_TOOL(down_id, main_window::on_down)
	EVT_TOOL(top_id, main_window::on_top)
	EVT_TOOL(bottom_id, main_window::on_bottom)
	EVT_TOOL(open_directory_id, main_window::on_open_directory)

	EVT_TOOL(all_torrents_id, main_window::on_category)
	EVT_TOOL(downloading_id, main_window::on_category)
	EVT_TOOL(uploading_id, main_window::on_category)
	EVT_TOOL(active_id, main_window::on_category)
	EVT_TOOL(inactive_id, main_window::on_category)
	EVT_TOOL(search_id, main_window::on_search)

	EVT_TOOL(pcon_id, main_window::on_pcon)

	EVT_UPDATE_UI(resume_id, main_window::on_torrent_tool_update)
	EVT_UPDATE_UI(pause_id, main_window::on_torrent_tool_update)
	EVT_UPDATE_UI(remove_id, main_window::on_torrent_tool_update)
	EVT_UPDATE_UI(recheck_id, main_window::on_torrent_tool_update)
	EVT_UPDATE_UI(move_storage_id, main_window::on_torrent_tool_update)
	EVT_UPDATE_UI(up_id, main_window::on_torrent_tool_update)
	EVT_UPDATE_UI(down_id, main_window::on_torrent_tool_update)
	EVT_UPDATE_UI(top_id, main_window::on_torrent_tool_update)
	EVT_UPDATE_UI(bottom_id, main_window::on_torrent_tool_update)
	EVT_UPDATE_UI(open_directory_id, main_window::on_torrent_tool_update)

	EVT_UPDATE_UI(all_torrents_id, main_window::on_category_update)
	EVT_UPDATE_UI(downloading_id, main_window::on_category_update)
	EVT_UPDATE_UI(uploading_id, main_window::on_category_update)
	EVT_UPDATE_UI(active_id, main_window::on_category_update)
	EVT_UPDATE_UI(inactive_id, main_window::on_category_update)
	EVT_UPDATE_UI(search_id, main_window::on_search_update)

	EVT_UPDATE_UI(pcon_id, main_window::on_pcon_update)

	EVT_UPDATE_UI(log_dialog_id, main_window::on_log_update)

	EVT_BUTTON(bad_ext_yes_id, main_window::on_bad_ext_yes)
	EVT_BUTTON(bad_ext_no_id, main_window::on_bad_ext_no)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(main_window::update_notification, notification)
	EVT_BUTTON(update_not_now_id, main_window::update_notification::on_update_not_now)
	EVT_HYPERLINK(update_link_id, main_window::update_notification::on_update_link)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(main_window::assoc_notification, notification)
	EVT_BUTTON(assoc_yes_id, main_window::assoc_notification::on_yes)
	EVT_BUTTON(assoc_no_id, main_window::assoc_notification::on_no)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(main_window::power_notification, notification)
	EVT_BUTTON(ad_power_cancel_id, main_window::power_notification::on_cancel)
END_EVENT_TABLE()

}	// namespace baretorrent
