// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_MAIN_WINDOW_INCLUDED
#define BARETORRENT_MAIN_WINDOW_INCLUDED

#include "utility.h"
#include "locale.h"
#include "resources.h"
#include "widgets.h"
#include "torrent_list.h"

namespace baretorrent
{

class app;
class files_panel;
class info_panel;
class connections_panel;
class log_dialog;
class taskbar_icon;
class python_console;

// various ids shared between widgets
//
enum common_ids
{
	// torrents (for both menus and the toolbar)
	add_id = wxID_HIGHEST + 400,
	remove_id,
	recheck_id,
	move_storage_id,
	resume_id,
	pause_id,
	up_id,
	down_id,
	top_id,
	bottom_id,
	open_directory_id
};

// this is the main toplevel window
//
// it registers events on the manager and torrents to update the user
// interface; the handlers (such as on_torrent_added() and
// on_torrent_updated()) are called from the main thread (see
// manager::tick())
//
class main_window : public custom_frame
{
public:
	typedef boost::function<void ()> callback_fun;

	// creates the window and widgets
	//
	main_window(app& a);

	// saves the options and cleans up
	//
	~main_window();

	// shows, hides and/or iconize the window depending on the
	// options and sets the window bounds
	//
	void show_window();

	void add_torrents(const std::vector<utf8_string>& ps);
	status_bar& status();

	void async_call(callback_fun f);

	wxNotebook* tabs()
	{
		return tabs_;
	}

private:
	DECLARE_EVENT_TABLE()

	class dnd : public wxFileDropTarget
	{
	public:
		dnd(main_window& w);
		virtual bool OnDropFiles(
			wxCoord, wxCoord, const wxArrayString& fs);

	private:
		main_window& w_;
	};

	class update_notification : public notification
	{
	public:
		update_notification(notification_panel* p);
		void set(const baretorrent_version& v);

	private:
		DECLARE_EVENT_TABLE()

		wxHyperlinkCtrl* link_;

		// fired when the 'not now' button is clicked for updates
		//
		void on_update_not_now(wxCommandEvent&);

		// fired when the link is clicked for updates
		//
		void on_update_link(wxHyperlinkEvent& e);
	};

	class assoc_notification : public notification
	{
	public:
		assoc_notification(notification_panel* p);

		void set(const utf8_string& other);
	
	private:
		DECLARE_EVENT_TABLE()

		wxStaticText* label_;

		void on_yes(wxCommandEvent&);
		void on_no(wxCommandEvent&);
	};

	class power_notification : public notification
	{
	public:
		power_notification(notification_panel* p);

		void set(power_actions p);

	private:
		DECLARE_EVENT_TABLE()

		wxStaticText* label_;

		void on_cancel(wxCommandEvent&);
	};


	// used for getting the manager, reloading options, etc.
	app& app_;

	// list of images for the window and notification icons
	wxIconBundle icons_;

	// notification icon
	taskbar_icon* taskbar_;

	// status bar
	status_bar* sb_;

	
	// main list
	torrent_list* list_;
	
	// torrent information panel
	info_panel* info_;

	// file list panel
	files_panel* files_;

	// tracker and peers panel
	connections_panel* connections_;

	// has the panels
	wxNotebook* tabs_;

	// currently selected category
	torrent_categories cat_;

	// toolbar
	wxToolBar* tb_;

	// bottom bar for misbehaving extensions
	wxPanel* bad_ext_panel_;

	// top bar for notifications
	notification_panel* notification_;

	update_notification* update_;
	assoc_notification* assoc_;
	power_notification* power_;

	// label in the bottom bar
	wxStaticText* bad_ext_;

	// log window
	log_dialog* log_;

	// python console window
	python_console* pcon_;
		

	// one shot; when fired, asks libtorrent to post updates on
	// torrents and restarts the timer when updates are received
	timer update_timer_;

	// if libtorrent does not post any updates after a few seconds
	// (because no torrents have changed), restarts the update timer
	timer update_monitor_timer_;

	// when fired, starts checking for incoming connections
	timer incoming_timer_;

	// fired once a second, checks various things like misbehaving
	// extensions
	timer misc_timer_;

	// the currently selected torrents
	//torrents selection_;
	
	// when true, starts checking the incoming connections; this is so
	// the message isn't displayed right on startup
	bool check_incoming_;

	// 
	std::vector<callback_fun> async_;
	boost::mutex async_mutex_;

	// filled in by d&d or the add torrent dialog; mostly not to
	// block d&d
	std::vector<utf8_string> new_;



	// connects to the manager events
	//
	void connect_events();


	// calls the various widget create functions below
	//
	void create_ui();

	// creates the menu bar
	//
	void create_menu();

	// creates the toolbar
	//
	void create_toolbar();

	// creates the status bar
	//
	void create_status_bar();


	// creates the various panels
	//
	void create_panels(wxWindow* parent);
	
	// creates the notification icon (if enabled); also called when
	// changing the options
	//
	void create_taskbar();

	// creates the bottom bar for bad extensions
	//
	void create_bad_ext(wxWindow* parent);

	// creates the top bar for notifications
	//
	void create_notifications(wxWindow* parent);


	// creates and/or opens the log dialog
	//
	void open_log();

	// creates and/or opens the python console
	//
	void open_pcon();


	virtual void iconized(bool b);
	void install_ext(const utf8_path& p);
	void handle_new_files(const std::vector<utf8_string>& v);


	// loads the application icons (for the window and notification
	// icons)
	//
	void load_icons();

	// recreate the notification icon (if needed) and updates the
	// manager settings; called on startup and after the options dialog
	// has been shown
	//
	void update_options();

	// calls quit() on the application which stops the manager and
	// eventually quits the program; called from various handlers
	//
	void quit();


	// updates the state of the toolbar buttons depending the state of
	// the selected torrents; called when the selection changes or
	// torrents are updated (their state might change from running to
	// paused, for example)
	//
	void update_toolbar();

	// updates the titlebar with the given download and upload rates
	//
	void update_titlebar(int dl_rate, int ul_rate);

	// updates the various panels; called when the selection changes
	//
	void update_panels();

	// restarts the update timer after torrent updates have been
	// received or the monitor timer is fired
	//
	void restart_update_timer();

	// opens the directory where the given torrent is saved; what
	// happens is platform-specific, such as opening explorer or
	// nautilus
	//
	void open_directory(const torrent_ptr t);


	// sets the currently displayed category; rebuilds the list
	//
	void category(torrent_categories c);


	void on_extension_loaded(ext::instance_ptr i);
	void on_extension_changed(ext::instance_ptr i);
	void on_extension_changed_impl(ext::instance_ptr i);

	// fired from update_timer_; asks the manager to post updates and
	// updates the titlebar (since that information is cached)
	//
	// the timer is restarted after the updates are received or in the
	// update monitor timer (see inside on_update_monitor_timer())
	//
	// this also starts the update monitor timer in case updates are
	// never posted by libtorrent (because nothing has changed) so the
	// update timer can be restarted
	//
	bool on_update_timer();

	// fired from update_monitor_timer_; restarts the update timer
	// 
	bool on_update_monitor_timer();

	// fired from incoming_timer_; starts checking for incoming
	// connections
	//
	bool on_incoming_timer();

	// fired from misc_timer_; checks various things
	//
	bool on_misc_timer();
		
	void on_selection_changed(const torrents& t);
	void on_torrent_activated(const torrent_ptr t);


	// fired from the app when it is checking for updates
	//
	void on_checking_update();
	void on_checking_update_impl();

	// fired from the app when it is done checking for updates, with
	// 'b' being false if there was an error
	//
	void on_update_done(bool b);
	void on_update_done_impl(bool b);

	// fired from the app when an update is available
	//
	void on_update_available(const baretorrent_version& v);
	void on_update_available_impl(const baretorrent_version& v);

	// fired from the app when the associations are missing
	//
	void on_missing_association(const utf8_string& other);
	void on_missing_association_impl(const utf8_string& other);

	// fired when the close button on the window is clicked, alt-f4
	// is pressed, etc.; calls quit()
	//
	void on_close(wxCloseEvent&);


	// fired when the add torrent menu item is selected or the toolbar
	// button is clicked; shows the add torrent dialog, adds the torrent
	// to the manager and rebuilds the list (todo: why?
	// on_torrent_added() will be called)
	//
	void on_add(wxCommandEvent&);

	// fired when the quit menu item is selected; calls quit()
	//
	void on_quit(wxCommandEvent&);


	// fired when any category menu item is selected or the toolbar
	// buttons are clicked; calls category() with the new value
	//
	void on_category(wxCommandEvent&);

	// fired by wxWidgets when the menu is shown and regularly for
	// toolbar items; updates the toggle state of the category toolbar
	// buttons and items since the category may be changed from either
	// the toolbar (which needs to update the menu) or the menu (which
	// needs to update the toolbar)
	// 
	void on_category_update(wxUpdateUIEvent&);

	// updates the resume/pause items in the menu
	//
	void on_torrent_tool_update(wxUpdateUIEvent& e);

	// toggles the search panel
	//
	void on_search(wxCommandEvent& e);

	// updates the search item in the view menu
	//
	void on_search_update(wxUpdateUIEvent& e);

	// fired when the recheck menu item is selected
	//
	void on_recheck(wxCommandEvent&);

	// fired when the move storage menu item is selected
	//
	void on_move_storage(wxCommandEvent&);

	// fired when the resume menu item is selected or the toolbar button
	// is clicked; starts all the selected torrents (even errored ones)
	//
	void on_resume(wxCommandEvent&);

	// fired when the pause menu item is selected or the toolbar button
	// is clicked; pauses all the selected torrents
	//
	void on_pause(wxCommandEvent&);

	// fired when the up menu item is selected or the toolbar button is
	// clicked; moves all the selected torrents up one position in the
	// queue
	//
	void on_up(wxCommandEvent&);

	// fired when the down menu item is selected or the toolbar button
	// is clicked; moves all the selected torrents down one position in
	// the queue
	//
	void on_down(wxCommandEvent&);

	// fired when the top menu item is selected or the toolbar button
	// is clicked; moves all the selected torrents to the top of the
	// queue
	//
	void on_top(wxCommandEvent&);

	// fired when the bottom menu item is selected or the toolbar button
	// is clicked; moves all the selected torrents to the bottom of the
	// queue
	//
	void on_bottom(wxCommandEvent&);

	// fired when the remove menu item is selected or the toolbar button
	// is clicked; shows a confirmation dialog where the user can
	// specify whether to keep the files, removes the torrent from the
	// manager and updates the selection (because the selected torrents
	// are gone)
	//
	// the list itself is updated in on_torrent_removed(), which is
	// fired immediately by the manager
	//
	void on_remove(wxCommandEvent&);

	// fired when the user selects the entry from the menu
	//
	void on_open_directory(wxCommandEvent&);


	// fired when the options menu item is clicked; shows the options
	// dialog and updates the options if needed
	//
	void on_options(wxCommandEvent&);

	void on_ad_power_update(wxUpdateUIEvent& e);

	// fired when the shutdown after downloads item is clicked
	//
	void on_ad_shutdown(wxCommandEvent&);

	// fired when the suspend after downloads item is clicked
	//
	void on_ad_suspend(wxCommandEvent&);

	// fired when the hibernate after downloads item is clicked
	//
	void on_ad_hibernate(wxCommandEvent&);

	// fired when the install extension menu item is clicked; shows an
	// open file dialog and installs the selected extension file
	//
	void on_install_ext(wxCommandEvent&);

	// reloads all the dev extensions
	//
	void on_reload_ext(wxCommandEvent&);

	// fired when the about menu item is clicked; shows the about
	// dialog
	//
	void on_about(wxCommandEvent&);

	void on_log(wxCommandEvent&);
	void on_log_update(wxUpdateUIEvent& e);

	void on_pcon(wxCommandEvent&);
	void on_pcon_update(wxUpdateUIEvent& e);

	void toggle_after_downloads(power_actions p);
	void on_power_action_changed(power_actions p);
	void on_power_action_changed_impl(power_actions p);
	void on_power_action(power_actions p);
	void on_power_action_impl(power_actions p);

	// fired by the manager when a torrent has been added to libtorrent
	// but not acknowledged yet (see on_torrent_added()); updates the
	// status bar
	//
	void on_torrent_adding(torrent_ptr t);
	void on_torrent_adding_impl(torrent_ptr t);

	// fired by the manager when the new torrent is a duplicate; updates
	// the status bar
	//
	void on_torrent_duplicate(torrent_ptr t);
	void on_torrent_duplicate_impl(torrent_ptr t);

	// fired by the manager when libtorrent acknowledges a new torrent
	// (see on_torrent_adding()); adds the torrent to the list and
	// updates the status bar
	//
	void on_torrent_added(torrent_ptr t);
	void on_torrent_added_impl(torrent_ptr t);

	// fired immediately by the manager when a torrent is removed
	// without waiting for libtorrent to acknowledge; removes the list
	// item corresponding to the given torrent
	//
	void on_torrent_removed(torrent_ptr t);
	void on_torrent_removed_impl(torrent_ptr t);

	void on_add_progress(std::size_t c, std::size_t t);
	void on_add_progress_impl(std::size_t c, std::size_t t);


	void on_torrent_moving(torrent_ptr t);
	void on_torrent_moving_impl(torrent_ptr t);
	void on_torrent_move_failed(torrent_ptr t);
	void on_torrent_move_failed_impl(torrent_ptr t);
	void on_torrent_moved(torrent_ptr t);
	void on_torrent_moved_impl(torrent_ptr t);

	// fired by the manager when libtorrent posts updates for a torrent
	// (requested in on_update_timer()); this:
	//
	// 1) adds or remove the list item if needed (if the torrent's
	//    category changes);
	// 2) updates the list item, panels and toolbars;
	// 3) restarts the update timer
	//
	// this may never be fired if libtorrent doesn't post any updates,
	// in which case update_monitor_timer_ eventually kicks in and
	// restarts the update timer
	//
	void on_torrent_updated(torrent_ptr t);

	// fired when the 'no' button is clicked in the bad extension
	// bar
	//
	void on_bad_ext_no(wxCommandEvent&);

	// fired when the 'yes' button is clicked in the bad extension
	// bar
	//
	void on_bad_ext_yes(wxCommandEvent&);
};

void move_torrent_storage(torrents ts);

}	// namespace baretorrent

#endif	// BARETORRENT_MAIN_WINDOW_INCLUDED
