// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "manager.h"
#include "torrent.h"
#include "utility.h"
#include "locale.h"
#include "options.h"
#include "platform.h"


namespace baretorrent
{

// number of milliseconds between checks in the monitor thread function
const boost::posix_time::milliseconds monitor_thread_delay(5000);

// number of milliseconds to sleep when waiting for the xml queue file
// to be written; this may take time because libtorrent might not
// even have acknowledged some of the new torrents yet
const boost::posix_time::milliseconds queue_file_check_delay(500);

// maximum number of alarms that can be queued until discarded; this
// should never be hit, it's just in case there's a problem
const std::size_t max_queued_alarms = 100;

// time between saving resume data
const boost::posix_time::milliseconds save_timer_delay(60 * 1000);

// delay after the first monitor error message
const boost::posix_time::seconds monitor_error_first_delay(10);

// delay after the second monitor error message
const boost::posix_time::seconds monitor_error_second_delay(60);

// maximum monitor error message delay
const boost::posix_time::seconds monitor_error_max_delay(600);



// defined in main.cpp, called in manager::tick() to check whether
// another instance passed arguments
//
void check_ipc();


manager::manager() :
	started_(false), session_(make_fingerprint()), quitting_(false),
	ready_(false), con_(checking_connection), power_(no_power_action),
	progress_total_(0), progress_current_(0)
{
	log() << "peer id is " << make_fingerprint().to_string() << "\n";
}

manager::~manager()
{
	// waiting until all torrents are added
	for (;;)
	{
		if (save_queue())
			break;

		boost::this_thread::sleep(queue_file_check_delay);
	}

	save_session();

	log_v() << "destroying manager\n";
	quit();
}

void manager::start()
{
	boost::thread(boost::bind(&manager::start_fun, this));
}

void manager::start_fun()
{
	init_session();
	resume_torrents();
	ready_ = true;
}

void manager::init_session()
{
	try
	{
		utf8_path sf = paths::conf() / paths::session_filename();

		if (sf.exists())
			load_session(sf);
		else
			log() << "no session file in " << sf.as_utf8_string() << "\n";
	}
	catch(std::exception& e)
	{
		log_error() << "can't load session file, " << e.what() << "\n";
	}

	session_.set_alert_mask(
		bt::alert::error_notification |
		bt::alert::storage_notification |
		bt::alert::tracker_notification |
		bt::alert::status_notification |
		bt::alert::debug_notification |
		bt::alert::progress_notification);

	session_.add_extension(&bt::create_metadata_plugin);
	session_.add_extension(&bt::create_ut_metadata_plugin);
	session_.add_extension(&bt::create_ut_pex_plugin);
	session_.add_extension(&bt::create_smart_ban_plugin);

	reload_settings();

	session_.start_dht();
	session_.start_lsd();
	session_.start_upnp();
	session_.start_natpmp();
}

template <typename T>
typename boost::disable_if< boost::is_enum<T>, bool>::type 
	set_session_setting(const utf8_string& o, T& t)
{
	t = as<T>(o);
	return true;
}

template <>
bool set_session_setting<std::string>(
	const utf8_string& o, std::string& t)
{
	t = o.std_string();
	return true;
}

template <typename T>
typename boost::enable_if< boost::is_enum<T>, bool>::type 
	set_session_setting(const utf8_string& o, T& t)
{
	std::istringstream iss(o.std_string());
	int i;
	iss >> i;

	t = static_cast<T>(i);
	return true;
}

void manager::reload_settings()
{
	// current settings
	bt::session_settings ss = session_.settings();

#define s(w) set_session_setting(							\
		opt("bt.session." BOOST_PP_STRINGIZE(w)), ss.w);

	s(user_agent);
	s(tracker_completion_timeout);
	s(tracker_receive_timeout);
	s(stop_tracker_timeout);
	s(tracker_maximum_response_length);
	s(piece_timeout);
	s(request_queue_time);
	s(max_allowed_in_request_queue);
	s(max_out_request_queue);
	s(whole_pieces_threshold);
	s(peer_timeout);
	s(urlseed_timeout);
	s(urlseed_pipeline_size);
	s(file_pool_size);
	s(allow_multiple_connections_per_ip);
	s(max_failcount);
	s(min_reconnect_time);
	s(peer_connect_timeout);
	s(ignore_limits_on_local_network);
	s(connection_speed);
	s(send_redundant_have);
	s(lazy_bitfields);
	s(inactivity_timeout);
	s(unchoke_interval);
	s(optimistic_unchoke_interval);
	s(announce_ip);
	s(num_want);
	s(initial_picker_threshold);
	s(allowed_fast_set_size);
	s(suggest_mode);
	s(max_queued_disk_bytes);
	s(handshake_timeout);
	s(use_dht_as_fallback);
	s(free_torrent_hashes);
	s(upnp_ignore_nonrouters);
	s(send_buffer_watermark);
	s(send_buffer_watermark_factor);
	s(choking_algorithm);
	s(seed_choking_algorithm);
	s(use_parole_mode);
	s(cache_size);
	s(cache_buffer_chunk_size);
	s(cache_expiry);
	s(use_read_cache);
	s(explicit_read_cache);
	s(explicit_cache_interval);
	s(disk_io_write_mode);
	s(disk_io_read_mode);
	//s(outgoing_ports);
	s(peer_tos);
	s(active_downloads);
	s(active_seeds);
	s(active_dht_limit);
	s(active_tracker_limit);
	s(active_limit);
	s(auto_manage_prefer_seeds);
	s(dont_count_slow_torrents);
	s(auto_manage_interval);
	s(share_ratio_limit);
	s(seed_time_ratio_limit);
	s(seed_time_limit);
	s(peer_turnover_interval);
	s(peer_turnover);
	s(peer_turnover_cutoff);
	s(close_redundant_connections);
	s(auto_scrape_interval);
	s(auto_scrape_min_interval);
	s(max_peerlist_size);
	s(min_announce_interval);
	s(prioritize_partial_pieces);
	s(auto_manage_startup);
	s(rate_limit_ip_overhead);
	s(announce_to_all_trackers);
	s(announce_to_all_tiers);
	s(prefer_udp_trackers);
	s(strict_super_seeding);
	s(seeding_piece_quota);
	s(max_sparse_regions);
	s(lock_disk_cache);
	s(max_rejects);
	s(recv_socket_buffer_size);
	s(send_socket_buffer_size);
	s(optimize_hashing_for_speed);
	s(file_checks_delay_per_block);
	s(disk_cache_algorithm);
	s(read_cache_line_size);
	s(write_cache_line_size);
	s(optimistic_disk_retry);
	s(disable_hash_checks);
	s(max_suggest_pieces);
	s(drop_skipped_requests);
	s(low_prio_disk);
	s(local_service_announce_interval);
	s(dht_announce_interval);
	s(udp_tracker_token_expiry);
	s(volatile_read_cache);
	s(guided_read_cache);
	s(default_cache_min_age);
	s(num_optimistic_unchoke_slots);
	s(no_atime_storage);
	s(default_est_reciprocation_rate);
	s(increase_est_reciprocation_rate);
	s(decrease_est_reciprocation_rate);
	s(incoming_starts_queued_torrents);
	s(report_true_downloaded);
	s(strict_end_game_mode);
	s(broadcast_lsd);
	s(enable_outgoing_utp);
	s(enable_incoming_utp);
	s(enable_outgoing_tcp);
	s(enable_incoming_tcp);
	s(max_pex_peers);
	s(ignore_resume_timestamps);
	s(no_recheck_incomplete_resume);
	s(anonymous_mode);
	s(tick_interval);
	s(share_mode_target);
	s(upload_rate_limit);
	s(download_rate_limit);
	s(local_upload_rate_limit);
	s(local_download_rate_limit);
	s(dht_upload_rate_limit);
	s(unchoke_slots_limit);
	s(half_open_limit);
	s(connections_limit);
	s(utp_target_delay);
	s(utp_gain_factor);
	s(utp_min_timeout);
	s(utp_syn_resends);
	s(utp_num_resends);
	s(utp_connect_timeout);
	s(utp_delayed_ack);
	s(utp_dynamic_sock_buf);
	s(utp_loss_multiplier);
	s(mixed_mode_algorithm);
	s(rate_limit_utp);
	s(listen_queue_size);
	s(announce_double_nat);
	s(torrent_connect_boost);
	s(seeding_outgoing_connections);
	s(no_connect_privileged_ports);
	s(alert_queue_size);
	s(max_metadata_size);
	s(smooth_connects);
	s(always_send_user_agent);
	s(apply_ip_filter_to_trackers);
	s(read_job_every);
	s(use_disk_read_ahead);
	s(lock_files);
	s(ssl_listen);
	s(tracker_backoff);
	s(ban_web_seeds);

#undef s

	// changing settings based on the options

	if (opt<bool>("bt.torrent.active_downloads_enabled"))
		ss.active_downloads = opt<int>("bt.torrent.active_downloads");
	else
		ss.active_downloads = -1;

	if (opt<bool>("bt.torrent.active_uploads_enabled"))
		ss.active_seeds = opt<int>("bt.torrent.active_uploads");
	else
		ss.active_seeds = -1;

	ss.share_ratio_limit = opt<int>("bt.torrent.seed_ratio") / 100.0f;

	if (opt<bool>("bt.torrent.auto_active_limit"))
	{
		if (ss.active_downloads == -1 || ss.active_seeds == -1)
			ss.active_limit = -1;
		else
			ss.active_limit = ss.active_downloads + ss.active_seeds;
	}

	if (opt<bool>("bt.torrent.max_download_enabled"))
	{
		ss.download_rate_limit =
			opt<int>("bt.torrent.max_download") * 1024;
	}
	else
	{
		ss.download_rate_limit = 0;
	}

	if (opt<bool>("bt.torrent.max_upload_enabled"))
		ss.upload_rate_limit = opt<int>("bt.torrent.max_upload") * 1024;
	else
		ss.upload_rate_limit = 0;

	ss.outgoing_ports = parse_ports(opt("bt.network.listen"));
	ss.connections_limit = opt<int>("bt.network.num_connections");
	
	const utf8_string pt = opt("bt.network.proxy_type");
	bt::proxy_settings ps;

	if (opt<bool>("bt.network.proxy_enabled") && !pt.empty())
	{
		if (pt == utf8_string("SOCKS4"))
		{
			ps.type = bt::proxy_settings::socks4;
		}
		else if (pt == utf8_string("SOCKS5"))
		{
			if (!ps.username.empty() || !ps.password.empty())
				ps.type = bt::proxy_settings::socks5_pw;
			else
				ps.type = bt::proxy_settings::socks5;
		}
		else if (pt == utf8_string("HTTP"))
		{
			if (!ps.username.empty() || !ps.password.empty())
				ps.type = bt::proxy_settings::http_pw;
			else
				ps.type = bt::proxy_settings::http;
		}
		else
		{
			log_error() << "unknown proxy type '" << pt << "'\n";
		}

		if (ps.type != bt::proxy_settings::none)
		{
			utf8_string h = opt("bt.network.proxy");
			utf8_string::size_type p = h.find(utf8_string(":"));

			if (p != utf8_string::npos)
			{
				ps.port = as<int>(h.substr(p + 1));
				h = h.substr(0, p);
			}

			ps.hostname = h.std_string();
			ps.username = opt("bt.network.proxy_username").std_string();
			ps.password = opt("bt.network.proxy_password").std_string();

			session_.set_proxy(ps);
		}
	}

	// giving back new settings
	session_.set_settings(ss);
	session_.set_proxy(ps);

	// port
	boost::system::error_code ec;
	session_.listen_on(parse_ports(opt("bt.network.listen")), ec);

	if (ec)
		log_error() << "listen failed, " << ec << "\n";
	else
		log() << "listening on port " << session_.listen_port() << "\n";

	// dht
	bt::dht_settings ds;
	ds.max_peers_reply = opt<int>("bt.dht.max_peers_reply");
	ds.search_branching = opt<int>("bt.dht.search_branching");
	ds.max_fail_count = opt<int>("bt.dht.max_fail_count");
	ds.max_torrents = opt<int>("bt.dht.max_torrents");
	ds.restrict_routing_ips = opt<bool>("bt.dht.restrict_routing_ips");
	ds.restrict_search_ips = opt<bool>("bt.dht.restrict_search_ips");

	session_.set_dht_settings(ds);

	// pe
	bt::pe_settings pe;
	pe.out_enc_policy =
		enc_policy_from_string(opt("bt.pe.out_enc_policy"));
	pe.in_enc_policy =
		enc_policy_from_string(opt("bt.pe.in_enc_policy"));
	pe.allowed_enc_level =
		enc_level_from_string(opt("bt.pe.allowed_enc_level"));
	pe.prefer_rc4 = opt<bool>("bt.pe.prefer_rc4");

	session_.set_pe_settings(pe);

	// asking all the torrents to reload their own settings (such as
	// the max connections per torrent)
	foreach(torrent_info t, torrents_)
		t.t->reload_settings();

	// dumping to stdout
	dump_settings();
}

// predicate used when sorting the list of settings per name
//
bool sort_pair_first(
	const std::pair<utf8_string, utf8_string>& p1,
	const std::pair<utf8_string, utf8_string>& p2)
{
	return (p1.first < p2.first);
}

void manager::dump_settings()
{
	// this dumps all the libtorrent session settings to stdout

	// current settings
	const bt::session_settings ss = session_.settings();

	// will be filled with name/value pairs, sorted at the end and
	// output
	std::vector<std::pair<utf8_string, utf8_string> > vs;

	// used by the various macros below to create the strings
	std::ostringstream oss;

	
	// this is me trying to be not too messy adding a crapload of
	// name/value pairs to a vector; the following macros makes the
	// list below a bit cleaner

	// adds a pair when the value is a string
#define o(n)							\
		oss.str(""); oss << ss.n;		\
		vs.push_back(std::make_pair(	\
			utf8_string(#n), utf8_string(oss.str())))

	// adds a pair when the value is an enum that has two possible
	// values (e1 and e2)
#define e2(n, e1, e2)											\
		oss.str(""); oss <<										\
			(ss.n == bt::session_settings::e1 ? #e1 :			\
			(ss.n == bt::session_settings::e2 ? #e2 : "?"));	\
		vs.push_back(std::make_pair(							\
			utf8_string(#n), utf8_string(oss.str())))

	// adds a pair when the value is an enum that has two possible
	// values (e1, e2 and e3)
#define e3(n, e1, e2, e3)										\
		oss.str(""); oss <<										\
			(ss.n == bt::session_settings::e1 ? #e1 :			\
			(ss.n == bt::session_settings::e2 ? #e2 :			\
			(ss.n == bt::session_settings::e3 ? #e3 : "?")));	\
			vs.push_back(std::make_pair(						\
				utf8_string(#n), utf8_string(oss.str())))

	// adds a pair when the value is an enum that has four possible
	// values (e1, e2, e3 and e4)
#define e4(n, e1, e2, e3, e4)										\
		oss.str(""); oss <<											\
			(ss.n == bt::session_settings::e1 ? #e1 :				\
			(ss.n == bt::session_settings::e2 ? #e2 :				\
			(ss.n == bt::session_settings::e3 ? #e3 :				\
			(ss.n == bt::session_settings::e4 ? #e4 : "?"))));		\
		vs.push_back(std::make_pair(								\
			utf8_string(#n), utf8_string(oss.str())))

	// adds a pair when the value is a pair itself, separated by a
	// dash (currently only used for the listen port)
#define p(n)														\
		oss.str(""); oss << ss.n.first << "-" << ss.n.second;		\
		vs.push_back(std::make_pair(								\
			utf8_string(#n), utf8_string(oss.str())))

	// adds a pair when the value is an hex string
#define h(n)														\
		oss.str(""); oss << bt::to_hex(std::string(1, ss.n));		\
		vs.push_back(std::make_pair(								\
			utf8_string(#n), utf8_string(oss.str())))

	// adds all the values to the vector
	o(version); o(user_agent); o(tracker_completion_timeout);
	o(tracker_receive_timeout); o(stop_tracker_timeout);
	o(tracker_maximum_response_length); o(piece_timeout);
	o(request_queue_time); o(max_allowed_in_request_queue);
	o(max_out_request_queue); o(whole_pieces_threshold);
	o(peer_timeout); o(urlseed_timeout);
	o(urlseed_pipeline_size); o(file_pool_size);
	o(allow_multiple_connections_per_ip); o(max_failcount);
	o(min_reconnect_time); o(peer_connect_timeout);
	o(ignore_limits_on_local_network); o(connection_speed);
	o(send_redundant_have); o(lazy_bitfields);
	o(inactivity_timeout); o(unchoke_interval);
	o(optimistic_unchoke_interval); o(announce_ip);
	o(num_want); o(initial_picker_threshold);
	o(allowed_fast_set_size);
	e2(suggest_mode, no_piece_suggestions, suggest_read_cache);
	o(max_queued_disk_bytes); o(handshake_timeout);
	o(use_dht_as_fallback); o(free_torrent_hashes);
	o(upnp_ignore_nonrouters); o(send_buffer_watermark);
	o(send_buffer_watermark_factor);
	e4(choking_algorithm,
		fixed_slots_choker, auto_expand_choker,
		rate_based_choker, bittyrant_choker);
	e3(seed_choking_algorithm,
			round_robin, fastest_upload, anti_leech);
	o(use_parole_mode); o(cache_size);
	o(cache_buffer_chunk_size); o(cache_expiry);
	o(use_read_cache); o(explicit_read_cache);
	o(explicit_cache_interval);
	e3(disk_io_write_mode,
		enable_os_cache, disable_os_cache_for_aligned_files,
		disable_os_cache);
	e3(disk_io_read_mode,
		enable_os_cache, disable_os_cache_for_aligned_files,
		disable_os_cache);
	p(outgoing_ports); h(peer_tos); o(active_downloads);
	o(active_seeds); o(active_dht_limit);
	o(active_tracker_limit); o(active_limit);
	o(auto_manage_prefer_seeds); o(dont_count_slow_torrents);
	o(auto_manage_interval); o(share_ratio_limit);
	o(seed_time_ratio_limit); o(seed_time_limit);
	o(peer_turnover_interval); o(peer_turnover);
	o(peer_turnover_cutoff); o(close_redundant_connections);
	o(auto_scrape_interval); o(auto_scrape_min_interval);
	o(max_peerlist_size); o(min_announce_interval);
	o(prioritize_partial_pieces); o(auto_manage_startup);
	o(rate_limit_ip_overhead); o(announce_to_all_trackers);
	o(announce_to_all_tiers); o(prefer_udp_trackers);
	o(strict_super_seeding); o(seeding_piece_quota);
	o(max_sparse_regions); o(lock_disk_cache);
	o(max_rejects); o(recv_socket_buffer_size);
	o(send_socket_buffer_size); o(optimize_hashing_for_speed);
	o(file_checks_delay_per_block);
	e3(disk_cache_algorithm,
		lru, largest_contiguous, avoid_readback);
	o(read_cache_line_size); o(write_cache_line_size);
	o(optimistic_disk_retry); o(disable_hash_checks);
	o(max_suggest_pieces); o(drop_skipped_requests);
	o(low_prio_disk); o(local_service_announce_interval);
	o(dht_announce_interval); o(udp_tracker_token_expiry);
	o(volatile_read_cache); o(guided_read_cache);
	o(default_cache_min_age); o(num_optimistic_unchoke_slots);
	o(no_atime_storage); o(default_est_reciprocation_rate);
	o(increase_est_reciprocation_rate);
	o(decrease_est_reciprocation_rate);
	o(incoming_starts_queued_torrents);
	o(report_true_downloaded); o(strict_end_game_mode);
	o(broadcast_lsd); o(enable_outgoing_utp);
	o(enable_incoming_utp); o(enable_outgoing_tcp);
	o(enable_incoming_tcp); o(max_pex_peers);
	o(ignore_resume_timestamps);
	o(no_recheck_incomplete_resume); o(anonymous_mode);
	o(tick_interval); o(share_mode_target);
	o(upload_rate_limit); o(download_rate_limit);
	o(local_upload_rate_limit); o(local_download_rate_limit);
	o(dht_upload_rate_limit); o(unchoke_slots_limit);
	o(half_open_limit); o(connections_limit);
	o(utp_target_delay); o(utp_gain_factor);
	o(utp_min_timeout); o(utp_syn_resends);
	o(utp_num_resends); o(utp_connect_timeout);
	o(utp_delayed_ack); o(utp_dynamic_sock_buf);
	o(utp_loss_multiplier);
	e2(mixed_mode_algorithm,
		prefer_tcp, peer_proportional);
	o(rate_limit_utp); o(listen_queue_size);
	o(announce_double_nat); o(torrent_connect_boost);
	o(seeding_outgoing_connections);
	o(no_connect_privileged_ports); o(alert_queue_size);
	o(max_metadata_size); o(smooth_connects);
	o(always_send_user_agent); o(apply_ip_filter_to_trackers);
	o(read_job_every); o(use_disk_read_ahead); o(lock_files);
	o(ssl_listen); o(tracker_backoff); o(ban_web_seeds);

#undef h
#undef p
#undef e4
#undef e3
#undef e2
#undef o

	// sorting alphabetically by name
	std::sort(vs.begin(), vs.end(), sort_pair_first);

	log_vv() << "dumping session settings\n";
	std::pair<utf8_string, utf8_string> p;
	foreach (p, vs)
		log_vv() << p.first << ": " << p.second << "\n";
	log_vv() << "\n";
}

const bt::session& manager::session() const
{
	return session_;
}

void manager::power(power_actions p)
{
	if (p == no_power_action)
	{
		if (power_ != no_power_action)
		{
			power_ = no_power_action;
			log() << "manager: disabling power action\n";
		}
	}

	if (!downloading())
	{
		log()
			<< "manager: disabling power action because not "
			<< "downloading\n";

		power_ = no_power_action;
		return;
	}
	
	log() << "manager: power action now " << to_string(p) << "\n";
	power_ = p;

	fire(power_action_changed, power_);
}

power_actions manager::power() const
{
	return power_;
}

void manager::monitor_thread(utf8_path m)
{
	log_v()
		<< "starting monitor thread for " << m.as_utf8_string() << "\n";

	try
	{
		// when the monitor option changes, interrupt() is called on
		// this thread which is caught below and unlocks the mutex,
		// allowing the new thread to proceed
		boost::mutex::scoped_lock lock(monitor_.running_mutex);

		log_vv() << "monitor: running mutex locked\n"; 

		// here, there is only this thread that is monitoring

		{
			// clears the list of any previous files; bad timing may
			// make this ignore files from the previous directory, but
			// it's better than subsequently handling files that are
			// not where the user specified
			
			log_vv() << "monitor: clearing list\n";
			boost::mutex::scoped_lock lock(monitor_.mutex);
			monitor_.list.clear();
		}

		log_vv() << "monitor: main loop\n";
		for (;;)
		{
			check_monitored_directory(m);

			log_vv() << "monitor: sleeping\n";
			boost::this_thread::sleep(monitor_thread_delay);
		}
	}
	catch(boost::thread_interrupted&)
	{
	}

	log_v() << "monitor thread for " << m.as_utf8_string() << " done\n";
}

bool manager::check_file(const utf8_path& p)
{
	// skip anything other than a file
	if (!p.is_regular_file())
		return true;

	// bad suffix
	if (p.extension() != paths::torrent_file_suffix())
		return true;

	log() << "monitor: found " << p << "\n";

	// protects the list since it might be used by the main
	// thread at the same time in add_monitored()
	boost::mutex::scoped_lock lock(monitor_.mutex);

	// making sure the path is not added twice
	std::vector<utf8_path>::iterator itor = std::find(
		monitor_.list.begin(), monitor_.list.end(), p);

	if (itor == monitor_.list.end())
		monitor_.list.push_back(p);

	return true;
}

void manager::check_monitored_directory(const utf8_path& m)
{
	std::string error;

	try
	{
		log_vv() << "monitor: checking " << m << "\n";
		walk_directory(m, boost::bind(&manager::check_file, this, _1));
	}
	catch(path_exception& e)
	{
		std::ostringstream oss;

		oss
			<< "error while walking monitored directory "
			<< m.as_utf8_string() << ", " << e.what() << "\n";

		error = oss.str();
	}
	catch(std::exception& e)
	{
		std::ostringstream oss;

		oss
			<< "error while walking monitored directory "
			<< m.as_utf8_string() << ", " << e.what() << "\n";

		error = oss.str();
	}

	if (!error.empty())
	{
		bool log = false;
		boost::posix_time::ptime now =
			boost::posix_time::second_clock::universal_time();

		if (monitor_.next_error == boost::posix_time::min_date_time)
		{
			log = true;

			monitor_.next_error =
				now  + monitor_.next_error_delay;

			monitor_.next_error_delay = monitor_error_second_delay;
		}
		else  if (now >= monitor_.next_error)
		{
			log = true;
			monitor_.next_error = now += monitor_.next_error_delay;
			monitor_.next_error_delay *= 2;

			if (monitor_.next_error_delay > monitor_error_max_delay)
				monitor_.next_error_delay = monitor_error_max_delay;
		}

		if (log)
			log_error() << error;
	}
}

void manager::add_monitored()
{
	// protects the list since it might be used by the monitor thread
	// at the same time in check_monitored_directory()
	boost::mutex::scoped_lock lock(monitor_.mutex);

	// this handles the whole list while the mutex is locked because
	// these files need to be removed from disk before the thread can
	// do another check on the directory; this avoids re-adding the same
	// files again and getting duplicate warnings

	foreach (const utf8_path& p, monitor_.list)
	{
		log() << "auto-add: " << p.as_utf8_string() << "\n";

		try
		{
			add_torrent(p.as_utf8_string());

			log() << "auto-add: removing " << p.as_utf8_string() << "\n";
			p.remove();
		}
		catch(std::exception& e)
		{
			log_error() << "couldn't auto-add, " << e.what() << "\n";
		}
	}

	if (!monitor_.list.empty())
		log_v() << "auto-add: done\n";

	// clearing the list: all the valid files were added to the manager
	// and deleted from disk; this will re-add any files that were
	// invalid, but not much can be done except deleting those also
	// (which may not be a good idea)
	monitor_.list.clear();
}

void manager::monitor(const utf8_path& p)
{
	// this will force the current thread to exit as quickly as possible
	// and unlock the running mutex so the new thread can start
	monitor_.thread.interrupt();

	// if p is empty, no monitor thread is started
	if (!p.empty())
	{
		monitor_.thread = boost::thread(boost::bind(
			&manager::monitor_thread, this, p));
	}
}

void manager::add_torrent(const utf8_string& s)
{
	if (!started_)
	{
		to_add_.push_back(s);
		return;
	}

	// here, 's' may be a file or a url, which is handled by
	// torrent::add()
	torrent_ptr t = torrent::add(session_, s);

	add_torrent_impl(t);
}

void manager::add_torrent_impl(torrent_ptr t)
{
	// new torrents are not added to the torrents_ list right away
	// because they need to be distinct from old torrents until they
	// are added correctly (if not duplicated)
	//
	// when the new torrent has been acknowledged by libtorrent, it 
	// will be moved from new_torrents_ to torrents_ and the xml queue
	// file will be saved
	new_torrents_.push_back(torrent_info(t));

	++progress_total_;

	// firing the signal
	fire(adding, t);
}

void manager::remove_torrent(torrent_ptr t, bool keep_files)
{
	// looking for the torrent in the list
	foreach (torrent_info ti, torrents_)
	{
		if (ti.t == t)
		{
			// asking the torrent to remove itself from libtorrent
			t->remove(keep_files);

			return;
		}
	}

	// this shouldn't happen because the given torrent has to be in
	// the torrents_ list since this is the only thing users have access
	// to (new_torrents_ is not exposed)
	bt_assert(false);
}

void manager::resume_torrents()
{
	log_v() << "resuming\n";

	const utf8_path queue = paths::conf() / paths::queue_filename();
	log() << "using queue file " << queue.as_utf8_string() << "\n";

	try
	{
		if (!queue.exists())
			throw path_exception(utf8_string("file doesn't exist"));

		// loading the xml file
		const xml::dom dom = xml::dom::load(
			paths::conf() / paths::queue_filename());
		const xml::element& root = dom.root();
		
		load_queue(root);
		return;
	}
	catch(path_exception& e)
	{
		log_error() << "can't access queue file, " << e.what() << "\n";
	}
	catch(xml::bad_xml& e)
	{
		log_error() << "bad xml: " << e.what() << "\n";

		// if the queue file is badly formatted, it is removed from
		// disk because there's not much else that can be done with it,
		// except reading it and failing again
		boost::system::error_code ec;
		queue.remove(ec);
	}

	fire(queue_progress, 0, 0);
}

void manager::load_queue(const xml::element& root)
{
	// the torrents are separated into two nodes so the 'seeding'
	// parameter can be set correctly

	std::size_t total = 0;

	const xml::element* dl_ts = nullptr;
	const xml::element* ul_ts = nullptr;

	if (root.has(utf8_string("downloading")))
	{
		const xml::element& e = root[utf8_string("downloading")];
		
		if (e.has(utf8_string("torrents")))
		{
			dl_ts = &e[utf8_string("torrents")];
			total += dl_ts->size();
		}
	}

	if (root.has(utf8_string("uploading")))
	{
		const xml::element& e = root[utf8_string("uploading")];
		
		if (e.has(utf8_string("torrents")))
		{
			ul_ts = &e[utf8_string("torrents")];
			total += ul_ts->size();
		}
	}
	
	if (total == 0)
	{
		fire(queue_progress, 0, 0);
		return;
	}

	std::size_t current = 0;

	if (dl_ts)
		load_torrents(*dl_ts, false, current, total);

	if (ul_ts)
		load_torrents(*ul_ts, true, current, total);
}

void manager::load_torrents(
	const xml::element& root, bool seeding,
	std::size_t& current, std::size_t total)
{
	try
	{
		// for each torrent node in the root
		foreach (const xml::element& e, root)
		{
			resume_info ri;

			// from the xml node
			ri.name = e.safe_value(utf8_string("name"));
			ri.hash = e.safe_value(utf8_string("hash"));
			ri.url = e.safe_value(utf8_string("url"));
			ri.save_changed =
				e.safe_value<bool>(utf8_string("save_changed"));
			ri.save_path = e.safe_value(utf8_string("save_path"));
			ri.torrent_file_save_path =
				e.safe_value(utf8_string("torrent_file_save_path"));
			ri.paused = e.safe_value<bool>(utf8_string("paused"));

			// whether this torrent comes from the download or uploading
			// node
			ri.seeding = seeding;


			torrent_ptr t = make_torrent(ri);
			if (t)
				add_torrent_impl(t);

			++current;
			fire(queue_progress, current, total);
		}
	}
	catch(xml::bad_element&)
	{
		// ignore
		log_error() << "bad torrent xml\n";
	}
}

torrent_ptr manager::make_torrent(resume_info ri)
{
	// this is resuming a torrent

	if (ri.hash.empty() && ri.url.empty())
	{
		if (ri.name.empty())
		{
			log_error()
				<< "unnamed torrent with no hash or url in the queue, "
				<< "ignoring\n";
		}
		else
		{
			log_error()
				<< "torrent '" << ri.name << "' has no hash or url in "
				<< "the queue, ignoring\n";
		}

		return torrent_ptr();
	}
	
	// logging
	{
		std::ostringstream oss;

		if (ri.hash.empty())
			oss << ri.url;
		else
			oss << ri.hash;
	
		oss << ": ";
	
		if (ri.name.empty())
			oss << "(no name)";
		else
			oss << ri.name;
	
		log() << oss.str() << "\n";
	}	
	
	if (!ri.hash.empty())
	{
		ri.torrent_file =
			paths::conf() / (ri.hash + paths::torrent_file_suffix());

		ri.resume_file =
			paths::conf() / (ri.hash + paths::resume_file_suffix());

		log()
			<< "loading torrent file "
			<< "'" << ri.torrent_file.as_utf8_string() << "' "
			<< "with resume file "
			<< "'" << ri.resume_file.as_utf8_string() << "'\n";

		if (!verify_torrent_file(ri.torrent_file, ri.resume_file))
			return torrent_ptr();

		if (!verify_resume_file(ri.resume_file))
		{
			// ignore, will recheck
			ri.resume_file = utf8_path();
		}
	}
	else
	{
		log() << "loading url '" << ri.url << "'\n";
	}

	return torrent::resume(session_, ri);
}

bool manager::verify_torrent_file(
	const utf8_path& t, const utf8_path& f)
{
	try
	{
		if (!t.exists())
		{
			log_error() << "torrent file doesn't exist\n";

			try
			{
				// delete fast resume if the torrent file is gone
				if (f.exists())
					f.remove();
			}
			catch(path_exception& e)
			{
				// ignore
				log_error()
					<< "can't remove resume file " << f.as_utf8_string()
					<< ", " << e.what() << "\n";
			}

			// the torrent file doesn't exist and needs to be skipped
			return false;
		}

		// the torrent file exists and all is fine
		return true;
	}
	catch(path_exception& e)
	{
		// the torrent file couldn't be accessed and needs to be skipped
		log_error()
			<< "can't access torrent file " << t.as_utf8_string() << ", "
			<< e.what() << "\n";

		return false;
	}
}

bool manager::verify_resume_file(const utf8_path& p)
{
	try
	{
		if (!p.exists())
		{
			log_error()
				<< "resume file " << p.as_utf8_string() << " doesn't exist\n";
			return false;
		}

		return true;
	}
	catch(path_exception& e)
	{
		log_error() << "can't load resume file, " << e.what() << "\n";
		return false;
	}
}

void manager::start_all()
{
	foreach (torrent_info& t, torrents_)
		t.t->start();
}

bool manager::on_save_timer()
{
	save_resume();
	return true;
}

void manager::save_resume()
{
	if (!resume_saved())
	{
		log_error()
			<< "manager: trying to save resume data but "
			<< "some haven't finished yet\n";

		return;
	}

	log_vv() << "manager: saving resume\n";
	foreach (torrent_info& t, torrents_)
		t.t->save_resume();
}

bool manager::resume_saved() const
{
	foreach (const torrent_info& t, torrents_)
	{
		if (!t.t->resume_saved())
			return false;
	}

	return true;
}

void manager::quit()
{
	if (quitting_)
		return;

	log_v() << "quitting manager\n";
	quitting_ = true;
	
	session_.pause();
	save_resume();

	log_v() << "waiting for torrents\n";
}

bool manager::done() const
{
	bt_assert(quitting_);

	if (!resume_saved() || !new_torrents_.empty())
	{
		// waiting for all torrents to be acknowledged and their resume
		// data saved
		return false;
	}

	log_v() << "manager is done\n";
	return true;
}

void manager::save_session()
{
	try
	{
		const utf8_path out = paths::conf() / paths::session_filename();

		log() << "saving session in " << out.as_utf8_string() << "\n";

		// making session data
		bt::entry ss;
		session_.save_state(ss);
		
		std::vector<char> data;
		bt::bencode(std::back_inserter(data), ss);

		// saving
		std::ofstream of(out.native().c_str(), std::ios::binary);
		of.exceptions(std::ios::badbit | std::ios::failbit);
		of.write(data.data(), data.size());
	}
	catch(std::exception& e)
	{
		log_error() << "can't save session state, " << e.what() << "\n";
	}
}

void manager::load_session(const utf8_path& p)
{
	log() << "loading session from " << p.as_utf8_string() << "\n";

	// using the file size to create a large enough vector
	std::size_t s = static_cast<std::size_t>(p.file_size());

	if (s == 0)
	{
		log_error() << "session file exists but is empty; deleting\n";
		boost::system::error_code ec;

		p.remove(ec);

		if (ec)
		{
			log_error()
				<< "couldn't remove session file, "
				<< ec.message() << "\n";
		}

		return;
	}

	std::vector<char> rd(s);

	// reading everything in the vector
	std::ifstream out(p.as_utf8_string().c_str(), std::ios::binary);
	out.exceptions(std::ios::badbit | std::ios::failbit);
	out.read(&rd[0], s);

	bt::error_code ec;
	bt::lazy_entry e;

	// decoding
	if (bt::lazy_bdecode(&rd[0], &rd[0] + rd.size(), e, ec) == 0)
	{
		session_.load_state(e);
	}
	else
	{
		log_error()
			<< "couldn't load session state: " << ec.message() << "\n";
		
		// removing the bad state file
		p.remove(ec);

		if (ec)
		{
			// ignoring failure
			log_error()
				<< "couldn't removing the bad session sate file: "
				<< ec.message() << "\n";
		}
	}
}

bool manager::save_queue()
{	
	try
	{
		// don't save the queue if some of the torrents haven't been
		// acknowledged by libtorrent yet
		if (!new_torrents_.empty())
			return false;
		
		xml::dom dom = xml::dom::create(utf8_string("baretorrent"));

		// downloading node
		xml::element dl(utf8_string("torrents"));

		// uploading node
		xml::element ul(utf8_string("torrents"));

		foreach(const torrent_info& t, torrents_)
		{
			if (!t.t->added())
				continue;
			
			if (t.t->downloaded())
				ul.add(make_torrent_node(t.t));
			else
				dl.add(make_torrent_node(t.t));
		}

		xml::element e(utf8_string("downloading"));
		e.add(dl);
		dom.root().add(e);

		e = xml::element(utf8_string("uploading"));
		e.add(ul);
		dom.root().add(e);
	
		// saving
		const utf8_path p = paths::conf() / paths::queue_filename();
		log() << "saving queue file in " << p.as_utf8_string() << "\n";

		// saving
		dom.write(p);

		// asks all the torrents to save their state file (not the fast
		// resume), this is synchronous
		foreach(const torrent_info& t, torrents_)
		{
			// this happens when the torrent is being removed
			if (t.t->status() == torrent::removing_status)
				continue;

			t.t->save_state_file();
		}
	}
	catch(xml::bad_xml& e)
	{
		log_error() << "can't save queue file, " << e.what() << "\n";
	}
	catch(path_exception& e)
	{
		log_error() << "can't save queue file, " << e.what() << "\n";
	}

	return true;
}

xml::element manager::make_torrent_node(const torrent_ptr t)
{
	xml::element e(utf8_string("torrent"));
	t->queue_node(e);

	return e;
}

torrent_ptr manager::find_torrent(bt::torrent_handle h)
{
	foreach (torrent_info& t, torrents_)
	{
		if (t.t->is(h))
			return t.t;
	}

	return torrent_ptr();
}

int manager::download_rate() const
{
	int t = 0;
	foreach (const torrent_info& ti, torrents_)
	{
		const int r = ti.t->torrent_status().download_payload_rate;

		if (r > opt<int>("bt.torrent.stalled_rate"))
			t += r;
	}
		
	return t;
}

int manager::upload_rate() const
{
	int t = 0;
	foreach (const torrent_info& ti, torrents_)
	{
		const int r = ti.t->torrent_status().upload_payload_rate;
		
		if (r > opt<int>("bt.torrent.stalled_rate"))
			t += r;
	}		

	return t;
}

bt::size_type manager::total_remaining() const
{
	bt::size_type t = 0;

	foreach (const torrent_info& ti, torrents_)
		t += ti.t->remaining();

	return t;
}

bool manager::downloading() const
{
	foreach (const torrent_info& ti, torrents_)
	{
		if (!ti.t->downloaded())
			return true;
	}

	return false;
}

bool manager::remote_peer_connected() const
{
	foreach (const torrent_info& ti, torrents_)
	{
		if (ti.t->remote_peer_connected())
			return true;
	}

	return false;
}

bool manager::ready() const
{
	return (ready_ && started_);
}

void manager::post_update()
{
	// this will make the session post update alerts for all torrents
	session_.post_torrent_updates();
}

void manager::handle_torrent_added(bt::add_torrent_alert& ta)
{
	// the new torrent object was set as the userdata pointer when
	// added to the session
	torrent* t = reinterpret_cast<torrent*>(ta.params.userdata);
	
	// shouldn't happen
	bt_assert(t);
	if (!t)
		return;

	// libtorrent returns the same handle as a previous torrent if its
	// hash already existed
	
	// looking for that handle in the current list; note that the
	// torrent that was just added is still in the new_torrents_ list
	// and so won't be found by find_torrent()
	torrent_ptr dupe = find_torrent(ta.handle);
	
	// looking up the element in the new_torrents_ list that corresponds
	// to this new torrent
	torrent_infos::iterator itor;
	for (itor=new_torrents_.begin(); itor!=new_torrents_.end(); ++itor)
	{
		const torrent_info& ti = *itor;
		if (ti.t.get() == t)
			break;
	}
	
	// this shouldn't happen
	bt_assert(itor != new_torrents_.end());
	if (itor == new_torrents_.end())
	{
		log_error()
			<< "can't find torrent in the new torrent list\n";
		return;
	}

	if (dupe)
	{
		// if the handle is already in the current list, this is a
		// duplicate
		
		// logging
		utf8_string n = dupe->name();
		if (n.empty())
			n = dupe->hash();

		if (n.empty())
			n = dupe->url();
				
		log() << "duplicate torrent " << n << "\n";

		// removing from the list, the existing torrent object in 
		// torrents_ will still be used
		new_torrents_.erase(itor);

		// firing the signal
		fire(duplicate, dupe);
	}
	else
	{
		// the handle doesn't exist in the current list, so this is a
		// good new torrent
		torrent_info ti = *itor;

		// moving it from the temporary list to the current list
		torrents_.push_back(ti);
		new_torrents_.erase(itor);

		// notifying the torrent it was added
		ti.t->on_torrent_added(ta);

		// firing the signal
		fire(added, ti.t);

		// trying to process the alerts that (may) have been queued
		// for this torrent
		clear_alert_queue();

		// saving the xml queue file with this new torrent in it
		save_queue();
	}

	bt_assert(progress_total_ > 0);
	bt_assert(progress_current_ < progress_total_);

	++progress_current_;

	fire(add_progress, progress_current_, progress_total_);

	if (progress_current_ == progress_total_)
	{
		progress_current_ = 0;
		progress_total_ = 0;
	}
}

void manager::handle_torrent_removed(const bt::sha1_hash& hash)
{
	// at this point, the torrent handle is dead and so the right
	// element in the list must be looked up by hash

	log_v() << "manager: removing '" << to_utf8_string(hash) << "\n";

	// for each torrent
	torrent_infos::iterator itor;
	for (itor=torrents_.begin(); itor!=torrents_.end(); ++itor)
	{
		torrent_ptr t = itor->t;

		if (t->hash() == to_utf8_string(hash))
		{
			// don't shutdown when the user removes the torrent
			// manually
			if (torrents_.size() == 1 && power_ != no_power_action)
				power(no_power_action);

			torrents_.erase(itor);
			save_queue();

			// firing the signal
			fire(removed, t);

			return;
		}
	}

	// shouldn't happen
	bt_assert(false);
}

bool manager::handle_alert(bt::alert* a)
{
	// this is a large switch for all alert types; some of the alerts
	// are dispatched directly to the right torrent using the handle,
	// others are handled here instead
	//
	// some of the alerts are received in what could be considered the
	// wrong order: save_resume_data_failed_alert could be received
	// before add_torrent_alert, which is used for bookkeeping and is
	// required before processing anything else for that torrent
	//
	// if the torrent handle is not found in the torrent list, the alert
	// is queued and will be checked later
	
	switch (a->type())
	{
		// state update
		case bt::state_update_alert::alert_type:
		{
			bt::state_update_alert* ua =
				dynamic_cast<bt::state_update_alert*>(a);
			
			bt_assert(ua);
			if (!ua)
				break;

			// looking up each torrent in the update list
			foreach (const bt::torrent_status& ts, ua->status)
			{
				torrent_ptr t = find_torrent(ts.handle);
				if (!t)
				{
					// shouldn't happen
					// todo: er, looks like it does, after a torrent was
					// removed
					// bt_assert(false);
					break;
				}

				t->update(ts);
			}

			break;
		}

		// error
		case bt::torrent_error_alert::alert_type:
		{
			return dispatch<bt::torrent_error_alert>(
				a, &torrent::on_error);
		}

		// metadata
		case bt::metadata_received_alert::alert_type:
		{
			const bool b = dispatch<bt::torrent_alert>(
				a, &torrent::on_metadata_received);

			save_queue();

			return b;
		}

		// finished
		case bt::torrent_finished_alert::alert_type:
		{
			const bool b = dispatch<bt::torrent_alert>(
				a, &torrent::on_torrent_complete);
			
			save_queue();
			
			return b;
		}
		
		// resume saved
		case bt::save_resume_data_alert::alert_type:
		{
			bt::save_resume_data_alert* rd =
				dynamic_cast<bt::save_resume_data_alert*>(a);
			
			bt_assert(rd);
			if (!rd)
				break;

			torrent_ptr t = find_torrent(rd->handle);
			if (!t)
			{
				log_error()
					<< "received save resume alert for "
					<< rd->handle.name() << " but that torrent is not "
					<< "in the list\n";

				return false;
			}

			t->on_resume_saved(rd->resume_data);
			break;
		}

		// resume save failed
		case bt::save_resume_data_failed_alert::alert_type:
		{
			bt::save_resume_data_failed_alert* rd =
				dynamic_cast<bt::save_resume_data_failed_alert*>(a);
			bt_assert(rd);
				
			torrent_ptr t = find_torrent(rd->handle);
			if (!t)
			{
				log_error()
					<< "received save resume failed alert for "
					<< rd->handle.name() << " but that torrent is not "
					<< "in the list\n";

				return false;
			}

			log_error()
				<< "couldn't save resume data: "
				<< rd->error << "\n";

			// the torrent is still expecting this to be called even
			// though it failed
			t->on_resume_saved(boost::shared_ptr<bt::entry>());
			break;
		}

		// fastresume failed
		case bt::fastresume_rejected_alert::alert_type:
		{
			return dispatch<bt::fastresume_rejected_alert>(
				a, &torrent::on_resume_failed);
		}

		// torrent added
		case bt::add_torrent_alert::alert_type:
		{
			bt::add_torrent_alert* ta =
				dynamic_cast<bt::add_torrent_alert*>(a);
			bt_assert(ta);

			handle_torrent_added(*ta);
			break;
		}

		// torrent removed
		case bt::torrent_removed_alert::alert_type:
		{
			bt::torrent_removed_alert* ra =
				dynamic_cast<bt::torrent_removed_alert*>(a);
			bt_assert(ra);

			handle_torrent_removed(ra->info_hash);
			break;
		}

		// file completed
		case bt::file_completed_alert::alert_type:
		{
			return dispatch<bt::file_completed_alert>(
				a, &torrent::on_file_completed);
		}

		// storage moved
		case bt::storage_moved_alert::alert_type:
		{
			const bool b = dispatch<bt::storage_moved_alert>(
				a, &torrent::on_storage_moved);

			save_queue();

			return b;
		}

		// storage failed
		case bt::storage_moved_failed_alert::alert_type:
		{
			const bool b = dispatch<bt::storage_moved_failed_alert>(
				a, &torrent::on_storage_moved_failed);

			save_queue();

			return b;
		}

		// anything else
		default:
		{
			//std::cout << "other alert " << a->what() << "\n";
			break;
		}
	}

	// this event has been processed (or ignored) correctly; the
	// various dispatch() calls will have returned false if the event
	// has to be queued
	return true;
}

void manager::clear_alert_queue()
{
	if (alert_queue_.empty())
		return;

	log_vv()
		<< "clearing queue with " << alert_queue_.size() << " items\n";

	// unprocessed events
	std::vector<bt::alert*> still;
	
	foreach (bt::alert* a, alert_queue_)
	{
		log_vv() << "trying " << a->message() << "\n";
		if (!handle_alert(a))
		{
			log_vv() << "still can't handle it\n";
			still.push_back(a);
		}
		else
		{
			log_vv() << "handled\n";
			delete a;
		}
	}

	alert_queue_ = still;
}

void manager::queue_alert(bt::alert* a)
{
	// some arbitrary limit
	if (alert_queue_.size() > max_queued_alarms)
	{
		delete a;
		return;
	}

	alert_queue_.push_back(a);
	log_error()
		<< "couldn't handle " << a->message() << ", queueing\n";
}

bool manager::tick()
{
	if (ready_)
	{
		if (!started_)
		{
			start_all();

			save_timer_.start(boost::bind(
				&manager::on_save_timer, this), save_timer_delay);

			started_ = true;

			foreach (const utf8_string& s, to_add_)
				add_torrent(s);

			to_add_.clear();
		}
	}
	else
	{
		return true;
	}
	

	// this is called regularly by the application timer to process
	// events from libtorrent

	// adds all the files found by the monitor (if any)
	add_monitored();

	check_ipc();
	check_power();

	std::auto_ptr<bt::alert> a = session_.pop_alert();
	if (!a.get())
		return false;
	
	if (!handle_alert(a.get()))
		queue_alert(a.release());

	return true;
}

void manager::check_power()
{
	switch (power_)
	{
		case shutdown_power:	// fall-through
		case suspend_power:
		case hibernate_power:
		{
			if (!downloading())
			{
				power_actions p = power_;
				power_ = no_power_action;

				log()
					<< "manager: finished downloading, current action "
					<< "is " << to_string(p) << "\n";

				fire(power_action, p);
			}

			break;
		}

		case no_power_action:
		default:
		{
			break;
		}
	}
}

torrents::size_type manager::size() const
{
	return torrents_.size();
}

bool manager::empty() const
{
	return torrents_.empty();
}

bool manager::has_active() const
{
	foreach (const torrent_info& ti, torrents_)
	{
		if (ti.t->active())
			return true;
	}

	return false;
}

torrent_ptr manager::get(std::size_t i) const
{
	bt_assert(i < torrents_.size());
	return torrents_[i].t;
}

torrent_ptr manager::get(std::size_t i)
{
	bt_assert(i < torrents_.size());
	return torrents_[i].t;
}


manager::monitor_info::monitor_info() :
	next_error(boost::posix_time::min_date_time),
	next_error_delay(monitor_error_first_delay)
{
}

}	// namespace baretorrent
