// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_MANAGER_INCLUDED
#define BARETORRENT_MANAGER_INCLUDED

#include "utility.h"
#include "xml.h"
#include "torrent.h"

namespace baretorrent
{

struct resume_info;

// wraps the session, contains the list of torrents and has some
// high-level convenience functions such as getting the total download/
// upload rate
//
class manager : boost::noncopyable
{
public:
	typedef sig::signal<void (power_actions)> power_handler;
	typedef sig::signal<void (torrent_ptr)> torrent_handler;
	typedef sig::signal<void (std::size_t, std::size_t)>
		progress_handler;

	// fired when a torrent has been added to libtorrent but hasn't been
	// acknowledged yet
	torrent_handler adding;

	// fired when a torrent is a duplicate
	torrent_handler duplicate;

	// fired when the torrent was added to libtorrent
	torrent_handler added;

	// fired when a torrent was removed
	torrent_handler removed;

	// fired when a torrent is being moved
	torrent_handler moving;

	// fired when a torrent failed to move
	torrent_handler move_failed;

	power_handler power_action_changed;
	power_handler power_action;
	
	progress_handler queue_progress;
	progress_handler add_progress;


	enum connection_status
	{
		checking_connection = 1,
		working_connection,
		local_connection_only
	};

	
	manager();
	~manager();

	// initializes the session
	//
	void start();

	void power(power_actions p);
	power_actions power() const;

	// stops the manager; this is called when the application is
	// closing, it doesn't close the application itself
	//
	void quit();

	// after calling quit(), returns whether the manager has done its
	// cleanup, such as saving the resume data
	//
	bool done() const;

	// called from a timer in the application; processes events from
	// libtorrent
	//
	bool tick();

	// returns true when all torrents have been added and acknowledged
	// by libtorrent; this is used on startup
	//
	bool ready() const;


	// number of torrents
	//
	torrents::size_type size() const;

	// whether there are any torrents in the manager
	//
	bool empty() const;


	// returns the torrent having the given index; this assumes the
	// index is valid
	//
	torrent_ptr get(std::size_t i) const;

	// returns the torrent having the given index; this assumes the
	// index is valid
	//
	torrent_ptr get(std::size_t i);


	// called from the main window in a timer; this will ask all the
	// torrents to post updates
	//
	void post_update();

	// starts saving the resume data for all torrents
	//
	void save_resume();

	
	// adds the given torrent; 's' may be a URL or a file
	//
	void add_torrent(const utf8_string& s);

	// removes the given torrent; if 'keep_files' is false, the files
	// from the torrent will also be removed
	//
	void remove_torrent(torrent_ptr t, bool keep_files);


	// returns the libtorrent session
	//
	const bt::session& session() const;


	// returns the current global download rate
	//
	int download_rate() const;

	// returns the current global upload rate
	//
	int upload_rate() const;

	// total number of bytes remaining to download
	//
	bt::size_type total_remaining() const;

	// returns whether there is any torrent that is not completed
	//
	bool downloading() const;


	// returns the connection status
	//
	connection_status connection() const;

	// returns whether any peer connection was opened remotely, used
	// to check whether port forwarding is fine
	//
	bool remote_peer_connected() const;

	// returns whether at least one torrent is trying to download or
	// upload something
	//
	bool has_active() const;

	// reads all the options again; used on startup and when the
	// options change (such as from the preferences dialog)
	//
	void reload_settings();

	// outputs all the settings on stdout
	//
	void dump_settings();

	
	// starts a thread that will monitor the given path for new
	// torrent files and delete them when handled; if a path is already
	// monitored, this one takes over
	//
	void monitor(const utf8_path& p);
	
private:
	// information about a single torrent, kept in a list below
	//
	struct torrent_info
	{
		// the torrent object
		torrent_ptr t;

		torrent_info(torrent_ptr t)
			: t(t)
		{
		}
	};

	// information about monitoring a directory; this is filled from
	// the monitoring thread and used from tick() (which should be
	// called from the main thread)
	//
	struct monitor_info
	{
		// thread handle
		boost::thread thread;

		// list of files to add
		std::vector<utf8_path> list;

		// protects the list
		boost::mutex mutex;
		
		// makes sure the previous monitor thread exists before a new
		// one starts up; this is locked in the thread function and
		// unlocked when the function exits
		boost::mutex running_mutex;

		// next time an error is logged; this prevents spamming when
		// the directory is unavailable
		boost::posix_time::ptime next_error;

		// delay of the next error
		boost::posix_time::seconds next_error_delay;
		
		monitor_info();
	};

	typedef std::vector<torrent_info> torrent_infos;


	// whether start() has been called
	bool started_;

	// libtorrent session
	bt::session session_;

	// list of torrents added before the manager was started; this
	// happens when adding from the command line
	std::vector<utf8_string> to_add_;

	// list of torrents
	torrent_infos torrents_;

	// list of torrents that are in the process of being added; this is
	// used to make sure all torrents are added correctly before exiting
	torrent_infos new_torrents_;

	// whether quit() has been called
	bool quitting_;

	// whether all the torrents have been added successfully after
	// startup
	mutable bool ready_;

	// information about the monitor thread
	monitor_info monitor_;

	// current connection status
	connection_status con_;

	// list of alerts that need to be processed; this is needed because
	// some of the events may be out of order, such as a
	// state_update_alert before an add_torrent_alert
	std::vector<bt::alert*> alert_queue_;

	// when fired, saves resume data for all torrents
	timer save_timer_;

	power_actions power_;
	
	std::size_t progress_total_;
	std::size_t progress_current_;


	// returns the torrent having the given libtorrent handle
	//
	torrent_ptr find_torrent(bt::torrent_handle h);

	// called from handle_alert(), uses the handle in the alert to find
	// the right torrent and call the given function on it; this is
	// merely for convenience
	//
	template <class T>
	bool dispatch(bt::alert* a, void (torrent::*f)(T&))
	{
		// casting to the right alert type
		T* aa = dynamic_cast<T*>(a);
		bt_assert(aa);

		// looking for the torrent; this may fail if add_torrent_alert
		// hasn't been received yet, in which case handle_alert() will
		// queue this alert for later
		torrent_ptr t = find_torrent(aa->handle);
		if (!t)
			return false;

		// this happens just after removing the torrent when libtorrent
		// sends more alerts but the handle has become invalid
		if (t->status() == torrent::removing_status)
			return true;

		(t.get()->*f)(*aa);

		return true;
	}

	// loads the libtorrent session file and initializes the session
	// object
	//
	void init_session();

	// starts all the torrents that can be started (such as those that
	// are not errored)
	//
	void start_all();

	// fired from save_timer_, saves resume data for all torrents
	//
	bool on_save_timer();

	// returns whether all torrents have saved their resume data
	// correctly; saving the resume file is asynchronous
	//
	bool resume_saved() const;


	// loads the xml queue file and resumes all the files mentioned in
	// it
	//
	void resume_torrents();

	// called from resume_torrent(), handles the 'downloading' and
	// 'uploading' nodes (which are identical but discrete nodes)
	//
	void load_queue(const xml::element& root);

	// called from load_queue() for the two nodes; loops through all
	// the torrent nodes and resumes them
	//
	void load_torrents(
		const xml::element& root, bool seeding,
		std::size_t& current, std::size_t total);

	
	// loops through all the queued alerts (if any) and tries to
	// dispatch them; those that fail will stay in the queue
	//
	void clear_alert_queue();

	// queues the given alert
	//
	void queue_alert(bt::alert* a);

	// handles the given alert; this may either do some internal stuff
	// in the manager or dispatch the alert to the right torrent (or
	// both)
	//
	bool handle_alert(bt::alert* a);


	// thread function for the monitor; loops through all the files in
	// the directory and looks for the right suffix on filenames
	//
	void monitor_thread(utf8_path m);

	// called from monitor_thread(), does the actual looping
	//
	void check_monitored_directory(const utf8_path& m);

	// checks a single file from the monitor thread
	//
	bool check_file(const utf8_path& p);


	// called from tick(), loops through all the files that were found
	// by monitor_thread() (if any), adds them to the manager and then
	// deletes them from disk
	//
	void add_monitored();


	// creates a torrent object based on the given resume data
	//
	torrent_ptr make_torrent(resume_info ri);

	// called from add_torrent() (which is public and is called when
	// the user is adding a torrent) and load_torrents() (which is
	// called when the xml queue file is being read); this adds the
	// given torrent to the new_torrents_ list and fires signals
	//
	void add_torrent_impl(torrent_ptr t);

	// called from make_torrent() when resuming a torrent; this makes
	// sure the given torrent file exists and deletes the fast-resume
	// associated with it (if any)
	//
	bool verify_torrent_file(const utf8_path& t, const utf8_path& f);

	// called from make_torrent() when resuming a torrent; this makes
	// sure the given resume file exists
	//
	bool verify_resume_file(const utf8_path& p);


	// creates an xml node for the given torrent; this is used when
	// saving the queue information and contains the hash, name,
	// filename, etc.
	//
	xml::element make_torrent_node(const torrent_ptr t);
	

	// saves the xml queue file; this is called from several places,
	// such as when torrents are removed or added
	//
	// this might fail if one or more newly added torrents have not
	// been acknowledged by libtorrent yet (because some information is
	// missing, such as whether the torrent is seeding or not)
	//
	bool save_queue();

	// saves the libtorrent session state
	//
	void save_session();

	// loads the libtorrent session state from the given file
	//
	void load_session(const utf8_path& p);

	// called from handle_alert() for torrent_removed_alert; this
	// cleans up some stuff (like signal slots), removes it from the
	// internal list and saves the queue
	//
	void handle_torrent_removed(const bt::sha1_hash& hash);

	// called from handle_alert() for add_torrent_alert; this will
	// check for duplicate torrents and move the new torrent to the
	// torrent list if everything is fine
	//
	void handle_torrent_added(bt::add_torrent_alert& ta);

	void check_power();
	void start_fun();
};

}	// namespace baretorrent

#endif	// BARETORRENT_MANAGER_INCLUDED
