// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "options.h"
#include "utility.h"
#include "xml.h"
#include "platform.h"
#include "ext/ext.h"

namespace baretorrent
{

// filename of the sqlite database
const utf8_string sqlite_filename("baretorrent.sqlite");

static db::database g_database;

// used in get_database() as a sanity check to make sure it is not
// used outside of the main thread
static boost::thread::id g_main_thread_id;

struct updated_opt
{
	std::string guid;
	std::string name;
	utf8_string value;

	updated_opt(
		const std::string& guid, const std::string& name,
		const utf8_string& value)
			: guid(guid), name(name), value(value)
	{
	}

	updated_opt(const std::string& name, const utf8_string& value)
		: name(name), value(value)
	{
	}
};

static option_list g_map;
static std::vector<updated_opt> g_update;
static boost::recursive_mutex g_mutex;

typedef std::map<std::string, option_list> ext_map;
static ext_map g_ext_map;

void ext_load_options();


void check_options()
{
	// todo: check all the options to make sure they have decent values

	if (!opt("bt.ui.language").empty())
	{
		try
		{
			// making sure the language file exists

			const utf8_string fn = 
				opt("bt.ui.language") + paths::locale_suffix();

			const utf8_path language = absolute(paths::locales()) / fn;

			if (!language.exists())
			{
				log_error()
					<< "language file " << language
					<< " doesn't exist\n";

				opt("bt.ui.language", utf8_string());
			}
		}
		catch(path_exception& e)
		{
			log_error()
				<< "can't check language file "
				<< opt("bt.ui.language") << ", " << e.what() << "\n";

			opt("bt.ui.language", utf8_string());
		}
	}
}

template <class T>
void make_bool(option_list& map, const std::string& name, T v);

void make_bool(option_list& map, const std::string& name, bool v)
{
	const utf8_string s = (v ? utf8_string("1") : utf8_string("0"));
	
	map[name] = option(
		name, s, s,
		boost::shared_ptr<option_type>(new value_option_type<bool>));
}

template <class T>
void make_unsigned_int(
	option_list& map, const std::string& name, T v,
	T min = 0, T max = std::numeric_limits<T>::max());

void make_unsigned_int(
	option_list& map, const std::string& name, int v,
	int min = 0, int max = std::numeric_limits<int>::max())
{
	clamp(v, min, max);
	const utf8_string s = as_utf8_string(v);

	map[name] = option(
		name, s, s, boost::shared_ptr<option_type>(
			new number_option_type<int>(min, max)));
}

template <class T>
void make_int(
	option_list& map, const std::string& name, T v,
	T min = std::numeric_limits<T>::min(),
	T max = std::numeric_limits<T>::max());

void make_int(
	option_list& map, const std::string& name, int v,
	int min = std::numeric_limits<int>::min(),
	int max = std::numeric_limits<int>::max())
{
	clamp(v, min, max);
	const utf8_string s = as_utf8_string(v);

	map[name] = option(
		name, s, s, boost::shared_ptr<option_type>(
			new number_option_type<int>(min, max)));
}

template <class T>
void make_float(
	option_list& map, const std::string& name, T v,
	T min = -std::numeric_limits<T>::max(),
	T max = std::numeric_limits<T>::max());

void make_float(
	option_list& map, const std::string& name, float v,
	float min = -std::numeric_limits<float>::max(),
	float max = std::numeric_limits<float>::max())
{
	clamp(v, min, max);
	const utf8_string s = as_utf8_string(v);

	map[name] = option(
		name, s, s, boost::shared_ptr<option_type>(
			new number_option_type<float>(min, max)));
}

void make_string(
	option_list& map, const std::string& name, const utf8_string& v)
{
	map[name] = option(name, v, v);
}

void make_string(
	option_list& map, const std::string& name,
	const std::string& v)
{
	make_string(map, name, utf8_string(v));
}

void make_rect(
	option_list& map, const std::string& name, const wxRect& r)
{
	map[name] = option(
		name, as_utf8_string(r), as_utf8_string(r),
		boost::shared_ptr<option_type>(
			new value_option_type<wxRect>));
}

void make_enum(
	option_list& map, const std::string& name, const utf8_string& v,
	const std::vector<std::string>& vs)
{
	map[name] = option(
		name, v, v, boost::shared_ptr<option_type>(
			new enum_option_type(vs)));
}

void make_enum(
	option_list& map, const std::string& name, const std::string& v,
	const std::vector<std::string>& vs)
{
	make_enum(map, name, utf8_string(v), vs);
}

option_list create_default_options()
{
	// defaults
	bt::session_settings ss;
	bt::torrent_handle h;
	bt::add_torrent_params atp;
	bt::pe_settings pe;
	bt::dht_settings ds;

	option_list map;

	std::vector<std::string> storage_mode_values;
	storage_mode_values.push_back("sparse");
	storage_mode_values.push_back("allocate");

	std::vector<std::string> proxy_type_values;
	proxy_type_values.push_back("");
	proxy_type_values.push_back("SOCKS4");
	proxy_type_values.push_back("SOCKS5");
	proxy_type_values.push_back("HTTP");

	std::vector<std::string> enc_policy_values;
	enc_policy_values.push_back("forced");
	enc_policy_values.push_back("enabled");
	enc_policy_values.push_back("disabled");
	
	std::vector<std::string> enc_level_values;
	enc_level_values.push_back("rc4");
	enc_level_values.push_back("plaintext");
	enc_level_values.push_back("both");

	std::vector<std::string> suggest_mode_values;
	suggest_mode_values.push_back("none");
	suggest_mode_values.push_back("cache");

	std::vector<std::string> choking_algorithm_values;
	choking_algorithm_values.push_back("fixed_slots");
	choking_algorithm_values.push_back("auto_expand");
	choking_algorithm_values.push_back("rate_based");
	choking_algorithm_values.push_back("bittyrant");

	std::vector<std::string> seed_choking_algorithm_values;
	seed_choking_algorithm_values.push_back("round_robin");
	seed_choking_algorithm_values.push_back("fastest_upload");
	seed_choking_algorithm_values.push_back("anti_leech");

	std::vector<std::string> io_buffer_mode_values;
	io_buffer_mode_values.push_back("enable");
	io_buffer_mode_values.push_back("disable_for_aligned_files");
	io_buffer_mode_values.push_back("disable");

	std::vector<std::string> disk_cache_values;
	disk_cache_values.push_back("avoid_readback");
	disk_cache_values.push_back("lru");
	disk_cache_values.push_back("largest_contiguous");
		
	std::vector<std::string> bandwidth_mixed_values;
	bandwidth_mixed_values.push_back("peer_proportional");
	bandwidth_mixed_values.push_back("prefer_tcp");


	const utf8_string choking_algorithm = to_string(
		bt::session_settings::choking_algorithm_t(
			ss.choking_algorithm));

	const utf8_string seed_choking_algorithm = to_string(
		bt::session_settings::seed_choking_algorithm_t(
			ss.seed_choking_algorithm));

	const utf8_string write_mode = to_string(
		bt::session_settings::io_buffer_mode_t(
			ss.disk_io_write_mode));

	const utf8_string read_mode = to_string(
		bt::session_settings::io_buffer_mode_t(
			ss.disk_io_read_mode));

	const utf8_string mixed_mode = to_string(
		bt::session_settings::bandwidth_mixed_algo_t(
			ss.mixed_mode_algorithm));


	make_string(map,		"bt.version",									BARETORRENT_USER_VERSION);
	make_string(map,		"bt.peer_id",									"BE");
	make_string(map,		"bt.ui.language",								"");
	make_bool(map,			"bt.ui.minimized",								false);
	make_bool(map,			"bt.ui.minimize_notification",					true);
	make_bool(map,			"bt.first_run",									true);
	make_bool(map,			"bt.update",									true);
	make_unsigned_int(map,	"bt.update_delay",								86400);
	make_bool(map,			"bt.update_on_startup",							true);
	make_string(map,		"bt.update_url",								website_url() + utf8_string("/u.txt"));
	make_bool(map,			"bt.assoc",										true);
	make_rect(map,			"bt.ui.main_window.bounds",						wxRect(-1, -1, 1000, 700));
	make_bool(map,			"bt.ui.main_window.maximized",					false);
	make_int(map,			"bt.ui.window_splitter",						-350);
	make_int(map,			"bt.ui.connections_splitter",					0);
	make_string(map,		"bt.ui.main_list.widths",						"0.07 0.32 0.12 0.11 0.11 0.11 0.09 0.06");
	make_string(map,		"bt.ui.main_list.sort",							"0 1");
	make_string(map,		"bt.ui.peers_list.widths",						"0.49 0.24 0.24");
	make_string(map,		"bt.ui.peers_list.sort",						"1 1");
	make_string(map,		"bt.ui.trackers_list.widths",					"0.74 0.24");
	make_string(map,		"bt.ui.trackers_list.sort",						"0 1");
	make_string(map,		"bt.ui.ext_list.widths",						"0.26 0.11 0.62");
	make_string(map,		"bt.ui.ext_list.sort",							"0 1");
	make_int(map,			"bt.ui.ext_splitter",							200);
	make_string(map,		"bt.ui.advanced_list.widths",					"0.48 0.48");
	make_string(map,		"bt.ui.advanced_list.sort",						"0 1");
	make_int(map,			"bt.ui.advanced_splitter",						-70);
	make_unsigned_int(map,	"bt.ui.extension_time",							2000);
	make_bool(map,			"bt.ui.log",									false);
	make_string(map,		"bt.ui.log_list.widths",						"0.15 0.82");
	make_string(map,		"bt.ui.log_list.sort",							"0 1");
	make_string(map,		"bt.ui.log_ext.widths",							"0.98");
	make_string(map,		"bt.ui.log_ext.sort",							"0");
	make_rect(map,			"bt.ui.log_window.bounds",						wxRect(-1, -1, 1100, 730));
	make_bool(map,			"bt.ui.log_window.maximized",					false);
	make_int(map,			"bt.ui.log.split",								-300);
	make_bool(map,			"bt.ui.add_from_clipboard",						true);
	make_bool(map,			"bt.ui.remove_confirm",							true);
	make_bool(map,			"bt.ui.remove_keeps_files",						true);

	make_bool(map,			"pcon.enabled",									false);
	make_bool(map,			"pcon.visible",									false);
	make_int(map,			"pcon.max_history",								100);

	make_unsigned_int(map,	"bt.power_timeout",								30);

	make_bool(map,			"bt.torrent.add_top",							false);
	make_bool(map,			"bt.torrent.append_extension",					true);
	make_string(map,		"bt.torrent.auto_load",							"");
	make_bool(map,			"bt.torrent.auto_load_enabled",					false);

	make_unsigned_int(map,	"bt.torrent.active_downloads",					ss.active_downloads);
	make_bool(map,			"bt.torrent.active_downloads_enabled",			(ss.active_downloads == -1 ? false : true));
	make_unsigned_int(map,	"bt.torrent.active_uploads",					ss.active_seeds);
	make_bool(map,			"bt.torrent.active_uploads_enabled",			(ss.active_seeds == -1 ? false : true));
	make_bool(map,			"bt.torrent.auto_active_limit",					true);
	make_unsigned_int(map,	"bt.torrent.seed_ratio",						150);
	make_bool(map,			"bt.torrent.max_download_enabled",				(ss.download_rate_limit == 0 ? false : true));
	make_unsigned_int(map,	"bt.torrent.max_download",						ss.download_rate_limit);
	make_bool(map,			"bt.torrent.max_upload_enabled",				(ss.upload_rate_limit == 0 ? false : true));
	make_unsigned_int(map,	"bt.torrent.max_upload",						ss.upload_rate_limit);
	
	make_string(map,		"bt.torrent.default_trackers",					"udp://tracker.openbittorrent.com:80/announce");
	make_unsigned_int(map,	"bt.torrent.stalled_rate",						1024);
	make_enum(map,			"bt.torrent.storage_mode",						to_string(atp.storage_mode), storage_mode_values);

	make_string(map,		"bt.network.listen",							"");

	make_bool(map,			"bt.system.firewall_exception",					true);
	make_unsigned_int(map,	"bt.network.num_connections",					ss.connections_limit);
	make_int(map,			"bt.network.connections_per_torrent",			h.max_connections(), 2);
	make_bool(map,			"bt.network.connections_per_torrent_enabled",	(h.max_connections() > 0 ? true : false));
	make_bool(map,			"bt.network.proxy_enabled",						false);
	make_string(map,		"bt.network.proxy",								"");
	make_enum(map,			"bt.network.proxy_type",						"", proxy_type_values);
	make_string(map,		"bt.network.proxy_username",					"");
	make_string(map,		"bt.network.proxy_password",					"");

	make_unsigned_int(map,	"bt.dht.max_peers_reply",						ds.max_peers_reply);
	make_unsigned_int(map,	"bt.dht.search_branching",						ds.search_branching);
	make_unsigned_int(map,	"bt.dht.max_fail_count",						ds.max_fail_count);
	make_unsigned_int(map,	"bt.dht.max_torrents",							ds.max_torrents);
	make_bool(map,			"bt.dht.restrict_routing_ips",					(ds.restrict_routing_ips ? true : false));
	make_bool(map,			"bt.dht.restrict_search_ips",					(ds.restrict_search_ips ? true : false));

	make_enum(map,			"bt.pe.out_enc_policy",							to_string(bt::pe_settings::forced), enc_policy_values);
	make_enum(map,			"bt.pe.in_enc_policy",							to_string(bt::pe_settings::forced), enc_policy_values);
	make_enum(map,			"bt.pe.allowed_enc_level",						to_string(bt::pe_settings::rc4), enc_level_values);
	make_bool(map,			"bt.pe.prefer_rc4",								(pe.prefer_rc4 ? true : false));

	make_string(map,		"bt.path.complete_save_path",					platform::download_directory().as_utf8_string());
	make_string(map,		"bt.path.incomplete_save_path",					"");
	make_string(map,		"bt.path.complete_torrent_file_path",			"");
	make_string(map,		"bt.path.incomplete_torrent_file_path",			"");

	make_bool(map,			"bt.path.incomplete_save_path_enabled",			false);
	make_bool(map,			"bt.path.complete_torrent_file_path_enabled",	false);
	make_bool(map,			"bt.path.incomplete_torrent_file_path_enabled", false);
	make_string(map,		"bt.paths.conf_filename",						"baretorrent.conf.xml");
	make_string(map,		"bt.paths.backup_conf_suffix",					".backup");
	make_string(map,		"bt.paths.bad_conf_suffix",						".bad");
	make_string(map,		"bt.paths.locales",								"locales");
	make_string(map,		"bt.paths.locale_suffix",						".txt");
	make_string(map,		"bt.paths.incomplete_suffix",					".baretorrent");
	make_string(map,		"bt.paths.torrent_file_suffix",					".torrent");
	make_string(map,		"bt.paths.resume_file_suffix",					".fastresume");
	make_string(map,		"bt.paths.state_file_suffix",					".state");
	make_string(map,		"bt.paths.session_filename",					"baretorrent.session");
	make_string(map,		"bt.paths.queue_filename",						"baretorrent.queue.xml");

	make_string(map,		"bt.ext.search_url",							website_url() + utf8_string("/ext/?s=%1%"));
	make_string(map,		"bt.ext.fetch_url",								website_url() + utf8_string("/ext/?d=%1%"));

	make_string(map,		"bt.session.user_agent",						user_agent());
	make_unsigned_int(map,	"bt.session.tracker_completion_timeout",		ss.tracker_completion_timeout);
	make_unsigned_int(map,	"bt.session.tracker_receive_timeout",			ss.tracker_receive_timeout);
	make_unsigned_int(map,	"bt.session.stop_tracker_timeout",				ss.stop_tracker_timeout);
	make_unsigned_int(map,	"bt.session.tracker_maximum_response_length",	ss.tracker_maximum_response_length);
	make_unsigned_int(map,	"bt.session.piece_timeout",						ss.piece_timeout);
	make_unsigned_int(map,	"bt.session.request_queue_time",				ss.request_queue_time);
	make_unsigned_int(map,	"bt.session.max_allowed_in_request_queue",		ss.max_allowed_in_request_queue);
	make_unsigned_int(map,	"bt.session.max_out_request_queue",				ss.max_out_request_queue);
	make_unsigned_int(map,	"bt.session.whole_pieces_threshold",			ss.whole_pieces_threshold);
	make_unsigned_int(map,	"bt.session.peer_timeout",						ss.peer_timeout);
	make_unsigned_int(map,	"bt.session.urlseed_timeout",					ss.urlseed_timeout);
	make_unsigned_int(map,	"bt.session.urlseed_pipeline_size",				ss.urlseed_pipeline_size);
	make_unsigned_int(map,	"bt.session.file_pool_size",					ss.file_pool_size);
	make_bool(map,			"bt.session.allow_multiple_connections_per_ip",	ss.allow_multiple_connections_per_ip);
	make_unsigned_int(map,	"bt.session.max_failcount",						ss.max_failcount);
	make_unsigned_int(map,	"bt.session.min_reconnect_time",				ss.min_reconnect_time);
	make_unsigned_int(map,	"bt.session.peer_connect_timeout",				ss.peer_connect_timeout);
	make_bool(map,			"bt.session.ignore_limits_on_local_network",	ss.ignore_limits_on_local_network);
	make_unsigned_int(map,	"bt.session.connection_speed",					ss.connection_speed);
	make_bool(map,			"bt.session.send_redundant_have",				ss.send_redundant_have);
	make_bool(map,			"bt.session.lazy_bitfields",					ss.lazy_bitfields);
	make_unsigned_int(map,	"bt.session.inactivity_timeout",				ss.inactivity_timeout);
	make_unsigned_int(map,	"bt.session.unchoke_interval",					ss.unchoke_interval);
	make_unsigned_int(map,	"bt.session.optimistic_unchoke_interval",		ss.optimistic_unchoke_interval);
	make_string(map,		"bt.session.announce_ip",						ss.announce_ip);
	make_unsigned_int(map,	"bt.session.num_want",							ss.num_want);
	make_unsigned_int(map,	"bt.session.initial_picker_threshold",			ss.initial_picker_threshold);
	make_unsigned_int(map,	"bt.session.allowed_fast_set_size",				ss.allowed_fast_set_size);
	make_enum(map,			"bt.session.suggest_mode",						to_string(bt::session_settings::suggest_mode_t(ss.suggest_mode)), suggest_mode_values);
	make_unsigned_int(map,	"bt.session.max_queued_disk_bytes",				ss.max_queued_disk_bytes);
	make_unsigned_int(map,	"bt.session.handshake_timeout",					ss.handshake_timeout);
	make_bool(map,			"bt.session.use_dht_as_fallback",				ss.use_dht_as_fallback);
	make_bool(map,			"bt.session.free_torrent_hashes",				ss.free_torrent_hashes);
	make_bool(map,			"bt.session.upnp_ignore_nonrouters",			ss.upnp_ignore_nonrouters);
	make_unsigned_int(map,	"bt.session.send_buffer_watermark",				ss.send_buffer_watermark);
	make_unsigned_int(map,	"bt.session.send_buffer_watermark_factor",		ss.send_buffer_watermark_factor);
	make_enum(map,			"bt.session.choking_algorithm",					choking_algorithm, choking_algorithm_values);
	make_enum(map,			"bt.session.seed_choking_algorithm",			seed_choking_algorithm, seed_choking_algorithm_values);
	make_bool(map,			"bt.session.use_parole_mode",					ss.use_parole_mode);
	make_int(map,			"bt.session.cache_size",						ss.cache_size, -1);
	make_int(map,			"bt.session.cache_buffer_chunk_size",			ss.cache_buffer_chunk_size, 1);
	make_unsigned_int(map,	"bt.session.cache_expiry",						ss.cache_expiry);
	make_bool(map,			"bt.session.use_read_cache",					ss.use_read_cache);
	make_bool(map,			"bt.session.explicit_read_cache",				ss.explicit_read_cache);
	make_unsigned_int(map,	"bt.session.explicit_cache_interval",			ss.explicit_cache_interval);
	make_enum(map,			"bt.session.disk_io_write_mode",				write_mode, io_buffer_mode_values);
	make_enum(map,			"bt.session.disk_io_read_mode",					read_mode, io_buffer_mode_values);
	make_int(map,			"bt.session.peer_tos",							ss.peer_tos, 0, 255);
	make_int(map,			"bt.session.active_downloads",					ss.active_downloads, -1);
	make_int(map,			"bt.session.active_seeds",						ss.active_seeds, -1);
	make_unsigned_int(map,	"bt.session.active_dht_limit",					ss.active_dht_limit);
	make_unsigned_int(map,	"bt.session.active_tracker_limit",				ss.active_tracker_limit);
	make_unsigned_int(map,	"bt.session.active_limit",						ss.active_limit);
	make_bool(map,			"bt.session.auto_manage_prefer_seeds",			ss.auto_manage_prefer_seeds);
	make_bool(map,			"bt.session.dont_count_slow_torrents",			ss.dont_count_slow_torrents);
	make_unsigned_int(map,	"bt.session.auto_manage_interval",				ss.auto_manage_interval);
	make_float(map,			"bt.session.share_ratio_limit",					ss.share_ratio_limit, 0.0f);
	make_float(map,			"bt.session.seed_time_ratio_limit",				ss.seed_time_ratio_limit, 0.0f);
	make_unsigned_int(map,	"bt.session.seed_time_limit",					ss.seed_time_limit);
	make_unsigned_int(map,	"bt.session.peer_turnover_interval",			ss.peer_turnover_interval);
	make_float(map,			"bt.session.peer_turnover",						ss.peer_turnover, 0.0f, 1.0f);
	make_float(map,			"bt.session.peer_turnover_cutoff",				ss.peer_turnover_cutoff, 0.0f, 1.0f);
	make_bool(map,			"bt.session.close_redundant_connections",		ss.close_redundant_connections);
	make_unsigned_int(map,	"bt.session.auto_scrape_interval",				ss.auto_scrape_interval);
	make_unsigned_int(map,	"bt.session.auto_scrape_min_interval",			ss.auto_scrape_min_interval);
	make_unsigned_int(map,	"bt.session.max_peerlist_size",					ss.max_peerlist_size);
	make_unsigned_int(map,	"bt.session.min_announce_interval",				ss.min_announce_interval);
	make_bool(map,			"bt.session.prioritize_partial_pieces",			ss.prioritize_partial_pieces);
	make_unsigned_int(map,	"bt.session.auto_manage_startup",				ss.auto_manage_startup);
	make_bool(map,			"bt.session.rate_limit_ip_overhead",			ss.rate_limit_ip_overhead);
	make_bool(map,			"bt.session.announce_to_all_trackers",			ss.announce_to_all_trackers);
	make_bool(map,			"bt.session.announce_to_all_tiers",				ss.announce_to_all_tiers);
	make_bool(map,			"bt.session.prefer_udp_trackers",				ss.prefer_udp_trackers);
	make_bool(map,			"bt.session.strict_super_seeding",				ss.strict_super_seeding);
	make_unsigned_int(map,	"bt.session.seeding_piece_quota",				ss.seeding_piece_quota);
	make_unsigned_int(map,	"bt.session.max_sparse_regions",				ss.max_sparse_regions);
	make_bool(map,			"bt.session.lock_disk_cache",					ss.lock_disk_cache);
	make_unsigned_int(map,	"bt.session.max_rejects",						ss.max_rejects);
	make_unsigned_int(map,	"bt.session.recv_socket_buffer_size",			ss.recv_socket_buffer_size);
	make_unsigned_int(map,	"bt.session.send_socket_buffer_size",			ss.send_socket_buffer_size);
	make_bool(map,			"bt.session.optimize_hashing_for_speed",		ss.optimize_hashing_for_speed);
	make_unsigned_int(map,	"bt.session.file_checks_delay_per_block",		ss.file_checks_delay_per_block);
	make_enum(map,			"bt.session.disk_cache_algorithm",				to_string(ss.disk_cache_algorithm), disk_cache_values);
	make_unsigned_int(map,	"bt.session.read_cache_line_size",				ss.read_cache_line_size);
	make_unsigned_int(map,	"bt.session.write_cache_line_size",				ss.write_cache_line_size);
	make_unsigned_int(map,	"bt.session.optimistic_disk_retry",				ss.optimistic_disk_retry);
	make_bool(map,			"bt.session.disable_hash_checks",				ss.disable_hash_checks);
	make_unsigned_int(map,	"bt.session.max_suggest_pieces",				ss.max_suggest_pieces);
	make_bool(map,			"bt.session.drop_skipped_requests",				ss.drop_skipped_requests);
	make_bool(map,			"bt.session.low_prio_disk",						ss.low_prio_disk);
	make_unsigned_int(map,	"bt.session.local_service_announce_interval",	ss.local_service_announce_interval);
	make_unsigned_int(map,	"bt.session.dht_announce_interval",				ss.dht_announce_interval);
	make_unsigned_int(map,	"bt.session.udp_tracker_token_expiry",			ss.udp_tracker_token_expiry);
	make_bool(map,			"bt.session.volatile_read_cache",				ss.volatile_read_cache);
	make_bool(map,			"bt.session.guided_read_cache",					ss.guided_read_cache);
	make_unsigned_int(map,	"bt.session.default_cache_min_age",				ss.default_cache_min_age);
	make_unsigned_int(map,	"bt.session.num_optimistic_unchoke_slots",		ss.num_optimistic_unchoke_slots);
	make_bool(map,			"bt.session.no_atime_storage",					ss.no_atime_storage);
	make_unsigned_int(map,	"bt.session.default_est_reciprocation_rate",	ss.default_est_reciprocation_rate);
	make_unsigned_int(map,	"bt.session.increase_est_reciprocation_rate",	ss.increase_est_reciprocation_rate);
	make_unsigned_int(map,	"bt.session.decrease_est_reciprocation_rate",	ss.decrease_est_reciprocation_rate);
	make_bool(map,			"bt.session.incoming_starts_queued_torrents",	ss.incoming_starts_queued_torrents);
	make_bool(map,			"bt.session.report_true_downloaded",			ss.report_true_downloaded);
	make_bool(map,			"bt.session.strict_end_game_mode",				ss.strict_end_game_mode);
	make_bool(map,			"bt.session.broadcast_lsd",						ss.broadcast_lsd);
	make_bool(map,			"bt.session.enable_outgoing_utp",				ss.enable_outgoing_utp);
	make_bool(map,			"bt.session.enable_incoming_utp",				ss.enable_incoming_utp);
	make_bool(map,			"bt.session.enable_outgoing_tcp",				ss.enable_outgoing_tcp);
	make_bool(map,			"bt.session.enable_incoming_tcp",				ss.enable_incoming_tcp);
	make_int(map,			"bt.session.max_pex_peers",						ss.max_pex_peers);
	make_bool(map,			"bt.session.ignore_resume_timestamps",			ss.ignore_resume_timestamps);
	make_bool(map,			"bt.session.no_recheck_incomplete_resume",		ss.no_recheck_incomplete_resume);
	make_bool(map,			"bt.session.anonymous_mode",					ss.anonymous_mode);
	make_unsigned_int(map,	"bt.session.tick_interval",						ss.tick_interval);
	make_bool(map,			"bt.session.report_web_seed_downloads",			ss.report_web_seed_downloads);
	make_int(map,			"bt.session.share_mode_target",					ss.share_mode_target, 2);
	make_unsigned_int(map,	"bt.session.upload_rate_limit",					ss.upload_rate_limit);		// todo: not used
	make_unsigned_int(map,	"bt.session.download_rate_limit",				ss.download_rate_limit);	// todo: not used
	make_unsigned_int(map,	"bt.session.local_upload_rate_limit",			ss.local_upload_rate_limit);
	make_unsigned_int(map,	"bt.session.local_download_rate_limit",			ss.local_download_rate_limit);
	make_unsigned_int(map,	"bt.session.dht_upload_rate_limit",				ss.dht_upload_rate_limit);
	make_unsigned_int(map,	"bt.session.unchoke_slots_limit",				ss.unchoke_slots_limit);
	make_int(map,			"bt.session.half_open_limit",					ss.half_open_limit, -1);
	make_unsigned_int(map,	"bt.session.connections_limit",					ss.connections_limit);
	make_unsigned_int(map,	"bt.session.utp_target_delay",					ss.utp_target_delay);
	make_unsigned_int(map,	"bt.session.utp_gain_factor",					ss.utp_gain_factor);
	make_unsigned_int(map,	"bt.session.utp_min_timeout",					ss.utp_min_timeout);
	make_unsigned_int(map,	"bt.session.utp_syn_resends",					ss.utp_syn_resends);
	make_unsigned_int(map,	"bt.session.utp_num_resends",					ss.utp_num_resends);
	make_unsigned_int(map,	"bt.session.utp_connect_timeout",				ss.utp_connect_timeout);
	make_unsigned_int(map,	"bt.session.utp_delayed_ack",					ss.utp_delayed_ack);
	make_bool(map,			"bt.session.utp_dynamic_sock_buf",				ss.utp_dynamic_sock_buf);
	make_int(map,			"bt.session.utp_loss_multiplier",				ss.utp_loss_multiplier, 0, 100);
	make_enum(map,			"bt.session.mixed_mode_algorithm",				mixed_mode, bandwidth_mixed_values);
	make_bool(map,			"bt.session.rate_limit_utp",					ss.rate_limit_utp);
	make_unsigned_int(map,	"bt.session.listen_queue_size",					ss.listen_queue_size);
	make_bool(map,			"bt.session.announce_double_nat",				ss.announce_double_nat);
	make_unsigned_int(map,	"bt.session.torrent_connect_boost",				ss.torrent_connect_boost);
	make_bool(map,			"bt.session.seeding_outgoing_connections",		ss.seeding_outgoing_connections);
	make_bool(map,			"bt.session.no_connect_privileged_ports",		ss.no_connect_privileged_ports);
	make_unsigned_int(map,	"bt.session.alert_queue_size",					ss.alert_queue_size);
	make_unsigned_int(map,	"bt.session.max_metadata_size",					ss.max_metadata_size);
	make_bool(map,			"bt.session.smooth_connects",					ss.smooth_connects);
	make_bool(map,			"bt.session.always_send_user_agent",			ss.always_send_user_agent);
	make_bool(map,			"bt.session.apply_ip_filter_to_trackers",		ss.apply_ip_filter_to_trackers);
	make_unsigned_int(map,	"bt.session.read_job_every",					ss.read_job_every);
	make_bool(map,			"bt.session.use_disk_read_ahead",				ss.use_disk_read_ahead);
	make_bool(map,			"bt.session.lock_files",						ss.lock_files);
	make_int(map,			"bt.session.ssl_listen",						ss.ssl_listen, 0, 65535);
	make_unsigned_int(map,	"bt.session.tracker_backoff",					ss.tracker_backoff);
	make_bool(map,			"bt.session.ban_web_seeds",						ss.ban_web_seeds);

	return map;
}

void add_option(const std::string& name, const utf8_string& def)
{
	if (!opt_exists(name))
	{
		boost::recursive_mutex::scoped_lock lock(g_mutex);

		log()
			<< "an extension has created the option '" << name << "' "
			<< "with an initial value '" << def << "'\n";

		g_map[name] = option(name, def, def);
		g_update.push_back(updated_opt(name, def));
	}
}

void dump_options()
{
	log() << "\ndumping configuration:\n";

	option_list m;

	{
		boost::recursive_mutex::scoped_lock lock(g_mutex);
		m = g_map;
	}

	std::pair<std::string, option> p;
	foreach (p, m)
	{
		log()
			<< "  " << p.first << " = " << p.second.value()
			<< " (" << p.second.default_value() << ")\n";
	}
	
	log() << "\n";
}

option_list get_options()
{
	option_list m;

	{
		boost::recursive_mutex::scoped_lock lock(g_mutex);
		m = g_map;
	}

	return m;
}

void create_options_table()
{
	g_database.execute(utf8_string(
		"create table if not exists options ("
		"id integer primary key, "
		"name text not null, "
		"value text not null)"));
}

void create_extensions_table()
{
	g_database.execute(utf8_string(
		"create table if not exists extensions ("
		"id integer primary key, "
		"guid text not null, "
		"name text not null, "
		"author text not null, "
		"version text not null, "
		"short_description not null, "
		"long_description not null, "
		"icon text not null, "
		"email text not null, "
		"website text not null, "
		"enabled integer not null)"));
}

void create_extension_options_table()
{
	g_database.execute(utf8_string(
		"create table if not exists extension_options ("
		"id integer primary key, "
		"extension_id integer not null, "
		"name text not null, "
		"value text not null)"));
}

void create_tables()
{
	create_options_table();
	create_extensions_table();
	create_extension_options_table();

	try
	{
		g_database.query(
			utf8_string("select id, name, value from options"));
	}
	catch(db::exception& e)
	{
		log_error()
			<< "bad table 'options', " << e.what() << "; recreating\n";

		g_database.execute(utf8_string("drop table options"));
		create_options_table();
	}

	try
	{
		g_database.query(utf8_string(
			"select id, guid, name, author, version, "
			"short_description, long_description, icon, email, "
			"website, enabled from extensions"));
	}
	catch(db::exception& e)
	{
		log_error()
			<< "bad table 'extensions', " << e.what()
			<< "; recreating\n";

		g_database.execute(utf8_string("drop table extensions"));
		create_extensions_table();
	}

	try
	{
		g_database.query(utf8_string(
			"select id, extension_id, name, value "
			"from extension_options"));
	}
	catch(db::exception& e)
	{
		log_error()
			<< "bad table 'extension_options', " << e.what()
			<< "; recreating\n";

		g_database.execute(utf8_string("drop table extension_options"));
		create_extension_options_table();
	}
}

bool first_time_run()
{
	const utf8_path p = paths::conf() / sqlite_filename;
	return !p.exists();
}

void load_options()
{
	bt_assert(g_map.empty());
	bt_assert(g_update.empty());

	boost::recursive_mutex::scoped_lock lock(g_mutex);

	const bool first_time = first_time_run();

	g_main_thread_id = boost::this_thread::get_id();
	g_database = db::sqlite::connect(
		paths::conf() / sqlite_filename, false, true);

	create_tables();
	
	option_list m = create_default_options();
	std::pair<std::string, option> p;

	if (!first_time)
	{
		db::resultset rs = g_database
			.prepare(utf8_string(
				"select value from options where name=?"))
			.bind(utf8_string("bt.version"))
			.execute();

		// this is an old version (<0.3) that doesn't have bt.version
		if (rs.empty())
		{
			m["bt.force_update"] = option(
				"bt.force_update", utf8_string("1"), utf8_string("1"));
		}
	}

	db::transaction tn = g_database.transaction();

	db::prepared_statement get = g_database.prepare(utf8_string(
		"select value from options where name=?"));

	db::prepared_statement add = g_database.prepare(utf8_string(
		"insert into options (name, value) values (?, ?)"));

	db::prepared_statement set = g_database.prepare(utf8_string(
		"update options set value=? where name=?"));

	foreach (p, m)
	{
		db::resultset rs = get
			.bind(utf8_string(p.first))
			.execute();

		g_map[p.first] = p.second;

		if (rs.empty())
		{
			add
				.bind(utf8_string(p.first))
				.bind(p.second.value())
				.execute();
		}
		else
		{
			g_map[p.first].set(rs["value"], p.second.default_value());

			set
				.bind(g_map[p.first].value())
				.bind(utf8_string(p.first))
				.execute();
		}

		get.reset();
		add.reset();
		set.reset();
	}

	tn.commit();

	check_options();
	dump_options();

	ext_load_options();
}

void remove_option(const std::string& s)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);

	option_list::iterator itor = g_map.find(s);
	if (itor == g_map.end())
		return;

	g_map.erase(itor);

	try
	{
		g_database
			.prepare(utf8_string("delete from options where name=?"))
			.bind(utf8_string(s))
			.execute();
	}
	catch(db::exception& e)
	{
		// ignore

		log_error()
			<< "could not remove option '" << s << "', "
			<< e.what() << "\n";
	}
}

void save_options()
{
	try
	{
		boost::recursive_mutex::scoped_lock lock(g_mutex);
		if (g_update.empty())
			return;

		db::transaction tn = g_database.transaction();
	
		foreach (updated_opt o, g_update)
		{
			if (o.guid.empty())
			{
				db::resultset rs = g_database
					.prepare(utf8_string(
						"select id from options where name=?"))
					.bind(utf8_string(o.name))
					.execute();

				if (rs.empty())
				{
					g_database
						.prepare(utf8_string(
							"insert into options (name, value) "
							"values (?, ?)"))
						.bind(utf8_string(o.name))
						.bind(o.value)
						.execute();
				}
				else
				{
					g_database
						.prepare(utf8_string(
							"update options set value=? where name=?"))
						.bind(o.value)
						.bind(utf8_string(o.name))
						.execute();
				}
			}
			else
			{
				db::resultset rs = g_database
					.prepare(utf8_string(
						"select id from extensions where guid=?"))
					.bind(utf8_string(o.guid))
					.execute();

				if (rs.empty())
				{
					log_error()
						<< "couldn't find extension with guid "
						<< o.guid << " while trying to save options\n";

					continue;
				}
				
				db::primary_key id = rs.get<db::primary_key>("id");

				rs = g_database
					.prepare(utf8_string(
						"select id from extension_options "
						"where extension_id=? and name=?"))
					.bind(id)
					.bind(utf8_string(o.name))
					.execute();

				if (rs.empty())
				{
					g_database
						.prepare(utf8_string(
							"insert into extension_options "
							"(extension_id, name, value) "
							"values (?, ?, ?)"))
						.bind(id)
						.bind(utf8_string(o.name))
						.bind(o.value)
						.execute();
				}
				else
				{
					g_database
						.prepare(utf8_string(
							"update extension_options set value=? "
							"where extension_id=? and name=?"))
						.bind(o.value)
						.bind(id)
						.bind(utf8_string(o.name))
						.execute();
				}
			}
		}

		tn.commit();
	}
	catch(db::exception& e)
	{
		log_error() << "couldn't save options: " << e.what() << "\n";
	}

	g_update.clear();
}

bool opt_exists(const std::string& name)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);
	return (g_map.find(name) != g_map.end());
}

utf8_string opt(const std::string& n)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);
	bt_assert(opt_exists(n));
	return g_map[n].value();
}

option get_option(const std::string& name)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);
	bt_assert(opt_exists(name));
	return g_map[name];
}

void opt(const std::string& s, const utf8_string& v)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);
	bt_assert(opt_exists(s));

	if (g_map[s].value() != v)
	{
		g_map[s].set(v);
		g_update.push_back(updated_opt(s, v));
		log_vv() << "option " << s << " changed to '" << v << "'\n";
	}
}


void ext_load_options()
{
	db::resultset rs = g_database
		.prepare(utf8_string(
			"select "
			"	guid, extension_options.name, extension_options.value "
			"from extension_options "
			"join extensions "
			"on extension_options.extension_id=extensions.id"))
		.execute();

	foreach (db::row r, rs)
	{
		const std::string guid = r["guid"].std_string();
		const std::string name = r["name"].std_string();
		const utf8_string value = r["value"];

		g_ext_map[guid][name] = option(name, value, utf8_string());
	}
}

void add_ext_opt(
	const std::string& guid,
	const std::string& name, const utf8_string& def)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);

	if (g_ext_map[guid].find(name) == g_ext_map[guid].end())
		g_ext_map[guid][name] = option(name, def, def);

	g_ext_map[guid][name].default_value(def);
}

utf8_string ext_opt(const std::string& guid, const std::string& name)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);
	return g_ext_map[guid][name].value();
}

option get_ext_option(const std::string& guid, const std::string& name)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);
	return g_ext_map[guid][name];
}

void ext_opt(
	const std::string& guid, const std::string& name,
	const utf8_string& value)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);

	g_ext_map[guid][name].set(value);
	g_update.push_back(updated_opt(guid, name, value));

	ext::fire(utf8_string("option_changed"), name, value);

}

option_list get_ext_options(const std::string& guid)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);
	return g_ext_map[guid];
}

bool ext_installed(const std::string& guid)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);

	db::resultset rs = g_database
		.prepare(utf8_string("select id from extensions where guid=?"))
		.bind(utf8_string(guid))
		.execute();
		
	if (rs.empty())
		return false;
	
	return true;
}

bool ext_enabled(const std::string& guid)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);

	db::resultset rs = g_database
		.prepare(utf8_string(
			"select enabled from extensions where guid=?"))
		.bind(utf8_string(guid))
		.execute();
		
	// this is a new extension
	if (rs.empty())
		return true;

	return rs.get<bool>("enabled");
}

void ext_enabled(const std::string& guid, bool b)
{
	boost::recursive_mutex::scoped_lock lock(g_mutex);

	db::resultset rs = g_database
		.prepare(utf8_string(
			"select enabled from extensions where guid=?"))
		.bind(utf8_string(guid))
		.execute();
		
	if (rs.empty())
	{
		log_error()
			<< "could not set enabled state of extension "
			<< guid << " to " << b
			<< ", guid not found in the database\n";

		return;
	}

	g_database
		.prepare(
			utf8_string("update extensions set enabled=? where guid=?"))
		.bind(b)
		.bind(utf8_string(guid))
		.execute();
}


void add_extension(const ext::instance* i)
{
	bt_assert(i);

	log() << "adding extension " << i->guid() << " to database\n";

	boost::recursive_mutex::scoped_lock lock(g_mutex);

	db::resultset rs = g_database
		.prepare(utf8_string("select id from extensions where guid=?"))
		.bind(utf8_string(i->guid()))
		.execute();

	if (rs.empty())
	{
		g_database
			.prepare(utf8_string(
				"insert into extensions "
				"(guid, name, author, version, short_description, "
				"long_description, icon, email, website, enabled) "
				"values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"))
			.bind(utf8_string(i->guid()))
			.bind(utf8_string(i->name()))
			.bind(i->author())
			.bind(i->version())
			.bind(i->short_description())
			.bind(i->long_description())
			.bind(i->icon_data())
			.bind(i->email())
			.bind(i->website())
			.bind(true)
			.execute();
	}
	else
	{
		g_database
			.prepare(utf8_string(
				"update extensions set "
				"name=?, author=?, version=?, short_description=?, "
				"long_description=?, icon=?, email=?, website=? "
				"where guid=?"))
			.bind(utf8_string(i->name()))
			.bind(i->author())
			.bind(i->version())
			.bind(i->short_description())
			.bind(i->long_description())
			.bind(i->icon_data())
			.bind(i->email())
			.bind(i->website())
			.bind(utf8_string(i->guid()))
			.execute();
	}
}

void remove_extension(const std::string& guid)
{
	log() << "removing extension " << guid << " from database\n";

	db::resultset rs = g_database
		.prepare(utf8_string("select id from extensions where guid=?"))
		.bind(utf8_string(guid))
		.execute();

	if (rs.empty())
	{
		log_error()
			<< "trying to remove extension " << guid
			<< ", but it is not in the database; ignoring\n";

		return;
	}

	db::primary_key id = rs.get<db::primary_key>("id");

	g_database
		.prepare(utf8_string("delete from extensions where id=?"))
		.bind(id)
		.execute();

	g_database
		.prepare(utf8_string(
			"delete from extension_options where id=?"))
		.bind(id)
		.execute();
}

extension_info cached_extension_info(const std::string& guid)
{
	extension_info i;
	
	db::resultset rs = g_database
		.prepare(utf8_string(
			"select "
			"name, author, version, short_description, "
			"long_description, icon, email, website "
			"from extensions "
			"where guid=?"))
		.bind(utf8_string(guid))
		.execute();

	if (rs.empty())
		return i;

	i.name = rs["name"];
	i.author = rs["author"];
	i.version = extension_version::parse(rs["version"]);
	i.short_description = rs["short_description"];
	i.long_description = rs["long_description"];
	i.icon_data = rs.get<byte_array>("icon");
	i.email = rs["email"];
	i.website = rs["website"];

	return i;
}

db::database& get_database()
{
	bt_assert(g_main_thread_id == boost::this_thread::get_id());
	return g_database;
}

}	// namespace baretorrent
