// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_OPTIONS_INCLUDED
#define BARETORRENT_OPTIONS_INCLUDED

#include "utility.h"
#include "db/sqlite.h"

namespace baretorrent
{

namespace ext
{
	class instance;
}

class option_type : boost::noncopyable
{
public:
	virtual ~option_type()
	{
	}

	virtual void adjust(const std::string& name, utf8_string& s) = 0;
};

class enum_option_type : public option_type
{
public:
	enum_option_type(const std::vector<std::string>& v)
		: v_(v)
	{
	}

	virtual void adjust(const std::string& name, utf8_string& s)
	{
		if (v_.empty())
			return;

		if (std::find(v_.begin(), v_.end(), s.std_string()) == v_.end())
		{
			log_error()
				<< "bad option " << name << " with value "
				<< "'" << s << "', changing to '" << v_[0] << "'\n";

			s = utf8_string(v_[0]);
		}
	}

private:
	std::vector<std::string> v_;
};

template <class T>
class number_option_type : public option_type
{
public:
	T min, max;

	number_option_type(T min, T max)
		: min(min), max(max)
	{
	}

	virtual void adjust(const std::string& name, utf8_string& s)
	{
		std::istringstream iss(s.std_string());
		
		T t;
		iss >> t;

		T nt = t;
		clamp(nt, min, max);

		if (t < min || t > max)
		{
			log_error()
				<< "bad option " << name << " with value "
				<< "'" << t << "', changing to '" << nt << "'\n";

			s = as_utf8_string(nt);
		}
	}
};

template <class T>
class value_option_type;

template <>
class value_option_type<bool> : public option_type
{
public:
	virtual void adjust(const std::string& name, utf8_string& s)
	{
		if (s == utf8_string("0") || s.empty())
			return;

		if (s == utf8_string("1"))
			return;

		const utf8_string n("1");
		log_error()
			<< "bad option " << name << " with value "
			<< "'" << s << "', changing to '" << n << "'\n";
		
		s = n;
	}
};

template <>
class value_option_type<wxRect> : public option_type
{
	virtual void adjust(const std::string& name, utf8_string& s)
	{
		std::istringstream iss(s.std_string());

		wxRect r(0, 0, 0, 0);
		iss >> r;

		std::ostringstream oss;
		oss << r;

		if (s != utf8_string(oss.str()))
		{
			log_error()
				<< "bad option " << name << " with value "
				<< "'" << s << "', changing to '" << oss.str() << "'\n";

			s = utf8_string(oss.str());
		}
	}
};


class option
{
public:
	option()
	{
	}

	option(
		const std::string& n, const utf8_string& v,
		const utf8_string& d,
		boost::shared_ptr<option_type> t =
			boost::shared_ptr<option_type>()) :
				name_(n), value_(v), default_(d), type_(t)
	{
	}

	const std::string& name() const
	{
		return name_;
	}

	const utf8_string& value() const
	{
		return value_;
	}

	const utf8_string& default_value() const
	{
		return default_;
	}

	void default_value(const utf8_string& s)
	{
		default_ = s;
	}

	void set(utf8_string s)
	{
		if (type_)
			type_->adjust(name_, s);

		value_ = s;
	}

	void set(utf8_string s, utf8_string def)
	{
		if (type_)
		{
			type_->adjust(name_, s);
			type_->adjust(name_, def);
		}

		value_ = s;
		default_ = def;
	}

private:
	std::string name_;
	utf8_string value_;
	utf8_string default_;
	boost::shared_ptr<option_type> type_;
};


typedef std::map<std::string, option> option_list;


// loads the options from the configuration file
//
void load_options();

// saves all the modified options to the database
//
void save_options();

// outputs all the options using log()
//
void dump_options();

// returns a map of all the options
//
option_list get_options();

// returns whether the name exists in the current options
//
bool opt_exists(const std::string& name);


// returns the value of the given option from the current configuration
//
utf8_string opt(const std::string& s);

option get_option(const std::string& name);

// sets the value of the given option in the current configuration
//
void opt(const std::string& s, const utf8_string& v);

// returns the value of the given option from the current
// configuration; the value is converted using operator>>
//
template <class T>
T opt(const std::string& s)
{
	const utf8_string v = opt(s);
	return as<T>(v);
}

// sets the value of the given option in the current configuration; the
// value is converted using operator<<
//
template <class T>
void opt(const std::string& s, const T& v)
{
	std::ostringstream oss;
	oss << v;
	opt(s, utf8_string(oss.str()));
}

// removes the given option from the configuration; noop if the option
// doesn't exist
//
void remove_option(const std::string& s);

// adds the given option with a default value if it doesn't exist
//
void add_ext_opt(
	const std::string& guid,
	const std::string& name, const utf8_string& def);

// returns the value of the given option of an extension; if the option
// doesn't exist, adds it to the list with an empty value
//
utf8_string ext_opt(const std::string& guid, const std::string& name);

option get_ext_option(const std::string& guid, const std::string& name);

// sets the value of the given option for an extension
//
void ext_opt(
	const std::string& guid, const std::string& name,
	const utf8_string& value);

// returns the value of the given option for the extension; the value
// is converted using operator>>
//
template <class T>
T ext_opt(const std::string& guid, const std::string& s)
{
	const utf8_string v = ext_opt(guid, s);
	return as<T>(v);
}

// sets the value of the given option for the extension; the value
// is converted using operator>>
//
template <class T>
void ext_opt(const std::string& guid, const std::string& s, const T& v)
{
	std::ostringstream oss;
	oss << v;
	ext_opt(guid, s, oss.str());
}

// returns a map of all the options for the given extension
//
option_list get_ext_options(const std::string& guid);


// returns whether the given extension is in the database
//
bool ext_installed(const std::string& guid);

// returns whether the given extension is enabled
//
bool ext_enabled(const std::string& guid);

// sets whether the given extension is enabled
//
void ext_enabled(const std::string& guid, bool b);

// adds the given extension to the database; noop if the extension
// already exists
//
void add_extension(const ext::instance* i);

// removes the extension from the database
//
void remove_extension(const std::string& guid);

// loads the info from the database for this extension; these are saved
// in the database every time the extension is loaded
//
// this is used to display information about disabled extensions since
// those are never imported
//
extension_info cached_extension_info(const std::string& guid);

// returns a reference to the database; this shouldn't be called outside
// of the main thread
//
db::database& get_database();

}	// namespace baretorrent

#endif	// BARETORRENT_OPTIONS_INCLUDED
