// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "utility.h"
#include "options_dialog.h"
#include "locale.h"
#include "manager.h"
#include "ext/ext.h"

namespace baretorrent
{

// used in the language combobox; the language option is the filename of
// the language file this strips the extension
//
utf8_string language_stringer(const locale_file& lf)
{
	return lf.name;
}

utf8_string language_value(const locale_file& lf)
{
	return lf.ietf;
}

// used for any kind of combobox, this assumes 'n' is a locale string
//
utf8_string loc_stringer(const std::string& n)
{
	if (n.empty())
		return utf8_string();

	return loc(n);
}


options_dialog::options_dialog(wxWindow* parent) :
	wxDialog(
		parent, 0, loc_wx("options.title"),
		wxDefaultPosition, wxDefaultSize,
		wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER),
	advanced_(nullptr)
{
	wxNotebook* tabs = new wxNotebook(this, -1);
	
	advanced_ = new option_list_widget(tabs, "");
	advanced_->changed.connect(boost::bind(
		&options_dialog::on_advanced_changed, this, _1, _2));

	// panels
	tabs->InsertPage(0, general(tabs), loc_wx("options.general"));
	tabs->InsertPage(1, torrents(tabs), loc_wx("options.torrents"));
	tabs->InsertPage(2, network(tabs), loc_wx("options.network"));
	tabs->InsertPage(3, ext(tabs), loc_wx("options.extensions"));
	tabs->InsertPage(4, advanced_, loc_wx("options.advanced"));

	if (!ext::python_initialized())
		tabs->RemovePage(3);


	// ok/cancel
	wxSizer* buttons = CreateSeparatedButtonSizer(wxOK | wxCANCEL);

	wxFlexGridSizer* root = new wxFlexGridSizer(1);
	root->Add(tabs, 1, wxEXPAND);
	root->AddSpacer(5);
	root->Add(buttons, 1, wxEXPAND);
	
	root->AddGrowableCol(0);
	root->AddGrowableRow(0);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(root, 1, wxEXPAND | wxALL, 5);

	SetSizerAndFit(box);
	SetSize(630, 550);
	CentreOnParent();
}

const std::map<std::string, option_list>&
options_dialog::options() const
{
	return new_;
}


wxPanel* options_dialog::general(wxWindow* parent)
{
	wxPanel* p = new wxPanel(parent);
	wxPanel* g = new wxPanel(p);

	// list of files in the locale directory
	std::vector<locale_file> lf = locale_files();

	wxComboBox* languages = make_combobox<locale_file>(
		g, lf, "bt.ui.language", &language_stringer, &language_value);

	for (std::size_t i=0; i<lf.size(); ++i)
	{
		if (lf[i].path.stem() == opt("bt.ui.language"))
		{
			languages->SetSelection(static_cast<int>(i));
			break;
		}
	}

	wxCheckBox* minimized = make_checkbox(
		g, loc("options.minimized"), "bt.ui.minimized");
	
	wxCheckBox* notification = make_checkbox(
		g, loc("options.notification"), "bt.ui.minimize_notification");
	
	wxFlexGridSizer* sizer = new wxFlexGridSizer(1, 5, 5);
	sizer->Add(make_label(g, loc("options.language")));
	sizer->Add(languages);
	sizer->AddSpacer(5);
	sizer->Add(minimized);
	sizer->Add(notification);
	sizer->Add(make_checkbox(g, loc("options.update"), "bt.update"));

	g->SetSizer(sizer);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(g, 0, wxEXPAND | wxALL, 5);

	p->SetSizer(box);

	return p;
}

wxPanel* options_dialog::torrents(wxWindow* parent)
{
	wxPanel* p = new wxPanel(parent);
	wxPanel* torrents = new wxPanel(p);

	wxGridBagSizer* grid = new wxGridBagSizer(5, 10);
	grid->SetEmptyCellSize(wxSize(10, 5));
	
	// this is messy, as grid bags always are

	// this is for widgets that expand (all of them except for browse
	// buttons)
#define g(c, row, col, rs, cs)				\
			grid->Add(						\
				c, wxGBPosition(row, col),	\
				wxGBSpan(rs, cs), wxEXPAND | wxALIGN_CENTER_VERTICAL)

	// this is for browse buttons which never expand
#define gf(c, row, col, rs, cs)	\
			grid->Add(c, wxGBPosition(row, col), wxGBSpan(rs, cs), 0)

	// some of the controls are created here so the grid bag setup
	// below at least fits on one line

	wxTextCtrl* csp = make_text(
		torrents, "bt.path.complete_save_path");
	wxButton* csp_br = make_browser(torrents, csp);

	wxTextCtrl* isp = make_text(
		torrents, "bt.path.incomplete_save_path");
	wxButton* isp_br = make_browser(torrents, isp);
	wxCheckBox* isp_cb = make_checkbox(
		torrents, loc("options.save_incomplete_in"),
		"bt.path.incomplete_save_path_enabled",
		isp, isp_br);

	wxTextCtrl* ctfp = make_text(
		torrents, "bt.path.complete_torrent_file_path");
	wxButton* ctfp_br = make_browser(torrents, ctfp);
	wxCheckBox* ctfp_cb =  make_checkbox(
		torrents, loc("options.save_complete_torrent_file_in"),
		"bt.path.complete_torrent_file_path_enabled",
		ctfp, ctfp_br);

	wxTextCtrl* itfp = make_text(
		torrents, "bt.path.incomplete_torrent_file_path");
	wxButton* itfp_br = make_browser(torrents, itfp);
	wxCheckBox* itfp_cb = make_checkbox(
		torrents,
		loc("options.save_incomplete_torrent_file_in"),
		"bt.path.incomplete_torrent_file_path_enabled",
		itfp, itfp_br);

	wxTextCtrl* al = make_text(torrents, "bt.torrent.auto_load");
	wxButton* al_br = make_browser(torrents, al);
	wxCheckBox* al_cb = make_checkbox(
		torrents, loc("options.auto_load"),
		"bt.torrent.auto_load_enabled", al, al_br);
		
	
	// this allows for moving widgets to different lines without having
	// to change everything
	int r = 0;

	g(make_checkbox(
		torrents, loc("options.add_top"), "bt.torrent.add_top"),
		r, 0, 1, 4);
	++r;
	g(make_checkbox(
		torrents,
		locf("options.append_extension") % paths::incomplete_suffix(),
		"bt.torrent.append_extension"),
		r, 0, 1, 4);
	
	r += 2;

	g(make_label(torrents, loc("options.paths"), bold), r, 0, 1, 1);
	++r;
	g(make_label(
		torrents, loc("options.save_complete_in")), r, 0, 1, 1);
	g(isp_cb, r, 2, 1, 1);
	++r;
	g(csp, r, 0, 1, 1);
	gf(csp_br, r, 1, 1, 1);
	g(isp, r, 2, 1, 1);
	gf(isp_br, r, 3, 1, 1);

	r += 2;

	g(ctfp_cb, r, 0, 1, 1);
	g(itfp_cb, r, 2, 1, 1);
	++r;
	g(ctfp, r, 0, 1, 1);
	gf(ctfp_br, r, 1, 1, 1);
	g(itfp, r, 2, 1, 1);
	gf(itfp_br, r, 3, 1, 1);

	r += 2;

	g(al_cb, r, 0, 1, 1);
	++r;
	g(al, r, 0, 1, 1);
	g(al_br, r, 1, 1, 1);

	r += 2;

	g(make_label(torrents, loc("options.changing_paths"), bold),
		r, 0, 1, 4);

#undef gf
#undef g

	torrents->SetSizer(grid);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(torrents, 0, wxEXPAND | wxALL, 5);

	grid->AddGrowableCol(0);
	grid->AddGrowableCol(2);
	grid->SetEmptyCellSize(wxSize(10, 5));

	p->SetSizer(box);

	return p;
}

wxPanel* options_dialog::network(wxWindow* parent)
{
	wxPanel* p = new wxPanel(parent);
	wxPanel* network = new wxPanel(p);

	// again, sorry for the mess

	wxGridBagSizer* grid = new wxGridBagSizer(5, 10);

	grid->SetEmptyCellSize(wxSize(10, 5));

	// this is for widgets that expand (all of them except for spinners)
#define g(c, row, col, rs, cs)				\
			grid->Add(c,					\
				wxGBPosition(row, col),		\
				wxGBSpan(rs, cs), wxEXPAND | wxALIGN_CENTER_VERTICAL)

	// this is for widgets that don't expand (all the spinners)
#define gf(c, row, col, rs, cs)	\
			grid->Add(c, wxGBPosition(row, col), wxGBSpan(rs, cs), \
			wxALIGN_CENTER_VERTICAL)


	// some of the controls are created here so the grid bag setup
	// below at least fits on one line

	wxSpinCtrl* active_downloads = make_spinner(
		network, "bt.torrent.active_downloads");
	wxCheckBox* active_downloads_enabled = make_checkbox(
		network, loc("options.downloads_queued"),
		"bt.torrent.active_downloads_enabled", active_downloads);

	wxSpinCtrl* active_uploads = make_spinner(
		network, "bt.torrent.active_uploads");
	wxCheckBox* active_uploads_enabled = make_checkbox(
		network, loc("options.upload_queued"),
		"bt.torrent.active_uploads_enabled", active_uploads);

	wxSpinCtrl* seed_ratio = make_spinner(
		network, "bt.torrent.seed_ratio");

	wxSpinCtrl* max_dl_rate = make_spinner(
		network, "bt.torrent.max_download");
	wxCheckBox* max_dl_rate_enabled = make_checkbox(
		network, loc("options.max_download_rate"),
		"bt.torrent.max_download_enabled", max_dl_rate);

	wxSpinCtrl* max_ul_rate = make_spinner(
		network, "bt.torrent.max_upload");
	wxCheckBox* max_ul_rate_enabled = make_checkbox(
		network, loc("options.max_upload_rate"),
		"bt.torrent.max_upload_enabled", max_ul_rate);

	wxSpinCtrl* num_connections = make_spinner(
		network, "bt.network.num_connections");

	wxSpinCtrl* connections_per_torrent = make_spinner(
		network, "bt.network.connections_per_torrent");
	wxCheckBox* connections_per_torrent_enabled = make_checkbox(
		network, loc("options.connections_per_torrent"),
		"bt.network.connections_per_torrent_enabled",
		connections_per_torrent);

	std::vector<std::string> proxy_types;
	proxy_types.push_back("");
	proxy_types.push_back("proxy.socks4");
	proxy_types.push_back("proxy.socks5");
	proxy_types.push_back("proxy.http");
	
	wxTextCtrl* proxy = make_text(network, "bt.network.proxy");
	wxComboBox* proxy_type = make_combobox<std::string>(
		network, proxy_types, "bt.network.proxy_type", &loc_stringer);
	wxTextCtrl* proxy_username = make_text(
		network, "bt.network.proxy_username");
	wxTextCtrl* proxy_password = make_text(
		network, "bt.network.proxy_password");

	wxCheckBox* proxy_enabled = make_checkbox(
		network, loc("options.proxy_enabled"),
		"bt.network.proxy_enabled",
		proxy, proxy_type, proxy_username, proxy_password, bold);
	

	// this allows for moving widgets to different lines without having
	// to change everything
	int r = 0;


	// queue/speeds

	g(make_label(network, loc("options.queue"), bold), r, 0, 1, 2);
	g(make_label(network, loc("options.speeds"), bold), r, 2, 1, 2);
	
	++r;
	
	g(active_downloads_enabled, r, 0, 1, 1);
	gf(active_downloads, r, 1, 1, 1);

	g(max_dl_rate_enabled, r, 2, 1, 1);
	gf(max_dl_rate, r, 3, 1, 1);

	++r;

	g(active_uploads_enabled, r, 0, 1, 1);
	gf(active_uploads, r, 1, 1, 1);

	g(max_ul_rate_enabled, r, 2, 1, 1);
	gf(max_ul_rate, r, 3, 1, 1);

	++r;

	g(make_label(network, loc("options.stop_seed")), r, 0, 1, 1);
	gf(seed_ratio, r, 1, 1, 1);	

	r += 3;

	// connections/proxy

	g(make_label(network, loc("options.connections"), bold), r, 0, 1, 2);
	g(proxy_enabled, r, 2, 1, 2);

	++r;

	g(make_label(network, loc("options.num_connections")), r, 0, 1, 1);
	gf(num_connections, r, 1, 1, 1);

	g(make_label(network, loc("options.proxy.host")), r, 2, 1, 1);
	g(proxy, r, 3, 1, 1);

	++r;

	g(connections_per_torrent_enabled, r, 0, 1, 1);
	gf(connections_per_torrent, r, 1, 1, 1);

	g(make_label(network, loc("options.proxy.type")), r, 2, 1, 1);

	wxBoxSizer* ptb = new wxBoxSizer(wxVERTICAL);
	ptb->Add(proxy_type, 0, wxEXPAND);

	g(ptb, r, 3, 1, 1);

	++r;

	g(make_label(network, loc("options.proxy.username")), r, 2, 1, 1);
	g(proxy_username, r, 3, 1, 1);

	++r;

	g(make_label(network, loc("options.proxy.password")), r, 2, 1, 1);

	g(proxy_password, r, 3, 1, 1);

	r += 3;

	#undef g

	network->SetSizer(grid);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(network, 0, wxEXPAND | wxALL, 5);

	grid->AddGrowableCol(0);
	grid->AddGrowableCol(1);
	grid->AddGrowableCol(2);
	grid->AddGrowableCol(3);

	p->SetSizer(box);

	return p;
}

wxPanel* options_dialog::ext(wxWindow* parent)
{
	wxPanel* p = new wxPanel(parent);

	ext_.panel = new extension_panel(p, extension_panel::options_mode);
	ext_.panel->update(ext::instances());

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(ext_.panel, 1, wxEXPAND | wxALL, 5);

	p->SetSizer(root);

	return p;
}

void options_dialog::on_close(wxCommandEvent&)
{
	EndModal(wxID_CANCEL);
}


wxStaticText* options_dialog::make_label(
	wxWindow* parent, const utf8_string& name, text_flags f)
{
	wxStaticText* t = new wxStaticText(parent, -1, to_wxstring(name));

	if (f != no_text_flag)
	{
		wxFont font = t->GetFont();

		if (f & bold)
			font.SetWeight(wxFONTWEIGHT_BOLD);

		if (f & italic)
			font.SetStyle(wxFONTSTYLE_ITALIC);

		 t->SetFont(font);
	}

	return t;
}

wxTextCtrl* options_dialog::make_text(
	wxWindow* parent, const std::string& option, bool pw)
{
	long style = 0;
	if (pw)
		style |= wxTE_PASSWORD;

	wxTextCtrl* c = new wxTextCtrl(
		parent, -1, wxEmptyString,
		wxDefaultPosition, wxDefaultSize, style);

	c->SetValue(to_wxstring(opt(option)));

	// on_text() will be fired when the text changes
	c->Connect(
		wxEVT_COMMAND_TEXT_UPDATED,
		wxCommandEventHandler(options_dialog::on_text),
		new handler(c, option), this);

	return c;
}

wxCheckBox* options_dialog::make_checkbox(
	wxWindow* parent,
	const boost::format& caption, const std::string& option,
	wxWindow* buddy, wxWindow* buddy2, wxWindow* buddy3,
	wxWindow* buddy4, text_flags f)
{
	return make_checkbox(
		parent, utf8_string(caption.str()), option,
		buddy, buddy2, buddy3, buddy4, f);
}

wxCheckBox* options_dialog::make_checkbox(
	wxWindow* parent,
	const utf8_string& caption, const std::string& option,
	wxWindow* buddy, wxWindow* buddy2, wxWindow* buddy3,
	wxWindow* buddy4, text_flags f)
{
	wxCheckBox* c = new wxCheckBox(parent, -1, to_wxstring(caption));

	const bool b = opt<bool>(option);

	c->SetValue(b);

	// on_checkbox() will be fired when the checkbox is toggled
	c->Connect(
		wxEVT_COMMAND_CHECKBOX_CLICKED,
		wxCommandEventHandler(options_dialog::on_checkbox),
		new handler(c, option, buddy, buddy2, buddy3, buddy4), this);

	if (f != no_text_flag)
	{
		wxFont font = c->GetFont();

		if (f & bold)
			font.SetWeight(wxFONTWEIGHT_BOLD);

		if (f & italic)
			font.SetStyle(wxFONTSTYLE_ITALIC);

		 c->SetFont(font);
	}

	if (buddy)
		buddy->Enable(b);

	if (buddy2)
		buddy2->Enable(b);

	if (buddy3)
		buddy3->Enable(b);

	if (buddy4)
		buddy4->Enable(b);

	return c;
}

wxButton* options_dialog::make_browser(
	wxWindow* parent, wxTextCtrl* out)
{
	wxButton* b = new wxButton(
		parent, -1, loc_wx("browse"),
		wxDefaultPosition, wxDefaultSize, wxBU_EXACTFIT);

	// on_browser() will be fired when clicked
	b->Connect(
		wxEVT_COMMAND_BUTTON_CLICKED,
		wxCommandEventHandler(options_dialog::on_browser),
		new handler(b, "", out), this);

	return b;
}

wxSpinCtrl* options_dialog::make_spinner(
	wxWindow* parent, const std::string& option, int max)
{
	wxSpinCtrl* c = new wxSpinCtrl(parent);
	
	if (max == -1)
		max = std::numeric_limits<int>::max();

	c->SetRange(0, max);
	c->SetValue(opt<int>(option));

	// on_spinner() will be fired when the value changes
	c->Connect(
		wxEVT_COMMAND_SPINCTRL_UPDATED,
		wxCommandEventHandler(options_dialog::on_spinner),
		new handler(c, option), this);

	return c;
}


void options_dialog::on_text(wxCommandEvent& e)
{
	handler* h = dynamic_cast<handler*>(e.m_callbackUserData);
	bt_assert(h);

	wxTextCtrl* c = dynamic_cast<wxTextCtrl*>(h->w);
	bt_assert(c);

	new_[""][h->option] = option(
		"", from_wxstring(c->GetValue()), utf8_string());
}

void options_dialog::on_checkbox(wxCommandEvent& e)
{
	handler* h = dynamic_cast<handler*>(e.m_callbackUserData);
	bt_assert(h);

	wxCheckBox* cb = dynamic_cast<wxCheckBox*>(h->w);
	bt_assert(cb);

	new_[""][h->option].set(
		(cb->GetValue() ? utf8_string("1") : utf8_string("0")));

	// enabling/disabling buddies
	if (h->buddy)
		h->buddy->Enable(cb->GetValue());

	if (h->buddy2)
		h->buddy2->Enable(cb->GetValue());

	if (h->buddy3)
		h->buddy3->Enable(cb->GetValue());

	if (h->buddy4)
		h->buddy4->Enable(cb->GetValue());
}

void options_dialog::on_browser(wxCommandEvent& e)
{
	handler* h = dynamic_cast<handler*>(e.m_callbackUserData);
	bt_assert(h);

	// setting the buddy textbox to the selected path
	wxTextCtrl* c = dynamic_cast<wxTextCtrl*>(h->buddy);
	bt_assert(c);

	wxDirDialog dlg(
		this, loc_wx("browse.open_directory"), c->GetValue());

	if (dlg.ShowModal() != wxID_OK)
		return;

	c->SetValue(dlg.GetPath());
}

void options_dialog::on_spinner(wxCommandEvent& e)
{
	handler* h = dynamic_cast<handler*>(e.m_callbackUserData);
	bt_assert(h);

	wxSpinCtrl* c = dynamic_cast<wxSpinCtrl*>(h->w);
	bt_assert(c);

	new_[""][h->option].set(as_utf8_string(c->GetValue()));
}

void options_dialog::on_advanced_changed(std::string guid, option o)
{
	new_[guid][o.name()] = o;
}


BEGIN_EVENT_TABLE(options_dialog, wxDialog)
	EVT_BUTTON(wxID_CLOSE, options_dialog::on_close)
END_EVENT_TABLE()

}	// namespace baretorrent
