// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_OPTIONS_DIALOG_INCLUDED
#define BARETORRENT_OPTIONS_DIALOG_INCLUDED

#include "widgets.h"
#include "options.h"
#include "dialogs.h"
#include "ext_ui.h"

namespace baretorrent
{

class option_list_widget;

// allows the user to change the options; this creates a copy of the
// current options and returns it in options()
//
// each widget is paired with an option which is updated in the various
// "on change" events for those widgets; some widgets also control
// other widgets, such as checkboxes enabling/disabling textboxes
//
// widget/option pairs are stored in handlers, along with buddy widgets
// and the handler is saved as the callback user data
//
// wxWidgets doesn't seem to have any public member function for getting
// the user data, so m_callbackUserData is used directly
//
class options_dialog : public wxDialog
{
public:
	// creates the dialog and widgets
	//
	options_dialog(wxWindow* parent);

	// returns the modified options
	//
	const std::map<std::string, option_list>& options() const;

private:
	DECLARE_EVENT_TABLE()

	// used in make_label()
	//
	enum text_flags
	{
		// regular text
		no_text_flag	= 0x00,
		bold			= 0x01,
		italic			= 0x02
	};

	// a widget and its option, plus some additional widgets to enable/
	// disable if needed
	//
	struct handler : wxObject
	{
		// the widget
		wxWindow* w;

		// the option that this widget controls
		std::string option;

		// two possible buddies (some options have a textbox and a
		// button to disable)
		wxWindow *buddy, *buddy2, *buddy3, *buddy4;

		handler(
			wxWindow* w, const std::string& o,
			wxWindow* buddy=nullptr, wxWindow* buddy2=nullptr,
			wxWindow* buddy3=nullptr, wxWindow* buddy4=nullptr) :
				w(w), option(o), buddy(buddy), buddy2(buddy2),
				buddy3(buddy3), buddy4(buddy4)
		{
		}
	};

	template <class T>
	struct combobox_handler : public handler
	{
		boost::function<utf8_string (const T&)> stringer;
		boost::function<utf8_string (const T&)> value_maker;
		std::vector<T> items;

		combobox_handler(
			std::vector<T> items, wxWindow* w, const std::string& o,
			wxWindow* buddy=nullptr, wxWindow* buddy2=nullptr,
			wxWindow* buddy3=nullptr, wxWindow* buddy4=nullptr) :
				handler(w, o, buddy, buddy2, buddy3, buddy4),
				items(items)
		{
		}
	};

	// widgets in the extensions tab
	//
	struct extension_widgets
	{
		extension_panel* panel;

		extension_widgets()
		{
			std::memset(this, 0, sizeof(extension_widgets));
		}
	};


	// options that changed
	std::map<std::string, option_list> new_;

	// extension widgets
	extension_widgets ext_;

	// advanced options list
	option_list_widget* advanced_;


	// creates the general panel
	//
	wxPanel* general(wxWindow* parent);

	// creates the torrents panel
	//
	wxPanel* torrents(wxWindow* parent);

	// creates the network panel
	//
	wxPanel* network(wxWindow* parent);

	// creates the extensions panel
	//
	wxPanel* ext(wxWindow* parent);
	

	// fired when the user closes the dialog with the X button, escape,
	// etc.; cancels the dialog
	//
	void on_close(wxCommandEvent&);


	// creates a label with the given caption; this doesn't interact
	// with options
	//
	wxStaticText* make_label(
		wxWindow* parent, const utf8_string& caption,
		text_flags f = no_text_flag);

	// creates a textbox that gets its initial value from the given
	// option and saves its content whenever it changes; if pw is true,
	// this is a password textbox
	//
	wxTextCtrl* make_text(
		wxWindow* parent, const std::string& option, bool pw=false);

	// creates a checkbox that gets its initial checked value from the
	// given option and saves its content whenever it is toggled; the
	// caption must already be localized
	//
	// the two buddies (if specified) are enabled or disabled depending
	// on the checkbox state
	//
	wxCheckBox* make_checkbox(
		wxWindow* parent,
		const utf8_string& caption, const std::string& option,
		wxWindow* buddy = nullptr, wxWindow* buddy2=nullptr,
		wxWindow* buddy3 = nullptr, wxWindow* buddy4=nullptr,
		text_flags f = no_text_flag);

	wxCheckBox* make_checkbox(
		wxWindow* parent,
		const boost::format& caption, const std::string& option,
		wxWindow* buddy = nullptr, wxWindow* buddy2=nullptr,
		wxWindow* buddy3 = nullptr, wxWindow* buddy4=nullptr,
		text_flags f = no_text_flag);

	// creates a button that opens a directory selection dialog and
	// sets the value of the given textbox to the selected directory
	// (unless canceled)
	//
	wxButton* make_browser(wxWindow* parent, wxTextCtrl* out);

	// creates a combobox from a list of items
	// 
	// its initial selection is the item with the same caption as the
	// option; if there are multiple items with the same caption, the
	// first one is selected
	//
	// the option is saved with the caption of the selected item when
	// the selection changes
	//
	// a function to convert the items to utf-8 strings must be
	// specified
	//
	template <class T>
	wxComboBox* make_combobox(
		wxWindow* parent,
		const std::vector<T>& items,
		const std::string& option,
		boost::function<utf8_string (const T&)> stringer,
		boost::function<utf8_string (const T&)> value_maker =
			boost::function<utf8_string (const T&)>())
	{
		if (value_maker.empty())
			value_maker = stringer;

		wxComboBox* c = new wxComboBox(
			parent, -1, wxEmptyString,
			wxDefaultPosition, wxDefaultSize,
			0, nullptr, wxCB_DROPDOWN | wxCB_READONLY);
	
		int i=0;

		// this string will be selected if it exists
		const utf8_string o = opt(option);
		
		foreach (const T& f, items)
		{
			const utf8_string s = stringer(f);
			const utf8_string v = value_maker(f);

			c->Append(to_wxstring(s));

			// makes sure the first item is selected if there are
			// more than one with the same caption
			if (c->GetSelection() == -1 && v == o)
			{
				// same caption, this is the selected item
				c->SetSelection(i);
			}

			++i;
		}

		// if the option's value was not in the item list, the first
		// item is selected
		if (c->GetSelection() == -1)
			c->SetSelection(0);

		// on_combobox() will be called when the selection changes
		combobox_handler<T>* ch =
			new combobox_handler<T>(items, c, option);

		ch->stringer = stringer;
		ch->value_maker = value_maker;

		// http://gcc.gnu.org/bugzilla/show_bug.cgi?id=11407
		typedef void (options_dialog::*handler)(wxCommandEvent&);

		handler h = &options_dialog::on_combobox<T>;
		wxCommandEventFunction f = (wxCommandEventFunction)h;

		c->Connect(
			wxEVT_COMMAND_COMBOBOX_SELECTED,
			wxObjectEventFunction(f), ch, this);

		return c;
	}

	// creates a spinner that gets its initial value from the given
	// option converted to an int and saves its content whenever it
	// changes; if max is -1, there is no limit
	//
	wxSpinCtrl* make_spinner(
		wxWindow* parent, const std::string& options, int max=-1);


	// fired when the content of a textbox changes; updates the option
	// associated with it
	//
	void on_text(wxCommandEvent& e);

	// fired when the selection of a combobox changes; updates the
	// option associated with it
	//
	void on_checkbox(wxCommandEvent& e);

	// opens the directory selection dialog and sets the value of its
	// buddy textbox and the option associated with it
	//
	void on_browser(wxCommandEvent&);

	// fired when the spinner is clicked; updates the option associated
	// with it with the integer value converted to a string
	//
	void on_spinner(wxCommandEvent&);

	// fired when the combobox selection changes; updates the option
	// associated with it with the caption of the selected item
	//
	template <class T>
	void on_combobox(wxCommandEvent& e)
	{
		combobox_handler<T>* h = dynamic_cast<combobox_handler<T>*>(
			e.m_callbackUserData);
		bt_assert(h);

		wxComboBox* cb = dynamic_cast<wxComboBox*>(h->w);
		bt_assert(cb);

		const std::size_t i = cb->GetSelection();
		if (i < h->items.size())
			new_[""][h->option].set(h->value_maker(h->items[i]));
	}


	// fired from the options_list_widget when an advanced option
	// has been changed
	//
	void on_advanced_changed(std::string guid, option o);
};

}	// namespace baretorrent

#endif	// BARETORRENT_OPTIONS_DIALOG_INCLUDED
