// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "path.h"
#include "locale.h"
#include "utility.h"

namespace fs = boost::filesystem;

namespace baretorrent
{

#if BARETORRENT_TARGET_IS_UTF16
fs::path string_to_path(const utf8_string& p)
{
	try
	{
		return fs::path(utf8_to_utf16(p).std_string());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("string to path failed"), e.code());
	}
}

utf8_string path_to_utf8_string(const fs::path& p)
{
	try
	{
		return utf16_to_utf8(utf16_string(p.native()));
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("path to utf8 string failed"), e.code());
	}
}

utf16_string path_to_native_string(const fs::path& p)
{
	try
	{
		return utf16_string(p.native());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("path to native string failed"), e.code());
	}
}

#else
fs::path string_to_path(const utf8_string& p)
{
	return p.std_string();
}

utf8_string path_to_utf8_string(const fs::path& p)
{
	try
	{
		return utf8_string(p.native());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("path to utf8 string failed"), e.code());
	}
}

utf8_string path_to_native_string(const fs::path& p)
{
	try
	{
		return utf8_string(p.native());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("path to native string failed"), e.code());
	}
}
#endif


#if BARETORRENT_VISUALC
#pragma warning(push)
#pragma warning(disable: 4245)	// signed/unsigned mismatch
#endif

template <>
const basic_utf_string<char, utf8_encoding>::size_type
basic_utf_string<char, utf8_encoding>::npos(-1);

template <>
const basic_utf_string<wchar_t, utf16_encoding>::size_type
basic_utf_string<wchar_t, utf16_encoding>::npos(-1);

template <>
const basic_utf_string<wchar_t, utf32_encoding>::size_type
basic_utf_string<wchar_t, utf32_encoding>::npos(-1);

#if BARETORRENT_VISUALC
#pragma warning(pop)
#endif

utf8_path::utf8_path()
{
}

utf8_path::utf8_path(const boost::filesystem::path& p)
	: what_(p)
{
}

utf8_path::utf8_path(const utf8_string& s)
	: what_(string_to_path(s))
{
}

utf8_path utf8_path::internal_create(const fs::path& p)
{
	return utf8_path(p);
}

utf8_path& utf8_path::operator=(const utf8_string& s)
{
	what_ = string_to_path(s);
	return *this;
}

bool utf8_path::exists() const
{
	try
	{
		return fs::exists(what_);
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not check if path exists"),
			*this, utf8_path(), e.code());
	}
}

void utf8_path::remove() const
{
	try
	{
		fs::remove(what_);
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not remove file"),
			*this, utf8_path(), e.code());
	}
}

void utf8_path::remove(boost::system::error_code& ec) const
{
	fs::remove(what_, ec);
}

void utf8_path::remove_all() const
{
	try
	{
		fs::remove_all(what_);
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not remove all"),
			*this, utf8_path(), e.code());
	}
}

void utf8_path::remove_all(boost::system::error_code& ec) const
{
	fs::remove_all(what_, ec);
}

void utf8_path::rename(
	const utf8_path& new_path, boost::system::error_code& ec)
{
	fs::rename(what_, new_path.what_, ec);
}

bool utf8_path::empty() const
{
	return what_.empty();
}

utf8_string utf8_path::as_utf8_string() const
{
	return path_to_utf8_string(what_);
}

native_string_t utf8_path::native() const
{
	return path_to_native_string(what_);
}

utf8_string utf8_path::stem() const
{
	try
	{
		return path_to_utf8_string(what_.stem());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not get stem"),
			*this, utf8_path(), e.code());
	}
}

utf8_string utf8_path::extension() const
{
	try
	{
		return path_to_utf8_string(what_.extension());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not get extension"),
			*this, utf8_path(), e.code());
	}
}

bool utf8_path::is_directory() const
{
	try
	{
		return fs::is_directory(what_);
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("is directory failed"),
			*this, utf8_path(), e.code());
	}
}

bool utf8_path::is_regular_file() const
{
	try
	{
		return fs::is_regular_file(what_);
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("is regular file failed"),
			*this, utf8_path(), e.code());
	}
}

uintmax_t utf8_path::file_size() const
{
	try
	{
		return fs::file_size(what_);
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not get flie size"),
			*this, utf8_path(), e.code());
	}
}

void utf8_path::copy_file(const utf8_path& to) const
{
	try
	{
		fs::copy_file(what_, to.what_);
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not copy file"),
			*this, to, e.code());
	}
}

utf8_path utf8_path::parent() const
{
	try
	{
		return utf8_path(what_.parent_path());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not get parent path"),
			*this, utf8_path(), e.code());
	}
}

utf8_string utf8_path::leaf() const
{
	try
	{
		return path_to_utf8_string(what_.leaf());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not get leaf"),
			*this, utf8_path(), e.code());
	}
}

fs::path utf8_path::internal_fs_path() const
{
	return what_;
}


utf8_path current_path()
{
	try
	{
		return utf8_path::internal_create(fs::current_path());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not get current path"), e.code());
	}
}

void current_path(const utf8_path& p, boost::system::error_code& ec)
{
	try
	{
		fs::current_path(p.internal_fs_path(), ec);
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not set current path"),
			p, utf8_path(), e.code());
	}
}

utf8_path absolute(const utf8_path& p)
{
	try
	{
		return utf8_path::internal_create(
			fs::absolute(p.internal_fs_path()));
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not get absolute path"),
			p, utf8_path(), e.code());
	}
}

utf8_path unique_path()
{
	try
	{
		return utf8_path::internal_create(fs::unique_path());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not get unique path"), e.code());
	}
}

utf8_path temp_directory_path()
{
	try
	{
		return utf8_path::internal_create(fs::temp_directory_path());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not get temp directory path"), e.code());
	}
}

bool operator==(const utf8_path& left, const utf8_path& right)
{
	return left.internal_fs_path() == right.internal_fs_path();
}

bool operator!=(const utf8_path& left, const utf8_path& right)
{
	return !(left == right);
}

std::ostream& operator<<(std::ostream& out, const utf8_path& s)
{
	out << s.as_utf8_string();
	return out;
}

utf8_path operator/(const utf8_path& left, const utf8_path& right)
{
	return utf8_path::internal_create(
		left.internal_fs_path() / right.internal_fs_path());
}

utf8_path operator/(const utf8_path& left, const utf8_string& right)
{
	return utf8_path::internal_create(
		left.internal_fs_path() / string_to_path(right));
}

void create_directories(
	const utf8_path& p, boost::system::error_code& ec)
{
	fs::create_directories(p.internal_fs_path(), ec);
}

void create_directories(const utf8_path& p)
{
	try
	{
		fs::create_directories(p.internal_fs_path());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not create directories"),
			p, utf8_path(), e.code());
	}
}

void create_directory(const utf8_path& p)
{
	try
	{
		fs::create_directory(p.internal_fs_path());
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not create directory"),
			p, utf8_path(), e.code());
	}
}

void walk_directory(
	const utf8_path& root_p, boost::function<bool (utf8_path)> f)
{
	try
	{
		const fs::path root = root_p.internal_fs_path();

		fs::directory_iterator itor(root), end;
	
		// not recursive
		for (; itor != end; ++itor)
		{
			boost::this_thread::interruption_point();

			const fs::path pp = *itor;
			utf8_path p(path_to_utf8_string(pp));

			if (!f(p))
				break;
		}
	}
	catch(fs::filesystem_error& e)
	{
		throw path_exception(
			utf8_string("could not walk directory"),
			root_p, utf8_path(), e.code());
	}
}


std::ostream& operator<<(std::ostream& out, const utf8_string& s)
{
	out << s.std_string();
	return out;
}


path_exception::path_exception(const utf8_string& s)
	: what_(s)
{
}

path_exception::path_exception(
	const utf8_string& s, boost::system::error_code ec)
{
	make(s, utf8_path(), utf8_path(), ec);
}

path_exception::path_exception(
	const utf8_string& what,
	const utf8_path& from, const utf8_path& to,
	const boost::system::error_code& ec)
{
	make(what, from, to, ec);
}

void path_exception::make(
	const utf8_string& what,
	const utf8_path& from, const utf8_path& to,
	const boost::system::error_code& ec)
{
	std::ostringstream oss;

	oss << what;

	if (!from.empty() || !to.empty())
		oss << "(";

	if (!from.empty())
		oss << "'" << from.as_utf8_string() << "'";

	if (!to.empty())
	{
		if (!from.empty())
			oss << "and ";

		oss << "'" << to.as_utf8_string() << "'";
	}
	
	if (!from.empty() || !to.empty())
		oss << ")";

	oss << ": " << ec.message();

	what_ = utf8_string(oss.str());
}

utf8_string path_exception::what() const
{
	return what_;
}


byte_array::byte_array()
{
}

byte_array::byte_array(const std::string& data)
	: data_(data)
{
}

byte_array::byte_array(const unsigned char* p, std::size_t s)
	: data_(p, p + s)
{
}

const char* byte_array::data() const
{
	bt_assert(!empty());
	return data_.data();
}

std::size_t byte_array::size() const
{
	return data_.size();
}

bool byte_array::empty() const
{
	return data_.empty();
}

}	// namespace baretorrent
