// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_PATH_INCLUDED
#define BARETORRENT_PATH_INCLUDED

namespace baretorrent
{

enum encodings
{
	utf8_encoding = 1,
	utf16_encoding,
	utf32_encoding,
	current_locale_encoding
};

template <class C, encodings e>
class basic_utf_string
{
	typedef std::basic_string<C> impl_string;

public:
	typedef C char_t;
	typedef typename impl_string::size_type size_type;
	typedef typename impl_string::iterator iterator;
	typedef typename impl_string::const_iterator const_iterator;

	static const size_type npos;
	static const encodings encoding = e;


	basic_utf_string()
	{
	}

	explicit basic_utf_string(const impl_string& s)
		: what_(s)
	{
	}

	explicit basic_utf_string(const char_t* s)
		: what_(s)
	{
	}

	basic_utf_string(const char_t* c, size_type s)
		: what_(c, s)
	{
	}

	basic_utf_string(size_type n, char_t c)
		: what_(n, c)
	{
	}


	impl_string std_string() const
	{
		return impl_string(begin(), end());
	}


	iterator begin()
	{
		return what_.begin();
	}

	iterator end()
	{
		return what_.end();
	}

	const_iterator begin() const
	{
		return what_.begin();
	}

	const_iterator end() const
	{
		return what_.end();
	}


	size_type size() const
	{
		return what_.size();
	}

	size_type length() const
	{
		return what_.length();
	}

	size_type size_in_bytes() const
	{
		return size() * sizeof(char_t);
	}

	void clear()
	{
		what_.clear();
	}

	bool empty() const
	{
		return what_.empty();
	}

	const char_t* c_str() const
	{
		return what_.c_str();
	}


	const char_t& operator[](size_type p) const
	{
		return what_[p];
	}

	char_t& operator[](size_type p)
	{
		return what_[p];
	}


	basic_utf_string& insert(size_type p, const basic_utf_string& s)
	{
		what_.insert(p, s.what_);
		return *this;
	}

	template <class II>
	void insert(iterator p, II first, II last)
	{
		what_.insert(p, first, last);
	}

	basic_utf_string& append(size_type n, char_t c)
	{
		what_.append(n, c);
		return *this;
	}

	template <class II>
	basic_utf_string& append(II first, II last)
	{
		what_.append(first, last);
		return *this;
	}

	basic_utf_string& erase(size_type off, size_type n = npos)
	{
		what_.erase(off, n);
		return *this;
	}

	iterator erase(iterator p)
	{
		return what_.erase(p);
	}

	iterator erase(iterator first, iterator last)
	{
		return what_.erase(first, last);
	}


	basic_utf_string substr(size_type off, size_type n = npos) const
	{
		return basic_utf_string(what_.substr(off, n));
	}


	size_type find(const basic_utf_string& s, size_type off = 0) const
	{
		return what_.find(s.what_, off);
	}
	
	size_type find_first_of(
		const basic_utf_string& s, size_type off) const
	{
		return what_.find_first_of(s.what_, off);
	}


	size_type copy(char_t* s, size_type n, size_type off=0) const
	{
		return what_.copy(s, n, off);
	}

	basic_utf_string& operator+=(const basic_utf_string& s)
	{
		what_ += s.what_;
		return *this;
	}

private:
	impl_string what_;
};

typedef basic_utf_string<char, utf8_encoding> utf8_string;
typedef basic_utf_string<wchar_t, utf16_encoding> utf16_string;
typedef basic_utf_string<wchar_t, utf32_encoding> utf32_string;

std::ostream& operator<<(std::ostream& out, const utf8_string& s);


template <class T, encodings e>
inline bool operator==(
	const basic_utf_string<T, e>& left,
	const basic_utf_string<T, e>& right)
{
	return left.std_string() == right.std_string();
}

template <class T, encodings e>
inline bool operator!=(
	const basic_utf_string<T, e>& left,
	const basic_utf_string<T, e>& right)
{
	return !(left == right);
}

template <class T, encodings e>
inline basic_utf_string<T, e> operator+(
	const basic_utf_string<T, e>& left,
	const basic_utf_string<T, e>& right)
{
	return basic_utf_string<T, e>(
		left.std_string() + right.std_string());
}

template <class T, encodings e>
inline bool operator<(
	const basic_utf_string<T, e>& left,
	const basic_utf_string<T, e>& right)
{
	return left.std_string() < right.std_string();
}


class byte_array
{
public:
	byte_array();
	explicit byte_array(const std::string& data);
	byte_array(const unsigned char* p, std::size_t s);

	const char* data() const;
	std::size_t size() const;
	bool empty() const;

private:
	std::string data_;
};


#if BARETORRENT_POSIX
	typedef utf8_string native_string_t;
#else
	typedef utf16_string native_string_t;
#endif

class utf8_path
{
public:
	static utf8_path internal_create(const boost::filesystem::path& p);

	utf8_path();
	explicit utf8_path(const utf8_string& s);

	utf8_path& operator=(const utf8_string& s);

	bool exists() const;
	void remove() const;
	void remove(boost::system::error_code& ec) const;
	void remove_all() const;
	void remove_all(boost::system::error_code& ec) const;
	void rename(const utf8_path& new_path, boost::system::error_code& ec);

	bool empty() const;

	utf8_string as_utf8_string() const;
	native_string_t native() const;
	utf8_string stem() const;

	utf8_string extension() const;
	
	bool is_directory() const;
	bool is_regular_file() const;
	uintmax_t file_size() const;
	void copy_file(const utf8_path& to) const;

	utf8_path parent() const;
	utf8_string leaf() const;

	boost::filesystem::path internal_fs_path() const;

private:
	boost::filesystem::path what_;

	utf8_path(const boost::filesystem::path& p);
};

void walk_directory(
	const utf8_path& p, boost::function<bool (utf8_path)> f);

utf8_path current_path();
void current_path(const utf8_path& p, boost::system::error_code& ec);
utf8_path absolute(const utf8_path& p);
utf8_path unique_path();
utf8_path temp_directory_path();

bool operator==(const utf8_path& left, const utf8_path& right);
bool operator!=(const utf8_path& left, const utf8_path& right);

std::ostream& operator<<(std::ostream& out, const utf8_path& s);

utf8_path operator/(const utf8_path& left, const utf8_path& right);
utf8_path operator/(const utf8_path& left, const utf8_string& right);

void create_directories(const utf8_path& p);
void create_directories(
	const utf8_path& p, boost::system::error_code& ec);

void create_directory(const utf8_path& p);

class path_exception
{
public:
	path_exception(const utf8_string& s = utf8_string());
	path_exception(const utf8_string& s, boost::system::error_code ec);
	path_exception(
		const utf8_string& what,
		const utf8_path& from, const utf8_path& to,
		const boost::system::error_code& ec);

	utf8_string what() const;

private:
	utf8_string what_;

	void make(
		const utf8_string& what,
		const utf8_path& from, const utf8_path& to,
		const boost::system::error_code& ec);
};

}	// namespace baretorrent

#endif	// BARETORRENT_PATH_INCLUDED
