// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "platform.h"
#include "log.h"

#if BARETORRENT_POSIX
	#include <sys/types.h>
	#include <pwd.h>
	#include "app.h"
#endif

namespace baretorrent
{
namespace platform
{

// returns the name of the platform on which baretorrent was compiled
//
utf8_string compile_platform()
{
#if BARETORRENT_WINDOWS
	return utf8_string("Windows");
#elif BARETORRENT_APPLE
	return utf8_string("Apple");
#elif BARETORRENT_LINUX
	return utf8_string("Linux");
#else
	#error unrecognized platform
#endif
}

// returns the compiler name and version
//
utf8_string compiler()
{
	std::ostringstream oss;

#if BARETORRENT_VISUALC
	// uses the compiler version instead of the ide
	oss << "Microsoft Visual C++ " << _MSC_VER;

#elif BARETORRENT_GCC
	oss
	   << "gcc "
	   << __GNUC__ << "." << __GNUC_MINOR__ << "."
	   << __GNUC_PATCHLEVEL__;

#else
	#error unrecognized compiler

#endif

	return utf8_string(oss.str());
}

// returns the architecture (x86 or x64)
//
utf8_string architecture()
{
	utf8_string s;

#if BARETORRENT_X86
	s = utf8_string("x86");
#elif BARETORRENT_X64
	s = utf8_string("x64");
#else
	#error unrecognized architecture
#endif

#if BARETORRENT_WINDOWS && !BARETORRENT_VISTA_PLUS
	s += utf8_string(" XP");
#endif

	return s;
}

// returns either "(debug)" or "(release)"
//
utf8_string compile_variant()
{
#if BARETORRENT_DEBUG
	return utf8_string("(debug)");
#else
	return utf8_string("(release)");
#endif
}

utf8_string compile_target()
{
	std::ostringstream oss;
	oss
		<< compile_platform() << ", " << compiler() << " "
		<< architecture() << " " << compile_variant();

	return utf8_string(oss.str());
}

#if BARETORRENT_POSIX

bool executable_exists(const utf8_string& cmd)
{
	const int sr = system(cmd.c_str());
	const int r = WEXITSTATUS(sr);

	if (r == 0)
	{
		log() << "checking for `" << cmd << "`... okay (" << r << ")\n";
		return true;
	}
	else
	{
		log() << "checking for `" << cmd << "`... failed (" << r << ")\n";
		return false;
	}
}

bool has_xdg()
{
	const utf8_string c("which xdg-mime");
	return executable_exists(c);
}

utf8_string shell_output(const utf8_string& c)
{
	FILE* f = popen(c.c_str(), "r");
	if (!f)
		return utf8_string("");

	char line[100] = {0};
	fgets(line, sizeof(line) - 1, f);

	fclose(f);

	utf8_string s(line);
	boost::trim(s);

	return s;
}

// translates errno to a system message
//
utf8_string system_message(long r)
{
	return utf8_string(std::strerror(r));
}

void signal_handler(int s)
{
	if (s == SIGINT)
	{
		log() << "sigint, quitting\n";
		quit();
	}
}

void set_signals()
{
	signal(SIGINT, &signal_handler);
}

#endif	// BARETORRENT_POSIX

std::vector<std::pair<utf8_string, utf8_string> > versions()
{
	std::vector<std::pair<utf8_string, utf8_string> > v;

	utf8_string b(BOOST_LIB_VERSION);
	boost::replace_all(b, utf8_string("_"), utf8_string("."));

	std::ostringstream iconv;

#ifdef _LIBICONV_VERSION
	int iconv_minor = (_LIBICONV_VERSION & 0xff);
	int iconv_major = ((_LIBICONV_VERSION >> 8) & 0xff);

	iconv << iconv_major << "." << iconv_minor;
#else
	iconv << "?";
#endif

#define a(a, b) v.push_back( \
			std::make_pair(utf8_string(a), utf8_string(b)))

	a("boost", b);
	a("wxWidgets", wxVERSION_NUM_DOT_STRING);
	a("libtorrent", LIBTORRENT_VERSION);
	a("curl", LIBCURL_VERSION);
	a("libxml", LIBXML_DOTTED_VERSION);
	a("openssl", SHLIB_VERSION_NUMBER);
	a("python", PY_VERSION);
	a("sqlite", SQLITE_VERSION);
	a("iconv", iconv.str());

#undef a

	return v;
}

}	// namespace platform
}	// namespace baretorrent
