// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_PLATFORM_INCLUDED
#define BARETORRENT_PLATFORM_INCLUDED

#include "resources.h"
#include "log.h"
#include "utility.h"

namespace baretorrent
{
namespace platform
{

// these are various platform-specific functions (such as getting
// known directories) or functions that may be generic but relate to
// the platform (such as the compiler name)

// thrown when something went wrong in the platform-specific stuff
//
class exception
{
public:
	explicit exception(const utf8_string& s=utf8_string())
		: what_(s)
	{
	}

	const utf8_string& what() const
	{
		return what_;
	}

private:
	utf8_string what_;
};

// platform-specific object that will be created when the program
// starts and destroyed after everything has closed
//
struct platform : boost::noncopyable
{
	platform();
	~platform();
};


// some of the default styling of wxwidgets is not native to the
// platform; this is called every time the appropriate control is
// created so its theme looks alright

// returns the the native style of a toolbar
//
long toolbar_style();

// some platforms need to do stuff when the list is created, such as
// gtk for sort indicators
//
void init_list_control(wxListCtrl* w);

// some platforms need to do stuff when the list is destroyed, such
// gtk for sort indicators; note that this doesn't delete the list
// itself
//
void destroy_list_control(wxListCtrl* w);

// sets the given list so it looks fine on the current platform
//
void set_list_control_style(wxListCtrl*);

// sets the given tree so it looks fine on the current platform
//
void set_tree_control_style(wxTreeCtrl*);

// windows in particular has native up/down arrows on column headers;
// this sets them
//
void set_list_sort_column(wxListCtrl* list, std::size_t i, int dir);



// these are various filesystem utilities

// opens the path in whatever graphical program the platform uses
// (explorer, nautilus, finder, etc.)
//
void open_directory(const utf8_path& p);

// executes the given files
//
void open_file(const utf8_path& p);

// returns the preferred download directory for the current user; this
// may be something like /Users/me/Downloads or /home/me/Downloads
//
utf8_path download_directory();

// returns the preferred configuration directory for the current user;
// this may be something like /home/me or /Users/me/AppData/Roaming
//
utf8_path conf_directory();

// returns the filename of the configuration directory that will be
// created inside of conf_directory(); the directory might have a dot
// as a prefix so it's hidden on some platforms
//
utf8_string conf_directory_name();

// returns the directory of the executable
//
utf8_path executable_directory();


struct assoc
{
	bool okay;
	utf8_string name;

	assoc()
		: okay(false)
	{
	}

	explicit assoc(bool b, const utf8_string& n = utf8_string())
		: okay(b), name(n)
	{
	}
};

// returns whether the file and scheme associations are okay
//
assoc associated();

// sets file and scheme associations
//
void associate();


// not all platforms handle image resources the same, nor do they have
// the same means of getting system icons

// adds the image handlers needed by wxwidgets on this platform; apple
// in particular needs tiff
//
void add_image_handlers();

// creates the bitmap corresponding to the given image
//
wxBitmap make_image(images::image i);

// creates the icon corresponding to the given image
//
wxIcon make_icon(images::image i);

// calls make_image() but returns a bad wxBitmap in case of failure
// instead of the default bad image
//
wxBitmap make_image_unsafe(images::image i);

// returns a timestamp as a string using the given format
//
utf8_string make_time(const utf8_string& format, std::time_t s);


utf8_string get_env(const utf8_string& name);
void set_env(const utf8_string& name, const utf8_string& value);
void add_to_path(const utf8_path& p);

// these are miscellaneous platform-specific calls

// translates the error number to a localized system message; on posix,
// this is expected to be errno, while Windows wants an HRESULT
//
utf8_string system_message(long r);

// adds an exception for the ports used by libtorrent to the system
// firewall (if any)
//
void add_firewall_exception();

// sets the signal stuff for this platform
// 
void set_signals();

void power(power_actions t);
bool has_power(power_actions t);
bool has_power_management();


#if BARETORRENT_POSIX

// executes the given command in the shell and returns its output; this
// is shared between apple and linux implementations and uses popen()
//
utf8_string shell_output(const utf8_string& c);

bool executable_exists(const utf8_string& cmd);

// returns whether xdg is available on this system
//
bool has_xdg();

#endif	// BARETORRENT_POSIX

// returns a list of all the different components used by baretorrent
// along with their compile-time version
//
std::vector<std::pair<utf8_string, utf8_string> > versions();

// meh, this is for windows: baretorrent is compiled as a console
// application mainly so it can use the console from which it is called
// instead of detaching itself automatically and creating a new one
//
// the problem is that without -v, the console needs to disappear, which
// is what detach_console() does here
//
void detach_console();

// again, for windows: recreates the console if -v is passed through
// ipc and the console had been detached earlier
//
void create_console();

// logs the given string to a stream that outputs to the console; some
// platforms may change the color of the text depending on the log level
// for example
//
void log_to_console(
	std::ostream* out, log_t::levels lv, const utf8_string& s);

// the target for which the program was compiled (such as "Microsoft
// Visual C++ 1700 x64")
//
utf8_string compile_target();

// the name of the platform on which the program is running (such as
// "Microsoft Windows 8 Service Pack 1 (build 7601) 64-bit")
//
utf8_string name();

}	// namespace platform
}	// namespace baretorrent

#endif	// BARETORRENT_PLATFORM_INCLUDED
