// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "python_console.h"
#include "ext/ext.h"
#include "app.h"
#include "main_window.h"

namespace baretorrent
{

std::size_t python_console::no_history =
	std::numeric_limits<std::size_t>::max();

python_console::python_console(wxWindow* parent) :
	custom_dialog(parent, loc("pcon.title")),
	extensions_(nullptr), out_(nullptr), in_(nullptr),
	indicator_(nullptr), more_(false),
	current_history_(no_history)
{
	wxBoxSizer* top = new wxBoxSizer(wxHORIZONTAL);

	extensions_ = new wxComboBox(
		this, -1, wxEmptyString, wxDefaultPosition, wxDefaultSize,
		0, nullptr, wxCB_READONLY);

	extensions_->Connect(
		wxEVT_COMBOBOX,
		wxCommandEventHandler(python_console::on_ext_selected),
		nullptr, this);

	top->Add(
		new wxStaticText(this, -1, loc_wx("pcon.extensions")),
		0, wxALIGN_CENTER_VERTICAL);
	top->Add(extensions_);

	wxTextAttr attr;

	out_ = new multiline_text(this, -1, "", wxTE_READONLY);

	attr.SetFontFamily(wxFONTFAMILY_TELETYPE);
	out_->SetDefaultStyle(attr);

	in_ = new wxTextCtrl(
		this, -1, "", wxDefaultPosition, wxDefaultSize,
		wxTE_RICH2 | wxTE_PROCESS_ENTER);

	attr.SetFontFamily(wxFONTFAMILY_TELETYPE);
	in_->SetDefaultStyle(attr);

	in_->Connect(
		wxEVT_TEXT_ENTER,
		wxCommandEventHandler(python_console::on_input),
		nullptr, this);

	in_->Connect(
		wxEVT_KEY_DOWN,
		wxKeyEventHandler(python_console::on_input_key),
		nullptr, this);

	indicator_ = new wxStaticText(this, -1, ">>> ");
	
	wxFont f = indicator_->GetFont();
	f.SetFamily(wxFONTFAMILY_TELETYPE);
	indicator_->SetFont(f);

	wxBoxSizer* bottom = new wxBoxSizer(wxHORIZONTAL);
	bottom->Add(indicator_, 0, wxALIGN_CENTER_VERTICAL);
	bottom->Add(in_, 1, wxEXPAND);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(top, 0, wxEXPAND);
	box->Add(out_, 1, wxEXPAND);
	box->Add(bottom, 0, wxEXPAND);

	SetSizer(box);

	refresh_extensions();

	gcons_.add(g_std_log.flushed.connect(boost::bind(
			&python_console::on_log, this, _1)));
}

python_console::~python_console()
{
	ext::python_lock lock;
	console_ = bpl::object();
}

void python_console::refresh_extensions()
{
	extensions_->Clear();

	ext::instance_list is = ext::instances();
	foreach (const ext::instance_ptr i, is)
		extensions_->Append(to_wxstring(i->name()), new ext_info(i));

	if (!is.empty())
	{
		extensions_->SetSelection(0);
		set_ext(is[0]);
	}
}

void python_console::on_ext_selected(wxCommandEvent&)
{
	const int i = extensions_->GetSelection();
	if (i == wxNOT_FOUND)
		return;

	const ext_info* ei = dynamic_cast<const ext_info*>(
		extensions_->GetClientObject(i));
	if (!ei)
		return;

	set_ext(ei->i);
	out_->SetValue(wxEmptyString);
}

void python_console::set_ext(ext::instance_ptr i)
{
	bt_assert(i.get());
	if (!i)
		return;

	ext_ = i;

	econs_.clear();
	econs_.add(ext_->loaded.connect(boost::bind(
		&python_console::on_ext_loaded, this)));

	create_interactive_console(i);
}

void python_console::on_ext_loaded()
{
	create_interactive_console(ext_);
}

void python_console::create_interactive_console(ext::instance_ptr i)
{
	ext::python_lock lock;

	try
	{
		bpl::object mod = i->module();

		if (!mod)
		{
			log_error(i)
				<< "extension has not been loaded properly; no "
				<< "interactive console available\n";
			return;
		}

		bpl::object dict = mod.attr("__dict__");
	
		bpl::exec("dir()", dict, dict);
		bpl::exec("import code", dict, dict);

		bpl::object code = dict["code"];
		bpl::object ic = code.attr("InteractiveConsole");

		console_ = ic(dict);
		bt_assert(console_);
	}
	catch(bpl::error_already_set&)
	{
		log_error(i)
			<< "could not create interactive console, "
			<< ext::python_error() << "\n";

		PyErr_Clear();
	}
}

void python_console::on_log(const log_t::entry& e)
{
	get_app().window().async_call(boost::bind(
		&python_console::on_log_impl, this, e));
}

void python_console::on_log_impl(const log_t::entry& e)
{
	if (ext_ && (ext_ == e.i))
		out_->AppendText(to_wxstring(e.s) + "\n");
}

void python_console::on_input_key(wxKeyEvent& e)
{
	const std::size_t prev = current_history_;

	if (e.GetKeyCode() == WXK_UP)
	{
		if (current_history_ == no_history)
		{
			if (!history_.empty())
				current_history_ = history_.size() - 1;
		}
		else if (current_history_ > 0)
		{
			--current_history_;
		}
	}
	else if (e.GetKeyCode() == WXK_DOWN)
	{
		if (current_history_ != no_history)
		{
			if (current_history_ < (history_.size() - 1))
				++current_history_;
			else
				current_history_ = no_history;
		}
	}
	else
	{
		e.Skip();
		return;
	}

	if (prev != current_history_)
	{
		if (current_history_ == no_history)
		{
			in_->SetValue(wxEmptyString);
		}
		else
		{
			bt_assert(current_history_ < history_.size());

			const wxString s = to_wxstring(history_[current_history_]);
			in_->SetValue(s);
			in_->SetSelection(
				static_cast<long>(s.Length()),
				static_cast<long>(s.Length()));
		}
	}
}

void python_console::on_input(wxCommandEvent&)
{
	std::string s = in_->GetValue().ToStdString();
	boost::trim(s);

	if (s.empty())
		return;

	const utf8_string line(s);

	add_history(line);
	output_command(line);
	execute(line);

	if (more_)
		indicator_->SetLabel("... ");
	else
		indicator_->SetLabel(">>> ");

	clear_input();
	current_history_ = no_history;
}

void python_console::clear_input()
{
	in_->SetValue("");

	wxTextAttr attr;
	attr.SetFontFamily(wxFONTFAMILY_TELETYPE);
	in_->SetDefaultStyle(attr);

}

void python_console::add_history(const utf8_string& s)
{
	const int max_history = opt<int>("pcon.max_history");
	if (max_history == 0)
		return;

	history_.push_back(s);

	if (max_history > 0)
	{
		while (history_.size() > max_history)
			history_.pop_front();
	}
}

void python_console::output_command(const utf8_string& s)
{
	if (more_)
		out_->AppendText("... ");
	else
		out_->AppendText(">>> ");

	out_->AppendText(to_wxstring(s) + "\n");
}

void python_console::execute(const utf8_string &s)
{
	try
	{
		ext::instance_ptr i =
			ext::find("4A73F6F6-5010-4F32-BB67-4BB2994CC7D7");
		bt_assert(i.get());

		if (!i)
			return;

		if (!more_ && execute_builtin(s))
			return;

		ext::python_lock lock;

		if (!console_)
		{
			log_error(ext_) << "no interactive console\n";
			return;
		}

		ext::scoped_output_redirector sor;
		bpl::object r = console_.attr("push")(s.c_str());
		sor.stop();

		more_ = handle_result(r, sor.out(), sor.err());
	}
	catch(ext::exception&)
	{
	}
	catch(bpl::error_already_set&)
	{
		log_error(ext_) << ext::python_error() << "\n";
		PyErr_Clear();
	}
}

bool python_console::execute_builtin(utf8_string s)
{
	if (s[0] != ':')
		return false;

	s = s.substr(1);

	utf8_string cmd;

	std::string::size_type p = s.find(utf8_string(" "));
	if (p == std::string::npos)
	{
		cmd = s;
	}
	else
	{
		cmd = s.substr(0, p);
		s = s.substr(p + 1);
	}

	if (cmd == utf8_string("reload"))
		ext_->reload();
	else
		log_error(ext_) << "unrecognized command '" << cmd << "'\n";

	return true;
}

bool python_console::handle_result(
	bpl::object r, const utf8_string& out, const utf8_string& err)
{
	if (!out.empty())
		log(ext_) << out;

	if (!err.empty())
		log(ext_) << err;

	const utf8_string pe = ext::python_error();
	if (!pe.empty())
	{
		log_error(ext_) << pe << "\n";
		PyErr_Clear();
		return false;
	}
	else
	{
		return bpl::extract<bool>(r);
	}
}

}	// namespace baretorrent
