// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "torrent.h"
#include "utility.h"
#include "manager.h"
#include "locale.h"
#include "options.h"
#include "ext/ext.h"

namespace baretorrent
{

// time to wait after the torrent is added before starting the stalled
// check
const boost::posix_time::seconds initial_stalled_timer_delay(30);

// time to wait once the initial stalled timer delay has passed
const boost::posix_time::seconds stalled_timer_delay(1);


torrent::torrent(bt::session& s) :
	session_(s), ti_(nullptr), status_(no_status), save_changed_(false),
	resumed_(false), name_(""), saved_(true), paused_(false),
	total_size_(0), ready_(false)
{
}

torrent_ptr torrent::add(bt::session& s, const utf8_string& what)
{
	torrent_ptr t(new torrent(s));

	// here, 'what' may be either a path or a url

	try
	{
		// trying it as a path

		utf8_path p(what);
		if (p.exists())
		{
			// this is a valid path and the file exists
			t->add_impl(p);
			return t;
		}
	}
	catch(path_exception&)
	{
		// ignore, means this is not a valid path
	}

	// this is either not a path or the file doesn't exist; assuming
	// this is a url
	t->add_impl(what);

	return t;
}

torrent_ptr torrent::resume(bt::session& s, const resume_info& ri)
{
	try
	{
		torrent_ptr t(new torrent(s));
		t->resume_impl(ri);

		return t;
	}
	catch(bt::invalid_torrent_file& e)
	{
		log_error() << "invalid torrent file, " << e.what() << "\n";
		return torrent_ptr();
	}
}

void torrent::add_impl(const utf8_string& url)
{
	bt_assert(!handle_.is_valid());

	log() << "adding url '" << url << "'\n";

	bt::add_torrent_params p;

	// libtorrent will load the metadata from the given url
	p.url = url.std_string();

	// the initial name is based on the url (could be the last component
	// of the url or the dn field of a magnet link)
	resume_.name = name_from_url(url);

	// the initial hash based on the url (could be the dtih field of a
	// magnet link or an empty string for file urls)
	resume_.hash = hash_from_url(url);

	// this is the url used to fetch the metadata; this is saved in the
	// state file
	resume_.url = url;

	// save path for this torrent, based on the url
	save_ = save_path(false);
	p.save_path = save_.as_utf8_string().std_string();
	
	boost::system::error_code ec;
	create_directories(save_, ec);

	if (ec)
	{
		log_error()
			<< "failed to create directory, " << ec.message() << "\n";
	}

	// adds the default trackers, if any
	std::vector<utf8_string> temp = explode(
		opt("bt.torrent.default_trackers"), utf8_string(" "));

	foreach (const utf8_string& s, temp)
		p.trackers.push_back(s.std_string());

	log() << "save path is " << save_ << "\n";

	// gives the parameters to libtorrent
	add_torrent(p);
}

void torrent::add_impl(const utf8_path& file)
{
	bt_assert(!handle_.is_valid());

	log() << "adding torrent file " << file << "\n";

	bt::add_torrent_params p;

	// the initial name is the filename without extensions
	resume_.name = file.stem();

	// loading the metadata
	p.ti = new bt::torrent_info(file.as_utf8_string().std_string());
	ti_ = p.ti.get();

	if (!ti_->priv())
	{
		// adds the default trackers, if any, but not to private
		// torrents
		std::vector<utf8_string> temp = explode(
			opt("bt.torrent.default_trackers"), utf8_string(" "));

		foreach (const utf8_string& s, temp)
			p.trackers.push_back(s.std_string());
	}

	// save path for this torrent, based on the initial name on whether
	// it needs its own directory
	save_ = save_path(false);
	p.save_path = save_.as_utf8_string().std_string();
	create_directories(save_);

	log() << "save path is " << save_ << "\n";
	
	// gives the parameters to libtorrent
	add_torrent(p);
}

void torrent::resume_impl(const resume_info& ri)
{
	bt_assert(!handle_.is_valid());

	{
		// logging
		std::ostringstream oss;

		if (ri.hash.empty())
		{
			oss << "resuming url " << ri.url;
		}
		else
		{
			oss << "resuming torrent file " << ri.torrent_file;
	
			if (!ri.resume_file.empty())
				oss << " using fastresume " << ri.resume_file;
			else
				oss << " without fastresume";
		}

		log() << oss.str() << "\n";
	}

	bt::add_torrent_params p;

	resumed_ = true;

	// these all come from the xml file
	paused_ = ri.paused;
	resume_ = ri;
	name_ = ri.name;
	
	// loading the metadata
	if (ri.torrent_file.empty())
	{
		p.url = ri.url.std_string();
	}
	else
	{
		p.ti = new bt::torrent_info(
			ri.torrent_file.as_utf8_string().std_string());
	
		ti_ = p.ti.get();
	}

	// save path for this torrent, comes from the xml file (because the
	// save path might have been changed after this torrent was added)
	save_changed_ = ri.save_changed;
	save_ = ri.save_path;

	// this shouldn't happen, but this will use the default save path
	// if none was specified in the xml file
	if (save_.empty())
	{
		save_changed_ = false;
		save_ = save_path(ri.seeding);
	}

	p.save_path = save_.as_utf8_string().std_string();

	log() << "save path is " << save_ << "\n";

	if (!save_.exists())
	{
		log_error()
			<< "trying to resume, but save path is not accessible\n";

		set_status(bad_save_path_status);
	}
	else
	{
		// everything seems fine, libtorrent will start resuming this
		// torrent
	
		if (read_resume_file(ri.resume_file))
			p.resume_data = &resume_data_;

		set_status(resuming_status);
	}

	// gives the parameters to libtorrent
	add_torrent(p);
}

bool torrent::read_resume_file(const utf8_path& p)
{
	// no fastresume file
	if (p.empty())
		return false;

	// file size
	uintmax_t s = p.file_size();

	if (s > 0)
	{
		// the fastresume file was found and is not empty

		// resizes the vector so all the data can be read in it
		resume_data_.resize(static_cast<std::size_t>(s));

		try
		{
			// reading the file
			std::ifstream in(p.native().c_str(), std::ios::binary);
			in.exceptions(std::ios::badbit | std::ios::failbit);
			in.read(&resume_data_[0], s);
		}
		catch(std::exception& e)
		{
			log_error()
				<< "couldn't read fastresume, " << e.what() << "\n";
			resume_data_.clear();
		}
	}
	else
	{
		log_error() << "fastresume exists but is empty\n";
	}

	// in case of errors, resume_data_ is empty

	if (resume_data_.empty())
	{
		// if the fastresume file couldn't be read correctly, it is
		// faulty and should be removed

		boost::system::error_code ec;
		p.remove(ec);

		if (ec)
		{
			log_error()
				<< "couldn't remove bad resume file, "
				<< ec.message() << "\n";
		}

		return false;
	}
	else
	{
		return true;
	}
}

bool torrent::stalled() const
{
	if (ts_.download_payload_rate <= opt<int>("bt.torrent.stalled_rate"))
		return true;
	return false;
}

bool torrent::on_stalled_timer()
{
	if (status_ == downloading_status)
	{
		if (stalled())
			set_status(stalled_status);
	}
	else if (status_ == stalled_status)
	{
		if (!stalled())
			set_status(downloading_status);
	}

	stalled_timer_.start(
		boost::bind(&torrent::on_stalled_timer, this),
		stalled_timer_delay);

	return false;
}

void torrent::add_torrent(bt::add_torrent_params& p)
{
	// this will be picked up in the manager in the add_torrent_alert
	// alert; the problem is that the manager looks up torrents by
	// handle, but this torrent doesn't have a handle yet
	p.userdata = this;

	// this overrides the pause state because all torrents start
	// paused and only those that are not errored will start again
	p.flags =
		bt::add_torrent_params::flag_apply_ip_filter |
		bt::add_torrent_params::flag_update_subscribe | 
		bt::add_torrent_params::flag_paused |
		bt::add_torrent_params::flag_override_resume_data;

	p.storage_mode =
		storage_mode_from_string(opt("bt.torrent.storage_mode"));

	// the status might have already been set (such as when resuming),
	// but not necessarily (such as when everything went well while
	// adding a new torrent)
	if (status_ == no_status)
		set_status(starting_status);

	// gives the parameters to libtorrent; this will eventually fire
	// an add_torrent_alert that will be picked up by the manager, 
	// handled there and forwarded to on_torrent_added() below
	session_.async_add_torrent(p);
}

boost::system::error_code torrent::error() const
{
	return error_;
}

bool torrent::is(bt::torrent_handle h) const
{
	prevent_remove pr(this);
	return (handle_ == h);
}

bt::torrent_handle torrent::handle() const
{
	return handle_;
}

void torrent::up()
{
	prevent_remove pr(this);
	if (!handle_.is_valid())
		return;

	if (ts_.queue_position >= 0)
		handle_.queue_position_up();
}

void torrent::down()
{
	prevent_remove pr(this);
	if (!handle_.is_valid())
		return;

	if (ts_.queue_position >= 0)
		handle_.queue_position_down();
}

void torrent::top()
{
	prevent_remove pr(this);
	if (!handle_.is_valid())
		return;

	if (ts_.queue_position >= 0)
		handle_.queue_position_top();
}

void torrent::bottom()
{
	prevent_remove pr(this);
	if (!handle_.is_valid())
		return;

	if (ts_.queue_position >= 0)
		handle_.queue_position_bottom();
}

void torrent::recheck()
{
	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;

		log() << "rechecking " << name() << "\n";
		
		rename_incomplete_files();
		handle_.force_recheck();
		start(true);
	}

	set_status(checking_status);
}

bt::size_type torrent::total_size() const
{
	prevent_remove pr(this);

	if (total_size_ == 0)
	{
		boost::mutex::scoped_lock lock(total_size_mutex_);

		if (total_size_ == 0)
		{
			if (handle_.is_valid() && ti_)
			{
				const int n = ti_->num_files();

				// counting all the files
				for (int i=0; i<n; ++i)
				{
					const bt::file_entry& f = ti_->file_at(i);
					total_size_ += f.size;
				}
			}
		}
	}

	return total_size_;
}

bt::size_type torrent::remaining() const
{
	prevent_remove pr(this);

	if (!handle_.is_valid())
		return 0;

	return ts_.total_wanted - ts_.total_wanted_done;
}

bool torrent::errored() const
{
	const int s = static_cast<int>(status_);
	const int fe = static_cast<int>(first_error_status);

	return (s >= fe);
}

void torrent::start(bool force)
{
	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;
	}

	// cannot implicitly start an errored torrent
	if (!force && errored())
		return;

	if (!errored() && !paused_)
	{
		// this torrent is already running; this may be because 
		//  1) the torrent was just added (torrents are not added
		//     paused)
		//  2) the torrent was resumed on startup and wasn't paused
		//     in the fastresume (this is the normal case)
		//  3) the torrent was resumed as paused/errored and the user
		//     already started it
		//
		// for 1) and 2), the torrent is not auto managed yet and needs
		// to be set here; for 3), this is a noop
		{
			prevent_remove pr(this);
			if (handle_.is_valid())
				handle_.auto_managed(true);
		}

		set_status(starting_status);

		// nothing else to do, the torrent is already running
		return;
	}

	if (errored())
	{
		// if the torrent was errored, the files may have changed and
		// some may need to be renamed
		rename_incomplete_files();
	}

	// if this torrent is errored, 'force' was specified and so the
	// error must be cleared
	clear_error();
	set_status(starting_status);

	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;

		paused_ = false;
		handle_.resume();
		handle_.auto_managed(true);
	}

	start_stalled_timer();
}

void torrent::pause()
{
	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;
	}

	// an errored or already paused torrent cannot be paused because
	// it's not running
	if (errored() || paused_)
		return;

	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;
		
		handle_.pause();
		handle_.auto_managed(false);
	}

	paused_ = true;
	set_status(paused_status);
}

void torrent::stop()
{
	prevent_remove pr(this);
	if (!handle_.is_valid())
		return;

	if (status_ != stopped_status)
	{		
		paused_ = true;
		handle_.pause();
		handle_.auto_managed(false);

		set_status(stopped_status);
	}
}

void torrent::remove(bool keep_files)
{
	if (!handle_.is_valid())
		return;

	bt::torrent_handle temp_handle;

	{
		boost::unique_lock<boost::shared_mutex> remove_lock(mutex_);
		// removing the torrent from libtorrent
		temp_handle = handle_;
		handle_ = bt::torrent_handle();
	}

	// flag, whether to delete the files
	int f = 0;
	if (!keep_files)
		f |= bt::session::delete_files;

	// libtorrent takes care of deleting the files from inside the
	// torrent, but the copy of the .torrent file that was given
	// to the user must also be deleted
	if (!keep_files)
		delete_user_torrent_file(downloaded());

	delete_internal_files();

	// this torrent will keep this state until libtorrent acknowledges
	set_status(removing_status);

	// the torrent must be running in libtorrent to get removed
	temp_handle.resume();
	session_.remove_torrent(temp_handle, f);
}

void torrent::delete_user_torrent_file(bool complete)
{
	// path of the torrent file
	const utf8_path p = torrent_file_save_path(complete);

	// disabled by the user
	if (p.empty())
		return;

	if (!p.exists())
		return;

	log() << "removing user torrent file " << p << "\n";

	boost::system::error_code ec;
	p.remove(ec);
			
	if (ec)
	{
		log_error()
			<< "can't remove user torrent file, "
			<< ec.message() << "\n";
	}
}

void torrent::delete_internal_files()
{
	// a torrent without metadata hasn't saved anything yet, although
	// it has been saved in the xml queue file (which is handled by the
	// manager and so is not a concern here)
	if (!ts_.has_metadata)
		return;

	// fastresume file
	{
		const utf8_path p = internal_resume_file_save_path();
		log_v() << "removing internal resume file " << p << "\n";

		boost::system::error_code ec;
		p.remove(ec);

		if (ec)
		{
			log_error()
				<< "can't remove internal resume file, "
				<< ec.message() << "\n";
		}
	}

	// torrent file
	{
		const utf8_path p = internal_torrent_file_save_path();
		log_v() << "removing internal torrent file " << p << "\n";

		boost::system::error_code ec;
		p.remove(ec);

		if (ec)
		{
			log_error()
				<< "can't remove internal torrent file, "
				<< ec.message() << "\n";
		}
	}

	// torrent state
	{
		const utf8_path p = internal_torrent_state_save_path();
		log_v() << "removing internal torrent state file " << p << "\n";

		boost::system::error_code ec;
		p.remove(ec);

		if (ec)
		{
			log_error()
				<< "can't remove internal torrent state file, "
				<< ec.message() << "\n";
		}
	}
}

void torrent::save_torrent_file(bool complete)
{
	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;
	}

	// creating the buffer with data from libtorrent
	bt::create_torrent ct(*ti_);
	bt::entry e = ct.generate();

	std::vector<char> buffer;
	bt::bencode(std::back_inserter(buffer), e);

	// this saves both the user and internal .torrent files

	// user .torrent file
	try
	{
		const utf8_path f = torrent_file_save_path(complete);

		// f is empty if the user has not enabled this
		if (!f.empty())
		{
			log() << "saving user torrent file in " << f << "\n";

			create_directories(f.parent());

			// saving the file
			std::ofstream out(f.native().c_str(), std::ios::binary);
			out.exceptions(std::ios::badbit | std::ios::failbit);
			out.write(&buffer[0], buffer.size());
		}
	}
	catch(path_exception& e)
	{
		log_error()
			<< "couldn't save user torrent file, " << e.what() << "\n";
	}
	catch(std::exception& e)
	{
		log_error()
			<< "couldn't save user torrent file, " << e.what() << "\n";
	}

	// internal .torrent file
	try
	{
		const utf8_path f = internal_torrent_file_save_path();
		log_v() << "saving internal torrent file in " << f << "\n";

		create_directories(f.parent());

		// saving the file
		std::ofstream out(f.native().c_str(), std::ios::binary);
		out.exceptions(std::ios::badbit | std::ios::failbit);
		out.write(&buffer[0], buffer.size());
	}
	catch(path_exception& e)
	{
		log_error()
			<< "couldn't save user torrent file, " << e.what() << "\n";
	}
	catch(std::exception& e)
	{
		log_error()
			<< "couldn't save internal torrent file, "
			<< e.what() << "\n";
	}
}

void torrent::save_state_file()
{
	std::vector<int> priorities;

	{
		prevent_remove pr(this);
		
		if (!handle_.is_valid())
			return;

		priorities = handle_.file_priorities();
	}

	try
	{
		// this happens when the torrent has no metadata yet, in which
		// case the state file cannot be saved yet
		if (!ts_.has_metadata)
			return;

		xml::dom dom = xml::dom::create(utf8_string("state"));

		// priorities
		const utf8_string ps = implode(priorities, utf8_string(" "));
		dom.root().add(xml::element(utf8_string("priorities"), ps));

		const utf8_path p = internal_torrent_state_save_path();
		log_v() << "saving state file in " << p << "\n";

		create_directories(p.parent());
		dom.write(p);
	}
	catch(path_exception& e)
	{
		log_error() << "can't save sate file, " << e.what() << "\n";
	}
	catch(xml::bad_xml& e)
	{
		log_error() << "can't save state file, " << e.what() << "\n";
	}
}

void torrent::load_state()
{
	{
		prevent_remove pr(this);
		
		if (!handle_.is_valid())
			return;
	}

	try
	{
		const utf8_path p = internal_torrent_state_save_path();
		log_v() << "loading state file for " << name() << "\n";

		if (p.exists())
		{
			const xml::dom dom = xml::dom::load(
				internal_torrent_state_save_path());

			if (dom.root().has(utf8_string("priorities")))
			{
				const xml::element& e =
					dom.root()[utf8_string("priorities")];

				std::vector<int> v =
					explode<int>(e.value(), utf8_string(" "));

				if (ti_->num_files() != static_cast<int>(v.size()))
				{
					// something's wrong with the state file
					log_error()
						<< "bad state file, number of priorities is "
						<< v.size() << " but there's "
						<< ti_->num_files() << " files in the "
						<< "torrent; deleting\n";

					boost::system::error_code ec;
					p.remove(ec);

					if (ec)
					{
						log_error()
							<< "couldn't remove the state file, "
							<< ec.message() << "\n";
					}
				}
				else
				{
					prevent_remove pr(this);
					if (handle_.is_valid())
						handle_.prioritize_files(v);
				}
			}

			return;
		}
	}
	catch(xml::bad_xml& e)
	{
		log_error() << " can't load state file, " << e.what() << "\n";
	}
	catch(path_exception& e)
	{
		log_error() << "can't load state file, " << e.what() << "\n";
	}

	// here, the state file either didn't exist or wasn't loaded
	// properly; therefore, the default priorities are set
	default_priorities();
}

void torrent::save_resume()
{
	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;
	}

	// a previous attempt still ongoing
	if (!saved_)
	{
		log_v()
			<< name() << ": not saving resume because it's still "
			<< "waiting for the last request\n";
		return;
	}

	// this will be set to false below only if the request is sent
	// correctly to libtorrent
	saved_ = true;

	if (status_ == resume_failed_status)
	{
		bt_assert(!resume_data_.empty());

		// fastresume failed earlier and the torrent hasn't been
		// started by the user yet; this will write the bad data again
		//
		// because the data is already available, this is synchronous
		// and saved_ remains true
		if (!resume_data_.empty())
			save_resume_file(resume_data_);
	}
	else
	{
		if (ts_.has_metadata && ts_.need_save_resume)
		{
			log_v() << "saving resume data for '" << name() << "'\n";
			
			{
				prevent_remove pr(this);

				if (handle_.is_valid())
				{
					// libtorrent will notify the manager when the data
					// is available, at which point on_resume_saved()
					// will be called, the data will be written and
					// saved_ will be set to true
					handle_.save_resume_data();
					saved_ = false;
				}
			}
		}
		else
		{
			log_v() << name() << ": not saving resume because ";

			if (!ts_.has_metadata)
				log_v() << "it doesn't have metadata\n";
			else if (!ts_.need_save_resume)
				log_v() << "it doesn't need to\n";
		}
	}
}

bool torrent::downloaded() const
{
	if (status_ == starting_status)
		return resume_.seeding;

	if (ts_.is_seeding || ts_.is_finished)
		return true;

	return false;
}

bool torrent::resume_saved() const
{
	return saved_;
}

int torrent::eta() const
{
	{
		prevent_remove pr(this);

		// not added yet
		if (!handle_.is_valid())
			return 0;
	}

	// finished downloading
	if (downloaded())
		return 0;

	// this is the instantaneous eta based on the current speed, there's
	// no averaging

	const int r = ts_.download_payload_rate;
	
	bt::size_type s = ts_.total_wanted;
	if (s < ts_.total_wanted_done)
		s = 0;
	else
		s -= ts_.total_wanted_done;

	// don't bother with the 1000 days left
	if (r < opt<int>("bt.torrent.stalled_rate") || s == 0)
		return 0;

	int v = static_cast<int>(s / r);

	// again, more than 10 days
	if (v > 60 * 60 * 24 * 10)
		return 0;

	return v;
}

bool torrent::active() const
{
	switch(status())
	{
		case torrent::metadata_status:		// fall-through
		case torrent::downloading_status:
		case torrent::seeding_status:
		case torrent::stalled_status:
			return true;

		case torrent::checking_status:		// fall-through
		case torrent::resuming_status:
		case torrent::starting_status:
		case torrent::queued_status:
		case torrent::moving_status:
		case torrent::removing_status:
		case torrent::no_status:
		case torrent::first_error_status:
		case torrent::bad_save_path_status:
		case torrent::resume_failed_status:
		case torrent::move_failed_status:
		case torrent::generic_error_status:
		case torrent::paused_status:
		case torrent::stopped_status:
		default:
			return false;
	}
}

void torrent::default_priorities()
{
	{
		prevent_remove pr(this);

		if (!handle_.is_valid())
			return;
	}
	
	if (!ti_)
		return;

	// this is in case normal_priority ever gets different than
	// libtorrent's default priority

	std::vector<int> v(ti_->num_files());
	foreach (int& i, v)
		i = static_cast<int>(normal_priority);

	{
		prevent_remove pr(this);

		if (handle_.is_valid())
			handle_.prioritize_files(v);
	}
}

void torrent::on_error(bt::torrent_error_alert& ea)
{
	log_error()
		<< "torrent " << name() << " is errored: "
		<< ea.error.message() << " (" << ea.error << ")\n";

	set_error(ea.error);
	set_status(generic_error_status);
}

void torrent::on_metadata_received(bt::torrent_alert& ta)
{
	// this may happen (I think) if the metadata was received while the
	// torrent was being added and ti_ got set in on_torrent_added()
	if (ti_)
		return;

	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;

		bt_assert(ta.handle == handle_);
	}

	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;

		// todo: this should be merged with on_torrent_added(), they do
		// the same thing

		// this depends on the name of the torrent and so must be
		// checked again
		if (!save_changed_)
		{
			const utf8_path ns = save_path(false);

			if (save_ != ns)
			{
				save_ = ns;
				handle_.move_storage(
					save_.as_utf8_string().std_string());
			}
		}

		resume_.hash = to_utf8_string(handle_.info_hash());
		ti_ = &handle_.get_torrent_info();
	}

	on_ready();
	ext::fire(utf8_string("torrent.metadata"), boost::ref(*this));
}

void torrent::on_torrent_complete(bt::torrent_alert& ta)
{
	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;

		bt_assert(handle_ == ta.handle);
	}
	
	// removing the old torrent file, if any
	delete_user_torrent_file(false);

	// changing the save path to the complete one but only if it
	// hasn't changed and it wasn't resumed seeding (in which case the
	// path has already been changed)

	if (!resume_.seeding && !save_changed_)
	{
		save_ = save_path(true);

		{
			prevent_remove pr(this);
			if (!handle_.is_valid())
				return;

			handle_.move_storage(save_.as_utf8_string().std_string());
		}
	}

	save_torrent_file(true);
	save_state_file();
	save_resume();

	ext::fire(utf8_string("torrent.complete"), boost::ref(*this));
}

void torrent::on_resume_saved(boost::shared_ptr<bt::entry> data)
{
	if (data)
	{
		std::vector<char> v;
		bt::bencode(std::back_inserter(v), *data);
		save_resume_file(v);
	}
}

void torrent::on_torrent_added(bt::add_torrent_alert& ta)
{
	prevent_remove pr(this);
	
	bt_assert(!handle_.is_valid());

	// this is called from the manager when libtorrent acknowledges the
	// new torrent

	handle_ = ta.handle;
	bt_assert(handle_.is_valid());

	if (!resumed_)
	{
		// don't touch the position for already existing torrents
		if (opt<bool>("bt.torrent.add_top"))
			handle_.queue_position_top();
	}

	if (!paused_)
	{
		// the torrent was not added as paused (which would be the case
		// if the user paused the torrent and restarted the program)
		//
		// this will set the automanaged flag
		start();
	}
	else
	{
		// the torrent was added as paused
		set_status(paused_status);
	}

	// this will apply the options to the torrent
	reload_settings();

	ext::fire(utf8_string("torrent.added"), boost::ref(*this));
	bt_assert(handle_.is_valid());
}

void torrent::start_stalled_timer()
{
	// starts the stalled timer; this will set the stalled status when
	// the timer fires if nothing is being transferred
	stalled_timer_.start(
		boost::bind(&torrent::on_stalled_timer, this),
		initial_stalled_timer_delay);
}

void torrent::on_ready()
{
	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;

		name_ = utf8_string(handle_.name());
	}

	if (!resumed_)
	{
		// torrent added, has metadata, not resumed: this is the
		// first time the files are available
		rename_incomplete_files();
	}

	load_state();
	save_torrent_file(false);

	fire(updated);
	ext::fire(utf8_string("torrent.updated"), boost::ref(*this));

	start_stalled_timer();

	ext::fire(utf8_string("torrent.ready"), boost::ref(*this));

	ready_ = true;
}

void torrent::on_resume_failed(bt::fastresume_rejected_alert& ra)
{
	log_error()
		<< "loading fastresume failed for " << name() << ": "
		<< ra.error.message() << " (" << ra.error << ")\n";

	set_error(ra.error);
	set_status(resume_failed_status);
}

void torrent::on_file_completed(bt::file_completed_alert& a)
{
	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;
	}

	if (!ti_)
		return;

	bt_assert(a.index >= 0 && a.index < ti_->num_files());
	if (a.index < 0 || a.index >= ti_->num_files())
		return;

	const bt::file_entry& f = ti_->file_at(a.index);
	const bt::file_entry& o = ti_->orig_files().at(a.index);
		
	const utf8_string p = utf8_string(o.path);

	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;

		if (utf8_string(f.path) != p)
		{
			log_v()
				<< "renaming '" << f.path << "' to "
				<< "'" << p << "'\n";

			handle_.rename_file(a.index, p.std_string());
		}
	}

	ext::fire(
		utf8_string("torrent.file_completed"), boost::ref(*this), o);
}

void torrent::on_storage_moved(bt::storage_moved_alert&)
{
	log() << "storage moved to " << save_ << "\n";

	save_changed_ = true;
	start(true);
	
	{
		prevent_remove pr(this);

		if (handle_.is_valid())
			refresh_status(ts_.state);
	}

	fire(moved);
}

void torrent::on_storage_moved_failed(bt::storage_moved_failed_alert& a)
{
	log_error()
		<< "could not move storage to " << save_ << ": "
		<< a.error.message() << " (" << a.error << ")\n";

	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;

		save_ = utf8_path(utf8_string(handle_.save_path()));
	}

	set_error(a.error);
	set_status(move_failed_status);
	fire(move_failed);
}

void torrent::save_resume_file(const std::vector<char>& data)
{
	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;
	}

	std::ofstream out(
		internal_resume_file_save_path().native().c_str(),
		std::ios::binary);

	out.unsetf(std::ios::skipws);
	out.write(&data[0], data.size());

	if (!out)
	{
		log_error()
			<< "couldn't save resume data for '" << name() << "'\n";
	}
	else
	{
		log_v() << "resume data saved for '" << name() << "'\n";
	}

	// mark this as done even if it failed; this flag is merely there
	// to indicate that libtorrent hasn't responded yet, not that saving
	// succeeded or failed
	saved_ = true;
}

void torrent::reload_settings()
{
	prevent_remove pr(this);
	if (!handle_.is_valid())
		return;

	if (opt<bool>("bt.network.connections_per_torrent_enabled"))
	{
		handle_.set_max_connections(
			opt<int>("bt.network.connections_per_torrent"));
	}
	else
	{
		handle_.set_max_connections(-1);
	}
}

void torrent::rename_incomplete_files()
{
	if (!opt<bool>("bt.torrent.append_extension"))
		return;

	prevent_remove pr(this);

	if (!handle_.is_valid())
		return;

	if (!ti_)
		return;

	bt::file_storage ofs = ti_->orig_files();
	bt::file_storage nfs = ti_->files();
	std::vector<bt::size_type> pg = file_progress();
	
	bt_assert(ofs.num_files() == nfs.num_files());

	for (int i=0; i<ofs.num_files(); ++i)
	{
		const bt::file_entry& of = ofs.at(i);
		const bt::file_entry& nf = nfs.at(i);

		// already changed
		if (of.path != nf.path)
			continue;
		
		// file complete; pg is empty when metadata has just been
		// received
		if (i < static_cast<int>(pg.size()) && pg[i] == nf.size)
			continue;

		const utf8_string p =
			utf8_string(of.path) + paths::incomplete_suffix();

		handle_.rename_file(i, p.std_string());
	}
}

utf8_path torrent::save_path(bool complete) const
{
	const bool incomplete_enabled =
		opt<bool>("bt.path.incomplete_save_path_enabled");

	// if the incomplete save path is disabled, the complete one is
	// always used
	if (complete || !incomplete_enabled)
		return utf8_path(opt("bt.path.complete_save_path"));
	else
		return utf8_path(opt("bt.path.incomplete_save_path"));
}

bool torrent::has_metadata() const
{
	return ts_.handle.is_valid() && ts_.has_metadata;
}

void torrent::queue_node(xml::element& e) const
{
	const utf8_string& n = name();
	if (n.empty())
		return;

	e.add(xml::element(utf8_string("name"), n));
		
	utf8_string h;
	if (ts_.has_metadata)
		h = hash();
	else if (resumed_)
		h = resume_.hash;

	if (!h.empty())
		e.add(xml::element(utf8_string("hash"), h));

	if (!url().empty())
		e.add(xml::element(utf8_string("url"), url()));

	e.add(xml::element(utf8_string("save_path"), save_path()));
	e.add(xml::element(utf8_string("save_changed"), save_changed_));

	e.add(xml::element(
		utf8_string("torrent_file_save_path"),
		torrent_file_save_path(downloaded())));

	e.add(xml::element(
		utf8_string("paused"), (paused_ ? "1" : "0")));
}

utf8_path torrent::torrent_file_save_path(bool complete) const
{
	const utf8_string nn = name();
	if (nn.empty())
		return utf8_path();

	// filename of the torrent file
	const utf8_string n = nn + paths::torrent_file_suffix();

	if (complete)
	{
		const bool enabled = opt<bool>(
			"bt.path.complete_torrent_file_path_enabled");

		if (!enabled)
		{
			// the torrent is complete but that option is disabled
			return utf8_path();
		}
		else
		{
			// the torrent is complete, use the path specified by
			// the user
			const utf8_path p(
				opt("bt.path.complete_torrent_file_path"));

			return p / n;
		}
	}
	else
	{
		const bool enabled = opt<bool>(
			"bt.path.incomplete_torrent_file_path_enabled");

		if (!enabled)
		{
			// the torrent is incomplete but that option is disabled by
			// the user
			return utf8_path();
		}
		else
		{
			// the torrent is incomplete, use the path specified by the
			// user
			const utf8_path p(opt(
				"bt.path.incomplete_torrent_file_path"));

			return p / n;
		}
	}
}

utf8_path torrent::internal_torrent_file_save_path() const
{
	const utf8_string& h = hash();
	if (h.empty())
		return utf8_path();

	return paths::conf() / (h + paths::torrent_file_suffix());
}

utf8_path torrent::internal_resume_file_save_path() const
{
	const utf8_string& h = hash();
	if (h.empty())
		return utf8_path();

	return paths::conf() / (h + paths::resume_file_suffix());
}

utf8_path torrent::internal_torrent_state_save_path() const
{
	const utf8_string& h = hash();
	if (h.empty())
		return utf8_path();

	return paths::conf() / (h + paths::state_file_suffix());
}

utf8_string torrent::name() const
{
	// this is either from the xml queue file or libtorrent, but can
	// be empty if the torrent has no metadata
	if (!name_.empty())
		return name_;
	
	// this is from the url or the file name, but can also be empty in
	// certain cases, such as when adding a magnet link with no
	// information but the hash
	if (!resume_.name.empty())
		return resume_.name;

	return hash();
}

utf8_string torrent::hash() const
{
	{
		prevent_remove pr(this);

		if (handle_.is_valid())
		{
			if (ti_)
				return to_utf8_string(ti_->info_hash());
			else
				return to_utf8_string(handle_.info_hash());
		}
	}

	// this will be empty for urls to files before metadata has been
	// received
	return resume_.hash;
}

utf8_string torrent::url() const
{
	return resume_.url;
}

torrent::statuses torrent::status() const
{
	// the pause flag and status are independent in the sense that
	// the flag might be true while the status isn't paused_status
	//
	// this is because libtorrent is asynchronous and may take time
	// before acknowledging a status change but the user needs to be
	// shown that the torrent has been paused correctly even though
	// stuff might still be happening in the background

	// because errored torrents are always paused, the real status
	// needs to be returned so the error can be shown to the user
	if (errored())
		return status_;

	if (status_ == removing_status)
		return status_;

	// this torrent is not errored but has been paused, whatever its
	// current status
	if (paused_ && status_ != stopped_status)
		return paused_status;

	// messy: queued torrents have the pause flag on, which is the same
	// as just being paused; however, paused_ will be false in this case
	if (ts_.paused && !paused_ && status_ != stopped_status)
		return queued_status;

	// this torrent is not errored and is running
	return status_;
}

bt::sha1_hash torrent::sha1_hash() const
{
	return ts_.info_hash;
}

const utf8_path& torrent::save_path() const
{
	return save_;
}

utf8_path torrent::root_path() const
{
	if (!ts_.has_metadata || !ti_)
		return save_;
	
	bt::file_storage fs;

	{
		prevent_remove pr(this);

		if (!handle_.is_valid())
			return save_;

		fs = ti_->orig_files();
	}

	if (fs.num_files() <= 1)
		return save_;

	const std::string path = fs.at(0).path;
	const std::string::size_type p =
		path.find(directory_separator().std_string());
	if (p == std::string::npos)
		return save_;

	return save_ / utf8_string(path.substr(0, p));
}

void torrent::move(const utf8_path& p)
{
	if (move_impl(p))
		save_changed_ = true;
}

bool torrent::move_impl(const utf8_path& p)
{
	{
		prevent_remove pr(this);

		if (!handle_.is_valid())
			return false;
		
		if (save_ == p)
		{
			log()
				<< "not moving storage because " << p
				<< " is already the save path\n";

			return false;
		}

		log() << "moving storage from " << save_ << " to " << p << "\n";

		save_ = p;
		set_status(moving_status);
		fire(moving);

		handle_.move_storage(p.as_utf8_string().std_string());
	}

	return true;
}

int torrent::position() const
{
	return ts_.queue_position;
}

std::vector<bt::size_type> torrent::file_progress() const
{
	std::vector<bt::size_type> p;

	if (!ts_.has_metadata)
		return p;

	{
		prevent_remove pr(this);

		if (!handle_.is_valid())
			return p;

		handle_.file_progress(p, bt::torrent_handle::piece_granularity);
	}

	return p;
}

std::vector<int> torrent::file_priorities() const
{
	std::vector<int> v;

	if (!ts_.has_metadata)
		return v;

	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return v;
		
		v = handle_.file_priorities();
	}

	return v;
}

void torrent::file_priorities(const std::vector<int>& v)
{
	if (!ts_.has_metadata)
		return;

	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;

		handle_.prioritize_files(v);
	}
}

std::vector<torrent::file_info> torrent::files() const
{
	std::vector<file_info> v;

	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return v;

		if (ts_.has_metadata && ti_)
		{
			const bt::file_storage& ofiles = ti_->orig_files();
			const std::size_t n = ti_->num_files();

			// for each file in the torrent
			for (int i=0; i<static_cast<int>(n); ++i)
			{
				file_info fi;
				fi.original = ofiles.at(i);
				fi.current = ti_->file_at(i);

				v.push_back(fi);
			}
		}
	}

	return v;
}

std::vector<bt::announce_entry> torrent::trackers() const
{
	std::vector<bt::announce_entry> v;

	{
		prevent_remove pr(this);
		if (handle_.is_valid())
			v = handle_.trackers();
	}

	return v;
}

void torrent::trackers(const std::vector<bt::announce_entry>& v)
{
	{
		prevent_remove pr(this);
		if (handle_.is_valid())
			handle_.replace_trackers(v);
	}
}

void torrent::add_tracker(const utf8_string& s)
{
	{
		prevent_remove pr(this);
		if (handle_.is_valid())
			handle_.add_tracker(bt::announce_entry(s.std_string()));
	}
}

std::vector<bt::peer_info> torrent::peers() const
{
	std::vector<bt::peer_info> p;

	if (!ts_.has_metadata)
		return p;

	{
		prevent_remove pr(this);
		if (handle_.is_valid())
			handle_.get_peer_info(p);
	}

	return p;
}

bool torrent::added() const
{
	prevent_remove pr(this);
	return handle_.is_valid();
}

void torrent::set_status(statuses s)
{
	if (s != status_)
	{
		status_ = s;
		fire(updated);
		ext::fire(utf8_string("torrent.updated"), boost::ref(*this));
	}
}

void torrent::set_error(boost::system::error_code e)
{
	{
		prevent_remove pr(this);
		if (!handle_.is_valid())
			return;

		handle_.pause();
		handle_.auto_managed(false);
	}

	paused_ = true;
	error_ = e;
	ext::fire(utf8_string("torrent.error"), boost::ref(*this));
}

void torrent::clear_error()
{
	if (errored())
	{
		error_ = boost::system::error_code();
		status_ = no_status;
	}
}

void torrent::refresh_status(bt::torrent_status::state_t s)
{
	if (errored())
		return;

	switch(s)
	{
		case bt::torrent_status::checking_files:
		{
			set_status(checking_status);
			break;
		}

		case bt::torrent_status::queued_for_checking:
		{
			set_status(queued_status);
			break;
		}

		case bt::torrent_status::checking_resume_data:
		{
			set_status(resuming_status);
			break;
		}

		case bt::torrent_status::downloading_metadata:
		{
			set_status(metadata_status);
			break;
		}

		case bt::torrent_status::allocating:
		{
			set_status(starting_status);
			break;
		}

		case bt::torrent_status::downloading:
		{
			if (ts_.state == bt::torrent_status::checking_files)
			{
				// this is after checking the files because of a bad
				// fastresume so the new fastresume has to be saved
				save_resume();
			}


			if (status_ != stalled_status)
				set_status(downloading_status);

			break;
		}

		case bt::torrent_status::finished:	// fall-through
		case bt::torrent_status::seeding:
		{
			set_status(seeding_status);
			break;
		}

		default:
		{
			bt_assert(false);
			set_status(paused_status);
			break;
		}
	}

}

const bt::torrent_status& torrent::torrent_status() const
{
	return ts_;
}

double torrent::ratio() const
{
	if (ts_.all_time_download == 0)
		return 0.0;

	return
		static_cast<double>(ts_.all_time_upload) /
		ts_.all_time_download;
}

bool torrent::can_resume() const
{
	if (errored())
		return true;
	
	if (status() == torrent::paused_status ||
		status() == torrent::stopped_status)
	{
		return true;
	}

	return false;
}

bool torrent::can_pause() const
{
	return !can_resume();
}

bool torrent::remote_peer_connected() const
{
	return ts_.has_incoming;
}

void torrent::update(const bt::torrent_status& ts)
{
	if (!ready_)
	{
		resume_.hash = to_utf8_string(ts.info_hash);

		// although the torrent has been added, it doesn't necessarily
		// have metadata yet
		if (ts.has_metadata)
		{
			if (!ti_)
				ti_ = &handle_.get_torrent_info();

			on_ready();
		}
	}

	refresh_status(ts.state);
	ts_ = ts;

	const float max = opt<int>("bt.torrent.seed_ratio") / 100.0f;

	if (downloaded() && ratio() >= max)
		stop();

	fire(updated);
	ext::fire(utf8_string("torrent.updated"), boost::ref(*this));
}

utf8_string status_to_loc(torrent::statuses s)
{
	switch (s)
	{
		case torrent::no_status:
			return utf8_string("");


		case torrent::paused_status:
			return loc("torrent.status.paused");

		case torrent::stopped_status:
			return loc("torrent.status.stopped");

		case torrent::checking_status:
			return loc("torrent.status.checking");

		case torrent::starting_status:
			return loc("torrent.status.starting");

		case torrent::metadata_status:
			return loc("torrent.status.metadata");

		case torrent::queued_status:
			return loc("torrent.status.queued");

		case torrent::stalled_status:
			return loc("torrent.status.stalled");

		case torrent::downloading_status:
			return loc("torrent.status.downloading");

		case torrent::seeding_status:
			return loc("torrent.status.seeding");

		case torrent::resuming_status:
			return loc("torrent.status.resuming");

		case torrent::moving_status:
			return loc("torrent.status.moving");

		case torrent::removing_status:
			return loc("torrent.status.removing");


		case torrent::bad_save_path_status:
			return loc("torrent.status.bad_save_path");

		case torrent::resume_failed_status:
			return loc("torrent.status.resume_failed");

		case torrent::move_failed_status:
			return loc("torrent.status.move_failed_status");

		case torrent::generic_error_status:
			return loc("torrent.status.generic_error");

		case torrent::first_error_status:	// fall-through
		default:
			bt_assert(false);
			return utf8_string("unknown status");
	}
}

}	// namespace baretorrent
