// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_TORRENT_INCLUDED
#define BARETORRENT_TORRENT_INCLUDED

#include "utility.h"
#include "xml.h"

namespace baretorrent
{

// loaded from the xml file and passed to the manager when resuming
// torrents
//
struct resume_info
{
	// display name of the torrent
	utf8_string name;

	// hash
	utf8_string hash;

	// the original url used to get the torrent (even though the torrent
	// file is now saved locally)
	utf8_string url;

	// whether this torrent was downloaded
	bool seeding;
	
	// whether this torrent was paused
	bool paused;

	// whether the torrent path was changed manually
	bool save_changed;

	// path where the torrent file was saved
	utf8_path torrent_file;

	// path where the resume file was saved
	utf8_path resume_file;
		
	// path where the torrent was saved
	utf8_path save_path;

	// path where the user's torrent file was saved
	utf8_path torrent_file_save_path;


	resume_info() :
		name(""), hash(""), url(""), seeding(false), paused(false),
		save_changed(false)
	{
	}

	bool empty() const
	{
		return name.empty() && hash.empty() && url.empty();
	}
};


// has a libtorrent torrent object and handles status, priorities, etc.
//
class torrent : boost::noncopyable
{
public:
	typedef sig::signal<void ()> handler;

	// fired when the status of the torrent changes
	handler updated;
	handler moving;
	handler move_failed;
	handler moved;


	struct file_info
	{
		bt::file_entry original;
		bt::file_entry current;
	};

	// download priority of a file
	enum priorities
	{
		skip_priority = 0,
		normal_priority,
		high_priority
	};

	// status of this torrent
	enum statuses
	{
		no_status = 1,

		// paused
		paused_status,

		// stopped; used when the seed ratio has reached the maximum
		stopped_status,

		// checking the data because fastresume was rejected
		checking_status,

		// this is an already existing torrent that is currently
		// being loaded and its fastresume data checked
		resuming_status,


		// the torrent has been added but isn't transferring yet
		starting_status,

		// queued
		queued_status,

		// fetching metadata; only used for urls
		metadata_status,

		// could be downloading but no peers are available
		stalled_status,

		// the torrent is currently downloading data
		downloading_status,

		// the torrent is finished and is (or could be) uploading data
		seeding_status,

		// the torrent is currently being moved
		moving_status,

		
		// the user asked to remove the torrent but libtorrent hasn't
		// acknowledged it yet
		removing_status,
		

		// never set, but all remaining values are errors
		first_error_status,

		// one or more files cannot be saved
		bad_save_path_status,

		// the fastresume failed
		resume_failed_status,

		// moving failed
		move_failed_status,

		// any other kind of error; error() can be used to retrieve
		// the exact error
		generic_error_status
	};


	// creates a new torrent from the given file or url and adds it to
	// the session
	//
	static torrent_ptr add(bt::session& s, const utf8_string& what);

	// resumes a torrent from the given information and adds it to the
	// session
	//
	static torrent_ptr resume(bt::session& s, const resume_info& ri);


	// whether the given handle is the same as this torrent
	//
	bool is(bt::torrent_handle h) const;

	// returns the libtorrent torrent_handle 
	//
	bt::torrent_handle handle() const;

	// total size in bytes of this torrent; this include anything that
	// is skipped
	//
	bt::size_type total_size() const;

	// total number of bytes left to download
	//
	bt::size_type remaining() const;


	// amount uploaded / downloaded
	//
	double ratio() const;

	// whether this torrent is paused/stopped and could be resumed
	//
	bool can_resume() const;

	// whether this torrent is running and can be paused
	//
	bool can_pause() const;


	// saves the fast resume data for this torrent; this is asynchronous
	// and the manager will eventually call on_resume_saved() when
	// libtorrent has the data
	//
	void save_resume();

	// saves the state of this torrent; this is synchronous
	//
	void save_state_file();


	// whether this torrent is done downloading
	//
	bool downloaded() const;

	// whether the fastresume data has been received from libtorrent
	// and saved; this is false between a call to save_resume() and
	// when libtorrent fires an alert with the data (which makes the
	// manager call on_resume_saved())
	//
	bool resume_saved() const;

	// whether this torrent has been added by libtorrent; this is false
	// between creation and when libtorrent fires an alert (which makes
	// the manager call on_torrent_added())
	//
	bool added() const;

	// whether this torrent ran into any error and is currently paused
	//
	bool errored() const;

	// if errored(), this returns the error code
	//
	boost::system::error_code error() const;

	// called when the torrent has metadata and is ready to start
	// download
	//
	void on_ready();

	// returns whether this torrent's download rate is too low; this
	// doesn't necessarily mean that status() would return
	// stalled_status since there is a delay before the torrent
	// becomes actually stalled
	//
	bool stalled() const;



	// starts this torrent; if 'force' is false, this won't do anything
	// if errored(); noop if the torrent is already started
	//
	void start(bool force=false);

	// pauses this torrent; noop if the torrent is already paused
	//
	void pause();

	// removes this torrent from the session and deletes any external
	// files related to it, such as the fastresume; this is called from
	// the manager and shouldn't be called directly as the manager also
	// has stuff to do
	//
	void remove(bool keep_files);


	// moves this torrent up one in the priority list
	//
	void up();

	// moves this torrent down one in the priority list
	//
	void down();

	// moves this torrent to the top of the priority list
	//
	void top();

	// moves this torrent to the bottom of the priority list
	//
	void bottom();

	// rechecks this torrent
	//
	void recheck();


	// called from the manager for torrent_error_alert alerts; sets the
	// error status and code
	//
	void on_error(bt::torrent_error_alert& ea);

	// called from the manager for on_metadata_received alerts; sets up
	// the various paths because they might depend on the torrent name
	// and other various things that depend on the file list
	//
	void on_metadata_received(bt::torrent_alert& ta);

	// called from the manager for on_torrent_complete alerts; this
	// moves files around if the paths for complete/incomplete are
	// different
	//
	void on_torrent_complete(bt::torrent_alert& ta);

	// called from the manager for on_resume_saved alerts; this writes
	// the actual fastresume data to disk
	//
	void on_resume_saved(boost::shared_ptr<bt::entry> data);

	// called from the manager for on_torrent_added alerts; this
	// remembers some information such as the handle and status and
	// sets up some stuff if the torrent has metadata (which would
	// otherwise be done in on_metadata_received())
	//
	void on_torrent_added(bt::add_torrent_alert& ta);

	// called from the manager for on_resume_failed alerts; this sets
	// the error status and code
	//
	void on_resume_failed(bt::fastresume_rejected_alert& ra);

	// called from the manager for file_completed_alert; removes the
	// suffix from the file
	//
	void on_file_completed(bt::file_completed_alert& a);

	// called from the manager when the storage has been moved
	//
	void on_storage_moved(bt::storage_moved_alert& a);

	// called from the manager when the storage couldn't be moved
	//
	void on_storage_moved_failed(bt::storage_moved_failed_alert& a);


	// this returns a display name for this torrent; its content depends
	// on the state of the torrent:
	//   1) for newly added torrents, this is constructed from the
	//      stem of the path (for files or urls), the dn field (for
	//      magnets) or the hash (btih) if dn wasn't in the url)
	//   2) once libtorrent receives metadata for this torrent (either
	//      immediately for .torrent files, after downloading the
	//      .torrent from the network or after receiving metadata from
	//      peers for magnet links), this uses whatever libtorrent
	//      returns
	//   3) for torrents that are resumed, name() was saved in the xml
	//      file and so is reused until libtorrent loads the metadata
	//      (which should be immediately anyways)
	//
	utf8_string name() const;

	// this returns a hex string for the torrent's hash; its content
	// depends on the state of the torrent:
	//   1) for newly added torrent, this comes from the content of
	//      the .torrent file or the btih field of a magnet link; in
	//      case of a .torrent that needs to be fetched, hash() will
	//      return an empty string until metadata has been received
	//   2) once libtorrent receives metadata for this torrent, this
	//      uses whatever libtorrent returns
	//   3) for torrents that are resumed, hash() was saved in the xml
	//      file and so is reused until libtorrent loads the metadata
	//      (which should be immediately anyways)
	//
	utf8_string hash() const;

	// this returns the url used to fetch the metadata, either a url
	// to a .torrent or a magnet link; in case of a local file, this is
	// always empty
	//
	// the value of url() is saved in the xml file and so will always
	// be remembered
	//
	utf8_string url() const;

	// the status of this torrent; this is not a bitfield
	//
	statuses status() const;

	// returns the info_hash field from libtorrent, which may be empty
	// if metadata hasn't been received yet
	//
	bt::sha1_hash sha1_hash() const;

	// returns the directory in which the files are saved; this may
	// change over time because completed torrents may be moved to a
	// different directory
	//
	const utf8_path& save_path() const;

	// returns the root directory; this is save_path() for single-file
	// torrents and the root directory for multiple-file torrents
	//
	utf8_path root_path() const;

	// changes the save path of this torrent
	//
	void move(const utf8_path& p);

	// fills the given xml element with information that can be saved in
	// the xml queue file
	//
	void queue_node(xml::element& e) const;

	// whether this torrent has metadata
	//
	bool has_metadata() const;


	// the position of this torrent in the queue
	//
	int position() const;

	// the number of bytes downloaded so far for each file; this will
	// be empty if the torrent has no metadata
	//
	std::vector<bt::size_type> file_progress() const;

	// the priorities for each file in the torrent; this comes directly
	// from libtorrent and therefore may have other values than the
	// priorities enum (which is there merely for convenience)
	//
	std::vector<int> file_priorities() const;
	
	// sets the priorities for each file in the torrent; this is passed
	// directly to libtorrent and there can have other values than the
	// ones in the priorities enum
	//
	void file_priorities(const std::vector<int>& v);

	// returns the list of files in the torrent; may be empty if the
	// torrent has no metadata
	//
	std::vector<file_info> files() const;

	// returns the list of trackers for this torrent
	//
	std::vector<bt::announce_entry> trackers() const;

	// sets the list of trackers for this torrent
	//
	void trackers(const std::vector<bt::announce_entry>& v);

	// adds the given url to the list of trackers
	//
	void add_tracker(const utf8_string& s);

	// returns the list of all peers connected to this torrent
	//
	std::vector<bt::peer_info> peers() const;

	// returns the libtorrent torrent status object
	//
	const bt::torrent_status& torrent_status() const;

	// returns whether at least one peer connection was opened remotely,
	// used to check whether port forwarding is fine
	//
	bool remote_peer_connected() const;

	// called when the options have changed; this resets some of the
	// global options such as the maximum number of connections per
	// torrent
	//
	void reload_settings();

	// called from the manager in response to a state_update_alert alert
	//
	void update(const bt::torrent_status& ts);

	// the estimated number of seconds until the torrent is completed;
	// this will be 0 for torrents that are stalled
	//
	int eta() const;

	// whether this torrent is trying to download or upload stuff
	//
	bool active() const;

private:
	// locks mutex_ in the constructor, unlocks it in the destructor
	//
	class prevent_remove : boost::noncopyable
	{
	public:
		prevent_remove(const torrent* t)
			: this_(t)
		{
			this_->mutex_.lock_shared();
		}

		~prevent_remove()
		{
			this_->mutex_.unlock_shared();
		}

	private:
		const torrent* this_;
	};


	// libtorrent session from the manager
	bt::session& session_;

	// libtorrent handle of this torrent
	bt::torrent_handle handle_;

	// libtorrent metadata; this is set on creation when resuming or
	// when the metadata is received
	const bt::torrent_info* ti_;

	// libtorrent torrent status
	bt::torrent_status ts_;

	// status of this torrent
	statuses status_;

	// save path
	utf8_path save_;
	bool save_changed_;

	// last error code
	boost::system::error_code error_;

	// whether this torrent was resumed or is a new one
	bool resumed_;

	// checks whether anything is being transferred and set the stalled
	// status if not
	timer stalled_timer_;
	
	// prevents the torrent from being removed since anything can be
	// called from any thread, including extensions
	mutable boost::shared_mutex mutex_;


	resume_info resume_;

	// this is initial_name_ until libtorrent receives the metadata
	utf8_string name_;

	// set to false when libtorrent is asked to generate the fastresume
	// data and try when the data is either received and saved or failed
	// 
	bool saved_;

	// whether this torrent is paused; this corresponds to the user
	// action, not necessarily to the current status since it might take
	// time for libtorrent to react to a pause request but the user
	// interface needs to immediately show that this torrent is paused
	bool paused_;

	// this contains the data that was read from the fastresume file;
	// if the fastresume failed, the same bad data will be saved again
	// unless the torrent has been manually started
	//
	// this avoids the case where the program is started when the save
	// path is nonexistent (it hasn't been mounted yet, for example),
	// closed and then restarted again; in this case, the torrent won't
	// restart itself automatically because the bad fastresume is
	// still there
	//
	// note that "bad" doesn't mean ill-formed (in which case
	// read_resume_file() would have deleted the file immediately), it
	// is just rejected by libtorrent because it doesn't fit the current
	// state of the files (such as mismatched timestamps)
	std::vector<char> resume_data_;

	// cached size of the torrent
	mutable bt::size_type total_size_;
	mutable boost::mutex total_size_mutex_;
	
	bool ready_;


	// new torrent for the given session
	//
	torrent(bt::session& s);
	
	
	// depending on the options, this may add a suffix to all of the
	// incomplete files in the torrent
	//
	void rename_incomplete_files();

	// called when the torrent is added, sets the default priority for
	// all files
	//
	void default_priorities();

	// pauses the torrent and marks it as stopped_status; used when
	// the seed ratio is reached
	//	
	void stop();


	bool move_impl(const utf8_path& p);


	// returns the save path for this torrent based whether the torrent
	// is complete
	//
	utf8_path save_path(bool complete) const;

	// returns the save path for the copy of the torrent file that
	// belongs to the user; empty if the user has disabled this
	//
	utf8_path torrent_file_save_path(bool complete) const;

	// returns the save path of the .torrent file that is saved
	// internally by the program; this is used when resuming the torrent
	//
	utf8_path internal_torrent_file_save_path() const;

	// returns the save path of the fastresume file that is saved
	// internally by the program; this is used when resuming the torrent
	//
	utf8_path internal_resume_file_save_path() const;

	// returns the save path of the state file that is saved internally
	// by the program; this is used when resuming the torrent
	//
	utf8_path internal_torrent_state_save_path() const;


	// creates an add_torrent_params structure to resume the torrent
	// based on the given information and passes it to add_torrent()
	//
	void resume_impl(const resume_info& ri);

	// creates an add_torrent_params structure to load a new torrent
	// file from disk and passes it to add_torrent()
	//
	void add_impl(const utf8_path& file);

	// creates an add_torrent_params structure to load a new torrent
	// file/magnet from the given url and passes it to add_torrent()
	//
	void add_impl(const utf8_string& url);

	// sets some default values and adds the torrent to the libtorrent
	// session
	//
	void add_torrent(bt::add_torrent_params& p);

	// reads the fastresume file from the given path and saves the
	// data in resume_; deletes the file if something went wrong
	//
	bool read_resume_file(const utf8_path& p);


	// starts the initial stalled timer
	//
	void start_stalled_timer();

	// fired by stalled_timer_, checks whether anything is being
	// transferred and sets the stalled status
	//
	bool on_stalled_timer();

	// saves both a copy of the torrent file for the user in the right
	// path and an internal copy for resuming, based on whether the
	// torrent is complete or not
	//
	void save_torrent_file(bool complete);

	// called from on_resume_saved() (which is called from the manager
	// for a fastresume saved/failed alert) and saves the given data
	// (if any)
	//
	void save_resume_file(const std::vector<char>& data);

	// deletes the copy of the torrent file for the user, called when
	// removing the torrent and the user chose not to keep the files
	//
	void delete_user_torrent_file(bool complete);

	// deletes all the internal files related to this torrent; called
	// when removing the torrent
	//
	void delete_internal_files();

	
	// called from update() (which is called from the manager for a
	// state alert), saves the torrent status and fires signals
	//
	void refresh_status(bt::torrent_status::state_t s);

	// reads the xml state file for this torrent and loads its settings
	//
	void load_state();


	// sets the status and fires signals
	//
	void set_status(statuses s);

	// clears the error state; this does not fire a signal because 
	// the new status is temporarily invalid
	//
	void clear_error();

	// sets the error state and code
	//
	void set_error(boost::system::error_code e);
};

// converts the given status to a localized string
//
utf8_string status_to_loc(torrent::statuses s);

}	// namespace baretorrent

#endif	// BARETORRENT_TORRENT_INCLUDED
