// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "torrent_list.h"
#include "main_window.h"
#include "manager.h"

namespace baretorrent
{
		
// various widget ids
//
enum ids
{
	// torrent list
	list_id = wxID_HIGHEST + 600,

	search_id,
	search_close_id
};


#define s(name) int name##_sorter(						\
					torrent_list::torrent_item* left,	\
					torrent_list::torrent_item* right)

s(queue);
s(name);
s(status);
s(progress);
s(total_size);
s(download_rate);
s(upload_rate);
s(peers);
s(eta);

#undef s

// time after which the new selection is handled; prevents from handling
// multiple messages in quick succession
const boost::posix_time::milliseconds selection_delay(100);


// popup menu for the torrent list
//
class popup_menu : public wxMenu
{
public:
	popup_menu()
	{
#define a(id, n) Append( \
				id, loc_wx("menu." n), loc_wx("menu." n ".help"));

		a(resume_id, "torrent.resume");
		a(pause_id, "torrent.pause");
		AppendSeparator();
		a(up_id, "torrent.up");
		a(down_id, "torrent.down");
		a(top_id, "torrent.top");
		a(bottom_id, "torrent.bottom");
		AppendSeparator();
		a(remove_id, "torrent.remove");
		a(recheck_id, "torrent.recheck");
		a(move_storage_id, "torrent.move_storage");
		AppendSeparator();
		a(open_directory_id, "torrent.open_directory");

#undef a
	}
};

torrent_list::torrent_list(wxWindow* parent, manager& m) :
	wxPanel(parent), manager_(m), cat_(all_categories),
	list_(nullptr), search_panel_(nullptr), search_(nullptr)
{
	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	SetSizer(box);

	list_ = new custom_list<torrent_item>(
		this, list_id, "bt.ui.main_list");
	box->Add(list_, 1, wxEXPAND);
	
	search_panel_ = new wxPanel(this);

	wxButton* close_search = new wxButton(
		search_panel_, search_close_id, wxEmptyString,
		wxDefaultPosition, wxDefaultSize,
		wxBU_EXACTFIT);// | wxBU_NOTEXT);
	close_search->SetBitmap(
		platform::make_image(images::close_search));

	search_ = new wxTextCtrl(search_panel_, search_id);
	
	search_->Connect(
		wxEVT_KEY_DOWN,
		wxKeyEventHandler(torrent_list::on_search_keydown),
		nullptr, this);

	wxBoxSizer* search_sizer = new wxBoxSizer(wxHORIZONTAL);
	search_sizer->Add(
		close_search, 0, wxSHAPED | wxALIGN_CENTER_VERTICAL);
	search_sizer->AddSpacer(10);
	search_sizer->Add(
		new wxStaticText(search_panel_, -1, loc_wx("search.label")),
		0, wxALIGN_CENTER_VERTICAL);
	search_sizer->AddSpacer(5);
	search_sizer->Add(search_, 1, wxEXPAND | wxALIGN_CENTER_VERTICAL);

	search_panel_->SetSizer(search_sizer);
	search_panel_->Hide();

	box->Add(search_panel_, 0, wxALL | wxEXPAND, 3);


	list_->context(new popup_menu);
	list_->selected.connect(boost::bind(
		&torrent_list::on_selection, this, _1));
	list_->deselected.connect(boost::bind(
		&torrent_list::on_selection, this, _1));
	list_->activated.connect(boost::bind(
		&torrent_list::on_activated, this, _1));

	// this assumes that for a column C, a locale "torrent.C.caption"
	// has the column caption

#define a(c, s) \
			list_->add_column(loc("torrent." c ".caption"), s);

	// adds all the columns along with their sorter (used when clicking
	// columns)

	a("position", &queue_sorter);
	a("name", &name_sorter);
	a("status", &status_sorter);
	a("progress", &progress_sorter);
	a("total_size", &total_size_sorter);
	a("download_rate", &download_rate_sorter);
	a("upload_rate", &upload_rate_sorter);
	a("eta", &eta_sorter);

#undef a

	list_->sort();
}

torrent_list::~torrent_list()
{
	list_->delete_data();
}

void torrent_list::rebuild()
{
	if (!list_)
		return;

	// prevents flickering
	freezer list_freezer(list_);

	list_->clear();

	// for each torrent
	for (std::size_t i=0; i<manager_.size(); ++i)
	{
		torrent_ptr t = manager_.get(i);

		if (displayable(t))
			add(t);
	}

	list_->sort();
	handle_selection();
}

void torrent_list::category(torrent_categories cat)
{
	cat_ = cat;
	rebuild();
}

int torrent_list::find_item(torrent_ptr t)
{
	for (int i=0; i<list_->GetItemCount(); ++i)
	{
		torrent_item* item = list_->data(i);
		if (item->t == t)
			return i;
	}

	return -1;
}

void torrent_list::update(torrent_categories cat, torrent_ptr t)
{
	freezer list_freezer(list_);

	int i = find_item(t);

	if (i == -1)
	{
		// not in the list and it shouldn't be displayed
		if (!in_category(cat, t))
			return;

		add(t);
	}
	else
	{
		// in the list but shouldn't be displayed anymore
		if (!in_category(cat, t))
		{
			remove(t);
			return;
		}

		update_list_item(i, t);
	}

	// resort the list because some of information on the selected
	// column may have changed
	list_->sort();
}

torrents torrent_list::selection() const
{
	if (!list_)
		return torrents();
	
	std::vector<const torrent_item*> s = list_->selection();

	torrents ts;
	foreach(const torrent_item* i, s)
		ts.push_back(i->t);

	return ts;
}

bool torrent_list::is_selected(torrent_ptr t) const
{
	return has(selection(), t);
}

bool torrent_list::selection_can_resume() const
{
	foreach (const torrent_ptr t, selection())
	{
		if (t->can_resume())
			return true;
	}

	return false;
}

bool torrent_list::selection_can_pause() const
{
	foreach (const torrent_ptr t, selection())
	{
		if (t->can_pause())
			return true;
	}

	return false;
}

bool torrent_list::searching() const
{
	return search_panel_->IsShown();
}

void torrent_list::search(bool b, const utf8_string& s)
{
	if (b)
	{
		search_panel_->Show();
		search_->SetFocus();
	}
	else
	{
		search_panel_->Hide();
	}

	search_->ChangeValue(to_wxstring(s));
	rebuild();

	Layout();
}

bool torrent_list::matches_search(const torrent_ptr t)
{
	const utf8_string s = search_string();
	if (s.empty())
		return true;
	
	const wxString a(to_wxstring(t->name()).Lower());
	const wxString b(to_wxstring(s).Lower());

	return (a.Find(b) != wxNOT_FOUND);
}

bool torrent_list::displayable(const torrent_ptr t)
{
	return in_category(cat_, t) && matches_search(t);
}

void torrent_list::set_item(int i, int c, const utf8_string& s)
{
	list_->col(i, c, s);
}

void torrent_list::update_list_item(int i, const torrent_ptr t)
{
	const bt::torrent_status& ts = t->torrent_status();

	// the following is building the strings for the various columns;
	// when values are unavailable or 0 (such as rates and eta), the
	// column is empty instead of showing zeros
	
	utf8_string pos;
	if (ts.handle.is_valid() && t->position() >= 0)
	{
		// the position might be -1 if the torrent hasn't been
		// acknowledged by libtorrent yet or if it is seeding, in which
		// case the column will be empty
		
		// the position is zero-based, but starting at one looks better
		pos = as_utf8_string(t->position() + 1);
	}
		
	const int p = make_percent(ts.progress);

	utf8_string size;
	if (ts.total_wanted > 0)
		size = make_loc_size(ts.total_wanted);

	utf8_string dl;
	if (ts.download_payload_rate > opt<int>("bt.torrent.stalled_rate"))
	{
		dl = utf8_string((locf("transfer_rate")
			% make_loc_size(ts.download_payload_rate)).str());
	}

	utf8_string ul;
	if (ts.upload_payload_rate > opt<int>("bt.torrent.stalled_rate"))
	{
		ul = utf8_string((locf("transfer_rate")
			% make_loc_size(ts.upload_payload_rate)).str());
	}

	utf8_string eta;
	if (t->eta() > 0)
		eta = make_loc_duration_abbr(t->eta());


	// setting the columns

	// allows for adding and removing columns without having to change
	// every line
	int c = 0;

	set_item(i, c++, pos);
	set_item(i, c++, t->name());
	set_item(i, c++, status_to_loc(t->status()));
	set_item(
		i, c++, 
		utf8_string((locf("torrent.progress") % p).str()));
	set_item(i, c++, size);
	set_item(i, c++, dl);
	set_item(i, c++, ul);
	set_item(i, c++, eta);

	// errored torrents are of a different color
	if (t->errored())
		list_->row_color(i, errored_text_color());
	else
		list_->row_color(i, normal_text_color());
}

void torrent_list::add(torrent_ptr t)
{
	if (!list_)
		return;

	if (displayable(t))
	{
		// the columns will be set by update_list_item()
		const int i = list_->append();
	
		update_list_item(i, t);
		list_->data(i, new torrent_item(t));
	}
}

void torrent_list::remove(torrent_ptr t)
{
	const int i = find_item(t);

	if (i == -1)
	{
		log_error()
			<< "torrent list: can't remove torrent, not found\n";
		return;
	}

	torrent_item* item = list_->data(i);

	delete item;
	list_->DeleteItem(i);

	handle_selection();
}

void torrent_list::on_selection(torrent_item*)
{
	// this is called for both selections and de-selections (because the
	// code is the same) and therefore will be called multiple times
	// even for single selections (current item deselected, new item
	// selected); this is even worse for multiple selections
	//
	// so a short timer is started here and constantly reset until
	// there's no more selection messages, at which point
	// on_selection_timer() is fired, which handles the actual selection

	selection_timer_.start(boost::bind(
		&torrent_list::on_selection_timer, this), selection_delay);
}

bool torrent_list::on_selection_timer()
{
	handle_selection();

	// this is a one-shot
	return false;
}

void torrent_list::on_activated(torrent_item*)
{
	torrents ts = selection();
	if (ts.empty())
		return;

	activated(ts[0]);
}

void torrent_list::on_search(wxCommandEvent&)
{
	rebuild();
}

void torrent_list::on_search_close(wxCommandEvent&)
{
	search(false);
}

void torrent_list::on_search_keydown(wxKeyEvent& e)
{
	if (e.GetKeyCode() == WXK_ESCAPE)
		search(false);
	
	e.Skip();
}

utf8_string torrent_list::search_string() const
{
	return from_wxstring(search_->GetValue());
}

void torrent_list::handle_selection()
{
	selection_changed(selection());
}

bool torrent_list::in_category(
	torrent_categories cat, const torrent_ptr t)
{
	if (cat == all_categories)
		return true;

	// todo: active_category doesn't take rates into account

	switch(t->status())
	{
		case torrent::checking_status:		// fall-through
		case torrent::resuming_status:
		case torrent::starting_status:
		case torrent::queued_status:
		case torrent::metadata_status:
		case torrent::downloading_status:
		case torrent::moving_status:
		{
			if (cat == downloading_category ||
				cat == active_category)
			{
				return true;
			}

			break;
		}

		case torrent::seeding_status:
		{
			if (cat == uploading_category ||
				cat == active_category)
			{
				return true;
			}
			
			break;
		}

		case torrent::removing_status:		// fall-through
		case torrent::stalled_status:
		case torrent::no_status:
		case torrent::first_error_status:
		case torrent::bad_save_path_status:
		case torrent::resume_failed_status:
		case torrent::move_failed_status:
		case torrent::generic_error_status:
		case torrent::paused_status:
		case torrent::stopped_status:
		default:
		{
			if (cat == inactive_category)
				return true;
			break;
		}
	}

	return false;
}


#define s(name) int name##_sorter(						\
					torrent_list::torrent_item* left,	\
					torrent_list::torrent_item* right)

s(queue)
{
	// position is -1 when the torrent hasn't been added yet or is
	// seeding, in which case will be sorted at the end
	if (left->t->position() == -1 && right->t->position() == -1)
		return status_sorter(left, right);

	// torrents not in the queue are sorted last
	if (left->t->position() == -1)
		return 1;
	else if (right->t->position() == -1)
		return -1;

	return compare(left->t->position(), right->t->position());
}

s(name)
{
	return natural_compare(left->t->name(), right->t->name());
}

s(status)
{
	return natural_compare(
		status_to_loc(left->t->status()),
		status_to_loc(right->t->status()));
}

s(progress)
{
	return compare(
		left->t->torrent_status().progress,
		right->t->torrent_status().progress);
}

s(total_size)
{
	return compare(left->t->total_size(), right->t->total_size());
}

s(download_rate)
{
	return compare(
		left->t->torrent_status().download_payload_rate,
		right->t->torrent_status().download_payload_rate);
}

s(upload_rate)
{
	return compare(
		left->t->torrent_status().upload_payload_rate,
		right->t->torrent_status().upload_payload_rate);
}

s(peers)
{
	return compare(
		left->t->torrent_status().num_peers,
		right->t->torrent_status().num_peers);
}

s(eta)
{
	return compare(left->t->eta(), right->t->eta());
}

#undef s

BEGIN_EVENT_TABLE(torrent_list, wxPanel)
	EVT_TEXT(search_id, torrent_list::on_search)
	EVT_BUTTON(search_close_id, torrent_list::on_search_close)
END_EVENT_TABLE()

}	// namespace baretorrent
