// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_TORRENT_LIST_INCLUDED
#define BARETORRENT_TORRENT_LIST_INCLUDED

#include "utility.h"
#include "custom_list.h"

namespace baretorrent
{

class manager;
class main_window;

class torrent_list : public wxPanel
{
public:
	typedef sig::signal<void (torrents)> multiple_torrents_handler;
	typedef sig::signal<void (torrent_ptr)> single_torrent_handler;

	multiple_torrents_handler selection_changed;
	single_torrent_handler activated;


	// the user data of items in the list
	//
	struct torrent_item
	{
		// the torrent
		torrent_ptr t;

		torrent_item(torrent_ptr t)
			: t(t)
		{
		}
	};


	torrent_list(wxWindow* parent, manager& m);
	~torrent_list();


	void update(torrent_categories cat, const torrent_ptr t);

	// clears the torrent list and adds all the torrents for the
	// current directory; called on startup, when torrents are
	// added (todo: why?) and when the category changes
	//
	void rebuild();

	// sets the visible category and calls rebuild()
	//
	void category(torrent_categories cat);


	// adds a new item in the list for the given torrent and connects
	// the updated signal; called when rebuilding or torrents are added
	// or updated (if they now must be displayed because their category
	// has changed)
	//
	void add(torrent_ptr t);

	// removes the list item corresponding to the given torrent; called
	// when torrents are removed or updated (if they now must be hidden
	// because their category has changed)
	//
	void remove(torrent_ptr t);


	// returns all the selected torrents in the list; this may be
	// different from selection_ when handling selection change, until
	// selection_ has been updated
	//
	torrents selection() const;

	bool is_selected(torrent_ptr t) const;

	// returns whether at least one selected torrent can resume
	//
	bool selection_can_resume() const;

	// returns whether at least one selected torrent can pause
	//
	bool selection_can_pause() const;

	
	// returns whether the search panel is currently visible
	//
	bool searching() const;

	// sets whether the search panel is currently visible, with an
	// options search string
	//
	void search(bool b, const utf8_string& s = utf8_string());

private:
	DECLARE_EVENT_TABLE()

	manager& manager_;

	torrent_categories cat_;

	// main torrent list
	custom_list<torrent_item>* list_;

	// search panel
	wxPanel* search_panel_;
	wxTextCtrl* search_;

	// started and restarted every time the selection changes because
	// multiple selection messages may be sent quickly; see
	// on_selection()
	timer selection_timer_;


	// updates the given item in the list with the torrent information;
	// called when torrents are updated or added
	//
	void update_list_item(int i, const torrent_ptr t);

	// wrapper; sets the text of the given item and column if it has
	// changed
	//
	void set_item(int i, int c, const utf8_string& s);

	// returns the search string, if any
	//
	utf8_string search_string() const;

	// returns whether the given torrent matches the current search
	//
	bool matches_search(const torrent_ptr t);

	
	// updates the various panels; called when the selection has changed
	// or when torrents are removed (since they might have been
	// selected)
	//
	// see inside on_selection()
	//
	void handle_selection();

	// returns the index of the given torrent in the list or -1
	//
	int find_item(torrent_ptr t);

	// returns whether the given torrent is part of the given category
	//
	bool in_category(torrent_categories cat, const torrent_ptr t);

	// returns whether the given torrent should be displayed in the
	// list, based on the category and search
	//
	bool displayable(const torrent_ptr t);


	// fired from selection_timer_; actually handles the current
	// selection (see inside on_selection())
	//
	bool on_selection_timer();
	
	// fired by selection_timer_ after selection events stop coming in
	//
	void on_selection(torrent_item*);

	// fired when double clicking a torrent in the list; opens its
	// download directory
	//
	void on_activated(torrent_item*);

	// fired when the text in the search box changes
	//
	void on_search(wxCommandEvent& e);

	// fired when the close button on the search panel is clicked
	//
	void on_search_close(wxCommandEvent&);

	// fired when keys are pressed in the search textbox; handles
	// escape 
	//
	void on_search_keydown(wxKeyEvent& e);
};

}	// namespace baretorrent

#endif	// BARETORRENT_TORRENT_LIST_INCLUDED
