// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "log.h"
#include "options.h"

namespace baretorrent
{

baretorrent_version make_version(
	int major, int minor, int maint, const std::string& tag = "")
{
	baretorrent_version v;

	v.major = major;
	v.minor = minor;
	v.maintenance = maint;
	v.tag = tag;

	return v;
}

void update_from_0_4_3()
{
	// user-agent
	if (opt("bt.session.user_agent") == utf8_string("baretorrent/0.4.3 libtorrent/0.16.10.0"))
		opt("bt.session.user_agent", user_agent());

	opt("bt.version", "0.4.4");
}

void update_from_0_4_2()
{
	if (opt("bt.torrent.default_trackers") == utf8_string("http://tracker.openbittorrent.com"))
		opt("bt.torrent.default_trackers", utf8_string(""));

	// user-agent
	if (opt("bt.session.user_agent") == utf8_string("baretorrent/0.4.2 libtorrent/0.16.10.0"))
		opt("bt.session.user_agent", "baretorrent/0.4.3 libtorrent/0.16.10.0");

	opt("bt.version", "0.4.3");
}

void update_from_0_4_1_alpha()
{
	// new proxy enabled option
	if (!opt("bt.network.proxy_type").empty())
		opt("bt.network.proxy_enabled", true);

	if (opt("bt.pe.out_enc_policy") == utf8_string("enabled"))
		opt("bt.pe.out_enc_policy", "forced");

	if (opt("bt.pe.in_enc_policy") == utf8_string("enabled"))
		opt("bt.pe.in_enc_policy", "forced");

	opt("bt.session.peer_tos", 0);

	if (opt<int>("bt.network.listen") == 6881)
	{
		opt("bt.network.listen", "");
	}
	else
	{
		std::ostringstream oss;
		oss
			<< opt<int>("bt.network.listen") << " "
			<< opt<int>("bt.network.listen");

		opt("bt.network.listen", oss.str());
	}

	// user-agent
	if (opt("bt.session.user_agent") == utf8_string("baretorrent/0.4.1-alpha libtorrent/0.16.9.0"))
		opt("bt.session.user_agent", user_agent());

	// version
	opt("bt.version", "0.4.2-alpha");
}

void update_from_0_4_0_alpha()
{
	// max download/upload rates were initially enabled but set to 0;
	// this disables them
	if (opt<bool>("bt.torrent.max_download_enabled"))
	{
		if (opt<int>("bt.torrent.max_download") == 0)
			opt("bt.torrent.max_download_enabled", "0");
	}

	if (opt<bool>("bt.torrent.max_upload_enabled"))
	{
		if (opt<int>("bt.torrent.max_upload") == 0)
			opt("bt.torrent.max_upload_enabled", "0");
	}

	// seed ratio has been changed from floating point (2.0) to
	// percentages (200)
	if (opt<int>("bt.torrent.seed_ratio") < 10)
	{
		opt(
			"bt.torrent.seed_ratio",
			opt<int>("bt.torrent.seed_ratio") * 100);
	}

	// libtorrent changed its default alert queue size
	if (opt<int>("bt.session.alert_queue_size") == 1000)
		opt("bt.session.alert_queue_size", "6000");

	// update url
	opt("bt.update_url", "http://baretorrent.org/u.txt");

	// user-agent
	opt("bt.session.user_agent", "baretorrent/0.4.1-alpha libtorrent/0.16.9.0");

	// version
	opt("bt.version", "0.4.1-alpha");
}

void update_options(
	const baretorrent_version& from, const baretorrent_version& to)
{
	bool force = false;
	if (opt_exists("bt.force_update"))
	{
		force = true;
		remove_option("bt.force_update");
	}

	if (!force && from == to)
		return;

	if (!force && from > to)
	{
		log_error()
			<< "update: " << from << " is more recent than " << to
			<< ", not doing anything\n";

		return;
	}

	typedef std::map<baretorrent_version, boost::function<void()> >
		updaters;
	updaters us;

#define vt(major, minor, maint, tag) \
		us.insert(std::make_pair( \
			make_version(major, minor, maint, #tag), \
			&update_from_##major##_##minor##_##maint##_##tag))

#define v(major, minor, maint) \
		us.insert(std::make_pair( \
			make_version(major, minor, maint), \
			&update_from_##major##_##minor##_##maint))

	vt(0, 4, 0, alpha);
	vt(0, 4, 1, alpha);
	v(0, 4, 2);
	v(0, 4, 3);

#undef v

	updaters::iterator first = us.end();

	if (force)
	{
		first = us.begin();
	}
	else
	{
		for (updaters::iterator itor=us.begin(); itor!=us.end(); ++itor)
		{
			if (itor->first == from)
			{
				first = itor;
				break;
			}
		}
	}

	if (first == us.end())
	{
		log_error()
			<< "bad current version " << from << ", assuming old one "
			<< us.begin()->first << "\n";

		first = us.begin();
	}

	for (updaters::iterator itor=first; itor!=us.end(); ++itor)
	{
		log() << "updating from " << itor->first << "\n";
		(itor->second)();
	}

	dump_options();
	log() << "up to date\n";
}

}	// namespace baretorrent
