// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "utility.h"
#include "torrent.h"
#include "locale.h"
#include "manager.h"
#include "options.h"
#include "platform.h"

namespace baretorrent
{

// remembers the command line so it can be displayed elsewhere, such as
// in the log dialog; set in set_cmd_line(), called from main()
static utf8_string g_cmd;

struct remote_file_data
{
	std::ostringstream oss;
	download_callback cb;
	double p;

	remote_file_data(download_callback f)
		: cb(f), p(0.0)
	{
	}
};

size_t remote_file_write(char* p, size_t s, size_t n, void* d)
{
	remote_file_data* rfd = reinterpret_cast<remote_file_data*>(d);
	bt_assert(rfd);

	if (rfd->cb && !rfd->cb(rfd->p))
	{
		log_v()
			<< "curl: aborting transfer from write function "
			<< "because stop is true\n";
		
		return 0;
	}

	rfd->oss.write(p, s * n);

	return s * n;
}

int remote_file_progress(
	void* d, double dl_total, double dl_now, double, double)
{
	remote_file_data* rfd = reinterpret_cast<remote_file_data*>(d);
	bt_assert(rfd);

	if (dl_total == 0.0)
		rfd->p = 0.0;
	else
		rfd->p = dl_now / dl_total;

	if (rfd->cb && !rfd->cb(rfd->p))
	{
		log_v()
			<< "curl: aborting transfer from progress "
			<< "because stop is true\n";

		return 1;
	}

	return 0;
}

void set_proxy(curl_handle& h)
{
	const utf8_string pt = opt("bt.network.proxy_type");
	const utf8_string ph = opt("bt.network.proxy");

	if (!opt<bool>("bt.network.proxy_enabled"))
	{
		log_vv()
			<< "remote file: not setting proxy because disabled\n";
		return;
	}

	if (pt.empty())
	{
		log_vv()
			<< "remote file: not setting proxy because type is empty\n";
		return;
	}
	else if (ph.empty())
	{
		log_vv()
			<< "remote file: not setting proxy because type is "
			<< pt << " but host is empty\n";
		return;
	}

	int t = -1;

	if (pt == utf8_string("SOCKS4"))
		t = CURLPROXY_SOCKS4;
	else if (pt == utf8_string("SOCKS5"))
		t = CURLPROXY_SOCKS5;
	else if (pt == utf8_string("HTTP"))
		t = CURLPROXY_HTTP;
	else
		log_error() << "unknown proxy type '" << pt << "'\n";

	if (t != -1)
	{
		const utf8_string u = opt("bt.network.proxy_username");
		const utf8_string p = opt("bt.network.proxy_password");

		log_vv()
			<< "remote file: proxy is " << ph << " (" << t << "), "
			<< "url is " << ph << ", "
			<< "username is '" << u << "' "
			<< "and password is '" << p << "'\n";

		curl_easy_setopt(h, CURLOPT_PROXY, ph.c_str());
		curl_easy_setopt(h, CURLOPT_PROXYTYPE, t);
		curl_easy_setopt(h, CURLOPT_PROXYUSERNAME, u.c_str());
		curl_easy_setopt(h, CURLOPT_PROXYPASSWORD, p.c_str());
	}
}

byte_array get_remote_file(const utf8_string& url, download_callback f)
{
	curl_handle h;

	char error[CURL_ERROR_SIZE + 1] = {0};
	remote_file_data rfd(f);
	
	log_v() << "getting remote file " << url << "\n";

	curl_easy_setopt(h, CURLOPT_URL, url.c_str());
	curl_easy_setopt(h, CURLOPT_TIMEOUT, 300);
	curl_easy_setopt(h, CURLOPT_CONNECTTIMEOUT, 300);
	curl_easy_setopt(h, CURLOPT_WRITEFUNCTION, &remote_file_write);
	curl_easy_setopt(h, CURLOPT_WRITEDATA, &rfd);
	curl_easy_setopt(h, CURLOPT_FOLLOWLOCATION, 1);
	curl_easy_setopt(h, CURLOPT_PROGRESSFUNCTION, &remote_file_progress);
	curl_easy_setopt(h, CURLOPT_PROGRESSDATA, &rfd);
	curl_easy_setopt(h, CURLOPT_ERRORBUFFER, error);
	curl_easy_setopt(h, CURLOPT_NOPROGRESS, 0);
	curl_easy_setopt(h, CURLOPT_NOSIGNAL, 1);
	set_proxy(h);

	CURLcode c = curl_easy_perform(h);
	if (c != CURLE_OK)
	{
		log_error()
			<< "could not get file at " << url << ", "
			<< "curl says '" << error << "' (" << c << ")\n";

		return byte_array();
	}
	
	long code = 0;
	curl_easy_getinfo(h, CURLINFO_RESPONSE_CODE, &code);

	if (code < 200 || code >= 300)
	{
		log_error()
			<< "could not get file at " << url << ", "
			<< "http code is " << code << "\n";

		return byte_array();
	}

	return byte_array(rfd.oss.str());
}

utf8_string urlencode(const utf8_string& s)
{
	curl_handle h;

	char* e = curl_easy_escape(
		h, s.c_str(), static_cast<int>(s.size()));
	
	utf8_string r(e);
	curl_free(e);
	
	boost::replace_all(r, utf8_string(" "), utf8_string("+"));

	return r;
}

utf8_string urldecode(const utf8_string& s)
{
	curl_handle h;

	int outlength = 0;
	char* e = curl_easy_unescape(
		h, s.c_str(), static_cast<int>(s.size()), &outlength);
	
	utf8_string r(e, outlength);
	curl_free(e);

	boost::replace_all(r, "+", " ");

	return r;
}


clipboard::clipboard()
	: opened_(false)
{
	if (!wxTheClipboard->Open())
	{
		log_error() << "couldn't open clipboard\n";
		return;
	}

	opened_ = true;
}

clipboard::~clipboard()
{
	if (opened_)
		wxTheClipboard->Close();
}

utf8_string clipboard::get_text() const
{
	if (!opened_)
		return utf8_string();

	if (wxTheClipboard->IsSupported(wxDF_TEXT))
	{
		wxTextDataObject d;
		wxTheClipboard->GetData(d);

		return utf8_string(d.GetText().ToUTF8());
	}

	return utf8_string();
}

void clipboard::set_text(const utf8_string& s)
{
	if (!opened_)
		return;

	wxTheClipboard->SetData(new wxTextDataObject(to_wxstring(s)));
}


utf8_string to_utf8_string(const bt::sha1_hash& s)
{
	if (s.is_all_zeros())
		return utf8_string();

	return utf8_string(bt::to_hex(std::string(
		reinterpret_cast<const char*>(&s[0]),
		bt::sha1_hash::size)));
}

utf8_string round(double d, int p)
{
	std::ostringstream oss;
	oss << std::fixed << std::setprecision(p) << d;
	return utf8_string(oss.str());
}

utf8_string make_loc_size(uint64_t tt)
{
	const uint64_t b = 1024;
	const uint64_t kibi = b * 1024;
	const uint64_t mebi = kibi * 1024;
	const uint64_t gibi = mebi * 1024;

	if (tt < b)
		return utf8_string((locf("byte.abbr") % tt).str());

	const double s = static_cast<double>(tt);

	if (tt < kibi)
		return utf8_string((locf("kibi.abbr") % round(s / b)).str());

	if (s < mebi)
		return utf8_string((locf("mebi.abbr") % round(s / kibi)).str());

	if (s < gibi)
		return utf8_string((locf("gibi.abbr") % round(s / mebi)).str());

	return utf8_string((locf("tebi.abbr") % round(s / gibi)).str());
}

utf8_string make_loc_duration_impl(
	int t,
	boost::format second, boost::format seconds,
	boost::format minute, boost::format minutes,
	boost::format hour, boost::format hours,
	boost::format day, boost::format days)
{
	std::ostringstream oss;

	if (t <= 1)
	{
		oss << second % t;
	}
	else if (t < 60)
	{
		oss << seconds % t;
	}
	else if (t < 3600)
	{
		const int i = t / 60;
		if (i <= 1)
			oss << minute % i;
		else
			oss << minutes % i;
	}
	else if (t < 86400)
	{
		const int i = (t / 3600);
		if (i <= 1)
			oss << hour % i;
		else
			oss << hours % i;
	}
	else
	{
		const int i = (t / 86400);
		if (i <= 1)
			oss << day % i;
		else
			oss << days % i;
	}

	return utf8_string(oss.str());
}

utf8_string make_loc_duration(int s)
{
	return make_loc_duration_impl(
		s,
		locf("time.second"), locf("time.seconds"), locf("time.minute"),
		locf("time.minutes"), locf("time.hour"), locf("time.hours"),
		locf("time.day"), locf("time.days"));
}

utf8_string make_loc_duration_abbr(int s)
{
	return make_loc_duration_impl(
		s,
		locf("time.second.abbr"), locf("time.second.abbr"),
		locf("time.minute.abbr"), locf("time.minute.abbr"),
		locf("time.hour.abbr"), locf("time.hour.abbr"),
		locf("time.day.abbr"), locf("time.day.abbr"));
}


utf8_string make_loc_rate(int s)
{
	return utf8_string((locf("transfer_rate")
		% make_loc_size(s)).str());
}

utf8_string make_loc_time(std::time_t s)
{
	if (s == 0)
		return utf8_string("--");

	return platform::make_time(loc("time.full"), s);
}

utf8_string make_loc_version()
{
	return utf8_string((locf("about.version")
		% BARETORRENT_USER_VERSION
		% BARETORRENT_USER_BUILD_DATE).str());
}


utf8_string make_loc_priority(int i)
{
	switch (i)
	{
		case torrent::skip_priority:
			return loc("torrent.file.skip");

		case torrent::normal_priority:
			return utf8_string();

		case torrent::high_priority:	// fall-through
		default:
		{
			return loc("torrent.file.high");
		}
	}
}

std::vector<utf8_string> explode(
	const utf8_string& s, const utf8_string& sep)
{
	if (s.empty())
		return std::vector<utf8_string>();

	std::vector<std::string> v;
	std::string ss = s.std_string();
	std::string ssep = sep.std_string();

	boost::iter_split(v, ss, boost::first_finder(ssep)); 

	std::vector<utf8_string> r;
	foreach (const std::string& s, v)
		r.push_back(utf8_string(s));

	return r;
}

utf8_string hash_from_url(const utf8_string& url)
{
	// not a magnet link, there's no hash
	if (!boost::istarts_with(url, "magnet:"))
		return utf8_string();

	// magnet links should always have a btih field
	boost::regex e("(?:.*)urn:btih:(.*?)&(?:.*)");
	boost::smatch m;

	const std::string s = url.std_string();
	if (boost::regex_match(s, m, e))
		return utf8_string(m[1].str());

	return utf8_string();
}

utf8_string name_from_url(const utf8_string& url)
{
	utf8_string r = url;

	if (boost::istarts_with(url, "magnet:"))
	{
		// some magnet links have the torrent name in the dn field
		boost::regex e("(?:.*)dn=(.*?)&(?:.*)");
		boost::smatch m;

		const std::string s = url.std_string();
		if (boost::regex_match(s, m, e))
		{
			r = urldecode(utf8_string(m[1]));
		}
		else
		{
			// no dn field, just return the hash
			r = hash_from_url(url);
		}
	}
	else
	{
		boost::regex e("(?:.*)\\/(.*)");
		boost::smatch m;

		// getting the last path component
		const std::string s = url.std_string();
		if (boost::regex_match(s, m, e))
		{
			utf8_string fn(m[1]);

			// getting rid of the extension (if any)
			e = boost::regex("(.*)\\.(?:.*)");

			const std::string s2 = fn.std_string();
			if (boost::regex_match(s2, m, e))
				r = utf8_string(m[1]);
		}
	}

	boost::trim(r);
	return r;
}

bool plausible_torrent_string(const utf8_string& s)
{
	try
	{
		utf8_path p(s);
		if (p.exists())
			return true;
	}
	catch(path_exception&)
	{
		// ignore
	}

	wxURI u(to_wxstring(s).c_str());
	if (u.HasServer())
		return true;

	if (boost::istarts_with(s, "magnet:"))
		return true;

	return false;
}

#if BARETORRENT_VISUALC
#pragma warning(push)
#pragma warning(disable: 4706)
#endif

// adapted from http://www.davekoelle.com/alphanum.html
//
int natural_compare(const char *l, const char *r)
{
	enum mode_t
	{
		STRING,
		NUMBER
	};
	
	mode_t mode = STRING;

	while (*l && *r)
	{
		if (mode == STRING)
		{
			char l_char, r_char;

			while ((l_char=*l) && (r_char=*r))
			{
				if (l_char >= -1 && r_char >= -1)
				{
					// check if this are digit characters
					const bool l_digit = (std::isdigit(l_char) != 0);
					const bool r_digit = (std::isdigit(r_char) != 0);

					// if both characters are digits, we continue in
					// NUMBER mode
					if (l_digit && r_digit)
					{
						mode=NUMBER;
						break;
					}

					// if only the left character is a digit, we have a
					// result
					if (l_digit)
						return -1;

					// if only the right character is a digit, we have a
					// result
					if (r_digit)
						return +1;
				}

				// compute the difference of both characters
				const int diff = std::tolower(l_char) - std::tolower(r_char);

				// if they differ we have a result
				if (diff < 0)
					return -1;
				else if (diff > 0)
					return 1;

				// otherwise process the next characters
				++l;
				++r;
			}
		}
		else // mode==NUMBER
		{
			// get the left number
			char *end;
			unsigned long l_int=strtoul(l, &end, 0);
			
			if (l == end)
			{
				++l;
				++r;
				mode=STRING;
				continue;
			}

			l = end;

			// get the right number
			unsigned long r_int=strtoul(r, &end, 0);
			r = end;

			if (l == end)
			{
				++l;
				++r;
				mode=STRING;
				continue;
			}

			// if the difference is not equal to zero, we have a
			// comparison result
			const long diff=l_int-r_int;
			if (diff < 0)
				return -1;
			else if (diff > 0)
				return 1;

			// otherwise we process the next substring in STRING mode
			mode=STRING;
		}
	}

	if (*r)
		return -1;

	if (*l)
		return +1;

	return 0;
}

#if BARETORRENT_VISUALC
#pragma warning(pop)
#endif

int natural_compare(const utf8_string& a, const utf8_string& b)
{
	return natural_compare(a.c_str(), b.c_str());
}

int make_percent(double c, double t)
{
	const double p = c / t * 100.0;
	return static_cast<int>(p);
}

std::ostream& operator<<(
	std::ostream& out, bt::proxy_settings::proxy_type t)
{
	out << static_cast<int>(t);
	return out;
}

std::istream& operator>>(
	std::istream& in, bt::proxy_settings::proxy_type& t)
{
	int i = 0;
	in >> i;
	t = bt::proxy_settings::proxy_type(i);
	return in;
}

utf8_string directory_separator()
{
	return utf8_string(BARETORRENT_DIRECTORY_SEPARATOR);
}

utf8_string env_path_separator()
{
	return utf8_string(BARETORRENT_PATH_ENV_SEPARATOR);
}

bool real_copy_directory_impl(const utf8_path& p, const utf8_path& to)
{
	if (p.is_directory())
		real_copy_directory(p, to / p.leaf());
	else
		p.copy_file(to / p.leaf());

	return true;
}

void real_copy_directory(const utf8_path& from, const utf8_path& to)
{
	// dummy, used to throw
	boost::system::error_code ec;

	if (!from.exists())
	{
		throw path_exception(
			utf8_string("source doesn't exist"), from, to, ec);
	}

	if (!from.is_directory())
	{
		throw path_exception(
			utf8_string("source is not a directory"), from, to, ec);
	}

	if (to.exists())
	{
		throw path_exception(
			utf8_string("target already exists"), from, to, ec);
	}

	create_directory(to);
	walk_directory(
		from, boost::bind(&real_copy_directory_impl, _1, to));
}

byte_array read_file(const utf8_path& p)
{
	std::ifstream in(p.native().c_str(), std::ios::binary);

	std::ostringstream oss;
	oss << in.rdbuf();

	return byte_array(oss.str());
}

utf8_string to_string(bt::pe_settings::enc_policy p)
{
	switch (p)
	{
		case bt::pe_settings::enabled:
			return utf8_string("enabled");

		case bt::pe_settings::disabled:
			return utf8_string("disabled");

		case bt::pe_settings::forced:		// fall-through
		default:
			return utf8_string("forced");
	}
}

bt::pe_settings::enc_policy enc_policy_from_string(const utf8_string& s)
{
	const std::string ls = boost::to_lower_copy(s.std_string());

	if (ls == "enabled")
		return bt::pe_settings::enabled;
	else if (ls == "disabled")
		return bt::pe_settings::disabled;
	else
		return bt::pe_settings::forced;
}

utf8_string to_string(bt::pe_settings::enc_level lv)
{
	switch (lv)
	{
		case bt::pe_settings::plaintext:
			return utf8_string("plaintext");

		case bt::pe_settings::rc4:
			return utf8_string("rc4");

		case bt::pe_settings::both:		// fall-through
		default:
			return utf8_string("both");
	}
}

bt::pe_settings::enc_level enc_level_from_string(const utf8_string& s)
{
	const std::string ls = boost::to_lower_copy(s.std_string());

	if (ls == "plaintext")
		return bt::pe_settings::plaintext;
	else if (ls == "rc4")
		return bt::pe_settings::rc4;
	else
		return bt::pe_settings::both;
}

utf8_string to_string(power_actions p)
{
	switch (p)
	{
		case shutdown_power:
			return utf8_string("shutdown");

		case suspend_power:
			return utf8_string("suspend");

		case hibernate_power:
			return utf8_string("hibernate");

		case no_power_action:	// fall-through
		default:
			return utf8_string("no power action");
	}
}

utf8_string to_string(bt::storage_mode_t m)
{
	switch (m)
	{
		case bt::storage_mode_allocate:
			return utf8_string("allocate");
		
		case bt::storage_mode_sparse:	// fall-through
		case bt::internal_storage_mode_compact_deprecated:
		default:
			return utf8_string("sparse");
	}
}

bt::storage_mode_t storage_mode_from_string(const utf8_string& m)
{
	if (m == utf8_string("allocate"))
		return bt::storage_mode_allocate;
	else
		return bt::storage_mode_sparse;
}

utf8_string to_string(bt::session_settings::suggest_mode_t i)
{
	switch (i)
	{
		case bt::session_settings::suggest_read_cache:
			return utf8_string("cache");

		case bt::session_settings::no_piece_suggestions: // fall-through
		default:
			return utf8_string("none");

	}
}

bt::session_settings::suggest_mode_t suggest_mode_from_string(
	const utf8_string& s)
{
	if (s == utf8_string("cache"))
		return bt::session_settings::suggest_read_cache;
	else
		return bt::session_settings::no_piece_suggestions;
}

utf8_string to_string(bt::session_settings::choking_algorithm_t s)
{
	switch (s)
	{
		case bt::session_settings::auto_expand_choker:
			return utf8_string("auto_expand");

		case bt::session_settings::rate_based_choker:
			return utf8_string("rate_based");
	
		case bt::session_settings::bittyrant_choker:
			return utf8_string("bittyrant");

		case bt::session_settings::fixed_slots_choker:	// fall-through
		default:
			return utf8_string("fixed_slots");
	}
}

bt::session_settings::choking_algorithm_t choking_algorithm_from_string(
	const utf8_string& s)
{
	if (s == utf8_string("auto_expand"))
		return bt::session_settings::auto_expand_choker;
	else if (s == utf8_string("rate_based"))
		return bt::session_settings::rate_based_choker;
	else if (s == utf8_string("bittyrant"))
		return bt::session_settings::bittyrant_choker;
	else
		return bt::session_settings::fixed_slots_choker;
}

utf8_string to_string(bt::session_settings::seed_choking_algorithm_t s)
{
	switch (s)
	{
		case bt::session_settings::fastest_upload:
			return utf8_string("fastest_upload");

		case bt::session_settings::anti_leech:
			return utf8_string("anti_leech");

		case bt::session_settings::round_robin:		// fall-through
		default:
			return utf8_string("round_robin");
	}
}

bt::session_settings::seed_choking_algorithm_t
	seed_choking_algorithm_from_string(const utf8_string& s)
{
	if (s == utf8_string("fastest_upload"))
		return bt::session_settings::fastest_upload;
	else if (s == utf8_string("anti_leech"))
		return bt::session_settings::anti_leech;
	else
		return bt::session_settings::round_robin;
}

utf8_string to_string(bt::session_settings::io_buffer_mode_t s)
{
	switch (s)
	{
		case bt::session_settings::disable_os_cache_for_aligned_files:
			return utf8_string("disable_for_aligned_files");

		case bt::session_settings::disable_os_cache:
			return utf8_string("disable");

		case bt::session_settings::enable_os_cache:	// fall-through
		default:
			return utf8_string("enable");
			
	}
}

// returns the io buffer mode from the given string
//
bt::session_settings::io_buffer_mode_t io_buffer_mode_from_string(
	const utf8_string& s)
{
	if (s == utf8_string("disable_for_aligned_files"))
		return bt::session_settings::disable_os_cache_for_aligned_files;
	else if (s == utf8_string("disable"))
		return bt::session_settings::disable_os_cache;
	else
		return bt::session_settings::enable_os_cache;
}

utf8_string to_string(bt::session_settings::disk_cache_algo_t s)
{
	switch (s)
	{
		case bt::session_settings::lru:
			return utf8_string("lru");

		case bt::session_settings::largest_contiguous:
			return utf8_string("largest_contiguous");
			
		case bt::session_settings::avoid_readback:	// fall-through
		default:
			return utf8_string("avoid_readback");
	}
}

bt::session_settings::disk_cache_algo_t
	disk_cache_algorithm_from_string(const utf8_string& s)
{
	if (s == utf8_string("lru"))
		return bt::session_settings::lru;
	else if (s == utf8_string("largest_contiguous"))
		return bt::session_settings::largest_contiguous;
	else
		return bt::session_settings::avoid_readback;
}

utf8_string to_string(bt::session_settings::bandwidth_mixed_algo_t s)
{
	switch (s)
	{
		case bt::session_settings::prefer_tcp:
			return utf8_string("prefer_tcp");

		case bt::session_settings::peer_proportional:	// fall-through
		default:
			return utf8_string("peer_proportional");
	}
}

bt::session_settings::bandwidth_mixed_algo_t
	bandwidth_mixed_algo_from_string(const utf8_string& s)
{
	if (s == utf8_string("prefer_tcp"))
		return bt::session_settings::prefer_tcp;
	else
		return bt::session_settings::peer_proportional;
}


std::pair<int, int> parse_ports(const utf8_string& s)
{
	std::istringstream iss(s.std_string());

	int first = 0, last = 0;
	iss >> first >> last;

	if (first == 0)
		return std::make_pair(0, 0);
	else if (last == 0)
		return std::make_pair(first, first);
	else
		return std::make_pair(first, last);
}

byte_array base64_decode(const std::string& a)
{
	boost::shared_array<char> out(new char[a.size()]);
	std::memset(out.get(), 0, a.size());

	BIO* b64 = BIO_new(BIO_f_base64());
	BIO* bmem = BIO_new_mem_buf(
		const_cast<char*>(a.data()), static_cast<int>(a.size()));
	bmem = BIO_push(b64, bmem);
	BIO_set_flags(bmem, BIO_FLAGS_BASE64_NO_NL);

	const int s = BIO_read(bmem, out.get(), static_cast<int>(a.size()));

	BIO_free_all(bmem);

	return byte_array(reinterpret_cast<unsigned char*>(out.get()), s);
}


timer::timer()
	: restarted_(false)
{
}

void timer::start(fun f, boost::posix_time::time_duration t)
{
	restarted_ = true;
	f_ = f;
	delay_ = t;
	Start(static_cast<int>(t.total_milliseconds()));
}

void timer::Notify()
{
	// stops the timer if f() returned false
	if (!f_())
	{
		if (!restarted_)
			Stop();
	}

	restarted_ = false;
}


extension_version extension_version::parse(const utf8_string& s)
{
	extension_version v;

	if (s.empty())
		return v;

	std::size_t p = s.find(utf8_string("."));
	v.major = as<int>(s.substr(0, p));
	
	if (p == utf8_string::npos)
		return v;

	std::size_t last = p + 1;
	p = s.find(utf8_string("."), last);
	v.minor = as<int>(s.substr(last , last - p));

	if (p == utf8_string::npos)
		return v;

	last = p + 1;
	v.maintenance = as<int>(s.substr(last));

	return v;
}

void set_cmd_line(const utf8_string& s)
{
	g_cmd = s;
}

utf8_string cmd_line()
{
	return g_cmd;
}

utf8_string versions()
{
	std::vector<std::pair<utf8_string, utf8_string> > v =
		platform::versions();

	std::ostringstream oss;
	oss
		<< "baretorrent " << BARETORRENT_USER_VERSION << ", "
		<< "built " << BARETORRENT_USER_BUILD_DATE << "\n"
		<< "compiled on " << platform::compile_target() << "\n"
		<< "running on " << platform::name() << "\n";

	std::pair<utf8_string, utf8_string> p;
	foreach (p, v)
		oss << p.first << ": " << p.second << "\n";

	return utf8_string(oss.str());
}

utf8_string diag_info()
{
	std::ostringstream oss;

	oss
		<< "command line is '" << cmd_line() << "'\n"
		<< versions() << "\n"
		<< "configuration path: " << paths::conf() << "\n"
		<< "extensions path: " << paths::extensions() << "\n";

	return utf8_string(oss.str());
}

utf8_string website_url()
{
	return utf8_string("http://baretorrent.org");
}

utf8_string program_name()
{
	return utf8_string("baretorrent");
}

utf8_string user_agent()
{
	std::ostringstream oss;
	oss
		<< "baretorrent/" << BARETORRENT_USER_VERSION << " "
		<< "libtorrent/" << LIBTORRENT_VERSION;

	return utf8_string(oss.str());
}

bt::fingerprint make_fingerprint()
{
	const utf8_string s(
		boost::to_lower_copy(opt("bt.peer_id").std_string()));

	if (s == utf8_string("be"))
	{
		const char* id = "BE";

		const int major = BARETORRENT_VERSION_MAJOR;
		const int minor = BARETORRENT_VERSION_MINOR;

#ifdef BARETORRENT_VERSION_MAINTENANCE
		const int rev = BARETORRENT_VERSION_MAINTENANCE;
#else
		const int rev = 0;
#endif

		return bt::fingerprint(id, major, minor, rev, 0);
	}
	else
	{
		return bt::fingerprint(
			"LT", LIBTORRENT_VERSION_MAJOR, LIBTORRENT_VERSION_MINOR,
			0, 0);
	}
}

baretorrent_version baretorrent_version::parse(const std::string& v)
{
	baretorrent_version nv;

	if (v.empty())
		return nv;

	boost::regex e("(\\d+)(\\.\\d+)?(\\.\\d+)?(-.*)?");
	boost::smatch m;

	if (!boost::regex_match(v, m, e))
	{
		log_error() << "bad version '" << v << "'\n";
		return nv;
	}

	std::vector<std::string> cs;
	for (std::size_t i=1; i<m.size(); ++i)
	{
		std::string s = m[static_cast<int>(i)];
		if (!s.empty())
		{
			if (s[0] == '.')
				s = s.substr(1);

			cs.push_back(s);
		}
	}

	bt_assert(!cs.empty());
	

	if (boost::starts_with(cs.back(), "-"))
	{
		nv.tag = cs.back().substr(1);
		cs.pop_back();
	}
	
	if (cs.empty())
	{
		log_error() << "bad version '" << v << "'\n";
		return nv;
	}

	nv.major = as<int>(cs[0]);
	
	if (cs.size() >= 2)
		nv.minor = as<int>(cs[1]);
	
	if (cs.size() >= 3)
		nv.maintenance = as<int>(cs[2]);

	std::ostringstream oss;
	oss << nv;

	nv.string = oss.str();

	return nv;
}

bool new_version(const baretorrent_version& nv)
{
	if (nv.bad())
		return false;

	return (nv > version());
}

std::ostream& operator<<(std::ostream& o, const baretorrent_version& v)
{
	if (v.bad())
	{
		o << "(bad version)";
		return o;
	}

	o << v.major << "." << v.minor << "." << v.maintenance;

	if (!v.tag.empty())
		o << "-" << v.tag;

	return o;
}

bool operator==(
	const baretorrent_version& left, const baretorrent_version& right)
{
	if (left.major == right.major && left.minor == right.minor)
		if (left.maintenance == right.maintenance)
			if (left.tag == right.tag)
				return true;

	return false;
}

bool operator!=(
	const baretorrent_version& left, const baretorrent_version& right)
{
	return !(left == right);
}

bool operator<(
	const baretorrent_version& left, const baretorrent_version& right)
{
	if (left.major < right.major)
	{
		return true;
	}
	else if (left.major == right.major)
	{
		if (left.minor < right.minor)
		{
			return true;
		}
		else if (left.minor == right.minor)
		{
			if (left.maintenance < right.maintenance)
				return true;
		}
	}

	return false;
}

bool operator<=(
	const baretorrent_version& left, const baretorrent_version& right)
{
	return (left < right) || (left == right);
}

bool operator>(
	const baretorrent_version& left, const baretorrent_version& right)
{
	if (left.major > right.major)
	{
		return true;
	}
	else if (left.major == right.major)
	{
		if (left.minor > right.minor)
		{
			return true;
		}
		else if (left.minor == right.minor)
		{
			if (left.maintenance > right.maintenance)
				return true;
		}
	}

	return false;
}

bool operator>=(
	const baretorrent_version& left, const baretorrent_version& right)
{
	return (left > right) || (left == right);
}

namespace paths
{

static utf8_path g_configuration_path;
static utf8_string g_python_path;
static bool g_clear_sys_path = false;

utf8_path extensions()
{
	return conf() / utf8_string("extensions");
}

utf8_string conf_directory_name()
{
	return platform::conf_directory_name();
}

utf8_path conf()
{	
	if (g_configuration_path.empty())
	{
		// this happens when the path hasn't been specified on the
		// command line (set_configuration_path() would have been
		// called)
		g_configuration_path =
			platform::conf_directory() / conf_directory_name();
	}

	return g_configuration_path;
}

void set_configuration_path(const utf8_path& p)
{
	g_configuration_path = p;
}

void set_python_path(const std::vector<utf8_path>& ps)
{
	g_python_path = utf8_string();

	std::vector<utf8_string> ss;
	foreach (const utf8_path& s, ps)
		ss.push_back(absolute(s).as_utf8_string());		

	g_python_path = implode(ss, env_path_separator());
}

utf8_string python_path()
{
	if (g_python_path.empty())
	{
		const utf8_path p =
			platform::executable_directory() / utf8_string("Lib");

		g_python_path = p.as_utf8_string();
	}

	return g_python_path;
}

void clear_sys_path(bool b)
{
	g_clear_sys_path = b;
}

bool clear_sys_path()
{
	return g_clear_sys_path;
}

utf8_string python_zip()
{
	std::ostringstream oss;
	oss << "python" << PY_MAJOR_VERSION << PY_MINOR_VERSION << ".zip";

	const utf8_path p =
		platform::executable_directory() / utf8_string(oss.str());

	return p.as_utf8_string();
}


void create_all()
{
	if (!conf().exists())
	{
		log() << "creating conf directory " << conf() << "\n";
		create_directories(conf());
	}

	if (!extensions().exists())
	{
		log()
			<< "creating extensions directory "
			<< extensions() << "\n";

		create_directories(extensions());
	}
}

utf8_path locales()
{
	return utf8_path(opt("bt.paths.locales"));
}

utf8_string locale_suffix()
{
	return opt("bt.paths.locale_suffix");
}

utf8_string incomplete_suffix()
{
	return opt("bt.paths.incomplete_suffix");
}

utf8_string torrent_file_suffix()
{
	return opt("bt.paths.torrent_file_suffix");
}

utf8_string resume_file_suffix()
{
	return opt("bt.paths.resume_file_suffix");
}

utf8_string state_file_suffix()
{
	return opt("bt.paths.state_file_suffix");
}

utf8_string session_filename()
{
	return opt("bt.paths.session_filename");
}

utf8_string queue_filename()
{
	return opt("bt.paths.queue_filename");
}

}	// namespace paths

}	// namespace baretorrent
