// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_UTILITY_INCLUDED
#define BARETORRENT_UTILITY_INCLUDED

#include "locale.h"
#include "log.h"

// these are various utility things

namespace baretorrent
{

class torrent;

typedef boost::shared_ptr<torrent> torrent_ptr;
typedef std::vector<torrent_ptr> torrents;

// this will use DebugBreak() on visual c++ and the standard assert
// otherwise
#if BARETORRENT_DEBUG
	#if BARETORRENT_VISUALC 
		template <class T>
		inline void bt_assert(T t)
		{
			if (!t)
				DebugBreak();
		}
	#else
		template <class T>
		inline void bt_assert(T t)
		{
			if (!t)
				assert(false);
		}
	#endif

#else
	template <class T>
	inline void bt_assert(const T&) {}

#endif
	
enum power_actions
{
	no_power_action = 0,
	shutdown_power = 1,
	hibernate_power,
	suspend_power
};

class clipboard : boost::noncopyable
{
public:
	clipboard();
	~clipboard();

	utf8_string get_text() const;
	void set_text(const utf8_string& s);

private:
	bool opened_;
};

// initializes curl in the constructor, cleans is up in the destructor
///
class curl_handle : boost::noncopyable
{
public:
	curl_handle()
		: handle_(nullptr)
	{
		handle_ = curl_easy_init();
	}

	~curl_handle()
	{
		if (handle_)
			curl_easy_cleanup(handle_);
	}

	operator CURL*()
	{
		return handle_;
	}

private:
	CURL* handle_;
};

typedef boost::function<bool (double)> download_callback;

// connects to the given url and returns the data; returns an empty
// byte array if an error occured
//
byte_array get_remote_file(
	const utf8_string& url, download_callback f = download_callback());

// disconects the slots from the signal in the destructor
//
class signal_connections : boost::noncopyable
{
public:
	~signal_connections()
	{
		clear();
	}

	void clear()
	{
		foreach (sig::connection& c, cons_)
			c.disconnect();
	}

	void add(const sig::connection& c)
	{
		cons_.push_back(c);
	}

private:
	std::vector<sig::connection> cons_;
};

namespace db
{

// may be thrown by any database operation
//
class exception
{
public:
	exception(const utf8_string& s)
		: what_(s)
	{
	}

	const utf8_string& what() const
	{
		return what_;
	}

private:
	utf8_string what_;
};

}	// namespace db


// extension version
//
struct extension_version
{
	int major, minor, maintenance;

	extension_version()
		: major(0), minor(0), maintenance(0)
	{
	}

	// parses major[.minor[.maintenance]]
	//
	static extension_version parse(const utf8_string& s);
};


// information about an the extension
//
struct extension_info
{
	std::string guid;
	utf8_string name;
	utf8_string author;
	extension_version version;
	utf8_string short_description;
	utf8_string long_description;
	wxBitmap icon;
	byte_array icon_data;
	wxBitmap small_icon;
	byte_array small_icon_data;
	utf8_string email;
	utf8_string website;
	baretorrent_version requires;
};


// the type of torrent shown in the list
//
enum torrent_categories
{
	// all the torrents
	all_categories = 1,

	// torrents that are not completed
	downloading_category,

	// torrents that are completed
	uploading_category,

	// torrents that are currently transferring data
	active_category,

	// torrents that are not transferring data
	inactive_category
};


// parses two integers as a pair
//
std::pair<int, int> parse_ports(const utf8_string& s);

// uses curl; url-encodes the given string and replaces spaces by
// plus signs
//
utf8_string urlencode(const utf8_string& s);

// uses curl; url-decodes the given string and replaces plus signs by
// spaces
//
utf8_string urldecode(const utf8_string& s);


// returns a localized string for the given size in bytes, such as
// "42 KiB"
//
utf8_string make_loc_size(uint64_t s);

// returns a localized string for the given time duration, such as
// "34 minutes"
//
utf8_string make_loc_duration(int s);

// returns a localized string for the given time duration using
// abbreviated units, such as "3 h"
//
utf8_string make_loc_duration_abbr(int s);

// returns a localized string for the given transfer rate, such as
// "28 MiB/s"
//
utf8_string make_loc_rate(int s);

// returns a localized string for the given time, such as "Sun Nov 11 
// 11:25:04 2012" or "--" if 't' is 0
//
utf8_string make_loc_time(std::time_t t);

// returns a localized string for the given priority, such as "Skip",
// or an empty string for torrent::normal_priority
//
utf8_string make_loc_priority(int i);

// returns a localized string for the version of the program, such as
// "0.1, built Nov 11 2012"
//
utf8_string make_loc_version();

// returns the directory separators in paths for the platform against
// which the program was compiled
//
utf8_string directory_separator();

// returns the separator between paths in the PATH environment variable
//
utf8_string env_path_separator();


// returns the libtorrent encoding policy as a string for the options
//
utf8_string to_string(bt::pe_settings::enc_policy p);

// returns the libtorrent encoding policy from the given string
//
bt::pe_settings::enc_policy enc_policy_from_string(
	const utf8_string& s);

// returns the libtorrent encoding level as a string for the options
//
utf8_string to_string(bt::pe_settings::enc_level lv);

// returns the libtorrent encoding level from the given string
//
bt::pe_settings::enc_level enc_level_from_string(const utf8_string& s);

// returns the given power action as a string for logging
//
utf8_string to_string(power_actions p);

// returns the libtorrent storage mode as a string
//
utf8_string to_string(bt::storage_mode_t m);

// returns the libtorrent storage mode from the given string string
//
bt::storage_mode_t storage_mode_from_string(const utf8_string& m);

// returns the suggest mode as a string
//
utf8_string to_string(bt::session_settings::suggest_mode_t s);

// returns the suggest mode from the given string
//
bt::session_settings::suggest_mode_t suggest_mode_from_string(
	const utf8_string& s);

// returns the choking algorithm as a string
//
utf8_string to_string(bt::session_settings::choking_algorithm_t s);

// returns the choking algorithm from the given string
//
bt::session_settings::choking_algorithm_t choking_algorithm_from_string(
	const utf8_string& s);

// returns the seed choking algorithm as a string
//
utf8_string to_string(bt::session_settings::seed_choking_algorithm_t s);

// returns the seed choking algorithm from the given string
//
bt::session_settings::seed_choking_algorithm_t
	seed_choking_algorithm_from_string(const utf8_string& s);

// returns the io buffer mode as a string
//
utf8_string to_string(bt::session_settings::io_buffer_mode_t s);

// returns the io buffer mode from the given string
//
bt::session_settings::io_buffer_mode_t io_buffer_mode_from_string(
	const utf8_string& s);

// returns the disk cache algorithm as a string
//
utf8_string to_string(bt::session_settings::disk_cache_algo_t s);

// returns the disk cache algorithm from the given string
//
bt::session_settings::disk_cache_algo_t
	disk_cache_algorithm_from_string(const utf8_string& s);

// returns the mixed mode algorithm as a string
//
utf8_string to_string(bt::session_settings::bandwidth_mixed_algo_t s);

// returns the mixed mode algorithm from the given string
//
bt::session_settings::bandwidth_mixed_algo_t
	bandwidth_mixed_algo_from_string(const utf8_string& s);

byte_array base64_decode(const std::string& a);


// compares the two given string by considering consecutive digits as
// a number, so that
//
//		12 something
//
// comes after
//
//		1 something
//
int natural_compare(const utf8_string& a, const utf8_string& b);

template <class T>
void clamp(T& t, T min, T max)
{
	if (t < min)
		t = min;
	else if (t > max)
		t = max;
}

// returns -1, 0 or 1 for the given two values
//
template <class T>
int compare(T left, T right)
{
	if (left < right)
		return -1;
	else if (left > right)
		return 1;
	else
		return 0;
}

// returns whether the given container contains 't', merely convenience
//
template <class Cont>
bool has(const Cont& c, const typename Cont::value_type& t)
{
	if (std::find(c.begin(), c.end(), t) == c.end())
		return false;
	return true;
}

// undefined on purpose, use as_utf8_string
template <class T>
utf8_string as(const T& t);


// converts 't' to a U using operator>> and operator<<
template <class U, class T>
inline U as(const T& t)
{
	std::stringstream ss;
	ss << t;

	U u = U();
	ss >> u;

	return u;
}

// converts 't' to a string using operator<<
//
template <class T>
utf8_string as_utf8_string(const T& t)
{
	std::ostringstream oss;
	oss << t;

	return utf8_string(oss.str());
}

// splits 's' on all separators and returns them in a vector of strings
//
std::vector<utf8_string> explode(
	const utf8_string& s, const utf8_string& sep);

// splits 's' on all separators and returns them in a vector of Ts using
// operator<< to convert the strings
//
template <class T>
std::vector<T> explode(const utf8_string& s, const utf8_string& sep)
{
	std::vector<T> v;

	if (s.empty())
		return v;

	typename utf8_string::size_type pos = 0;
	typename utf8_string::size_type last = 0;

	while (pos != utf8_string::npos)
	{
		pos = s.find_first_of(sep, last);
		if (pos == utf8_string::npos)
		{
			if (last == s.length())
				break;
		}

		std::size_t len = pos-last;
		if (len != 0)
		{
			const utf8_string vs(s.substr(last, pos-last));
			v.push_back(as<T>(vs));
		}
	
		last = pos + 1;
	}

	return v;
}

// merges all the vector elements into a string by using operator<< and
// separates them with 'sep'
//
template <class T>
inline utf8_string implode(
	const std::vector<T>& v, const utf8_string& sep)
{
	utf8_string r("");

	foreach(const T& s, v)
		r += as_utf8_string(s) + sep;

	if (r.length() > 0)
		r = r.substr(0, r.length() - sep.length());

	return r;
}
/*
// merges all the vector elements into a string by using operator<< and
// separates them with 'sep'
//
template <class T>
inline std::wstring implode(
	const std::vector<T>& v, const std::wstring& sep)
{
	std::wstring r;

	foreach(const T& s, v)
		r += as_wstring(s) + sep;

	if (r.length() > 0)
		r = r.substr(0, r.length() - sep.length());

	return r;
}
*/

// parses the given url and tries to figure out the torrent name
//
// for magnet links, this will use the dn or btih field; for other
// urls, this will use the last path component without the extension
//
// if nothing interesting is found, 'url' is returned
//
utf8_string name_from_url(const utf8_string& url);

// returns btih for magnet links or an empty string
//
utf8_string hash_from_url(const utf8_string& url);

// whether the given string could be a torrent url or path
//
bool plausible_torrent_string(const utf8_string& s);


// outputs the proxy type as an int
//
std::ostream& operator<<(
	std::ostream& out, bt::proxy_settings::proxy_type t);

// inputs the proxy type as an int
//
std::istream& operator>>(
	std::istream& in, bt::proxy_settings::proxy_type& t);

inline std::ostream& operator<<(std::ostream& out, const wxRect& r)
{
	out << r.x << " " << r.y << " " << r.width << " " << r.height;
	return out;
}

inline std::istream& operator>>(std::istream& in, wxRect& r)
{
	in >> r.x >> r.y >> r.width >> r.height;
	return in;
}


// converts the given hash to a hex string, such as
// b9c43737f7d4eecf5cf1853fd965aa9902bdd711
//
utf8_string to_utf8_string(const bt::sha1_hash& s);

// returns an int between 0 and 100 by doing current/total
//
int make_percent(double current, double total=1.0);

// returns a string for the given floating point number where there are
// at most 'p' digits after the decimal point
//
utf8_string round(double d, int p=1);

// copies a directory to the given target; the target must be a non
// existent directory in which the content of the source will be
// copied
//
// this functionality does not exist in boost, copy_directory() just
// creates a new directory using the attributes of the source
//
void real_copy_directory(const utf8_path& from, const utf8_path& to);

// reads the content of the file and returns it
//
byte_array read_file(const utf8_path& p);

// displayed in the log dialog
//
utf8_string diag_info();
utf8_string versions();
void set_cmd_line(const utf8_string& s);
utf8_string cmd_line();

void update_options(
	const baretorrent_version& from, const baretorrent_version& to);

utf8_string user_agent();
bt::fingerprint make_fingerprint();

// a wrapper around wxTimer using a boost::function and posix_time
//
class timer : public wxTimer
{
public:
	typedef boost::function<bool ()> fun;


	timer();
	
	// will call 'f' after 't' has elapsed; 'f' should return false
	// to stop the timer
	//
	void start(fun f, boost::posix_time::time_duration t);
	
	void restart(boost::posix_time::time_duration t)
	{
		start(f_, t);
	}

	// fired from wxWidgets when the timer has elapsed
	//
	virtual void Notify();
	
	boost::posix_time::time_duration delay() const
	{
		return delay_;
	}

private:
	// callback
	fun f_;

	// delay
	boost::posix_time::time_duration delay_;

	// set to true inside start(), will prevent Stop() from being
	// called in Notify() if start() was called from the callback;
	// allows for restarting this timer from the callback
	bool restarted_;
};


// forwards to f(); this is just so that firing events is clearer by
// using "fire(event)" instead of "event()"
//
template <class F>
inline void fire(F& f)
{
	f();
}

// forwards to f(); this is just so that firing events is clearer by
// using "fire(event)" instead of "event()"
//
template <class F, class A1>
inline void fire(F& f, A1 a1)
{
	f(a1);
}

// forwards to f(); this is just so that firing events is clearer by
// using "fire(event)" instead of "event()"
//
template <class F, class A1, class A2>
inline void fire(F& f, A1 a1, A2 a2)
{
	f(a1, a2);
}

// forwards to f(); this is just so that firing events is clearer by
// using "fire(event)" instead of "event()"
//
template <class F, class A1, class A2, class A3>
inline void fire(F& f, A1 a1, A2 a2, A3 a3)
{
	f(a1, a2, a3);
}

// returns baretorrent's website
//
utf8_string website_url();

// returns 'baretorrent'
//
utf8_string program_name();

// returns whether the given version corresponds to a newer version of
// baretorrent
//
bool new_version(const baretorrent_version& v);


namespace paths
{

// path where the extensions are
//
utf8_path extensions();

// this is used when changing the configuration path from the command
// line
//
void set_configuration_path(const utf8_path& p);

// returns the configuration path
//
utf8_path conf();

// this is used when changing the python path from the command line
//
void set_python_path(const std::vector<utf8_path>& ps);

// returns the python libraries path
//
utf8_string python_path();

// retrnss whether sys.path should be cleared or the paths appended to
// it
//
bool clear_sys_path();

// sets whether sys.path should be cleared or the paths appended to it
//
void clear_sys_path(bool b);

// returns the pyhonXX.zip filename depending on the compile-time
// version
//
utf8_string python_zip();


// creates all the needed directories (configuration, extensions, etc.)
//
void create_all();

// returns the path of the locale files
//
utf8_path locales();

// returns the suffix of locale files (something like ".txt")
//
utf8_string locale_suffix();


// returns the suffix added to incomplete files in a torrent
//
utf8_string incomplete_suffix();

// returns the suffix of a torrent file (probably ".torrent")
//
utf8_string torrent_file_suffix();

// returns the suffix of resume files (probably ".fastresume")
//
utf8_string resume_file_suffix();

// returns the suffix of state files (probably ".state")
//
utf8_string state_file_suffix();

// returns the session filename
//
utf8_string session_filename();

// returns the queue filename
//
utf8_string queue_filename();

}	// namespace paths

}	// namespace baretorrent

#endif	// BARETORRENT_UTILITY_INCLUDED
