// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "utility.h"
#include "widgets.h"
#include "main_window.h"
#include "ext/ext.h"

namespace baretorrent
{

wxBitmap create_bitmap(const byte_array& data)
{
	// bad images pop up a message box, which sucks
	log_redirector lr;

	wxMemoryInputStream in(data.data(), data.size());
	wxImage im(in, wxBITMAP_TYPE_PNG);

	if (!im.IsOk())
	{
		log_error() << "couldn't create bitmap, " << lr.what() << "\n";
		return wxBitmap();
	}

	return im;
}

wxColour errored_text_color()
{
	return wxColour(255, 0, 0);
}

wxColour normal_text_color()
{
	return wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOWTEXT);
}

wxColour disabled_extension_text_color()
{
	return wxSystemSettings::GetColour(wxSYS_COLOUR_GRAYTEXT);
}


visible_timer::visible_timer()
	: widget_(nullptr)
{
}

void visible_timer::start(wxWindow* what, fun f)
{
	widget_	= what;
	fun_ = f;

	timer_.start(
		boost::bind(&visible_timer::on_timer, this),
		boost::posix_time::milliseconds(100));
}

bool visible_timer::on_timer()
{
	if (widget_->IsShownOnScreen())
	{
		fun_();
		return false;
	}

	return true;
}

status_bar::status_bar(wxFrame* parent) :
	wxStatusBar(parent), incoming_(false), remaining_(0),
	adding_(false)
{
	rebuild();
}

status_bar::icons::iterator status_bar::find_icon(ext::instance_ptr i)
{
	for (icons::iterator itor=icons_.begin(); itor!=icons_.end(); ++itor)
	{
		if (itor->i == i)
			return itor;
	}
	
	return icons_.end();
}

status_bar::icons::iterator status_bar::hit_test_icon(int mx, int my)
{
	if (icons_.empty())
		return icons_.end();
	
	wxRect r;
	GetFieldRect(2, r);

	int x = r.x + r.width - icon_width;
	const int y = r.y + 2;

	for (icons::iterator itor=icons_.begin(); itor!=icons_.end(); ++itor)
	{
		if (mx >= x && mx < (x + icon_width))
		{
			if (my >= y && my < (y + icon_height))
				return itor;
		}

		x -= (icon_width + icon_padding);
	}

	return icons_.end();
}

void status_bar::add_icon(
	ext::instance_ptr i, const utf8_path& p, bool greyed)
{
	icons::iterator itor = find_icon(i);

	if (itor == icons_.end())
	{
		wxStaticBitmap* b = new wxStaticBitmap(this, -1, wxBitmap());
		b->Connect(wxEVT_LEFT_UP, wxMouseEventHandler(status_bar::on_click), nullptr, this);

		itor = icons_.insert(icons_.end(), icon_info(i, b));
	}

	byte_array data = read_file(p);
	wxBitmap bmp = create_bitmap(data);
	wxImage im = bmp.ConvertToImage();
	
	if (im.GetWidth() != 16 || im.GetHeight() != 16)
		im = im.Scale(16, 16, wxIMAGE_QUALITY_HIGH);

	if (greyed)
		im = im.ConvertToGreyscale();

	itor->w->SetBitmap(wxBitmap(im));
	
	rebuild();
}

void status_bar::remove_icon(ext::instance_ptr i)
{
	for (icons::iterator itor=icons_.begin(); itor!=icons_.end(); ++itor)
	{
		if (itor->i == i)
		{
			itor->w->Destroy();
			icons_.erase(itor);
			rebuild();

			return;
		}
	}
}

void status_bar::rebuild()
{
	int fields_count = 2;

	if (!icons_.empty())
		++fields_count;

	const int widths[] = {-1, 150, static_cast<int>(icons_.size() * 20)};

	SetFieldsCount(fields_count);
	SetStatusWidths(fields_count, widths);
	
	set_icon_positions();
}

void status_bar::on_size(wxSizeEvent& e)
{
	set_icon_positions();
	e.Skip();
}

void status_bar::set_icon_positions()
{
	if (icons_.empty())
		return;

	wxRect r;
	GetFieldRect(2, r);

	int x = r.x + r.width - icon_width;
	const int y = r.y + 2;

	foreach (icon_info& i, icons_)
	{
		i.w->SetSize(x, y, icon_width, icon_height);
		x -= (icon_width + icon_padding);
	}
}

void status_bar::set_status(const boost::format& f)
{
	set_status(utf8_string(f.str()));
}

void status_bar::set_status(const utf8_string& s)
{
	if (s.empty())
	{
		if (incoming_)
			SetStatusText(loc_wx("connections.no_remote"), 0);
		else
			SetStatusText(loc_wx("status.ready"), 0);
	}
	else
	{
		SetStatusText(to_wxstring(s), 0);
	}
}

void status_bar::clear_status()
{
	set_status(utf8_string());
}

void status_bar::set_incoming_connection(bool b)
{
	incoming_ = b;
	clear_status();
}

void status_bar::set_remaining(bt::size_type s)
{
	remaining_ = s;

	if (adding_)
		return;

	utf8_string rm;
	if (s > 0)
	{
		rm = utf8_string(str(
			locf("status.remaining") % make_loc_size(s)));
	}

	SetStatusText(to_wxstring(rm), 1);
}

void status_bar::set_add_progress(std::size_t c, std::size_t t)
{
	if (c == t)
	{
		adding_ = false;
		set_remaining(remaining_);
		return;
	}
	
	const utf8_string rm(str(locf("status.adding") % c % t));
	SetStatusText(to_wxstring(rm), 1);

	adding_ = true;
}

void status_bar::on_click(wxMouseEvent& e)
{
	foreach (icon_info& i, icons_)
	{
		wxObject* o = e.GetEventObject();

		if (o == i.w)
		{
			i.i->fire(utf8_string("icon_clicked"));
			break;
		}
	}
}

BEGIN_EVENT_TABLE(status_bar, wxStatusBar)
	EVT_SIZE(status_bar::on_size)
END_EVENT_TABLE()


splitter::splitter(wxWindow* parent, const std::string& o) :
	wxSplitterWindow(
		parent, -1, wxDefaultPosition, wxDefaultSize,
		wxSP_LIVE_UPDATE),
	creating_(true),
	opt_(o)
{
	SetMinimumPaneSize(10);

	if (!opt_.empty())
	{
		// the sash position cannot be set here because the splitter
		// hasn't been laid out; the timer will fire repeatedly until
		// the window has been shown and the sash set
		vtimer_.start(this, boost::bind(&splitter::on_visible, this));
	}

	// saves the new position
	Connect(
		wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGED,
		wxSplitterEventHandler(splitter::on_sash_moved));
}

splitter::~splitter()
{
	Disconnect(
		wxEVT_COMMAND_SPLITTER_SASH_POS_CHANGED,
		wxSplitterEventHandler(splitter::on_sash_moved));
}

void splitter::on_visible()
{
	// the splitter has been laid out, the sash position can be used

	log_vv()
		<< "splitter visible, setting sash to "
		<< opt<int>(opt_) << "\n";

	// setting the position from the options
	SetSashPosition(opt<int>(opt_));
	creating_ = false;
}

bool splitter::on_timer()
{
//	if (!opt_.empty())
//	{
//	}

	return false;
}

void splitter::on_sash_moved(wxSplitterEvent& e)
{
	if (creating_)
	{
		e.SetSashPosition(-1);
		return;
	}

		log_vv()
			<< "sash " << opt_ << " moved to "
			<< GetSashPosition() << "\n";

		opt(opt_, GetSashPosition());
/*	timer_.start(
		boost::bind(&splitter::on_timer, this),
		boost::posix_time::milliseconds(500));*/

	e.Skip();
}

void splitter::OnDoubleClickSash(int, int)
{
	// noop
}


custom_dialog::custom_dialog(
	wxWindow* parent, const utf8_string& title, bool resizable)
		: wxDialog(
			parent, -1, to_wxstring(title), wxDefaultPosition, wxDefaultSize,
			wxDEFAULT_DIALOG_STYLE | (resizable ? wxRESIZE_BORDER : 0))
{
}


custom_frame::custom_frame(const std::string& o)
	: wxFrame(nullptr, -1, wxEmptyString), creating_(true), option_(o)
{
	if (!option_.empty())
	{
		Connect(wxEVT_ICONIZE,
			wxIconizeEventHandler(custom_frame::on_iconize));
		Connect(wxEVT_SIZE, wxSizeEventHandler(custom_frame::on_size));
		Connect(wxEVT_MOVE, wxMoveEventHandler(custom_frame::on_move));

		if (!opt(option_ + ".bounds").empty())
		{
			log_vv()
			   << "setting window size to "
			   << opt(option_ + ".bounds") << "\n";

			SetSize(opt<wxRect>(option_ + ".bounds"));
		}

		// the maximize state is independent of the minimize option since
		// an iconized window may well be maximized when restored
		if (opt<bool>(option_ + ".maximized"))
			Maximize();
	}
}

custom_frame::~custom_frame()
{
	if (!option_.empty())
	{
		Disconnect(wxEVT_ICONIZE,
			wxIconizeEventHandler(custom_frame::on_iconize));
		Disconnect(wxEVT_SIZE, wxSizeEventHandler(custom_frame::on_size));
		Disconnect(wxEVT_MOVE, wxMoveEventHandler(custom_frame::on_move));
	}
}

void custom_frame::on_iconize(wxIconizeEvent& e)
{
	// this happens when the window is being created
	if (creating_)
		return;

	// hook
	iconized(e.IsIconized());
}

bool custom_frame::on_timer()
{
	// GetRect() returns the bounds of either the restored or maximized
	// window; therefore, the bounds should only be saved when the
	// window is not maximized
	if (!IsMaximized())
	{
		const wxRect r = GetRect();

		log_vv()
			<< "window moved, changed to " << r.x << " " << r.y << " "
		   	<< r.width << " " << r.height << "\n";

		opt(option_ + ".bounds", GetRect());
	}

	opt(option_ + ".maximized", IsMaximized());

	return false;
}

void custom_frame::on_size(wxSizeEvent& e)
{
	// this happens when the window is being created
	if (creating_)
		return;

	timer_.start(
		boost::bind(&custom_frame::on_timer, this),
		boost::posix_time::milliseconds(500));

	e.Skip();
}

void custom_frame::on_move(wxMoveEvent&)
{
	// this happens when the window is being created
	if (creating_)
		return;

	timer_.start(
		boost::bind(&custom_frame::on_timer, this),
		boost::posix_time::milliseconds(500));
}


multiline_text::multiline_text(
	wxWindow* parent, wxWindowID id, const wxString& v, long style) :
		wxTextCtrl(
			parent, id, v, wxDefaultPosition, wxDefaultSize,
			style | wxTE_RICH2 | wxTE_MULTILINE)
{
#if BARETORRENT_WINDOWS
	Connect(
		wxEVT_CHAR,
		wxKeyEventHandler(multiline_text::on_char));
#endif
}

void multiline_text::on_char(wxKeyEvent& e)
{
	if (e.GetKeyCode() == 1)
		SetSelection(-1, -1);
	e.Skip();
}


notification_panel::notification_panel(wxWindow* parent)
	: wxPanel(parent), box_(nullptr)
{
	SetBackgroundColour(wxColour(255, 255, 225));

	box_ = new wxBoxSizer(wxVERTICAL);

	SetSizer(box_);

	Hide();
}

void notification_panel::add(notification* p)
{
	box_->Add(p, 0, wxEXPAND);
	p->Hide();
}

void notification_panel::close(notification* p)
{
	p->Hide();
	GetParent()->Layout();
}

void notification_panel::show(notification* p)
{
	p->Show();
	Show();
	GetParent()->Layout();
}


notification::notification(notification_panel* p) :
	wxPanel(
		p, wxID_ANY, wxDefaultPosition, wxDefaultSize,
		wxTAB_TRAVERSAL | wxBORDER_THEME),
	notif_(p)
{
	notif_->add(this);
}

void notification::close()
{
	notif_->close(this);
}

}	// namespace baretorrent
