// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_UI_WIDGETS_INCLUDED
#define BARETORRENT_UI_WIDGETS_INCLUDED

#include "options.h"

// these are various convenience classes and function for dealing with
// widgets

namespace baretorrent
{

class main_window;

// freezes updates for the given window in the constructor, thaws the
// widget in the destructor
//
class freezer : boost::noncopyable
{
public:
	freezer(wxWindow* w)
		: w_(w)
	{
		w_->Freeze();
	}

	~freezer()
	{
		w_->Thaw();
	}

private:
	wxWindow* w_;
};

class status_bar : public wxStatusBar
{
public:
	status_bar(wxFrame* parent);

	void add_icon(ext::instance_ptr i, const utf8_path& p, bool greyed);
	void remove_icon(ext::instance_ptr i);

	void clear_status();

	void set_status(const utf8_string& s);
	void set_status(const boost::format& f);
	void set_remaining(bt::size_type s);
	void set_incoming_connection(bool b);
	void set_add_progress(std::size_t c, std::size_t t);

private:
	static const int icon_width = 16;
	static const int icon_height = 16;
	static const int icon_padding = 4;

	struct icon_info
	{
		ext::instance_ptr i;
		wxStaticBitmap* w;

		icon_info(ext::instance_ptr i, wxStaticBitmap* w)
			: i(i), w(w)
		{
		}
	};

	DECLARE_EVENT_TABLE()

	typedef std::vector<icon_info> icons;
	
	icons icons_;
	bool incoming_;
	bt::size_type remaining_;
	bool adding_;

	void rebuild();
	void on_size(wxSizeEvent&);
	void on_click(wxMouseEvent& e);
	void set_icon_positions();
	icons::iterator find_icon(ext::instance_ptr i);
	icons::iterator hit_test_icon(int x, int y);
};

// fires a callback when the given window first becomes visible; this
// is used by various widgets to setup stuff that depends on the control
// being laid out
//
class visible_timer
{
public:
	typedef boost::function<void ()> fun;

	visible_timer();

	// starts the timer which will call 'f' when the given window
	// becomes visible
	//
	void start(wxWindow* what, fun f);

private:
	// the widget to check
	wxWindow* widget_;

	// the callback
	fun fun_;
	timer timer_;

	bool on_timer();	
};

class custom_dialog : public wxDialog
{
public:
	custom_dialog(
		wxWindow* parent, const utf8_string& title, bool resizable = true);
};

class custom_frame : public wxFrame
{
public:
	custom_frame(const std::string& o="");
	~custom_frame();

protected:
	// set to true when the window is being created; prevents from
	// saving options because some events may fired while actually
	// loading and applying the options
	bool creating_;

	virtual void iconized(bool) {}

private:
	std::string option_;
	timer timer_;

	// fired when the window has moved so the options may be updated
	//
	void on_move(wxMoveEvent& e);

	// fired when the window has been minimized so it can be hidden if
	// minimizing to tray
	//
	void on_iconize(wxIconizeEvent& e);

	// fired when the window is resized so the options may be updated
	//
	void on_size(wxSizeEvent& e);

	bool on_timer();
};

// a wrapper around a wxSplitterWindow that can remember its position
//
// the name of the option that has the position of the sash is given in
// the constructor
//
class splitter : public wxSplitterWindow
{
public:
	// creates a splitter
	//
	splitter(wxWindow* parent, const std::string& opt="");

	~splitter();

private:
	// set to true until the window has been shown; prevents from
	// overwriting the position with the default one
	bool creating_;

	// fires until the window is shown so its sash can be positioned
	visible_timer vtimer_;

	timer timer_;

	// option containing the sash position
	std::string opt_;


	// fired when the user moves the sash; saves the position
	//
	void on_sash_moved(wxSplitterEvent& e);

	// fired by timer_; sets the sash position when the window is
	// visible
	//
	void on_visible();

	bool on_timer();

	// prevents double-click on sash
	//
	virtual void OnDoubleClickSash(int x, int y);
};

class multiline_text : public wxTextCtrl
{
public:
	multiline_text(
		wxWindow* parent, wxWindowID id, const wxString& v = "",
		long style = 0);

private:
	void on_char(wxKeyEvent& k);
};

class notification_panel;

class notification : public wxPanel
{
public:
	notification(notification_panel* p);
	
protected:
	void close();

private:
	notification_panel* notif_;
};

class notification_panel : public wxPanel
{
	friend class notification;

public:
	notification_panel(wxWindow* parent);

	void show(notification* p);
	void close(notification* p);

private:
	wxBoxSizer* box_;

	void add(notification* p);
};

// outputs in order the x, y, w and h separated by a single space
//
std::ostream& operator<<(std::ostream& out, const wxRect& r);

// expects in order the x, y, w and h separated by whitespaces
//
std::istream& operator>>(std::istream& in, wxRect& r);


// creates a bitmap from the given binary data
//
wxBitmap create_bitmap(const byte_array& data);


// returns the color used in the torrent list when a torrent is errored
//
wxColour errored_text_color();

// returns the color used in the torrent list for torrents in a normal
// state
//
wxColour normal_text_color();

// returns the color used for text if an extension is disabled
//
wxColour disabled_extension_text_color();


// changes the wxLog target to a stream and restores the previous
// one in the destructor
//
class log_redirector : boost::noncopyable
{
public:
	// careful: wxLogStream takes a pointer to oss_ and therefore must
	// be declared afterwards below
	//
	log_redirector()
		: log_(&oss_), old_(nullptr)
	{
		old_ = wxLog::SetActiveTarget(&log_);
	}

	~log_redirector()
	{
		try
		{
			wxLog::SetActiveTarget(old_);
		}
		catch(...)
		{
			// ignore
		}
	}

	utf8_string what() const
	{
		utf8_string s(oss_.str());
		boost::trim(s);

		return s;
	}

private:
	// careful: wxLogStream takes a pointer to oss_ and therefore must
	// be declared afterwards
	std::ostringstream oss_;
	wxLogStream log_;

	wxLog* old_;
};

}	// namespace baretorrent

#endif	// BARETORRENT_UI_WIDGETS_INCLUDED
