// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"

#if BARETORRENT_WINDOWS

#include "platform.h"
#include "utility.h"
#include "resources.h"
#include "app.h"

// openssl has a structure named ACCESS_DESCRIPTION and it conflicts
// with an enum in PowrProf.h
#define ACCESS_DESCRIPTION ACCESS_DESCRIPTION_
#include <PowrProf.h>
#undef ACCESS_DESCRIPTION

#if _MSC_VER == 1700
	// hack for visual c++ 11, should be fixed in rtm
	void operator delete(void*, int, const char*, int)
	{
		baretorrent::bt_assert(false);
	}
#endif


namespace baretorrent
{
namespace platform
{

const utf8_string prog_id("baretorrent.1");
const utf8_string torrent_mime("application/x-bittorrent");

typedef std::map<images::image, utf8_string> image_map;

// translates the hresult to a localized system message
//
utf8_string system_message(long r)
{
	wchar_t* buffer = nullptr;
	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS,
		nullptr, r, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		reinterpret_cast<wchar_t*>(&buffer), 0, nullptr);

	utf16_string s(buffer);

	LocalFree(buffer);

	return utf16_to_utf8(s);
}

// logs and throws a platform::exception if the given hresult is failed
//
void throw_if_failed(HRESULT r, const utf8_string& s)
{
	if (FAILED(r))
	{
		log_error() << s << ", " << system_message(r) << "\n";
		throw baretorrent::platform::exception();
	}
}

void initialize_com()
{
	HRESULT r = ::CoInitializeEx(
		nullptr, COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE);

	if (r != RPC_E_CHANGED_MODE)
		throw_if_failed(r, utf8_string("couldn't initialize com"));
}

void cleanup_com()
{
	CoUninitialize();
}

// used with shared_ptr as a deleted for com objects, calls Release() on
// the object
//
template <class T>
void com_deleter(T* t)
{
	try
	{
		if (t)
			t->Release();
	}
	catch(...)
	{
	}
}

// used with shared_ptr as a deleter for task objects, calls
// CoTaskMemFree() on the pointer
//
void task_deleter(void* p)
{
	try
	{
		if (p)
			CoTaskMemFree(p);
	}
	catch(...)
	{
	}
}

// used with shared_ptr as a deleter for bstr's, calls SysFreeString()
// on the object
//
void bstr_deleter(BSTR s)
{
	SysFreeString(s);
}


// creates a shared_ptr from the given pointer for com objects
//
template <class T>
boost::shared_ptr<T> com_ptr(T* t)
{
	if (!t)
		return boost::shared_ptr<T>();

	return boost::shared_ptr<T>(t, com_deleter<T>);
}

// creates a shared_ptr from the given pointer for task objects
//
template <class T>
boost::shared_ptr<T> task_ptr(T* t)
{
	if (!t)
		return boost::shared_ptr<T>();

	return boost::shared_ptr<T>(t, task_deleter);
}

// a BSTR is a pointer to an OLECHAR; the typedef is used instead
// because OLECHAR is ugly
typedef boost::shared_ptr<OLECHAR> bstr;

// creates a shared_ptr from the given bstr
//
bstr bstr_ptr(BSTR s)
{
	if (!s)
		return bstr();

	return bstr(s, bstr_deleter);
}


platform::platform()
{
	initialize_com();
}

platform::~platform()
{
	try
	{
		cleanup_com();
	}
	catch(...)
	{
	}
}

	
long toolbar_style()
{
	// todo: this looks okay on the classic theme on windows 7, haven't
	// tested it on anything else
	return wxBORDER_NONE | wxTB_FLAT;
}

void init_list_control(wxListCtrl*)
{
	// noop
}

void destroy_list_control(wxListCtrl*)
{
	// noop
}

void set_list_control_style(wxListCtrl* list)
{
	// double-buffered lists so they don't flicker when resized
	ListView_SetExtendedListViewStyleEx(
		(HWND)list->GetHWND(),
		LVS_EX_DOUBLEBUFFER | LVS_EX_HEADERDRAGDROP,
		LVS_EX_DOUBLEBUFFER);

	// for whatever reason (wxwidgets?), list headers have an offset
	// on the left side
	HWND h = ListView_GetHeader(
		reinterpret_cast<HWND>(list->GetHWND()));
	Header_SetBitmapMargin(h, 0);
}

void set_tree_control_style(wxTreeCtrl* tree)
{
#if BARETORRENT_VISTA_PLUS
	// double-buffered trees so they don't flicker when resized
	TreeView_SetExtendedStyle(
		(HWND)tree->GetHWND(),
		TVS_EX_DOUBLEBUFFER, TVS_EX_DOUBLEBUFFER);
#else
	tree = tree;
#endif
}

void set_list_sort_column(wxListCtrl* list, std::size_t ci, int dir)
{
	HWND h = ListView_GetHeader(
		reinterpret_cast<HWND>(list->GetHWND()));

	// for each column
	for (int i=0; i<list->GetColumnCount(); ++i)
	{
		HDITEM col = {0};

		// getting the current format
		col.mask = HDI_FORMAT;
		Header_GetItem(h, i, &col);

		// removing the sort arrow, whatever the direction was
		col.fmt &= ~(HDF_SORTDOWN | HDF_SORTUP);

		if (static_cast<std::size_t>(i) == ci)
		{
			// setting the arrow on this column
			if (dir > 0)
				col.fmt |= HDF_SORTDOWN;
			else if (dir < 0)
				col.fmt |= HDF_SORTUP;
		}
		
		Header_SetItem(h, i, &col);
	}
}

void open_directory(const utf8_path& p)
{
	SHELLEXECUTEINFO si = {0};

	const utf16_string f = utf8_to_utf16(p.as_utf8_string());

	si.cbSize = sizeof(si);
	si.lpVerb = L"explore";
	si.lpFile = f.c_str();
	si.nShow = SW_SHOW;

	if (!ShellExecuteEx(&si))
	{
		log_error()
			<< "couldn't open directory " << p << ", "
			<< system_message(GetLastError()) << "\n";
	}
}

void open_file(const utf8_path& p)
{
	SHELLEXECUTEINFO si = {0};

	const utf16_string f = utf8_to_utf16(p.as_utf8_string());

	si.cbSize = sizeof(si);
	si.lpVerb = L"open";
	si.lpFile = f.c_str();
	si.nShow = SW_SHOW;

	if (!ShellExecuteEx(&si))
	{
		log_error()
			<< "couldn't open file " << p << ", "
			<< system_message(GetLastError()) << "\n";
	}
}

enum known_directories
{
	download_known_directory = 1,
	documents_known_directory,
	app_data_known_directory,
	local_app_data_known_directory
};

utf8_path known_directory(known_directories d)
{
	wchar_t* ss = 0;
	HRESULT r = 0;
	utf8_string name;

#if BARETORRENT_VISTA_PLUS
	KNOWNFOLDERID id;

	switch(d)
	{
		case download_known_directory:
			id = FOLDERID_Downloads;
			name = utf8_string("FOLDERID_Downloads");
			break;

		case documents_known_directory:
			id = FOLDERID_Documents;
			name = utf8_string("FOLDERID_Documents");
			break;

		case app_data_known_directory:
			id = FOLDERID_RoamingAppData;
			name = utf8_string("FOLDERID_RoamingAppData");
			break;

		case local_app_data_known_directory:
			id = FOLDERID_LocalAppData;
			name = utf8_string("FOLDERID_LocalAppData");
			break;
	}

	r = SHGetKnownFolderPath(id, KF_FLAG_CREATE, 0, &ss);
	boost::shared_ptr<wchar_t> s = task_ptr(ss);

#else
	wchar_t buffer[MAX_PATH + 1] = {0};
	int id = 0;

	switch(d)
	{
		case download_known_directory:
			// xp doesn't have a downloads directory, using my
			// documents
			id = CSIDL_MYDOCUMENTS;
			name = utf8_string("CSIDL_MYDOCUMENTS");
			break;

		case documents_known_directory:
			id = CSIDL_MYDOCUMENTS;
			name = utf8_string("CSIDL_MYDOCUMENTS");
			break;

		case app_data_known_directory:
			id = CSIDL_APPDATA;
			name = utf8_string("CSIDL_APPDATA");
			break;

		case local_app_data_known_directory:
			id = CSIDL_LOCAL_APPDATA;
			name = utf8_string("CSIDL_LOCAL_APPDATA");
			break;
	}

	r = SHGetFolderPath(
		0, id, nullptr, SHGFP_TYPE_CURRENT, buffer);

	ss = buffer;

#endif

	if (SUCCEEDED(r) && ss)
	{
		return utf8_path(utf16_to_utf8(utf16_string(ss)));
	}
	else
	{
		log_error()
			<< "couldn't get known folder " << name << ", "
			<< system_message(r) << "\n";
	}

	return utf8_path();
}

utf8_path download_directory()
{
	// trying downloads directory
	utf8_path p = known_directory(download_known_directory);
	if (!p.empty())
		return p;

	// this probably shouldn't fail, but fall back on the documents
	// directory just in case
	
	p = known_directory(documents_known_directory);
	if (!p.empty())
		return p;

	// again, this probably shouldn't fail, fall back on c:\ since the
	// user probably has bigger problems than a failing baretorrent

	return utf8_path(utf8_string("c:\\"));
}

utf8_path executable_path()
{
	const int tries = 5;
	
	DWORD size = MAX_PATH;
	boost::shared_array<wchar_t> buffer;
	utf16_string ws;

	for (int i=0; i<tries; ++i)
	{
		buffer.reset(new wchar_t[size]);
		
		DWORD s = ::GetModuleFileName(0, buffer.get(), size);
		
		if (s == 0)
		{
			log_error()
				<< "couldn't get module filename, "
				<< system_message(GetLastError()) << "\n";

			break;
		}
		else if (s == size)
		{
			size *= 2;
		}
		else
		{
			ws = utf16_string(buffer.get(), s);
			break;
		}
	}

	return utf8_path(utf16_to_utf8(ws));
}

utf8_path executable_directory()
{
	return executable_path().parent();
}

utf8_path conf_directory()
{
	utf8_path p = known_directory(app_data_known_directory);
	if (!p.empty())
		return p;

	// this shouldn't fail, but fall back on %LOCALAPPDATA% just in
	// case

	p = known_directory(local_app_data_known_directory);
	if (!p.empty())
		return p;

	// again, this shouldn't fail, but fall back on the executable
	// directory just in case

	p = executable_directory();
	if (!p.empty())
		return p;

	// just use the current path and be done with it

	p = current_path();
	if (!p.empty())
		return p;

	// whatever
	return utf8_path(utf8_string("c:\\"));
}

utf8_string conf_directory_name()
{
	return utf8_string("baretorrent");
}


class bad_registry {};

class hkey_holder : boost::noncopyable
{
public:
	hkey_holder(HKEY k = 0)
		: key_(k)
	{
	}

	~hkey_holder()
	{
		if (key_)
			RegCloseKey(key_);
	}

	operator HKEY()
	{
		return key_;
	}

private:
	HKEY key_;
};

HKEY open_key(const utf8_string& name)
{
	const utf16_string utf16_name = utf8_to_utf16(name);

	HKEY key = 0;
	DWORD disp = 0;

	LONG r = RegCreateKeyEx(
		HKEY_CURRENT_USER, utf16_name.c_str(), 0, 0, 0,
		KEY_ALL_ACCESS, 0, &key, &disp);
	
	if (r != ERROR_SUCCESS)
	{
		log()
			<< "registry: could not create " << name << ": "
			<< system_message(r) << "\n";

		throw bad_registry();
	}

	return key;
}

void create_subkey(const utf8_string& name)
{
	hkey_holder h = open_key(name);
}

bool has_key_value(const utf8_string& subkey, const utf8_string& value)
{
	const utf16_string utf16_value = utf8_to_utf16(value);

	hkey_holder k = open_key(subkey);

	LONG r = RegQueryValueEx(k, utf16_value.c_str(), 0, 0, 0, 0);

	if (r == ERROR_SUCCESS)
		return true;
	else
		return false;
}

bool has_key(const utf8_string& subkey)
{
	const utf16_string utf16_subkey = utf8_to_utf16(subkey);

	HKEY k = 0;

	LONG r = RegOpenKeyEx(
		HKEY_CURRENT_USER, utf16_subkey.c_str(), 0,
		KEY_QUERY_VALUE, &k);

	if (r == ERROR_SUCCESS)
		return true;
	else
		return false;

}

template <class T>
T get_key_value(
	const utf8_string& subkey,
	const utf8_string& value = utf8_string());

template <>
DWORD get_key_value<DWORD>(
	const utf8_string& subkey, const utf8_string& value)
{
	const utf16_string utf16_value = utf8_to_utf16(value);

	DWORD w = 0;
	DWORD type = 0;
	DWORD s = sizeof(DWORD);

	hkey_holder k = open_key(subkey);
	LONG r = RegQueryValueEx(
		k, utf16_value.c_str(), 0, &type,
		reinterpret_cast<LPBYTE>(&w), &s);

	if (r != ERROR_SUCCESS)
	{
		log_error()
			<< "couldn't get value '" << value << "' "
			<< "for registry key '" << subkey << "': "
			<< system_message(r) << "\n";
	}

	return w;
}

template <>
utf8_string get_key_value<utf8_string>(
	const utf8_string& subkey, const utf8_string& value)
{
	std::size_t size = 1000;

	const utf16_string utf16_subkey = utf8_to_utf16(subkey);
	const utf16_string utf16_value = utf8_to_utf16(value);

	for (int tries=0; tries<5; ++tries)
	{
		boost::shared_array<wchar_t> buffer;
		
		try
		{
			buffer.reset(new wchar_t[size + 1]);
		}
		catch(std::bad_alloc&)
		{
			throw bad_registry();
		}

		std::memset(buffer.get(), 0, (size + 1) * sizeof(wchar_t));

		DWORD s = static_cast<DWORD>(size * sizeof(wchar_t));
		DWORD type = 0;

		hkey_holder k = open_key(subkey);
		LONG r = RegQueryValueEx(
			k, utf16_value.c_str(), 0, &type,
			reinterpret_cast<LPBYTE>(buffer.get()), &s);

		if (r == ERROR_MORE_DATA)
		{
			size *= 2;
			continue;
		}
		else if (r == ERROR_FILE_NOT_FOUND)
		{
			return utf8_string();
		}
		else if (r != ERROR_SUCCESS)
		{
			log_error()
				<< "couldn't get value '" << value << "' "
				<< "for registry key '" << subkey << "': "
				<< system_message(r) << "\n";

			break;
		}
		else
		{
			return utf16_to_utf8(utf16_string(buffer.get(), s));
		}
	}

	throw bad_registry();
}

utf8_string get_key_value(
	const utf8_string& subkey, const utf8_string& value = utf8_string())
{
	return get_key_value<utf8_string>(subkey, value);
}

std::vector<utf8_string> get_reg_subkeys(const utf8_string& subkey)
{
	hkey_holder key = open_key(subkey);

	DWORD i = 0;
	std::size_t size = 1000;

	std::vector<utf8_string> v;

	bool done = false;

	for (;;)
	{
		bool ok = false;

		for (int tries=0; tries<5; ++tries)
		{
			boost::shared_array<wchar_t> buffer;
		
			try
			{
				buffer.reset(new wchar_t[size + 1]);
			}
			catch(std::bad_alloc&)
			{
				throw bad_registry();
			}

			std::memset(buffer.get(), 0, (size + 1) * sizeof(wchar_t));

			DWORD s = static_cast<DWORD>(size);
			LONG r = RegEnumValue(key, i, buffer.get(), &s, 0, 0, 0, 0);

			if (r == ERROR_NO_MORE_ITEMS)
			{
				done = true;
				break;
			}
			else if (r == ERROR_MORE_DATA)
			{
				size *= 2;
				continue;
			}
			else if (r != ERROR_SUCCESS)
			{
				log_error()
					<< "could enumerate keys in " << subkey << ": "
					<< system_message(r) << "\n";

				break;
			}

			utf16_string name(buffer.get(), s);
			v.push_back(utf16_to_utf8(name));
			ok = true;

			break;
		}

		if (done)
			break;

		if (!ok)
			throw bad_registry();

		++i;
	}

	return v;
}

void set_key_value_impl(
	const utf8_string& key, const utf8_string& name,
	DWORD type, const void* value, std::size_t size)
{
	const utf16_string utf16_name = utf8_to_utf16(name);

	hkey_holder k = open_key(key);

	LONG r = RegSetValueEx(
		k, utf16_name.c_str(), 0, type,
		reinterpret_cast<const BYTE*>(value), static_cast<DWORD>(size));

	if (r != ERROR_SUCCESS)
	{
		log_error()
			<< "registry: could not set key "
			<< "'" << name << "': " << system_message(r) << "\n";

		throw bad_registry();
	}
}

template <class T>
void set_key_value(
	const utf8_string& key,
	const utf8_string& name, const T& value);

template <>
void set_key_value<DWORD>(
	const utf8_string& key,
	const utf8_string& name, const DWORD& value)
{
	set_key_value_impl(
		key, name, REG_DWORD, &value, sizeof(DWORD));
}

template <>
void set_key_value<utf8_string>(
	const utf8_string& key,
	const utf8_string& name, const utf8_string& value)
{
	const utf16_string utf16_value = utf8_to_utf16(value);

	set_key_value_impl(
		key, name, REG_SZ,
		utf16_value.c_str(), utf16_value.size_in_bytes());
}

void set_key_value(
	const utf8_string& key,
	const utf8_string& name, const utf8_string& value)
{
	set_key_value<utf8_string>(key, name, value);
}

template <class T>
void set_key_if_needed(
	const utf8_string& subkey,
	const utf8_string& name, const T& value)
{
	const T def = get_key_value<T>(subkey, name);

	bool create = false;
	if (def == T())
	{
		log()
			<< "registry: '" << name << "' value doesn't exist, "
			<< "creating\n";
		create = true;
	}
	else if (def != value)
	{
		log()
			<< "app paths: wrong (default) value: got '" << def << "', "
			<< "was expecting '" << value << "'\n";
		create = true;
	}
	else
	{
		log() << "app paths: (default) value looks right\n";
	}

	if (create)
		set_key_value<T>(subkey, name, value);
}

void set_key_if_needed(
	const utf8_string& subkey,
	const utf8_string& name, const utf8_string& value)
{
	set_key_if_needed<utf8_string>(subkey, name, value);
}


void delete_key(const utf8_string& name)
{
	const utf16_string utf16_name = utf8_to_utf16(name);

	LONG r = 0;
	
#if BARETORRENT_VISTA_PLUS
	RegDeleteKeyEx(HKEY_CURRENT_USER, utf16_name.c_str(), 0, 0);
#else
	RegDeleteKey(HKEY_CURRENT_USER, utf16_name.c_str());
#endif

	if (r != ERROR_SUCCESS)
	{
		log_error()
			<< "could not delete key '" << name << "': "
			<< system_message(r) << "\n";
	}
}

bool check_app_paths(bool create)
{
	const utf8_string subkey(
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\"
			"baretorrent.exe");
	
	const utf8_string exe = executable_path().as_utf8_string();
	const utf8_string dir = executable_directory().as_utf8_string();

	log() << "app paths: checking " << subkey << "\n";

	if (get_key_value(subkey) != exe)
	{
		if (create)
		{
			set_key_if_needed(subkey, utf8_string(), exe);
		}
		else
		{
			log() << "app paths: doesn't exist\n";
			return false;
		}
	}

	if (get_key_value(subkey, utf8_string("Path")) != dir)
	{
		if (create)
		{
			set_key_if_needed(subkey, utf8_string("Path"), dir);
		}
		else
		{
			log() << "app paths: wrong Path\n";
			return false;
		}
	}

	if (get_key_value<DWORD>(subkey, utf8_string("UseUrl")) != 1)
	{
		if (create)
		{
			set_key_if_needed<DWORD>(subkey, utf8_string("UseUrl"), 1);
		}
		else
		{
			log() << "app paths: wrong UseUrl\n";
			return false;
		}
	}
	
	return true;
}

bool check_command(const utf8_string& subkey, bool create)
{
	const utf8_string exe =
		executable_path().as_utf8_string() + utf8_string(" \"%1\"");

	const utf8_string shell(subkey + utf8_string("\\shell"));
	const utf8_string open(shell + utf8_string("\\open"));
	const utf8_string cmd(open + utf8_string("\\command"));

	if (get_key_value(cmd) != exe)
	{
		if (create)
		{
			create_subkey(shell);
			create_subkey(open);
			set_key_if_needed(cmd, utf8_string(), exe);
		}
		else
		{
			log() << "progid: open command not set\n";
			return false;
		}
	}

	return true;
}

bool check_default_icon(const utf8_string& subkey, bool create)
{
	const utf8_string icon(subkey + utf8_string("\\DefaultIcon"));
	const utf8_string icon_path =
		executable_path().as_utf8_string() + utf8_string(",0");

	if (get_key_value(icon) != icon_path)
	{
		if (create)
		{
			set_key_if_needed(icon, utf8_string(), icon_path);
		}
		else
		{
			log() << "progid: icon not set\n";
			return false;
		}
	}

	return true;
}

bool check_progid(bool create)
{
	const utf8_string subkey(utf8_string("SOFTWARE\\Classes\\") + prog_id);
	
	log() << "progid: checking " << subkey << "\n";

	if (get_key_value(subkey) != program_name())
	{
		if (create)
		{
			set_key_if_needed(subkey, utf8_string(), program_name());
		}
		else
		{
			log() << "progid: not set\n";
			return false;
		}
	}

	if (!check_default_icon(subkey, false))
	{
		if (create)
			check_default_icon(subkey, true);
		else
			return false;
	}

	if (!check_command(subkey, false))
	{
		if (create)
			check_command(subkey, true);
		else
			return false;
	}

	return true;
}

bool check_openwith(const utf8_string& subkey_root, bool create)
{
	const utf8_string subkey(
		subkey_root + utf8_string("\\OpenWithProgids"));

	log() << "assoc: checking openwith\n";

	if (!has_key_value(subkey, prog_id))
	{
		if (create)
		{
			set_key_value(subkey, prog_id, utf8_string());
		}
		else
		{
			log() << "openwith: missing\n";
			return false;
		}
	}

	return true;
}

bool check_ext(const utf8_string& ext, bool create)
{
	const utf8_string subkey(utf8_string("SOFTWARE\\Classes\\") + ext);
	const utf8_string userchoice(utf8_string(
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\"
		"FileExts\\") + ext + utf8_string("\\UserChoice"));
	const utf8_string userchoice_progid(
		userchoice + utf8_string("\\Progid"));

	log() << "assoc: checking " << subkey << "\n";

	if (get_key_value(subkey) != prog_id)
	{
		if (create)
		{
			set_key_value(subkey, utf8_string(), prog_id);
		}
		else
		{
			log() << "assoc: " << ext << " not set\n";
			return false;
		}
	}

	if (get_key_value(subkey, utf8_string("Content Type")) != torrent_mime)
	{
		if (create)
		{
			set_key_value(
				subkey, utf8_string("Content Type"), torrent_mime);
		}
		else
		{
			log() << "assoc: bad content type\n";
			return false;
		}
	}

	if (!check_openwith(subkey, false))
	{
		if (create)
			check_openwith(subkey, true);
		else
			return false;
	}

	if (has_key(userchoice) && get_key_value(userchoice_progid) != prog_id)
	{
		if (create)
		{
			delete_key(userchoice);
		}
		else
		{
			log() << "assoc: existing userchoice\n";
			return false;
		}
	}

	return true;
}

bool check_file_associations(bool create)
{
	if (!check_ext(utf8_string(".torrent"), create))
		return false;

	if (!check_ext(utf8_string(".btx"), create))
		return false;

	return true;
}

bool check_scheme_associations(bool create)
{
	const utf8_string subkey("SOFTWARE\\Classes\\magnet");

	const utf8_string disp("Magnet");
	const utf8_string proto("URL Protocol");

	log() << "scheme: checking " << subkey << "\n";

	if (get_key_value(subkey) != disp)
	{
		if (create)
		{
			set_key_if_needed(subkey, utf8_string(), disp);
		}
		else
		{
			log() << "scheme: bad display name\n";
			return false;
		}
	}

	if (!has_key_value(subkey, proto))
	{
		if (create)
		{
			set_key_value(subkey, proto, utf8_string());
		}
		else
		{
			log() << "scheme: missing proto\n";
			return false;
		}
	}

	if (!check_default_icon(subkey, false))
	{
		if (create)
			check_default_icon(subkey, true);
		else
			return false;
	}

	if (!check_command(subkey, false))
	{
		if (create)
			check_command(subkey, true);
		else
			return false;
	}

	return true;
}

assoc associated()
{
	try
	{
		if (!check_app_paths(false))
			return assoc(false);

		if (!check_progid(false))
			return assoc(false);

		if (!check_file_associations(false))
			return assoc(false);

		if (!check_scheme_associations(false))
			return assoc(false);

		return assoc(true);
	}
	catch(bad_registry&)
	{
		log_error() << "registry problem, can't check associations\n";
		return assoc(true);
	}
}

void associate()
{
	try
	{
		check_app_paths(true);
		check_progid(true);
		check_file_associations(true);
		check_scheme_associations(true);

		SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST, 0, 0);
	}
	catch(bad_registry&)
	{
		log_error() << "registry problem, can't set associations\n";
	}
}

void add_image_handlers()
{
	// noop
}

image_map get_image_map()
{
	image_map m;

#define a(e, s) m.insert(std::make_pair(images::e, utf8_string(s)))

	// todo: harcoded to 16x16, this looks okay on the classic theme on
	// windows 7, haven't tested it on anything else
	a(add, "add_16");
	a(remove, "remove_16");
	a(resume, "resume_16");
	a(pause, "pause_16");
	a(up, "up_16");
	a(down, "down_16");
	a(top, "top_16");
	a(bottom, "bottom_16");
	a(active, "active_16");
	a(upload, "upload_16");
	a(download, "download_16");
	a(inactive, "inactive_16");
	a(all, "all_16");
	a(creeper, "creeper");
	a(icon_16, "icon_16");
	a(icon_24, "icon_24");
	a(icon_32, "icon_32");
	a(icon_48, "icon_48");
	a(icon_64, "icon_64");
	a(icon_128, "icon_128");
	a(close_search, "close_search");
	a(pcon_16, "pcon_16");

#undef a

	return m;
}

wxBitmap make_image_unsafe(images::image i)
{
	const image_map m = get_image_map();
	
	image_map::const_iterator itor = m.find(i);
	bt_assert(itor != m.end());
	if (itor == m.end())
		return wxBitmap();

	const utf8_string& name = itor->second;

	// this gets the image from the resources

	// baretorrent runs in a dll on windows, this gets the module
	// handle for it since the resources are in it (and not in the
	// executable)
	//
	// GetModuleHandleEx() can take any address in memory and find to
	// which module it belongs; &make_image is used, but it could be
	// anything as long as it's in the baretorrent dll

	HMODULE hmod = 0;
	if (!GetModuleHandleEx(
		GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
		GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
		reinterpret_cast<const wchar_t*>(&make_image),
		&hmod))
	{
		log_error()
			<< "couldn't get module handle, "
			<< system_message(GetLastError()) << "\n";
		return make_bad_image();
	}

	// getting the resource handle for the given image
	const utf16_string wname = utf8_to_utf16(name);

	HRSRC h = ::FindResource(hmod, wname.c_str(), L"PNGTYPE");
	bt_assert(h);
	if (!h)
		return wxBitmap();

	// size of the image
	DWORD s = ::SizeofResource(hmod, h);
	bt_assert(s > 0);
	if (s == 0)
		return wxBitmap();

	// getting a handle to the image
	HGLOBAL r = ::LoadResource(hmod, h);
	bt_assert(r);
	if (!r)
		return wxBitmap();

	// getting a pointer to the data
	const char *d = static_cast<const char *>(LockResource(r));
	bt_assert(r);
	if (!d)
		return wxBitmap();

	// nothing needs to be done on the various handles above, they're
	// not allocating anything

	wxMemoryInputStream in(d, s);
	return wxBitmap(wxImage(in, wxBITMAP_TYPE_PNG));
}

wxBitmap make_image(images::image i)
{
	wxBitmap b = make_image_unsafe(i);
	if (!b.IsOk())
		return make_bad_image();

	return b;
}

wxIcon make_icon(images::image i)
{
	// no difference between images and icons
	wxIcon icon;
	icon.CopyFromBitmap(make_image(i));

	return icon;
}

utf8_string make_time(const utf8_string& format, std::time_t s)
{
	const std::size_t buffer_size = 500;
	wchar_t buffer[buffer_size + 1] = {0};

	const utf16_string wf = utf8_to_utf16(format);
	const std::tm t = *std::localtime(&s);

	const std::size_t n =
		std::wcsftime(buffer, buffer_size, wf.c_str(), &t);
	
	if (n == 0)
	{
		log_error() << utf8_string("wcsftime failed\n");
		return utf8_string();
	}

	return utf16_to_utf8(utf16_string(buffer, n * sizeof(wchar_t)));
}

utf8_string env(const utf8_string& name)
{
	const utf16_string name_16 = utf8_to_utf16(name);

	const DWORD s = GetEnvironmentVariable(name_16.c_str(), nullptr, 0);
	boost::shared_array<wchar_t> b(new wchar_t[s + 1]);
	std::memset(b.get(), 0, sizeof(wchar_t) * (s + 1));

	GetEnvironmentVariable(name_16.c_str(), b.get(), s);

	return utf16_to_utf8(utf16_string(b.get()));
}

void env(const utf8_string& name, const utf8_string& value)
{
	const utf16_string name_16 = utf8_to_utf16(name);
	const utf16_string value_16 = utf8_to_utf16(value);

	SetEnvironmentVariable(name_16.c_str(), value_16.c_str());
}

void add_to_path(const utf8_path& p)
{
	const utf8_string cp = env(utf8_string("PATH"));
	const utf8_string np = cp + utf8_string(";") + p.as_utf8_string();

	env(utf8_string("PATH"), np);
}

void add_firewall_exception()
{
	// todo: this is untested (and probably bunk)
/*	try
	{
		bstr name(bstr_ptr(::SysAllocString(L"baretorrent")));
		bstr description(bstr_ptr(::SysAllocString(L"allows incoming connections to baretorrent")));
		bstr group(bstr_ptr(::SysAllocString(L"wut?")));
		bstr app(bstr_ptr(::SysAllocString(module_name().c_str())));
		bstr ports(::SysAllocString(to_utf16(opt<std::string>("bt.network.listen")).c_str()));

		INetFwPolicy2* policy_p = nullptr;
		HRESULT hr = ::CoCreateInstance(
			__uuidof(NetFwPolicy2), nullptr, CLSCTX_INPROC_SERVER,
			__uuidof(INetFwPolicy2),
			reinterpret_cast<void**>(&policy_p));
		throw_if_failed(hr, "can't create policy2 object");
		bt_assert(policy_p);
		boost::shared_ptr<INetFwPolicy2> policy(policy_p);

		INetFwRules* rules_p = nullptr;
		hr = policy->get_Rules(&rules_p);
		throw_if_failed(hr, "can't get rules");
		bt_assert(rules_p);
		boost::shared_ptr<INetFwRules> rules(rules_p);

		long b = 0;
		hr = policy->get_CurrentProfileTypes(&b);
		throw_if_failed(hr, "can't get profile types");

		const long public_profile = NET_FW_PROFILE2_PUBLIC;
		if ((b & public_profile) && (b != public_profile))
			b ^= public_profile;

		INetFwRule* rule_p = nullptr;
		hr = ::CoCreateInstance(
			__uuidof(NetFwRule), nullptr, CLSCTX_INPROC_SERVER,
			__uuidof(INetFwRule), reinterpret_cast<void**>(&rule_p));
		throw_if_failed(hr, "can't create rule");
		bt_assert(rule_p);
		boost::shared_ptr<INetFwRule> rule(rule_p);

		rule->put_Name(name.get());
		rule->put_Description(description.get());
		rule->put_Grouping(group.get());
		rule->put_ApplicationName(app.get());
		rule->put_RemotePorts(ports.get());

		hr = rules->Add(rule.get());
		throw_if_failed(hr, "can't add rule");

		log() << "added rule\n";
	}
	catch(bad_hresult&)
	{
	}*/
}

BOOL WINAPI ctrl_handler(DWORD t)
{ 
	if (t == CTRL_C_EVENT)
    {
		log() << "ctrl-c, quitting\n";
		quit();
    }

	return FALSE;
}

void set_signals()
{
	SetConsoleCtrlHandler(ctrl_handler, TRUE);
}

bool set_shutdown_token(bool set)
{
	HANDLE t = 0;
	TOKEN_PRIVILEGES tp = {};

	BOOL r = ::OpenProcessToken(
		GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &t);

	if (!r)
	{
		log_error()
			<< "can't open process token, "
			<< system_message(GetLastError()) << "\n";

		return false;
	}

	r = ::LookupPrivilegeValue(
		NULL, SE_SHUTDOWN_NAME, &tp.Privileges[0].Luid);

	if (!r)
	{
		log_error()
			<< "can't lookup privilege value, "
			<< system_message(GetLastError()) << "\n";

		return false;
	}

	tp.PrivilegeCount = 1;

	if (set)
		tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
	else
		tp.Privileges[0].Attributes = 0;

	r = ::AdjustTokenPrivileges(t, FALSE, &tp, 0, NULL, 0);

	if (!r)
	{
		log_error()
			<< "can't adjust token privileges, "
			<< system_message(GetLastError()) << "\n";

		return false;
	}

	return true;
}

void power(power_actions t)
{
	switch (t)
	{
		case shutdown_power:
		{
			if (!set_shutdown_token(true))
				return;

			DWORD reason =
				SHTDN_REASON_MAJOR_OTHER | SHTDN_REASON_MINOR_OTHER |
				SHTDN_REASON_FLAG_PLANNED;

			if (!::InitiateSystemShutdownEx(
				NULL, NULL, 0, FALSE, FALSE, reason))
			{
				log_error()
					<< "system shutdown failed, "
					<< system_message(GetLastError()) << "\n";
			}

			set_shutdown_token(false);

			break;
		}

		case hibernate_power:
		{
			if (!set_shutdown_token(true))
				return;

			if (!::SetSuspendState(TRUE, FALSE, FALSE))
			{
				log_error()
					<< "system hibernate failed, "
					<< system_message(GetLastError()) << "\n";
			}

			set_shutdown_token(false);

			break;
		}

		case suspend_power:
		{
			if (!set_shutdown_token(true))
				return;

			if (!::SetSuspendState(TRUE, FALSE, FALSE))
			{
				log_error()
					<< "system suspend failed, "
					<< system_message(GetLastError()) << "\n";
			}

			set_shutdown_token(false);

			break;
		}

		case no_power_action:
		default:
		{
			break;
		}
	}
}

bool has_power(power_actions t)
{
	SYSTEM_POWER_CAPABILITIES c = {};
	if (!GetPwrCapabilities(&c))
	{
		log_error()
			<< "can't get power capabilities, "
			<< system_message(GetLastError()) << "\n";

		return true;
	}

	switch (t)
	{
		case shutdown_power:
		{
			return true;
		}

		case hibernate_power:
		{
			if (c.SystemS4 && c.HiberFilePresent)
				return true;
			else
				return false;
		}

		case suspend_power:
		{
			if (c.SystemS1 || c.SystemS2 || c.SystemS3)
				return true;
			else
				return false;
		}

		case no_power_action:
		default:
		{
			return false;
		}
	}
}

bool has_power_management()
{
	return true;
}

void detach_console()
{
	FreeConsole();

	// todo, windows 8: for whatever reason, the handles are not closed the
	// same way as 7 and before and python fails while trying to initialize
	// its standard streams
	*stdin = *fopen("NUL", "r+");
	*stdout = *fopen("NUL", "r+");
	*stderr = *fopen("NUL", "r+");
}

void create_console()
{
	if (!AllocConsole())
		return;

	// setting bigger buffer
	CONSOLE_SCREEN_BUFFER_INFO coninfo = {0};
	GetConsoleScreenBufferInfo(
		GetStdHandle(STD_OUTPUT_HANDLE), &coninfo);
	coninfo.dwSize.Y = 999;
	SetConsoleScreenBufferSize(
		GetStdHandle(STD_OUTPUT_HANDLE), coninfo.dwSize);

	// redirecting stdout
	intptr_t lStdHandle = (intptr_t)(GetStdHandle(STD_OUTPUT_HANDLE));
	int hConHandle = _open_osfhandle(lStdHandle, _O_TEXT);
	FILE* fp = _fdopen(hConHandle, "w");
	*stdout = *fp;
	setvbuf(stdout, NULL, _IONBF, 0);

	// redirecting stdin
	lStdHandle = (intptr_t)GetStdHandle(STD_INPUT_HANDLE);
	hConHandle = _open_osfhandle(lStdHandle, _O_TEXT);
	fp = _fdopen(hConHandle, "r");
	*stdin = *fp;
	setvbuf(stdin, NULL, _IONBF, 0);

	// redirecting stderr
	lStdHandle = (intptr_t)GetStdHandle(STD_ERROR_HANDLE);
	hConHandle = _open_osfhandle(lStdHandle, _O_TEXT);
	fp = _fdopen(hConHandle, "w");
	*stderr = *fp;
	setvbuf(stderr, NULL, _IONBF, 0);

	std::ios::sync_with_stdio();
	
	g_std_log.out(&std::cerr);
}

void log_to_console(
	std::ostream* out, log_t::levels lv, const utf8_string& s)
{
	CONSOLE_SCREEN_BUFFER_INFO info = {0};

	// set to false if getting the screen buffer info fails
	bool has_console = true;

	const HANDLE h = GetStdHandle(STD_OUTPUT_HANDLE);
	const WORD c = FOREGROUND_RED | FOREGROUND_INTENSITY;

	// setting the foreground color to red if this is an error
	if (lv == log_t::error_level)
	{
		if (GetConsoleScreenBufferInfo(h, &info) == false)
			has_console = false;
		else
			SetConsoleTextAttribute(h, c);
	}

	if (out)
		*out << s;

	// resetting the original foreground color
	if (lv == log_t::error_level && has_console)
		SetConsoleTextAttribute(h, info.wAttributes);
}

utf8_string name()
{
	typedef void (WINAPI *GNSI)(LPSYSTEM_INFO);

	// getting the version
	OSVERSIONINFOEX osvi = {0};
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	if (!GetVersionEx((OSVERSIONINFO*) &osvi))
		return utf8_string("(some windows)");

	const int major = osvi.dwMajorVersion;
	const int minor = osvi.dwMinorVersion;
	const int type = osvi.wProductType;

	// getting the architecture on which baretorrent is currently
	// running (as opposed to the one on which is was compiled)
	//
	// an x86 application running on an x64 windows will always give
	// x86 as the architecture because it is running on wow64 unless
	// GetNativeSystemInfo() is used directly; since it is not
	// exported, it needs to be found manually in kernel32.dll

#pragma warning(push)
#pragma warning(disable: 4191)
	GNSI gnsi = reinterpret_cast<GNSI>(GetProcAddress(
		GetModuleHandle(TEXT("kernel32.dll")), "GetNativeSystemInfo"));
#pragma warning(pop)
	
	SYSTEM_INFO si = {0};

	// this shouldn't fail, but fall back on GetSystemInfo() just in
	// case (which is what happens on x64 in any case)
	if (gnsi)
		gnsi(&si);
	else
		GetSystemInfo(&si);

	const int arch = si.wProcessorArchitecture;

	std::wostringstream oss;
	oss << "Microsoft ";

	// translating the version numbers into a friendly name
	//
	// two different editions of windows can have the same version
	// number a different type, such as vista being a workstation and
	// server 2008 being a server

	if (major == 6)
	{
		if (minor == 0)
		{
			if (type == VER_NT_WORKSTATION)
				oss << L"Windows Vista";
			else
				oss << L"Windows Server 2008";
		}
		else if (minor == 1)
		{
			if (type == VER_NT_WORKSTATION)
				oss << L"Windows 7";
			else
				oss << L"Windows Server 2008 R2";
		}
		else if (minor == 2)
		{
			if (type == VER_NT_WORKSTATION)
				oss << L"Windows 8";
			else
				oss << L"Windows Server 2012";
		}
		else
		{
			// for future windows
			oss << L"Windows " << major << L"." << minor << L"?";
		}
	}
	else if (major == 5 && minor == 2)
	{
		if (type == VER_NT_WORKSTATION &&
			arch == PROCESSOR_ARCHITECTURE_AMD64)
		{
			oss << L"Windows XP Professional x64 Edition";
		}
		else
		{
			oss << L"Windows Server 2003";
		}
	}
	else if (major == 5 && minor == 1)
	{
		oss << L"Windows XP";
	}
	else if (major == 5 && minor == 0)
	{
		oss << L"Windows 2000";
	}
	else
	{
		// for old, future or weird windows
		oss << L"Windows " << major << L"." << minor << L"?";
	}

	// service packs and stuff
	if (std::wcslen(osvi.szCSDVersion) > 0)
		oss << L" " << osvi.szCSDVersion;

	// build number
	oss << L" (build " << osvi.dwBuildNumber << L")";

	// windows xp is a special case handled above and older versions
	// (that aren't supported anyways) are x86; starting from vista,
	// this information is valid
	if (major >= 6)
	{
		if (arch == PROCESSOR_ARCHITECTURE_AMD64)
			oss << L" 64-bit";
		else if (arch == PROCESSOR_ARCHITECTURE_INTEL)
			oss << L" 32-bit";
	}

	return utf16_to_utf8(utf16_string(oss.str()));
}

}	// namespace platform
}	// namespace baretorrent

#endif	// BARETORRENT_WINDOWS
