// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "xml.h"
#include "log.h"
#include "locale.h"
#include "utility.h"

namespace baretorrent
{
namespace xml
{

// wraps xmlDocPTr
//
class xml_doc : boost::noncopyable
{
public:
	xml_doc(xmlDocPtr p)
		: p_(p)
	{
	}

	~xml_doc()
	{
		xmlFreeDoc(p_);
	}

	operator xmlDocPtr()
	{
		return p_;
	}

private:
	xmlDocPtr p_;
};


dom::dom()
{
}

dom dom::load(const utf8_path& file)
{
	log_v() << "dom: loading " << file << "\n";

	xml_doc doc(xmlReadFile(
		file.as_utf8_string().c_str(), nullptr, XML_PARSE_NOBLANKS));

	if (!doc)
		throw bad_xml(utf8_string("can't open file"));

	xmlNodePtr r = xmlDocGetRootElement(doc);
	if (!r)
	{
		log_error() << "dom: no root\n";
		throw bad_xml(utf8_string("no root"));
	}

	dom d;
	d.root_ = element(r);

	return d;
}

dom dom::create(const utf8_string& root)
{
	dom d;
	d.root_ = element(root);

	return d;
}

const element& dom::root() const
{
	return root_;
}

element& dom::root()
{
	return root_;
}

void dom::write(const utf8_path& p) const
{
	xmlTextWriterPtr w = xmlNewTextWriterFilename(
		p.as_utf8_string().c_str(), 0);

	if (!w)
		throw bad_xml(utf8_string("can't open for writing"));

	xmlTextWriterSetIndent(w, 1);
	const xmlChar* i = reinterpret_cast<const xmlChar*>("\t");
	xmlTextWriterSetIndentString(w, i);

	root_.write(w);

	xmlFreeTextWriter(w);
}


element::element()
{
}

element::element(const utf8_string& name, const utf8_string& value)
	: name_(name), value_(value)
{
}

element::element(xmlNodePtr n)
	: name_(reinterpret_cast<const char*>(n->name))
{
	bt_assert(n);

	if (n->content)
		value_ = utf8_string(reinterpret_cast<const char*>(n->content));

	if (n->properties)
	{
		// attributes

		xmlNode* p = n->properties->children;
		while (p)
		{		
			utf8_string k =
				utf8_string(reinterpret_cast<const char*>(p->name));
			utf8_string v =
				utf8_string(reinterpret_cast<const char*>(p->content));

			attributes_.insert(std::make_pair(k, v));

			p = p->next;
		}
	}

	bool only_text = true;

	// if all the children are text nodes, set them as the value instead

	xmlNodePtr c = n->children;
	while (c)
	{
		element e(c);
		if (c->type != XML_TEXT_NODE)
			only_text = false;

		children_.push_back(element(c));
		c = c->next;
	}

	if (only_text && !children_.empty())
	{
		// todo: ? why not all of them?
		bt_assert(children_.size() == 1);
		value_ = children_[0].value_;
		children_.clear();
	}
}

void element::write(xmlTextWriterPtr w) const
{
	bt_assert(w);

	int r = xmlTextWriterStartElement(
		w, reinterpret_cast<const xmlChar*>(name_.c_str()));

	if (r == -1)
		throw bad_xml(utf8_string("couldn't write element"));

	// attributes
	std::pair<utf8_string, utf8_string> p;
	foreach(p, attributes_)
	{
		r = xmlTextWriterWriteAttribute(
			w,
			reinterpret_cast<const xmlChar*>(p.first.c_str()),
			reinterpret_cast<const xmlChar*>(p.second.c_str()));

		if (r == -1)
			throw bad_xml(utf8_string("couldn't write element"));
	}

	if (!children_.empty())
	{
		// children
		bt_assert(value_.empty());

		foreach (const element& e, children_)
			e.write(w);
	}
	else
	{
		// value
		if (!value_.empty())
		{
			r = xmlTextWriterWriteString(
				w, reinterpret_cast<const xmlChar*>(value_.c_str()));

			if (r == -1)
				throw bad_xml(utf8_string("couldn't write element"));
		}
	}

	r = xmlTextWriterEndElement(w);
	if (r == -1)
		throw bad_xml(utf8_string("couldn't write element"));
}

void element::add(const element& e)
{
	bt_assert(value_.empty());
	children_.push_back(e);
}

const element& element::operator[](const utf8_string& name) const
{
	static element bad;

	foreach (const element& e, children_)
	{
		if (e.name_ == name)
			return e;
	}

	throw bad_element(utf8_string("unknown element ") + name);
}

utf8_string element::safe_value(const utf8_string& child) const
{
	if (has(child))
		return (*this)[child].value();
	
	return utf8_string();
}

const utf8_string& element::value() const
{
	return value_;
}

elements::const_iterator element::begin() const
{
	return children_.begin();
}

elements::const_iterator element::end() const
{
	return children_.end();
}

bool element::empty() const
{
	return children_.empty();
}

std::size_t element::size() const
{
	return children_.size();
}

bool element::has(const utf8_string& child) const
{
	foreach (const element& e, children_)
	{
		if (e.name_ == child)
			return true;
	}

	return false;
}

}	// namespace xml
}	// namespace baretorrent
