// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#ifndef BARETORRENT_XML_INCLUDED
#define BARETORRENT_XML_INCLUDED

#include "utility.h"

// this is a wrapper around libxml

namespace baretorrent
{
namespace xml
{

class bad_xml : public std::runtime_error
{
public:
	bad_xml(const utf8_string& w)
		: std::runtime_error("bad xml"), what_(w)
	{
	}

	~bad_xml() throw()
	{
	}

	const char* what() const throw()
	{
		return what_.c_str();
	}

private:
	utf8_string what_;
};

class bad_element : public std::runtime_error
{
public:
	bad_element(const utf8_string& w)
		: std::runtime_error("bad element"), what_(w)
	{
	}

	~bad_element() throw()
	{
	}

	const char* what() const throw()
	{
		return what_.c_str();
	}

private:
	utf8_string what_;
};

class element;
class dom;
typedef std::vector<element> elements;

// an xml element
//
class element
{
	friend class dom;

public:
	// empty element
	//
	element();

	// an element with the given name and content
	//
	explicit element(
		const utf8_string& name,
		const utf8_string& value=utf8_string(""));
	
	// an element with the given name and content, converted using
	// operator<<
	//
	template <class T>
	element(const utf8_string& name, const T& value)
		: name_(name), value_(as_utf8_string(value))
	{
	}


	// returns a child element having the given name; throws if the
	// element doesn't exist
	//
	const element& operator[](const utf8_string& name) const;
	
	// returns the content of the given child element or an empty string
	// if that element doesn't exist
	//
	utf8_string safe_value(const utf8_string& child) const;

	// returns the content of the given child element converted using
	// operator<< or a T() if that element doesn't exist
	//
	template <class T>
	T safe_value(const utf8_string& child) const
	{
		if (!has(child))
			return T();

		return as<T>(safe_value(child));
	}
	
	// returns whether this element has the given child element
	//
	bool has(const utf8_string& child) const;

	// returns the content of this element
	//
	const utf8_string& value() const;
	

	// returns an iterator to the first child
	//
	elements::const_iterator begin() const;

	// returns an iterator to the one past the last child
	//
	elements::const_iterator end() const;

	// returns whether this element has any children
	//
	bool empty() const;

	// returns the number of children
	//
	std::size_t size() const;


	// adds the given element to the child list
	//
	void add(const element& e);

	// writes this element to the given text writer
	//
	void write(xmlTextWriterPtr w) const;

private:
	typedef std::map<utf8_string, utf8_string> attributes;
	
	// name of this element
	utf8_string name_;

	// text content
	utf8_string value_;

	// attributes
	attributes attributes_;

	// children
	elements children_;


	// an element from the libxml node
	//
	element(xmlNodePtr n);
};

// an xml document
//
class dom
{
public:
	// loads the dom from the given file
	//
	static dom load(const utf8_path& file);

	// creates a dom with the given root name
	//
	static dom create(const utf8_string& root);


	// returns the root element
	//
	element& root();

	// returns the root element
	//
	const element& root() const;

	
	// writes this dom to the given file
	//
	void write(const utf8_path& f) const;

private:
	// root element
	element root_;

	// an empty dom
	//
	dom();
};

}	// namespace xml
}	// namespace baretorrent


// for BOOST_FOREACH with elements
namespace boost
{
	template<>
	struct range_mutable_iterator< baretorrent::xml::element >
	{
		typedef baretorrent::xml::elements::iterator type;
	};

	template<>
	struct range_const_iterator< baretorrent::xml::element >
	{
		typedef baretorrent::xml::elements::const_iterator type;
	};
}

#endif	// BARETORRENT_XML_INCLUDED
